package ru.crp.lvms.managers;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.stereotype.Service;
import ru.crp.cmlb.model.entities.base.Named;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.managers.effective.characteristics.ECManager;
import ru.crp.lvms.managers.misc.BaseManagerSuggest;
import ru.crp.lvms.model.AccessRights;
import ru.crp.lvms.model.dao.BaseSpaceDao;
import ru.crp.lvms.model.dao.CompositeMaterialDao;
import ru.crp.lvms.model.dao.LayersDao;
import ru.crp.lvms.model.dao.ParameterValuesDao;
import ru.crp.lvms.model.entities.compositeMaterials.CompositeMaterialEntity;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.Layer;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.utils.MessageUtils;
import ru.crp.lvms.utils.ValidationHelper;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.function.Function;

import static java.util.stream.Collectors.toMap;
import static ru.crp.lvms.model.AccessRights.update;
import static ru.crp.lvms.utils.MessageUtils.*;


@Service
public class CompositeMaterialManager extends BaseManagerSuggest<CompositeMaterialEntity> {
    @Autowired
    private CompositeMaterialDao compositeMaterialDao;

    @Autowired
    private LayersDao layersDao;

    @Autowired
    private FileManager fileManager;

    @Autowired
    private ParameterValuesDao parameterValuesDao;

    @Autowired
    private ValidationHelper validationHelper;

    @Autowired
    private SimulationManager simulationManager;

    @Autowired
    private OptimizationManager optimizationManager;

    @Autowired
    private ECManager ecManager;


    @Override
    protected BaseSpaceDao<CompositeMaterialEntity> aliveDao() {
        return compositeMaterialDao;
    }


    public void extendMaterial(CompositeMaterialEntity material, UserEntity currentUser) {
        material.setLayers(layersDao.findByMaterialId(material.getId()));
        material.setMaterialFiles(fileManager.getMaterialFiles(true, material.getId(), currentUser));
        material.setSampleFiles(fileManager.getSampleFiles(material.getId(), currentUser));
        material.setParameterValues(parameterValuesDao.findByMaterialId(material.getId()));
    }


    public List<Layer> getLayers(int compositeMaterialId, boolean checkRights, UserEntity currentUser) {
        if (checkRights) {
            checkEntityExistsAndAccesible(compositeMaterialId, AccessRights.update, currentUser);
        }
        return layersDao.findByMaterialId(compositeMaterialId);
    }

    public void addLayer(int materialId, Layer newLayer, UserEntity currentUser) {
        checkEntityExistsAndAccesible(materialId, AccessRights.update, currentUser);
        if (newLayer.getPosition() < 0) {
            newLayer.setPosition(0);
        }
        List<Layer> existedLayers = getLayers(materialId, false, currentUser);
        if (existedLayers.size() < newLayer.getPosition()) {
            existedLayers.add(newLayer);
        } else {
            existedLayers.add(newLayer.getPosition(), newLayer);
        }
        setLayersForMaterial(materialId, existedLayers, currentUser);
    }

    public void deleteLayer(int materialId, int layerId, UserEntity currentUser) {
        checkEntityExistsAndAccesible(materialId, AccessRights.update, currentUser);
        List<Layer> existedLayers = getLayers(materialId, false, currentUser);
        existedLayers.removeIf(layer -> layer.getId() == layerId);
        setLayersForMaterial(materialId, existedLayers, currentUser);
    }


    public void setLayersForMaterial(int materialId, List<Layer> newLayers, UserEntity currentUser) {
        checkEntityExistsAndAccesible(materialId, AccessRights.update, currentUser);
        newLayers.sort(Comparator.comparing(Layer::getPosition));
        layersDao.deleteMaterialLayers(materialId);
        if (newLayers.isEmpty()) {
            return;
        }
        for (int i = 0; i < newLayers.size(); ++i) {
            newLayers.get(i).setPosition(i);
            newLayers.get(i).setCompositeMaterialId(materialId);
        }
        layersDao.batchInsertLayers(newLayers);
    }

    public void updateLayersOrder(int materialId, List<Integer> newOrder, UserEntity currentUser) {
        checkEntityExistsAndAccesible(materialId, AccessRights.update, currentUser);
        Map<Integer, Layer> existedLayers = getLayers(materialId, false, currentUser).stream()
                .collect(toMap(Layer::getId, Function.identity()));
        if (existedLayers.size() != newOrder.size()) {
            throw new BadRequestException("Неверное количество слоев");
        }
        List<Layer> orderedLayers = new ArrayList<>(existedLayers.size());
        newOrder.forEach(id -> {
            Layer layer = existedLayers.get(id);
            if (layer == null) {
                throw new BadRequestException("Несуществующий идентификатор слоя");
            }
            orderedLayers.add(layer);
        });
        setLayersForMaterial(materialId, orderedLayers, currentUser);
    }


    @Override
    public void checkDelete(CompositeMaterialEntity entity, @Nullable UserEntity currentUser) {
        validationHelper.checkDependencies(entity.getId(), COMPOSITE_MATERIAL_HAS_MATERIAL_FILES,
                this::getMaterialFiles, currentUser);
        validationHelper.checkDependencies(entity.getId(), COMPOSITE_MATERIAL_HAS_SAMPLE_FILES,
                fileManager::getSampleFiles, currentUser);
        validationHelper.checkDependencies(entity.getId(), COMPOSITE_MATERIAL_HAS_SIMULATIONS,
                simulationManager::getSimulationsByMaterialId, currentUser);
        validationHelper.checkDependencies(entity.getId(), COMPOSITE_MATERIAL_HAS_EFFECTIVE_CHARACTERISTICS,
                ecManager::getEffectiveCharacteristicsByCompositeMaterialId, currentUser);
        super.checkDelete(entity, currentUser);
    }

    private List<? extends Named> getMaterialFiles(Integer materialId, UserEntity currentUser) {
        return fileManager.getMaterialFiles(true, materialId, currentUser);
    }

    @Override
    public void delete(CompositeMaterialEntity entity, @Nullable UserEntity currentUser) {
        // todo delete parameter files from disk and db
        checkDelete(entity, currentUser);
        parameterValuesDao.deleteAllParametersForMaterial(entity.getId());
        layersDao.deleteMaterialLayers(entity.getId());
        aliveDao().delete(entity.getId());
    }

    public List<CompositeMaterialEntity> getCompositeMaterialsUsingMaterial(int materialId,
                                                                            @Nullable UserEntity currentUser) {
        return compositeMaterialDao.getCompositeMaterialsUsingMaterial(materialId);
    }

    public List<CompositeMaterialEntity> getCompositeMaterialsUsingTechnology(int technologyId,
                                                                              @Nullable UserEntity currentUser) {
        return compositeMaterialDao.getCompositeMaterialsUsingTechnology(technologyId);
    }

    @Nonnull
    public FileEntity deleteMaterialFile(@Nullable Integer materialId, int materialFileId,
                                         UserEntity currentUser) {
        if (materialId != null) {
            checkEntityExistsAndAccesible(materialId, update, currentUser);
        }
        validationHelper.checkDependencies(materialFileId, MessageUtils.FILE_IS_USED_IN_OPTIMIZATION,
                this::getConnectedOptimisations, currentUser);
        return fileManager.deleteFile(materialFileId, currentUser);
    }


    private List<? extends Named> getConnectedOptimisations(Integer materialFileId, UserEntity unused) {
        return optimizationManager.getOptimizationsThatUseFile(materialFileId, unused);
    }

    public List<CompositeMaterialEntity> getByTechnologyId(int technologyId, UserEntity user) {
        return compositeMaterialDao.getCompositeMaterialsUsingTechnology(technologyId);
    }

    public PageImpl<CompositeMaterialEntity> getPageByTechnologyId(int technologyId, Map<String, Object> filters,
                                                                   Pageable pageRequest, UserEntity user) {
        return compositeMaterialDao.getCompositeMaterialsPageUsingTechnology(filters, pageRequest, technologyId);
    }

    public void removeTechnologyEquipment(int technologyId, int equipmentId) {
        compositeMaterialDao.removeTechnologyEquipment(technologyId, equipmentId);
    }

    public CompositeMaterialEntity getCompositeMaterialUsingSimulation(int simulationId) {
        return compositeMaterialDao.getCompositeMaterialUsingSimulation(simulationId);
    }
}
