package ru.crp.lvms.managers;

import org.springframework.beans.factory.BeanInitializationException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.MessageSource;
import org.springframework.dao.CannotSerializeTransactionException;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.retry.annotation.Backoff;
import org.springframework.retry.annotation.Retryable;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Isolation;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.multipart.MultipartFile;
import ru.crp.cmlb.model.SuggestRequest;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.EntityDeleteException;
import ru.crp.lvms.exceptions.InternalException;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.model.AccessRights;
import ru.crp.lvms.model.dao.FileDao;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.model.entities.misc.ItemType;
import ru.crp.lvms.model.entities.misc.TreeType;
import ru.crp.lvms.model.entities.tree.TreeNode;
import ru.crp.lvms.utils.Utils;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.PostConstruct;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import java.util.Map;

import static ru.crp.lvms.model.AccessRights.update;
import static ru.crp.lvms.model.AccessRights.view;
import static ru.crp.lvms.utils.MessageUtils.*;

@Service
@Transactional(isolation = Isolation.SERIALIZABLE)
// Set retry on transaction isolation fail, use random exponential back off
@Retryable(include = CannotSerializeTransactionException.class, maxAttempts = 5,
        backoff = @Backoff(delay = 100, multiplier = 2, random = true))
public class FileManager {
    public static final String OPTIMIZATIONS_DIR = "optimizations";

    @Autowired
    private FileDao fileDao;

    @Autowired
    private MaterialManager materialManager;

    @Autowired
    private CompositeMaterialManager compositeMaterialManager;

    @Autowired
    private MessageSource msgSr;

    @Autowired
    private TreeManager treeManager;

    @Value("${storage.root}")
    private String filesFolder;

    private Path pathToFiles;


    @PostConstruct
    public void init() {
        pathToFiles = Paths.get(filesFolder);
        if (!Files.exists(pathToFiles)) {
            try {
                Files.createDirectories(pathToFiles);
            } catch (IOException ex) {
                throw new BeanInitializationException(ex.getMessage(), ex);
            }
        }
    }


    @Transactional(readOnly = true)
    @Nullable
    public FileEntity findOne(int fileId, @Nullable UserEntity currentUser) {
        return fileDao.findOne(fileId);
    }

    public FileEntity checkEntityExists(int fileId, @Nullable UserEntity currentUser) {
        FileEntity fileEntity = fileDao.findOne(fileId);
        if (fileEntity == null) {
            throw new NotFoundException(fileId, FileEntity.class);
        }
        return fileEntity;
    }


    // sample files api
    @Transactional(readOnly = true)
    @Nonnull
    public List<FileEntity> suggestSampleFiles(int materialId, SuggestRequest suggestRequest, UserEntity currentUser) {
        checkMaterialExistsAndAccessible(true, materialId, view, currentUser);
        suggestRequest.upsertQuery("pid", String.valueOf(materialId));
        suggestRequest.upsertQuery("file_type_id", String.valueOf(FileType.SampleFile.getId()));
        return fileDao.suggest(suggestRequest, currentUser.getId());
    }


    @Transactional(readOnly = true)
    @Nonnull
    public List<FileEntity> getSampleFiles(int materialId, @Nullable UserEntity currentUser) {
        checkMaterialExistsAndAccessible(true, materialId, view, currentUser);
        return getFiles(materialId, FileType.SampleFile);
    }


    @Transactional(readOnly = true)
    @Nonnull
    public PageImpl<FileEntity> getSampleFilesPage(int materialId,
                                                   Map<String, Object> filters,
                                                   Pageable pageRequest, UserEntity currentUser) {
        checkMaterialExistsAndAccessible(true, materialId, view, currentUser);
        return getFilesPage(materialId, FileType.SampleFile, filters, pageRequest);
    }


    @Transactional(readOnly = true)
    @Nonnull
    public FileEntity getSampleFile(int materialId, int materialFileId,
                                    UserEntity currentUser) {
        checkMaterialExistsAndAccessible(true, materialId, view, currentUser);
        return getFile(materialFileId);
    }


    @Nonnull
    public FileEntity deleteSampleFile(@Nullable Integer materialId, int materialFileId,
                                       UserEntity currentUser) {
        if (materialId != null) {
            checkMaterialExistsAndAccessible(true, materialId, update, currentUser);
        }
        return deleteFile(materialFileId, currentUser);
    }

    // material files api

    @Transactional(readOnly = true)
    @Nonnull
    public List<FileEntity> suggestMaterialFiles(boolean isComposite, int materialId, SuggestRequest suggestRequest,
                                                 UserEntity currentUser) {
        checkMaterialExistsAndAccessible(isComposite, materialId, view, currentUser);
        suggestRequest.upsertQuery("pid", String.valueOf(materialId));
        int fileTypeId = toFileType(isComposite).getId();
        suggestRequest.upsertQuery("file_type_id", String.valueOf(fileTypeId));
        return fileDao.suggest(suggestRequest, currentUser.getId());
    }


    @Transactional(readOnly = true)
    @Nonnull
    public List<FileEntity> getMaterialFiles(boolean isComposite, int materialId, UserEntity currentUser) {
        checkMaterialExistsAndAccessible(isComposite, materialId, view, currentUser);
        return getFiles(materialId, toFileType(isComposite));
    }


    @Transactional(readOnly = true)
    @Nonnull
    public PageImpl<FileEntity> getMaterialFilesPage(boolean isComposite, int materialId,
                                                     Map<String, Object> filters,
                                                     Pageable pageRequest, UserEntity currentUser) {
        checkMaterialExistsAndAccessible(isComposite, materialId, view, currentUser);
        return getFilesPage(materialId, toFileType(isComposite), filters, pageRequest);
    }


    @Transactional(readOnly = true)
    @Nonnull
    public FileEntity getMaterialFile(boolean isComposite, int materialId, int materialFileId,
                                      UserEntity currentUser) {
        checkMaterialExistsAndAccessible(isComposite, materialId, view, currentUser);
        return getFile(materialFileId);
    }

    @Nonnull
    public FileEntity createMaterialFile(boolean isComposite, int materialId, MultipartFile newMaterialFile,
                                         UserEntity currentUser) {
        checkMaterialExistsAndAccessible(isComposite, materialId, update, currentUser);
        checkPathForMaterialFile(isComposite, materialId, newMaterialFile.getOriginalFilename());
        return createFile(newMaterialFile, materialId, toFileType(isComposite), currentUser);
    }


    private void checkPathForMaterialFile(boolean isComposite, int materialId, String newFileName) {
        TreeNode parent;
        if (isComposite) {
            parent = treeManager.findOne(TreeType.compositeMaterials, materialId, null);
        } else {
            parent = treeManager.findOne(TreeType.materials, materialId, null);
        }
        String pathNames = parent.getPathNames() + "/" + newFileName;
        List<TreeNode> conflicted = treeManager.findByPathNames(ItemType.File, pathNames);
        if (!conflicted.isEmpty()) {
            throw new BadRequestException("Файл с таким именем уже используется в материале");
        }
    }


    @Nonnull
    public FileEntity createSampleFile(boolean isComposite, int materialId, MultipartFile newSampleFile,
                                       UserEntity currentUser) {
        checkMaterialExistsAndAccessible(isComposite, materialId, update, currentUser);
        checkPathForMaterialFile(true, materialId, newSampleFile.getOriginalFilename());
        return createFile(newSampleFile, materialId, FileType.SampleFile, currentUser);
    }


    @Nonnull
    public FileEntity updateMaterialFile(boolean isComposite, int materialId,
                                         FileEntity updatedMaterialFile, UserEntity currentUser) {
        checkMaterialExistsAndAccessible(isComposite, materialId, update, currentUser);
        return updateFileEntity(updatedMaterialFile, currentUser);
    }

    private FileType toFileType(boolean isCompositeMaterial) {
        return isCompositeMaterial ? FileType.CompositeMaterialFile : FileType.MaterialFile;
    }


    // parameter files
    @Nonnull
    public FileEntity deleteParameterFile(int fileId, @Nullable UserEntity currentUser) {
        return deleteFile(fileId, currentUser);
    }


    // common files api
    public void updateFileContent(Integer fileId, MultipartFile newFileContent, UserEntity currentUser) {
        FileEntity entityToUpdate = fileDao.findOne(fileId);
        checkNotNull(fileId, entityToUpdate);
        String oldPathToFile = entityToUpdate.getPathToFile();

        Path newPathToFile = writeFileToDisk(newFileContent, pathToFiles.resolve(oldPathToFile).getParent());

        entityToUpdate.setPathToFile(newPathToFile.toString());
        entityToUpdate.setName(newFileContent.getOriginalFilename());
        deleteFromDiskIfNeed(oldPathToFile);
        fileDao.update(entityToUpdate);
    }


    @Nonnull
    private List<FileEntity> getFiles(int pid, FileType fileType) {
        return fileDao.findByParentIdAndType(pid, fileType);
    }


    @Nonnull
    private PageImpl<FileEntity> getFilesPage(int pid, FileType fileType,
                                              Map<String, Object> filters,
                                              Pageable pageRequest) {
        filters.put("pid", pid);
        filters.put("fileTypeId", fileType.getId());
        return fileDao.find(filters, pageRequest);
    }


    @Nonnull
    private FileEntity getFile(int materialFileId) {
        FileEntity one = fileDao.findOne(materialFileId);
        checkNotNull(materialFileId, one);
        return one;
    }


    @Nonnull
    public FileEntity updateFileEntity(FileEntity updatedMaterialFile, UserEntity currentUser) {
        FileEntity existedFile = fileDao.findOne(updatedMaterialFile.getId());
        checkNotNull(updatedMaterialFile.getId(), existedFile);
        // prevent updating not changeable properties
        updatedMaterialFile.setPid(existedFile.getPid());
        updatedMaterialFile.setFileTypeId(existedFile.getFileTypeId());
        updatedMaterialFile.setOwnerId(existedFile.getOwnerId());

        fileDao.update(updatedMaterialFile);
        return fileDao.findOne(updatedMaterialFile.getId());
    }


    private void deleteFilesForMaterial(int materialId, FileType fileType) {
        List<FileEntity> deletedEntities = fileDao.findByParentIdAndType(materialId, fileType);
        fileDao.deleteAllForMaterial(materialId, fileType);
        deletedEntities.stream()
                .map(FileEntity::getPathToFile)
                .forEach(this::deleteFromDiskIfNeed);
    }


    @Nonnull
    public FileEntity deleteFile(int fileId, @Nullable UserEntity currentUser) {
        FileEntity fileToDelete = fileDao.findOne(fileId);
        checkNotNull(fileId, fileToDelete);
        fileDao.delete(fileId);
        deleteFromDiskIfNeed(fileToDelete.getPathToFile());
        return fileToDelete;
    }


    private void checkNotNull(int materialFileId, @Nullable FileEntity one) {
        if (one == null) {
            throw new NotFoundException(materialFileId, FileEntity.class);
        }
    }


    private void deleteFromDiskIfNeed(@Nullable String pathToFileStr) {
        if (pathToFileStr == null) {
            return;
        }
        List<FileEntity> entitiesWithSameFile = fileDao.findByRelativePath(pathToFileStr);
        if (entitiesWithSameFile.isEmpty()) {
            Path pathToFile = pathToFiles.resolve(pathToFileStr);
            try {
                Files.deleteIfExists(pathToFile);
            } catch (IOException ex) {
                throw new EntityDeleteException(msgSr.getMessage(MATERIAL_DELETING_IO, null, null), ex);
            }
        }
    }


    private void checkMaterialExistsAndAccessible(boolean isComposite, int materialId, AccessRights desiredAccess,
                                                  @Nullable UserEntity currentUser) {
        if (isComposite) {
            compositeMaterialManager.checkEntityExistsAndAccesible(materialId, desiredAccess, currentUser);
        } else {
            materialManager.checkEntityExistsAndAccesible(materialId, desiredAccess, currentUser);
        }
    }

    public FileEntity createOptimizationFile(MultipartFile newModel, int optimizationId, FileType fileType, UserEntity currentUser) {
        Path parentDir = pathToFiles.resolve(OPTIMIZATIONS_DIR + "/" + optimizationId);
        return createFile(newModel, null, parentDir, fileType, currentUser);
    }

    public FileEntity createFile(MultipartFile file, @Nullable Integer pid, FileType fileType, UserEntity currentUser) {
        return createFile(file, pid, pathToFiles, fileType, currentUser);
    }

    private FileEntity createFile(MultipartFile file, @Nullable Integer pid, Path parentDir, FileType fileType, UserEntity currentUser) {
        Path path = writeFileToDisk(file, parentDir);
        return createFile(path, file.getOriginalFilename(), pid, fileType, currentUser.getId());
    }

    public FileEntity createOptimizationFile(Path newFile, int optimizationId, FileType fileType, Integer userId) {
        Path parentDir = pathToFiles.resolve(OPTIMIZATIONS_DIR + "/" + optimizationId);
        Path fileInOptimizationDirectory = writeFileToDisk(newFile, parentDir);
        return createFile(fileInOptimizationDirectory, newFile.getFileName().toString(), null, fileType, userId);
    }

    private FileEntity createFile(Path file, String originalName, @Nullable Integer pid, FileType fileType, Integer userId) {
        try {
            String fileName = URLDecoder.decode(originalName, "UTF-8");
            FileEntity newFile = new FileEntity();
            newFile.setPid(pid);
            newFile.setOwnerId(userId);
            newFile.setFileTypeId(fileType.getId());
            newFile.setName(fileName);
            newFile.setPathToFile(file.toString());
            int newFileId = fileDao.insert(newFile);
            return fileDao.findOne(newFileId);
        } catch (UnsupportedEncodingException ex) {
            throw new RuntimeException(ex);
        }
    }


    @Nonnull
    private Path writeFileToDisk(MultipartFile file, Path parentDir) {
        try (InputStream is = file.getInputStream()) {
            if (!Files.exists(parentDir)) {
                Files.createDirectories(parentDir);
            }
            String uniqueFileName = Utils.createSHAHashString(is);
            Path newPath = parentDir.resolve(uniqueFileName);
            if (!Files.exists(newPath)) {
                file.transferTo(newPath.toFile());
            }
            return pathToFiles.relativize(newPath);
        } catch (IOException ex) {
            throw new InternalException(msgSr.getMessage(MATERIAL_CREATING_IO, null, null), ex);
        }
    }

    private Path writeFileToDisk(Path file, Path parentDir) {
        try (InputStream is = Files.newInputStream(file)) {
            if (!Files.exists(parentDir)) {
                Files.createDirectories(parentDir);
            }
            String uniqueFileName = Utils.createSHAHashString(is);
            Path newPath = parentDir.resolve(uniqueFileName);
            if (!Files.exists(newPath)) {
                Files.copy(file, newPath);
            }
            return pathToFiles.relativize(newPath);
        } catch (IOException ex) {
            throw new InternalException(msgSr.getMessage(MATERIAL_CREATING_IO, null, null), ex);
        }
    }


    private void checkIfThisFileUsedAsAnotherFileType(String relativePath, String newType) {
        List<FileEntity> entitiesUsingFile = fileDao.findByRelativePath(relativePath);
        entitiesUsingFile.forEach(entityUsingFile -> {
            if (!entityUsingFile.getFileType().equals(newType)) {
                throw new RuntimeException(msgSr.getMessage(FILE_EXSIST_WITH_DIFFERENT_ROLE, null, null));
            }
        });
    }

    @Nullable
    public Path resolvePathToFile(FileEntity file, @Nullable UserEntity currentUser) {
        if (file.getPathToFile() == null) {
            return null;
        }
        Path pathToFile = pathToFiles.resolve(file.getPathToFile());
        if (!Files.isRegularFile(pathToFile)) {
            return null;
        }
        return pathToFile;
    }

    @Nullable
    public Path resolvePathToFile(String file, @Nullable UserEntity currentUser) {
        Path pathToFile = pathToFiles.resolve(file);
        if (!Files.isRegularFile(pathToFile)) {
            return null;
        }
        return pathToFile;
    }

    public List<FileEntity> findByOwnerId(Integer ownerId, @Nullable UserEntity userEntity) {
        return fileDao.findByOwnerId(ownerId);
    }

    public List<FileEntity> findConstructionElementFiles(Integer constructionElementId, FileType fileType,
                                                         @Nullable UserEntity currentUser) {
        return fileDao.getConstructionElementFiles(constructionElementId, fileType);
    }

    public List<FileEntity> findConstructionElementFiles(Integer constructionElementId, @Nullable UserEntity currentUser) {
        return fileDao.getConstructionElementFiles(constructionElementId);
    }


    public PageImpl<FileEntity> findConstructionElementFilesPage(Integer constructionElementId, @Nullable FileType fileType,
                                                                 Map<String, Object> filters, Pageable pageable,
                                                                 @Nullable UserEntity currentUser) {
        return fileDao.getConstructionElementFiles(constructionElementId, fileType, filters, pageable);
    }

    public List<FileEntity> findOptimizationFiles(Integer optimizationId, @Nullable Integer fileTypeId,
                                                  @Nullable UserEntity userEntity) {
        return fileDao.findOptimizationFiles(optimizationId, fileTypeId);
    }

    public PageImpl<FileEntity> findOptimizationFilesPage(int id, Map<String, Object> filters, Pageable page,
                                                          @Nullable UserEntity currentUser) {
        return fileDao.findOptimizationFilesPage(id, filters, page);
    }

    public void createFile(FileEntity file){
        fileDao.insert(file);
    }
}
