package ru.crp.lvms.managers;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.stereotype.Service;
import org.springframework.web.multipart.MultipartFile;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.managers.misc.BaseManagerSuggest;
import ru.crp.lvms.model.AccessRights;
import ru.crp.lvms.model.dao.BaseSpaceDao;
import ru.crp.lvms.model.dao.PossibleDefectsDao;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.PossibleDefectsEntity;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.misc.FileType;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.List;
import java.util.Map;
import java.util.Optional;

@Service
public class PossibleDefectsManager extends BaseManagerSuggest<PossibleDefectsEntity> {

    @Autowired
    private PossibleDefectsDao possibleDefectsDao;

    @Autowired
    private TechnologyManager technologyManager;

    @Autowired
    private FileManager fileManager;

    @Override
    protected BaseSpaceDao<PossibleDefectsEntity> aliveDao() {
        return possibleDefectsDao;
    }


    public List<PossibleDefectsEntity> getTechnologyEquipment(int technologyId, @Nullable UserEntity currentUser) {
        technologyManager.checkEntityExistsAndAccesible(technologyId, AccessRights.view, currentUser);
        return possibleDefectsDao.getTechnologyEquipment(technologyId);
    }


    public PageImpl<PossibleDefectsEntity> getTechnologyDefectsPage(Integer technologyId, Map<String, Object> filters,
                                                                    Pageable pageable, UserEntity currentUser) {
        technologyManager.checkEntityExistsAndAccesible(technologyId, AccessRights.view, currentUser);
        filters.put("technologyId", technologyId);
        return possibleDefectsDao.find(filters, pageable);
    }

    public PossibleDefectsEntity createAndRead(PossibleDefectsEntity newDefect, @Nullable MultipartFile file,
                                               @Nullable UserEntity currentUser) {
        PossibleDefectsEntity existed = possibleDefectsDao
                .findByTechnologyIdAndName(newDefect.getTechnologyId(), newDefect.getName());
        if (existed != null) {
            throw new BadRequestException("Дефект с таким именем уже добавлен к технологии");
        }
        PossibleDefectsEntity createdDefect = super.createAndRead(newDefect, currentUser);
        if (file != null) {
            FileEntity newFile = fileManager.createFile(file, createdDefect.getId(), FileType.DefectActFile, currentUser);
            createdDefect.setDefectFileId(newFile.getId());
            possibleDefectsDao.update(createdDefect);
        }
        return createdDefect;
    }


    @Nonnull
    @Override
    public PossibleDefectsEntity updateAndRead(PossibleDefectsEntity updated, @Nullable UserEntity currentUser) {
        technologyManager.checkEntityExistsAndAccesible(updated.getTechnologyId(), AccessRights.update, currentUser);
        PossibleDefectsEntity existed = possibleDefectsDao
                .findByTechnologyIdAndName(updated.getTechnologyId(), updated.getName());
        if (existed != null && existed.getId() != updated.getId()) {
            throw new BadRequestException("Дефект с таким именем уже добавлен к технологии");
        }
        return super.updateAndRead(updated, currentUser);
    }

    @Override
    public void delete(PossibleDefectsEntity entity, @Nullable UserEntity currentUser) {
        checkDelete(entity, currentUser);

        Optional<Integer> defectFileId = Optional.ofNullable(entity.getDefectFileId());
        entity.setDefectFileId(null);

        update(entity, currentUser);

        defectFileId.ifPresent(id -> fileManager.deleteFile(id, currentUser));

        aliveDao().delete(entity.getId());
    }
}
