package ru.crp.lvms.managers;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.MessageSource;
import org.springframework.dao.CannotSerializeTransactionException;
import org.springframework.retry.annotation.Backoff;
import org.springframework.retry.annotation.Retryable;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Isolation;
import org.springframework.transaction.annotation.Transactional;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.EntityDeleteException;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.managers.construction.elements.ConstructionElementManager;
import ru.crp.lvms.model.AccessRights;
import ru.crp.lvms.model.dao.TreeDao;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.misc.ItemType;
import ru.crp.lvms.model.entities.misc.TreeType;
import ru.crp.lvms.model.entities.tree.OptimizationsTreeNode;
import ru.crp.lvms.model.entities.tree.TreeNode;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;

import javax.annotation.Nullable;
import java.util.EnumSet;
import java.util.List;
import java.util.Set;

import static ru.crp.lvms.utils.MessageUtils.*;

@Service
@Transactional(isolation = Isolation.SERIALIZABLE)
// Set retry on transaction isolation fail, use random exponential back off
@Retryable(include = CannotSerializeTransactionException.class, maxAttempts = 5,
        backoff = @Backoff(delay = 100, multiplier = 2, random = true))
public class TreeManager {
    @Autowired
    private TreeDao treeDao;

    @Autowired
    private OptimizationManager optimizationManager;

    @Autowired
    private MaterialManager materialManager;

    @Autowired
    private CompositeMaterialManager compositeMaterialManager;

    @Autowired
    private TechnologyManager technologyManager;

    @Autowired
    private MessageSource msgSrc;

    @Autowired
    private ConstructionElementManager constructionElementManager;

    public TreeNode findOne(TreeType treeType, int nodeId, @Nullable UserEntity user) {
        return treeDao.findOne(treeType, nodeId);
    }

    public List<? extends TreeNode> getRootNodes(TreeType treeType, @Nullable UserEntity currentUser) {
        return treeDao.getRootNodes(treeType);
    }


    public TreeNode updateNodeName(TreeType treeType, TreeNode nodeToUpdate, @Nullable UserEntity currentUser) {
        TreeNode oldNode = checkNodeExists(treeType, nodeToUpdate.getId());
        oldNode.setName(nodeToUpdate.getName());
        if (oldNode.getItemType() == ItemType.ConstructionElementFolder) {
            throw new BadRequestException("Названия директорий с типами файлов являются неизменяемыми");
        }
        return treeDao.updateAndRead(treeType, oldNode);
    }


    public TreeNode checkNodeExists(TreeType treeType, int nodeId) {
        TreeNode node = treeDao.findOne(treeType, nodeId);
        if (node == null) {
            throw new NotFoundException(msgSrc.getMessage(TREE_NODE_NOT_FOUND, new Object[]{nodeId}, null, null));
        }
        return node;
    }


    public TreeNode deleteNode(TreeType treeType, int nodeId, UserEntity currentUser) {
        TreeNode nodeToDelete = checkNodeExists(treeType, nodeId);
        switch (nodeToDelete.getItemType()) {
            case File:
                if (treeType == TreeType.materials || treeType == TreeType.compositeMaterials) {
                    materialManager.deleteMaterialFile(null, nodeId, currentUser);
                } else if (treeType == TreeType.constructionElements) {
                    constructionElementManager.deleteFile(nodeId, currentUser);
                } else {
                    throw new BadRequestException("Невозможно удалить файл из дерева этого типа");
                }
                break;
            case Material:
                materialManager.delete(nodeId, currentUser);
                break;
            case CompositeMaterial:
                compositeMaterialManager.delete(nodeId, currentUser);
                break;
            case Technology:
                technologyManager.delete(nodeId, currentUser);
            case Folder:
                deleteFolder(nodeToDelete, currentUser);
                break;
            case OptimizationFolder:
                optimizationManager.deleteOptimizationFolder(nodeId, currentUser);
                break;
            case ConstructionElement:
                constructionElementManager.delete(nodeId, currentUser);
                break;
            case ConstructionElementFolder:
                throw new BadRequestException("Запрещено удалять автоматически созданную папку");
            default:
                throw new IllegalArgumentException(msgSrc.getMessage(TREE_UNEXPECTED_NODE,
                        new Object[]{nodeToDelete.getItemType()}, null));
        }
        return nodeToDelete;
    }


    private void deleteFolder(TreeNode nodeToDelete, UserEntity currentUser) {
        if (treeDao.hasChildren(nodeToDelete.getId())) {
            throw new EntityDeleteException(msgSrc.getMessage(TREE_NODE_WITH_CHILDREN_DELETE, null, null), true);
        }
        treeDao.deleteFolder(nodeToDelete.getId());
    }


    public List<? extends TreeNode> getSubNodes(@Nullable Integer rootId, TreeType treeType, UserEntity currentUser) {
        if (rootId == null) {
            return treeDao.getAllNodesByTreeType(treeType);
        }
        checkNodeExists(treeType, rootId);
        return treeDao.getSubTree(rootId, treeType);
    }


    public TreeNode createFolder(TreeType treeType, TreeNode newNode, UserEntity currentUser) {
        checkAllowedTypes(newNode, EnumSet.of(ItemType.Folder));
        if (treeType == TreeType.technologies) {
            throw new BadRequestException("Запрещено создавать директории в дереве технологий");
        }
        checkParent(treeType, newNode);
        return treeDao.insertAndReadFolder(treeType, newNode, currentUser.getId(), false);
    }


    public TreeNode createOptimizationTreeNode(OptimizationsTreeNode newNode, UserEntity currentUser) {
        checkAllowedTypes(newNode, EnumSet.of(ItemType.Folder, ItemType.OptimizationFolder));
        checkParent(TreeType.optimizations, newNode);
        if (newNode.isFolder()) {
            return treeDao.insertAndReadFolder(TreeType.constructionElements, newNode, currentUser.getId(), false);
        }
        if (newNode.getPid() == null && !newNode.isFolder()) {
            throw new BadRequestException(msgSrc.getMessage(TREE_ROOT_OPTIMIZATION_FOLDER, null, null));
        }
        checkBaseOptimizationId(newNode);
        if (newNode.getItemType() == ItemType.OptimizationFolder && !treeDao.findByName(newNode.getItemType(), newNode.getName()).isEmpty()) {
            throw new BadRequestException("Имя должно быть уникально");
        }
        return treeDao.insertAndReadOptimizationFolder(newNode, currentUser.getId());
    }


    private void checkBaseOptimizationId(OptimizationsTreeNode newNode) {
        // if folder is not first child of CE then base optimization id must be used
        if (newNode.getPid() == null || newNode.getConstructionElementId() == null) {
            return;
        }
        if (!newNode.getPid().equals(newNode.getConstructionElementId()) && newNode.getBaseOptimizationId() == null) {
            throw new BadRequestException("Базовая оптимизация должна быть задана");
        }
        if (newNode.getBaseOptimizationId() != null) {
            optimizationManager.checkEntityExistsAndAccesible(newNode.getBaseOptimizationId(), AccessRights.view, null);
        }
    }


    private void checkAllowedTypes(TreeNode newNode, Set<ItemType> allowedTypes) {
        if (!allowedTypes.contains(newNode.getItemType())) {
            throw new BadRequestException(msgSrc.getMessage(TREE_WRONG_NEW_NODE_TYPE, null, null));
        }
    }


    private TreeNode checkParent(TreeType treeType, TreeNode newNode) {
        if (newNode.getPid() == null) {
            return null;
        }
        TreeNode parent = checkNodeExists(treeType, newNode.getPid());
        if (!newNode.getItemType().canBeChildOf(parent.getItemType())) {
            throw new BadRequestException(msgSrc.getMessage(TREE_WRONG_NODE_PARENT, null, null));
        }
        return parent;
    }

    public List<TreeNode> findByPathNames(ItemType itemType, String pathNames) {
        return treeDao.findByPathNames(itemType, pathNames);
    }
}
