package ru.crp.lvms.managers.construction.elements;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.stereotype.Service;
import org.springframework.web.multipart.MultipartFile;
import ru.crp.cmlb.model.SuggestRequest;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.InternalException;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.managers.TreeManager;
import ru.crp.lvms.managers.image.ImageHandler;
import ru.crp.lvms.managers.misc.BaseManagerSuggest;
import ru.crp.lvms.model.AccessRights;
import ru.crp.lvms.model.dao.BaseSpaceDao;
import ru.crp.lvms.model.dao.FileDao;
import ru.crp.lvms.model.dao.TreeDao;
import ru.crp.lvms.model.dao.construction.elements.ConstructionElementDao;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.construction.elements.ConstructionElementEntity;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.model.entities.misc.ItemType;
import ru.crp.lvms.model.entities.misc.TreeType;
import ru.crp.lvms.model.entities.tree.TreeNode;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.utils.ValidationHelper;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.nio.charset.StandardCharsets;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static java.util.Arrays.asList;
import static ru.crp.lvms.utils.MessageUtils.CE_USED_IN_OPTIMIZATIONS;
import static ru.crp.lvms.utils.MessageUtils.FILE_USED_IN_OPTIMIZATIONS;

@Service
public class ConstructionElementManager extends BaseManagerSuggest<ConstructionElementEntity>
        implements ImageHandler<ConstructionElementEntity> {
    @Autowired
    private ConstructionElementDao constructionElementDao;

    @Autowired
    private FileManager fileManager;

    @Autowired
    private TreeDao treeDao;

    @Autowired
    private TreeManager treeManager;

    @Autowired
    private ValidationHelper validationHelper;

    @Autowired
    private OptimizationManager optimizationManager;

    @Autowired
    private FileDao fileDao;

    @Override
    public BaseSpaceDao<ConstructionElementEntity> aliveDao() {
        return constructionElementDao;
    }


    @Nonnull
    @Override
    public ConstructionElementEntity createAndRead(ConstructionElementEntity entity, @Nullable UserEntity currentUser) {
        ConstructionElementEntity newEntity = super.createAndRead(entity, currentUser);
        createDefaultFolders(newEntity, currentUser);
        return newEntity;
    }


    private void createDefaultFolders(ConstructionElementEntity newEntity, UserEntity currentUser) {
        asList("CAD", "CAE", "CAO")
                .forEach(folderName -> createFolder(newEntity.getId(), folderName, currentUser));
    }


    private void createFolder(int pid, String folderName, UserEntity currentUser) {
        TreeNode folder = new TreeNode();
        folder.setItemType(ItemType.Folder);
        folder.setPid(pid);
        folder.setName(folderName);
        treeDao.insertAndReadFolder(TreeType.constructionElements, folder, currentUser.getId(), true);
    }


    public ConstructionElementEntity createAndRead(ConstructionElementEntity in, @Nullable MultipartFile image,
                                                   @Nullable MultipartFile cad, UserEntity user) {
        in.setOwnerId(user.getId());
        ConstructionElementEntity created = createAndRead(in, user);
        if (image != null) {
            FileEntity imageFile = fileManager.createFile(image, created.getId(), FileType.ImageFile, user);
            created.setImageFileId(imageFile.getId());
        }
        if (cad != null) {
            FileEntity initialCadModel = createNewFile(created.getId(), cad, FileType.CADFile, user);
            created.setInitialCadModelId(initialCadModel.getId());
        }
        if (image != null || cad != null) {
            return updateAndRead(created, user);
        }
        return created;
    }


    public FileEntity createNewFile(int ceId, MultipartFile fileContent, FileType type, UserEntity currentUser) {
        ConstructionElementEntity ce = checkEntityExistsAndAccesible(ceId, AccessRights.update, currentUser);
        TreeNode filesFolder = treeDao.getConstructionElementFileFolder(ceId, type);
        int pid = ceId;
        if (filesFolder != null) {
            pid = filesFolder.getId();
        }
        String parentPathNames = filesFolder != null ? filesFolder.getPathNames() : ce.getPathNames();
        String modelName = null;
        try {
            modelName = URLDecoder.decode(fileContent.getOriginalFilename(), StandardCharsets.UTF_8.name());
        } catch (UnsupportedEncodingException e) {
            throw new InternalException("Unsupported encoding ", e);
        }
        checkNewFilePathNames(ceId, parentPathNames + "/" + modelName);
        FileEntity newFile = fileManager.createFile(fileContent, pid, type, currentUser);
        constructionElementDao.addLinkToFile(ceId, newFile.getId());
        return newFile;
    }

    private void checkNewFilePathNames(Integer ceId, String pathNames) {
        PageRequest pageRequest = new PageRequest(0, 1);
        Map<String, Object> filters = new HashMap<>();
        filters.put("pathNames", pathNames);
        PageImpl<FileEntity> conflicted = fileManager.findConstructionElementFilesPage(ceId, null, filters, pageRequest, null);
        if (!conflicted.getContent().isEmpty()) {
            throw new BadRequestException("Элемент конструкции уже содержит файл с таким именем");
        }
    }


    public void updateFile(int constructionElementId, FileType fileType, MultipartFile newFileContent,
                           UserEntity currentUser) {
        ConstructionElementEntity constructionElementEntity = checkEntityExistsAndAccesible(constructionElementId, AccessRights.update, currentUser);
        Integer fileId = getFileByType(constructionElementEntity, fileType);
        if (fileId == null) {
            if (fileType == FileType.ImageFile) {
                FileEntity imageFile = fileManager.createFile(newFileContent, constructionElementEntity.getId(), FileType.ImageFile, currentUser);
                constructionElementEntity.setImageFileId(imageFile.getId());
            } else {
                FileEntity initialCadModel = createNewFile(constructionElementEntity.getId(), newFileContent, FileType.CADFile, currentUser);
                constructionElementEntity.setInitialCadModelId(initialCadModel.getId());
            }
            aliveDao().update(constructionElementEntity);
        } else {
            fileManager.updateFileContent(fileId, newFileContent, currentUser);
        }
    }


    public void deleteInitialCadModel(int constructionElementId, int fileId, UserEntity currentUser) {
        ConstructionElementEntity constructionElement = checkEntityExistsAndAccesible(constructionElementId, AccessRights.update, currentUser);
        constructionElement.setInitialCadModelId(null);
        constructionElementDao.update(constructionElement);
        deleteFileNoCheck(constructionElementId, fileId, currentUser);
    }


    public void deleteFile(int constructionElementId, int fileId, UserEntity currentUser) {
        checkEntityExistsAndAccesible(constructionElementId, AccessRights.update, currentUser);
        deleteFileNoCheck(constructionElementId, fileId, currentUser);
    }


    private void deleteFileNoCheck(int constructionElementId, int fileId, UserEntity currentUser) {
        validationHelper.checkDependencies(fileId, FILE_USED_IN_OPTIMIZATIONS,
                optimizationManager::getOptimizationsThatUseFile, currentUser);
        constructionElementDao.deleteLinkToFile(constructionElementId, fileId);
        fileManager.deleteFile(fileId, currentUser);
    }

    public void deleteFile(int fileId, UserEntity currentUser) {
        TreeNode fileToDelete = treeManager.findOne(TreeType.constructionElements, fileId, currentUser);
        TreeNode parent = treeManager.findOne(TreeType.constructionElements, fileToDelete.getPid(), currentUser);
        while (parent.getItemType() != ItemType.ConstructionElement) {
            parent = treeManager.findOne(TreeType.constructionElements, parent.getPid(), currentUser);
        }
        deleteFile(parent.getId(), fileId, currentUser);
    }


    private Integer getFileByType(ConstructionElementEntity entity, FileType fileType) {
        switch (fileType) {
            case CADFile:
                return entity.getInitialCadModelId();
            case ImageFile:
                return entity.getImageFileId();
            default:
                throw new IllegalArgumentException("Unexpected file type: " + fileType);
        }
    }


    public List<FileEntity> findConstructionElementFiles(Integer constructionElementId, FileType fileType,
                                                         UserEntity currentUser) {
        checkEntityExistsAndAccesible(constructionElementId, AccessRights.view, currentUser);
        return fileManager.findConstructionElementFiles(constructionElementId, fileType, currentUser);
    }


    public PageImpl<FileEntity> findConstructionElementFilesPage(Integer constructionElementId,
                                                                 FileType fileType, Map<String, Object> filters,
                                                                 Pageable pageable, UserEntity currentUser) {
        return fileManager.findConstructionElementFilesPage(constructionElementId, fileType,
                filters, pageable, currentUser);
    }

    public void validate(ConstructionElementEntity newCE, UserEntity currentUser) {
        TreeNode parent = checkParent(newCE.getPid());
        List<ConstructionElementEntity> conflicted = constructionElementDao.findByColumnName("name", newCE.getName());
//        if (!conflicted.isEmpty()) {
//            throw new BadRequestException("Имя должно быть уникально");
//        }
        checkPathNames(parent, newCE.getName());
    }

    private void checkPathNames(TreeNode parent, String name) {
        String pathNames = parent.getPathNames() + "/" + name;
        List<TreeNode> existedNodes = treeDao.findByPathNames(ItemType.ConstructionElement, pathNames);
        if (!existedNodes.isEmpty()) {
            throw new BadRequestException("В папке уже есть элемент с таким именем");
        }
    }

    private TreeNode checkParent(Integer pid) {
        TreeNode one = treeManager.findOne(TreeType.constructionElements, pid, null);
        if (one == null || !one.isFolder()) {
            throw new BadRequestException("Неверный путь");
        }
        return one;
    }


    @Override
    public void checkDelete(ConstructionElementEntity entity, @Nullable UserEntity currentUser) {
        validationHelper.checkDependencies(entity.getId(), CE_USED_IN_OPTIMIZATIONS,
                optimizationManager::getOptimizationsByConstructionElementId, currentUser);
    }

    @Override
    public void delete(ConstructionElementEntity entity, @Nullable UserEntity currentUser) {
        checkDelete(entity, currentUser);

        fileManager.findConstructionElementFiles(entity.getId(), currentUser).stream()
                .map(FileEntity::getId)
                .forEach(fileId -> deleteFileNoCheck(entity.getId(), fileId, currentUser));

        aliveDao().delete(entity.getId());
    }

    @Override
    public FileManager fileManager() {
        return fileManager;
    }


    public List<FileEntity> suggestFiles(Integer ceId, FileType fileType, SuggestRequest suggestRequest, UserEntity currentUser) {
        return fileDao.suggestConstructionElementFiles(ceId, fileType, suggestRequest, currentUser.getId());
    }
}
