package ru.crp.lvms.managers.image;

import org.springframework.dao.CannotSerializeTransactionException;
import org.springframework.retry.annotation.Backoff;
import org.springframework.retry.annotation.Retryable;
import org.springframework.transaction.annotation.Isolation;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.multipart.MultipartFile;
import ru.crp.cmlb.model.entities.base.Entity;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.model.AccessRights;
import ru.crp.lvms.model.dao.BaseSpaceDao;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.misc.FileType;

import javax.annotation.Nullable;
import java.util.stream.Stream;

@Transactional(isolation = Isolation.SERIALIZABLE)
// Set retry on transaction isolation fail, use random exponential back off
@Retryable(include = CannotSerializeTransactionException.class, maxAttempts = 5,
        backoff = @Backoff(delay = 100, multiplier = 2, random = true))
public interface ImageHandler<E extends ImageFileContainer & Entity> {
    FileManager fileManager();

    BaseSpaceDao<E> aliveDao();

    E checkEntityExistsAndAccesible(int entityId, AccessRights accessRights, UserEntity currentUser);


    default void updateImageFile(E e, @Nullable Integer newImageFileId, @Nullable UserEntity currentUser) {
        aliveDao().updateImageFile(e, newImageFileId);
    }


    default FileEntity getImageFile(Integer entityId, UserEntity currentUser) {
        E entity = checkEntityExistsAndAccesible(entityId, AccessRights.view, currentUser);
        Integer fileId = entity.getImageFileId();
        if (fileId == null) {
            return null;
        }
        return fileManager().findOne(fileId, currentUser);
    }


    default void updateImageFile(Integer entityId, MultipartFile imageFile, UserEntity currentUser) {
        E entity = checkEntityExistsAndAccesible(entityId, AccessRights.update, currentUser);
        checkIsImage(imageFile);
        Integer fileId = entity.getImageFileId();
        if (fileId == null) {
            FileEntity newFile = fileManager().createFile(imageFile, entityId, FileType.ImageFile, currentUser);
            updateImageFile(entity, newFile.getId(), currentUser);
        } else {
            fileManager().updateFileContent(fileId, imageFile, currentUser);
        }
    }


    // todo change to testing content type
    default void checkIsImage(MultipartFile imageFile) {
        String originalFilename = imageFile.getOriginalFilename().toLowerCase();
        Stream<String> supportedFileExtensions = Stream.of("jpeg", "jpg", "png", "gif");
        if (!supportedFileExtensions.anyMatch(originalFilename::endsWith)) {
            throw new BadRequestException("Недопустимый формат файла");
        }
    }


    default void deleteImageFile(Integer entityId, UserEntity currentUser) {
        E entity = checkEntityExistsAndAccesible(entityId, AccessRights.view, currentUser);
        Integer fileId = entity.getImageFileId();
        if (fileId == null) {
            return;
        }
        updateImageFile(entity, null, currentUser);
        fileManager().deleteFile(fileId, currentUser);
    }
}
