package ru.crp.lvms.managers.solvers;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import org.springframework.web.multipart.MultipartFile;
import ru.crp.cmls.calculations.solvers.config.InvalidConfigException;
import ru.crp.cmls.calculations.solvers.config.SolverConfigParser;
import ru.crp.cmls.calculations.solvers.xml.config.SolverConfig;
import ru.crp.cmls.calculations.solvers.xml.config.Type;
import ru.crp.cmls.calculations.solvers.xml.config.files.CompositeFileResolver;
import ru.crp.cmls.calculations.solvers.xml.config.placeholder.BasePlaceholdersResolver;
import ru.crp.cmls.calculations.solvers.xml.config.results.ResultsResolver;
import ru.crp.lvms.exceptions.BadRequestException;

import javax.annotation.Nullable;
import javax.annotation.PostConstruct;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.stream.Stream;

import static java.util.stream.Collectors.toList;

@Service
public class SolverConfigManager {
    private static final Logger LOGGER = LoggerFactory.getLogger(SolverConfigManager.class);

    public static final String MOCK_SOLVER_NAME = "Mock solver";
    public static final String ANSYS_FOR_SLURM_SOLVER_NAME = "Ansys for optimizations";
    public static final String LS_DYNA_FOR_SLURM_SOLVER_NAME = "LS-Dyna MPI";
    public static final String TOSCA_POSTPROC_ANIM_NAME = "Tosca Animation";
    public static final String TOSCA_POSTPROC_SMOOTH_NAME = "Tosca Smoothing";

    private String filesDirectoryString;

    private Path filesDirectory;

    private ConcurrentMap<SolverConfig, Path> availableSolvers = new ConcurrentHashMap<>();

    @Value("${solverConfigs}")
    public void setFilesDirectoryString(String filesDirectoryString) {
        this.filesDirectoryString = filesDirectoryString;
    }

    @PostConstruct
    public void init() {
        filesDirectory = Paths.get(filesDirectoryString);
        try {
            if (!Files.isDirectory(filesDirectory)) {
                Files.createDirectories(filesDirectory);
            }
            try (Stream<Path> solvers = Files.list(filesDirectory)) {
                solvers.forEach(pathToSolver -> {
                    try {
                        SolverConfig config = SolverConfigParser.deserialize(pathToSolver);
                        if (getSolverByName(config.getName()) != null) {
                            LOGGER.warn("Found duplicate solver config: " + pathToSolver);
                        } else {
                            availableSolvers.putIfAbsent(config, pathToSolver);
                        }
                    } catch (InvalidConfigException ex) {
                        LOGGER.error("Error while reading config file: " + pathToSolver, ex);
                    }
                });
            }
        } catch (IOException ex) {
            throw new InvalidConfigException("Unable to create directory for solver xml's", ex);
        }
    }


    @Nullable
    public Path getSolverFile(String solverFileName) {
        SolverConfig solver = getSolverByName(solverFileName);
        if (solver != null) {
            return availableSolvers.get(solver);
        }
        return null;
    }

    public void deleteSolver(String fileName) throws IOException {
        SolverConfig solverConfig = getSolverByName(fileName);
        Files.delete(availableSolvers.get(solverConfig));
        availableSolvers.remove(solverConfig);
    }

    // Returns true if file is newly created and false if it is just replaced
    public boolean saveFile(MultipartFile file) throws IOException, BadRequestException {
        SolverConfig newConfig = SolverConfigParser.deserializeFileIS(file.getInputStream());
        Path pathToNewFile = filesDirectory.resolve(file.getOriginalFilename());
        boolean fileIsNewlyCreated = !Files.exists(pathToNewFile);
        file.transferTo(pathToNewFile.toFile());
        if (fileIsNewlyCreated) {
            availableSolvers.putIfAbsent(newConfig, pathToNewFile);
        }
        return fileIsNewlyCreated;
    }

    public Set<SolverConfig> getAllConfigs() {
        return availableSolvers.keySet();
    }

    @Nullable
    public SolverConfig getSolverByName(String solverName) {
        return availableSolvers.keySet().stream()
                .filter(solver -> solver.getName().equals(solverName))
                .findAny().orElse(null);
    }

    public List<SolverConfig> toSolver(Set<String> requestedSolvers) {
        return availableSolvers.keySet().stream()
                .filter(solver -> requestedSolvers.contains(solver.getName()))
                .collect(toList());

    }

    @Nullable
    public ResultsResolver getResultsResolver(String solverName, Type type, Path inputDir, Path resultsDir) {
        SolverConfig solverConfig = getSolverByName(solverName);
        if (solverConfig == null) {
            return null;
        }
        BasePlaceholdersResolver placeholderResolver = new BasePlaceholdersResolver(inputDir, resultsDir);
        CompositeFileResolver filesResolver = new CompositeFileResolver(placeholderResolver);
        return new ResultsResolver(solverConfig, type, filesResolver);
    }

    public List<ResultsResolver> getResultsResolver(Type type, Path inputDir, Path resultsDir) {
        return availableSolvers.keySet().stream()
                .filter(solver -> solver.canHandleType(type))
                .map(config -> getResultsResolver(config.getName(), type, inputDir, resultsDir))
                .collect(toList());
    }


    @Nullable
    public SolverConfig getDefaultPostprocessor() {
        return availableSolvers.keySet().stream()
                .filter(solver -> solver.canHandleType(Type.postprocessing) && !MOCK_SOLVER_NAME.equals(solver.getName()))
                .findFirst().orElse(null);
    }
}
