package ru.crp.lvms.managers.solvers;

import com.fasterxml.jackson.databind.ObjectMapper;
import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.impl.client.HttpClientBuilder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.core.io.FileSystemResource;
import org.springframework.stereotype.Service;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;
import org.springframework.web.client.*;
import ru.crp.cmlb.model.entities.FileData;
import ru.crp.cmls.api.NodeStatus;
import ru.crp.cmls.api.ServiceTask;
import ru.crp.cmls.calculations.solvers.dto.SolverPresenceRequest;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.exceptions.solving.*;
import ru.crp.lvms.model.entities.tasks.TaskStatusResponse;
import ru.crplab.downloader.collector.FSSubtreeAggregationInfo;
import ru.crplab.downloader.transfer.FileHttpMessageConverter;
import ru.crplab.downloader.transfer.FilesDownloader;
import ru.crplab.downloader.transfer.RestTemplateTransfer;

import javax.annotation.PostConstruct;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

@Service
public class SolverConnectionManager {
    private static final Logger LOGGER = LoggerFactory.getLogger(SolverConnectionManager.class);

    public static final String TASKS_PATH = "rest/tasks";
    private static final String FILE_INTEGRITY_MESSAGE = "FileIntegrityException";
    private static final int MAX_ATTEMPTS = 3;
    private static final int SLEEP_TIME = 1000 * 2;

    private RestTemplate nodeStatusRestTemplate;

    @Autowired
    @Qualifier("solverConnectionRestTemplate")
    private RestTemplate restTemplate;

    @Autowired
    private ObjectMapper jacksonMapper;


    @PostConstruct
    public void addFileMessageConverter() {
        restTemplate.getMessageConverters().add(new FileHttpMessageConverter());
    }

    public NodeStatus getNodeStatus(String hostUrl, List<SolverPresenceRequest> solversToCheck) {
        String url = hostUrl + "rest/status";
        return nodeStatusRestTemplate.postForObject(url, solversToCheck, NodeStatus.class);
    }

    @Autowired
    @Qualifier("nodeStatusRestTemplate")
    public void setNodeStatusRestTemplate(RestTemplate nodeStatusRestTemplate) {
        this.nodeStatusRestTemplate = nodeStatusRestTemplate;
    }

    public void sendPrepareTaskRequest(String hostUrl, ServiceTask task) {
        String url = hostUrl + TASKS_PATH;
        try {
            restTemplate.postForEntity(url, task, String.class);
        } catch (HttpServerErrorException | HttpClientErrorException ex) {
            handleTaskPreparingError(ex);
        } catch (ResourceAccessException ex) {
            // throwing this exception because it is recoverable error
            throw new NodeNotResponseException(ex.getMessage(), ex);
        }
    }

    private void handleTaskPreparingError(HttpStatusCodeException ex) throws SolverException {
        switch (ex.getStatusCode()) {
            case TOO_MANY_REQUESTS:
                throw new BusyNodeException(ex.getResponseBodyAsString(), ex);
            case CONFLICT:
                throw new DuplicateTaskException(ex.getResponseBodyAsString(), ex);
            case NOT_FOUND:
                throw new SolverTypeNotFoundException(ex.getResponseBodyAsString(), ex);
            default:
                throw new TaskStartingException("Error while sending prepare task request to node", ex);
        }
    }

    public void sendTaskFiles(String hostUrl, int taskId, List<FileData> taskFiles) {
        String url = hostUrl + TASKS_PATH + "/" + taskId + "/files";
        taskFiles.forEach(taskFile -> sendFile(url, taskId, taskFile));
    }

    private void sendFile(String url, int taskId, FileData taskFile) {
        int attempt = 0;
        while (attempt <= MAX_ATTEMPTS) {
            try {
                MultiValueMap<String, Object> request = new LinkedMultiValueMap<>();
                request.add("file", new FileSystemResource(taskFile.getPathToFile().toFile()));
                request.add("relativePath", taskFile.getRelativePath());
                // using getBytes is important to save encoding
                request.add("name", taskFile.getName().getBytes(StandardCharsets.UTF_8));
                restTemplate.postForEntity(url, request, String.class);
                return;
            } catch (HttpClientErrorException ex) {
                handleSendFileError(ex);
            } catch (HttpServerErrorException ex) {
                String responseStr = ex.getResponseBodyAsString();
                if (!FILE_INTEGRITY_MESSAGE.equals(responseStr)) {
                    handleSendFileError(ex);
                } else {
                    LOGGER.error("file integrity exception for task " + taskId + ". File name = " + taskFile.getName(), ex);
                    sleepBeforeRetry(attempt, taskId);
                    attempt++;
                }
            } catch (ResourceAccessException ex) {
                LOGGER.error("connection exception for task " + taskId + ". File name = " + taskFile.getName(), ex);
                sleepBeforeRetry(attempt, taskId);
                attempt++;
            }
        }
    }

    private void handleSendFileError(HttpStatusCodeException ex) throws SolverException {
        switch (ex.getStatusCode()) {
            case TOO_MANY_REQUESTS:
                throw new BusyNodeException(ex.getResponseBodyAsString(), ex);
            case NOT_FOUND:
                throw new NotFoundException(ex.getResponseBodyAsString());
            default:
                throw new TaskStartingException("Error while sending file to node", ex);
        }
    }

    private void sleepBeforeRetry(int attempt, int taskId) throws TaskStartingException {
        if (attempt >= MAX_ATTEMPTS) {
            throw new TaskStartingException("Connection error while sending file for task " + taskId);
        }
        try {
            // sleeping 1,3,5 minutes before retry
            Thread.sleep(SLEEP_TIME + 2 * attempt * SLEEP_TIME);
        } catch (InterruptedException ex) {
            LOGGER.error("Interrupted while waiting connection problem fixing for task " + taskId, ex);
            throw new TaskStartingException("Connection error while sending file for task " + taskId, ex);
        }
    }

    public void sendStartTaskRequest(String hostUrl, int taskId) {
        String url = hostUrl + TASKS_PATH + "/" + taskId + "/start";
        try {
            restTemplate.postForEntity(url, null, String.class);
        } catch (HttpServerErrorException | HttpClientErrorException ex) {
            handleStartTaskError(ex);
        } catch (ResourceAccessException ex) {
            LOGGER.error("Connection exception while starting task " + taskId);
            throw new TaskStartingException("Connection error while starting task " + taskId, ex);
        }
    }

    private void handleStartTaskError(HttpStatusCodeException ex) throws SolverException {
        switch (ex.getStatusCode()) {
            case TOO_MANY_REQUESTS:
                throw new BusyNodeException(ex.getResponseBodyAsString(), ex);
            case NOT_FOUND:
                throw new NotFoundException(ex.getResponseBodyAsString(), ex);
            default:
                throw new TaskStartingException("Error while sending start task request to node.", ex);
        }
    }

    public TaskStatusResponse sendCheckTaskRequest(String hostUrl, int taskId) {
        String url = hostUrl + TASKS_PATH + "/" + taskId;
        return restTemplate.getForEntity(url, TaskStatusResponse.class).getBody();
    }

    public FilesDownloader createFileDownloader(String hostUrl, int taskId) {
        String url = hostUrl + TASKS_PATH + "/" + taskId + "/resultFiles";
        return new RestTemplateTransfer(url, restTemplate);
    }

    public FilesDownloader createResultFileUpdater(String hostUrl, int taskId) {
        String url = hostUrl + TASKS_PATH + "/" + taskId + "/currentResultFiles";
        return new RestTemplateTransfer(url, restTemplate);
    }

    public FSSubtreeAggregationInfo downloadResultsInfo(String hostUrl, int taskId) {
        String url = hostUrl + TASKS_PATH + "/" + taskId + "/resultsInfo";
        return restTemplate.getForObject(url, FSSubtreeAggregationInfo.class);
    }

    public Map<String, Object> sendStopTaskRequest(String hostUrl, int taskId) throws IOException {
        String url = hostUrl + TASKS_PATH + "/" + taskId + "/stop";
        HttpGet httpGet = new HttpGet(url);
        try {
            HttpClient httpClient = HttpClientBuilder.create().build();
            HttpResponse response = httpClient.execute(httpGet);
            HttpEntity responseEntity = response.getEntity();

            if (responseEntity != null) {
                ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
                responseEntity.writeTo(byteArrayOutputStream);
                if (response.getStatusLine().getStatusCode() == 200) {
                    return (Map<String, Object>) jacksonMapper.readValue(byteArrayOutputStream.toString("UTF-8"), Map.class);
                } else {
                    Map<String, Object> result = new HashMap<>();
                    result.put("status", 1);
                    result.put("message", byteArrayOutputStream.toString("UTF-8"));
                    return result;
                }
            } else {
                return null;
            }
        } finally {
            httpGet.reset();
        }
    }
}
