package ru.crp.lvms.managers.tasks;

import com.fasterxml.jackson.databind.ObjectMapper;
import org.apache.commons.io.FilenameUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.scheduling.concurrent.ThreadPoolTaskExecutor;
import org.springframework.stereotype.Service;
import ru.crp.cmls.calculations.solvers.xml.config.ResultType;
import ru.crp.cmls.calculations.solvers.xml.config.SolverConfig;
import ru.crp.cmls.calculations.solvers.xml.config.Type;
import ru.crp.cmls.calculations.solvers.xml.config.files.CompositeFileResolver;
import ru.crp.cmls.calculations.solvers.xml.config.placeholder.BasePlaceholdersResolver;
import ru.crp.cmls.calculations.solvers.xml.config.results.ResultsResolver;
import ru.crp.lvms.exceptions.solving.ResultProcessingException;
import ru.crp.lvms.managers.ResultsManager;
import ru.crp.lvms.managers.SimulationManager;
import ru.crp.lvms.managers.effective.characteristics.ECManager;
import ru.crp.lvms.managers.solvers.SolverConfigManager;
import ru.crp.lvms.managers.tasks.resultsHandlers.*;
import ru.crp.lvms.model.entities.ResultEntity;
import ru.crp.lvms.model.entities.misc.CreatedBy;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationBaseSetting;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationConstraint;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationTargetFunc;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationVaryingParam;
import ru.crp.lvms.optimizations.mf.enums.MFBaseSettingType;
import ru.crp.lvms.optimizations.mf.enums.MFSettingsType;
import ru.crp.lvms.optimizations.mf.managers.MFOptimizationSettingsManager;
import ru.crp.lvms.optimizations.mf.results.ModeFrontierParamsHandler;
import ru.crp.lvms.optimizations.optimus.entities.OptimusOptimizationBaseSetting;
import ru.crp.lvms.optimizations.optimus.entities.OptimusOptimizationConstraint;
import ru.crp.lvms.optimizations.optimus.entities.OptimusOptimizationTargetFunc;
import ru.crp.lvms.optimizations.optimus.entities.OptimusOptimizationVaryingParam;
import ru.crp.lvms.optimizations.optimus.enums.OptimusBaseSettingType;
import ru.crp.lvms.optimizations.optimus.enums.OptimusSettingsType;
import ru.crp.lvms.optimizations.optimus.managers.OptimusOptimizationSettingsManager;
import ru.crp.lvms.optimizations.optimus.results.OptimusParamsHandler;
import ru.crp.lvms.optimizations.optistruct.managers.OptistructOptimizationSettingsManager;
import ru.crp.lvms.optimizations.optistruct.results.OptistructParamsHandler;
import ru.crp.lvms.optimizations.tosca.dao.ToscaTargetFunctionParametersDao;
import ru.crp.lvms.optimizations.tosca.managers.ToscaOptimizationSettingsManager;
import ru.crp.lvms.optimizations.tosca.results.ToscaRawDataHandler;
import ru.crp.lvms.utils.GifSequenceWriter;
import ru.crp.lvms.utils.Mapper;
import ru.crp.lvms.utils.URLs;

import javax.annotation.Nullable;
import javax.annotation.PostConstruct;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static java.util.Collections.singletonList;
import static java.util.stream.Collectors.toList;

@Service
public class TaskResultsExtractor {
    public static final Logger LOGGER = LoggerFactory.getLogger(TaskResultsExtractor.class);

    private static final Pattern GIF_IMAGE_FILE_PATTERN = Pattern.compile("file[0-9]{3}");


    @Autowired
    private SimulationManager simulationManager;

    @Autowired
    private ECManager ecManager;

    @Autowired
    private OptimizationManager optimizationManager;

    @Autowired
    private ResultsManager resultsManager;

    @Autowired
    private ResultsFilesManager resultsFilesManager;

    @Autowired
    private ObjectMapper jacksonMapper;

    @Autowired
    private MFOptimizationSettingsManager MFOptimizationSettingsManager;

    @Autowired
    private OptimusOptimizationSettingsManager optimusOptimizationSettingsManager;

    @Autowired
    private ToscaOptimizationSettingsManager toscaOptimizationSettingsManager;

    @Autowired
    private ToscaTargetFunctionParametersDao toscaTargetFunctionParametersDao;

    @Autowired
    private OptistructOptimizationSettingsManager optistructOptimizationSettingsManager;

    @Autowired
    private SolverConfigManager solverConfigManager;

    @Autowired
    private GifSequenceWriter gifSequenceWriter;

    @Autowired
    private ThreadPoolTaskExecutor gifConverterExecutor;

    private List<KeyResultHandler> resultHandlers = new ArrayList<>();

    @PostConstruct
    public void init() {
        Path storage = resultsFilesManager.getStorageDir();
        resultHandlers.add(new PicturesResultHandler(storage));
        resultHandlers.add(new AnimationResultHandler(storage));
        resultHandlers.add(new PresentationResultHandler(storage));
        resultHandlers.add(new SlurmLogsResultHandler(storage));
        resultHandlers.add(new DatabaseResultHandler(storage));
        resultHandlers.add(new MicrostructureParamsHandler(storage));
        resultHandlers.add(new MezostructureParamsHandler(storage));
        resultHandlers.add(new CurvesResultHandler(storage, jacksonMapper));
        resultHandlers.add(new ValuesResultHandler(storage));
        resultHandlers.add(new LogFilesResultHandler(storage));
        resultHandlers.add(new StlVisualisationResultHandler(storage));

        resultHandlers.add(new ModeFrontierParamsHandler(storage, jacksonMapper,
                (id) -> {
                    Map<MFBaseSettingType, MFOptimizationBaseSetting> baseSettings = MFOptimizationSettingsManager.findBaseSettings(id, null);
                    return (List<MFOptimizationTargetFunc>) MFOptimizationSettingsManager.findOptimizationSettings(id, MFSettingsType.TargetFunc, null).stream()
                            .limit(baseSettings.get(MFBaseSettingType.NUM_TARGET_FUNCS).getCurrentValue())
                            .collect(toList());
                }, (id) -> {
            Map<MFBaseSettingType, MFOptimizationBaseSetting> baseSettings = MFOptimizationSettingsManager.findBaseSettings(id, null);
            return (List<MFOptimizationConstraint>) MFOptimizationSettingsManager.findOptimizationSettings(id, MFSettingsType.Constraint, null).stream()
                    .limit(baseSettings.get(MFBaseSettingType.NUM_CONSTRAINTS).getCurrentValue())
                    .collect(toList());
        }, (id) -> {
            Map<MFBaseSettingType, MFOptimizationBaseSetting> baseSettings = MFOptimizationSettingsManager.findBaseSettings(id, null);
            return (List< MFOptimizationVaryingParam>) MFOptimizationSettingsManager.findOptimizationSettings(id, MFSettingsType.VaryingParam, null).stream()
                    .limit(baseSettings.get(MFBaseSettingType.NUM_CAO_VARYING_PARAMS).getCurrentValue())
                    .collect(toList());
        }));
        resultHandlers.add(new OptimusParamsHandler(storage, jacksonMapper,
                (id) -> {
                    Map<OptimusBaseSettingType, OptimusOptimizationBaseSetting> baseSettings = optimusOptimizationSettingsManager.findBaseSettings(id, null);
                    return (List<OptimusOptimizationTargetFunc>) optimusOptimizationSettingsManager.findOptimizationSettings(id, OptimusSettingsType.TargetFunc, null).stream()
                            .limit(baseSettings.get(OptimusBaseSettingType.NUM_TARGET_FUNCS).getCurrentValue())
                            .collect(toList());
                }, (id) -> {
            Map<OptimusBaseSettingType, OptimusOptimizationBaseSetting> baseSettings = optimusOptimizationSettingsManager.findBaseSettings(id, null);
            return (List<OptimusOptimizationConstraint>) optimusOptimizationSettingsManager.findOptimizationSettings(id, OptimusSettingsType.Constraint, null).stream()
                    .limit(baseSettings.get(OptimusBaseSettingType.NUM_CONSTRAINTS).getCurrentValue())
                    .collect(toList());
        }, (id) -> {
            Map<OptimusBaseSettingType, OptimusOptimizationBaseSetting> baseSettings = optimusOptimizationSettingsManager.findBaseSettings(id, null);
            return (List< OptimusOptimizationVaryingParam>) optimusOptimizationSettingsManager.findOptimizationSettings(id, OptimusSettingsType.VaryingParam, null).stream()
                    .limit(baseSettings.get(OptimusBaseSettingType.NUM_CAO_VARYING_PARAMS).getCurrentValue())
                    .collect(toList());
        }));

        resultHandlers.add(new ToscaRawDataHandler(storage, jacksonMapper,
                toscaOptimizationSettingsManager, toscaTargetFunctionParametersDao));

        resultHandlers.add(new OptistructFEMFileHandler(storage));
        resultHandlers.add(new OptistructParamsHandler(storage, jacksonMapper, (id) -> optistructOptimizationSettingsManager.findBaseSettings(id, null)));
        resultHandlers.add(new OptistructResultHandler(storage));
    }


    public void extractResults(TaskEntity task) {
        Path resultsDir = resultsFilesManager.getResultsDir(task);
        CompositeFileResolver filesResolver =
                new CompositeFileResolver(new BasePlaceholdersResolver(resultsDir, resultsDir));
        SolverConfig solverConfig = solverConfigManager.getSolverByName(task.getSolverName());
        List<ResultsResolver> resultResolvers =
                singletonList(new ResultsResolver(solverConfig, task.getType(), filesResolver));

        ResultsParent resultsParent;
        switch (task.getParentType()) {
            case effectiveCharacteristic:
                resultsParent = Mapper.map(ecManager.findOne(task.getParentId(), null), ResultsParent.class);
                break;
            case simulation:
                resultsParent = Mapper.map(simulationManager.findOne(task.getParentId(), null), ResultsParent.class);
                break;
            case optimization:
                resultsParent = Mapper.map(optimizationManager.findOne(task.getParentId(), null), ResultsParent.class);
                break;
            default:
                throw new IllegalArgumentException("Unexpected enum value: " + task.getParentType());
        }

        try (Stream<Path> resultsStream = Files.walk(resultsDir)) {
            List<Path> allResults = resultsStream.filter(Files::isRegularFile).collect(toList());
            allResults.forEach(pathToResult ->
                    handleResultFile(pathToResult, resultsParent, resultResolvers, CreatedBy.Auto));
        } catch (IOException ex) {
            throw new ResultProcessingException("IO exception while processing results dir", ex);
        }

        if (task.getType() == Type.toscaAnimation) {
            makeGifFromImages(task);
        }
    }

    private void makeGifFromImages(TaskEntity task) {
        gifConverterExecutor.submit(() -> {
            Path animationResult = resultsFilesManager.getResultsDir(task).resolve("animation.gif");
            OptimizationEntity optimization = optimizationManager.findOne(task.getParentId(), null);
            ResultsParent resultParent = Mapper.map(optimization, ResultsParent.class);
            List<ResultEntity> pictures = resultsManager.getResultByParent(resultParent, ResultType.pictures);
            List<Path> paths = pictures.stream()
                    .filter(re -> {
                        Matcher matcher = GIF_IMAGE_FILE_PATTERN.matcher(re.getName());
                        return matcher.matches();
                    })
                    .sorted(Comparator.comparing(ResultEntity::getName))
                    .map(re -> resultsFilesManager.getPathToResult(re))
                    .collect(Collectors.toList());
            gifSequenceWriter.makeGifFromImages(paths, animationResult, true);
            ResultEntity newResult = new ResultEntity();
            newResult.setPathToFile(resultsFilesManager.getStorageDir().relativize(animationResult).toString());
            newResult.setParentId(resultParent.getId());
            newResult.setParentType(resultParent.getType());
            newResult.setOwnerId(resultParent.getOwnerId());
            newResult.setCreatedBy(CreatedBy.Auto);
            newResult.setCreationDate(new Date());
            newResult.setName(FilenameUtils.removeExtension(animationResult.getFileName().toString()));
            newResult.setType(ResultType.animations);
            resultsManager.saveAllWithoutDuplicates(singletonList(newResult));
            pictures.forEach(picture -> resultsManager.deleteResult(picture.getId()));
        });
    }

    private void handleResultFile(Path pathToFile, ResultsParent resultsParent, List<ResultsResolver> resultsResolvers,
                                  CreatedBy createdBy) {
        ResultType keyResultType = resultsResolvers.stream()
                .map(resultsResolver -> resultsResolver.getKeyResultType(pathToFile))
                .filter(Objects::nonNull)
                .findFirst().orElse(null);
        if (keyResultType == null) {
            return;
        }
        if (isStlFileFormat(pathToFile.toString())) {
            modifyStlAsciiToBinary(pathToFile);
        }
        KeyResultHandler resultHandler = getResultHandlerForType(keyResultType, pathToFile);
        if (resultHandler == null) {
            return;
        }
        try {
            List<ResultEntity> extractedResults = resultHandler
                    .handleResultsFile(keyResultType, pathToFile, resultsParent, createdBy);
            resultsManager.saveAllWithoutDuplicates(extractedResults);
        } catch (ResultProcessingException ex) {
            LOGGER.error("Exception while processing result file: " + pathToFile, ex);
        }
    }
    private void modifyStlAsciiToBinary(Path stlFilePath) {
        try{
            String rubyPath = "/usr/bin/ruby";
            String convertStlPath = "/home/space/ConvertStl/convertSTL.rb";
            Process p = Runtime.getRuntime().exec(new String[]{rubyPath,convertStlPath,stlFilePath.toString()});
            p.waitFor();
            String oldStlPath = stlFilePath.getParent().toString();
            String oldStlName = stlFilePath.getFileName().toString();
            File newStl = new File(oldStlPath + "/"+FilenameUtils.removeExtension(oldStlName)+"-binary.stl");
            if (newStl.exists()) {
                Files.delete(stlFilePath);
                File oldStl = new File(oldStlPath + "/" + oldStlName);
                newStl.renameTo(oldStl);
            }
        }
        catch(Exception ex){
            System.out.println(ex.getMessage());
        }
    }

    private boolean isStlFileFormat(String stlFileName){
        return stlFileName.contains(".stl");
    }

    @Nullable
    private KeyResultHandler getResultHandlerForType(ResultType keyResultType, Path pathToFile) {
        return resultHandlers.stream()
                .filter(resultHanlder -> resultHanlder.canHandle(keyResultType, pathToFile))
                .findFirst().orElse(null);
    }
}
