package ru.crp.lvms.managers.tasks.resultsHandlers;

import com.fasterxml.jackson.databind.ObjectMapper;
import org.apache.commons.csv.CSVFormat;
import org.apache.commons.csv.CSVParser;
import org.apache.commons.csv.CSVRecord;
import org.apache.commons.io.FilenameUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import ru.crp.cmls.calculations.solvers.xml.config.ResultType;
import ru.crp.lvms.chart.ChartFactory;
import ru.crp.lvms.chart.LineChart;
import ru.crp.lvms.chart.Point;
import ru.crp.lvms.exceptions.solving.ResultProcessingException;
import ru.crp.lvms.managers.tasks.ResultsParent;
import ru.crp.lvms.model.entities.ResultEntity;
import ru.crp.lvms.model.entities.misc.CreatedBy;

import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;

import static java.util.Arrays.asList;

public class CurvesResultHandler extends KeyResultHandler {
    public static final Logger LOGGER = LoggerFactory.getLogger(CurvesResultHandler.class);

    public static final String X_LABEL = "Предел прочности";
    public static final String X_DIMENSION = "МПа";

    public static final String FIBER_CURVE_NAME = "Волокно";
    public static final String Y_FIBER_LABEL = "oz в волокне";
    public static final String Y_FIBER_DIMENSION = "МПа";

    public static final String MATRIX_CURVE_NAME = "Матрица";
    public static final String Y_MATRIX_LABEL = "oz в матрице";
    public static final String Y_MATRIX_DIMENSION = "МПа";

    public static final String FIBER_SUFFIX = "_fiber";
    public static final String MATRIX_SUFFIX = "_matrix";
    private static final String JSON_EXTENSION = ".json";

    private ObjectMapper objectMapper;

    public CurvesResultHandler(Path pathToStorage, ObjectMapper objectMapper) {
        super(pathToStorage);
        this.objectMapper = objectMapper;
    }

    @Override
    public boolean canHandle(ResultType resultType, Path pathToFile) {
        return resultType == ResultType.curves && Files.isRegularFile(pathToFile);
    }

    @Override
    protected List<ResultEntity> doHandleResultsFile(Path pathToFile, ResultsParent resultsParent, CreatedBy createdBy) {
        ResultEntity fiberCurveEntity = createCurveEntity(pathToFile, resultsParent, createdBy, true);
        ResultEntity matrixCurveEntity = createCurveEntity(pathToFile, resultsParent, createdBy, false);

        parseFile(pathToFile);
        return asList(fiberCurveEntity, matrixCurveEntity);
    }

    private ResultEntity createCurveEntity(Path pathToFile, ResultsParent resultsParent, CreatedBy createdBy, boolean isFiber) {
        ResultEntity newCurve = initCommonResultProperties(pathToFile, resultsParent, createdBy);
        if (createdBy == CreatedBy.Auto) {
            extractDescription(pathToFile, newCurve);
        }
        String resultName = FilenameUtils.removeExtension(pathToFile.getFileName().toString()) + (isFiber ? FIBER_SUFFIX : MATRIX_SUFFIX);
        newCurve.setName(resultName);
        newCurve.setType(ResultType.curves);
        return newCurve;
    }

    private void parseFile(Path rawFile) {
        String rawFileName = rawFile.getFileName().toString();
        String fiberFileName = FilenameUtils.removeExtension(rawFileName) + FIBER_SUFFIX + JSON_EXTENSION;
        String matrixFileName = FilenameUtils.removeExtension(rawFileName) + MATRIX_SUFFIX + JSON_EXTENSION;

        try {
            Path fiberFile = rawFile.resolveSibling(fiberFileName);
            Path matrixFile = rawFile.resolveSibling(matrixFileName);
            Files.deleteIfExists(fiberFile);
            Files.deleteIfExists(matrixFile);
            Files.createFile(fiberFile);
            Files.createFile(matrixFile);

            LineChart fiberChart = ChartFactory.createCurve();
            fiberChart.setTitle(FIBER_CURVE_NAME);
            fiberChart.setxAxisLabel(X_LABEL);
            fiberChart.setxDimension(X_DIMENSION);
            fiberChart.setyAxisLabel(Y_FIBER_LABEL);
            fiberChart.setyDimension(Y_FIBER_DIMENSION);

            LineChart matrixChart = ChartFactory.createCurve();
            matrixChart.setTitle(MATRIX_CURVE_NAME);
            matrixChart.setxAxisLabel(X_LABEL);
            matrixChart.setxDimension(X_DIMENSION);
            matrixChart.setyAxisLabel(Y_MATRIX_LABEL);
            matrixChart.setyDimension(Y_MATRIX_DIMENSION);

            CSVFormat csvFormat = CSVFormat.DEFAULT.withDelimiter(',');
            CSVParser parser = CSVParser.parse(rawFile.toFile(), Charset.defaultCharset(), csvFormat);
            List<CSVRecord> records = parser.getRecords();
            boolean isExtendedFormat = false;
            for (CSVRecord record : records) {
                checkRecordSize(record);
                if (record.size() == 8) {
                    isExtendedFormat = true;
                }
                fiberChart.addPoint(Point.commonPoint(
                        Double.parseDouble(record.get(1)),
                        Double.parseDouble(record.get(2)))
                );

                matrixChart.addPoint(Point.commonPoint(
                        Double.parseDouble(record.get(1)),
                        Double.parseDouble(record.get(3)))
                );
            }
            if (isExtendedFormat) {
                for (CSVRecord record : records) {
                    fiberChart.addPoint(Point.commonPoint(
                            Double.parseDouble(record.get(5)),
                            Double.parseDouble(record.get(6)))
                    );

                    matrixChart.addPoint(Point.commonPoint(
                            Double.parseDouble(record.get(5)),
                            Double.parseDouble(record.get(7)))
                    );
                }
            }
            objectMapper.writeValue(fiberFile.toFile(), fiberChart.toChartsSet());

            objectMapper.writeValue(matrixFile.toFile(), matrixChart.toChartsSet());
        } catch (IOException ex) {
            throw new ResultProcessingException("Can't create json file for curves data", ex);
        } catch (NumberFormatException ex) {
            throw new ResultProcessingException("Unexpected csv file format: can not convert to double record values", ex);
        }
    }

    private void checkRecordSize(CSVRecord record) {
        if (record.size() != 4 && record.size() != 8) {
            throw new ResultProcessingException("Unexpected csv file format: Expected 4 or 8 columns but has " + record.size());
        }
    }

    public static String getJsonFileNameByCsv(String csvFileName, boolean isFiber) {
        String suffix = isFiber ? FIBER_SUFFIX : MATRIX_SUFFIX;
        return FilenameUtils.removeExtension(csvFileName) + suffix + JSON_EXTENSION;
    }
}
