package ru.crp.lvms.model.dao;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Primary;
import org.springframework.stereotype.Repository;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.model.dao.tree.*;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.model.entities.misc.ItemType;
import ru.crp.lvms.model.entities.misc.TreeType;
import ru.crp.lvms.model.entities.tree.OptimizationsTreeNode;
import ru.crp.lvms.model.entities.tree.TreeNode;

import javax.annotation.PostConstruct;
import java.util.EnumMap;
import java.util.HashMap;
import java.util.List;
import java.util.Map;


@Repository
@Primary
public class TreeDao extends BaseSpaceDao<TreeNode> {
    private static final String INSERT_FOLDER_SQL = "INSERT INTO folders(pid, name, tree_type) VALUES "
            + "(:pid, :name, :tree_type :: tree_type)";
    private static final String INSERT_CONSTRUCTION_ELEMENT_FILES_FOLDER_SQL = "INSERT INTO construction_element_file_types(pid, name, owner_id) VALUES "
            + "(:pid, :name, :ownerId)";
    private static final String DELETE_FOLDER_SQL = "DELETE FROM folders WHERE id = ?";
    private static final String DELETE_OPTIMIZATION_FOLDER_SQL = "DELETE FROM optimization_folders WHERE id = ?";

    private static final String INSERT_OPTIMIZATION_FOLDER_SQL = "INSERT INTO optimization_folders(" +
            "construction_element_id, parent_folder_id, name, optimization_type_id, base_optimization_id, owner_id) VALUES "
            + "(:constructionElementId, :parentFolderId, :name, :optimizationTypeId, :baseOptimizationId, :ownerId)";

    private static final String GET_SUBNODES_SQL = "WITH RECURSIVE recs(id, pid, name, path, path_names, item_type, children_types)"
            + " AS "
            + "(SELECT tr.id, tr.pid, tr.name, array_to_string(tr.path, ',') AS path, tr.path_names, tr.item_type, tr.children_types "
            + "FROM <TREE_PLACEHOLDER> tr WHERE tr.id = ? "
            + "UNION ALL "
            + "SELECT tr2.id, tr2.pid, tr2.name, array_to_string(tr2.path, ',') AS PATH, tr2.path_names, tr2.item_type, tr2.children_types "
            + "FROM <TREE_PLACEHOLDER> AS tr2, recs AS r "
            + "WHERE tr2.pid = r.id) "
            + "SELECT * FROM recs "
            + "WHERE NOT recs.id = ? "
            + "ORDER BY name, item_type";

    // we need to select some additional properties (like 'optimizationTypeId')
    // todo remove copy-paste
    private static final String GET_OPTIMIZATION_SUBNODES_SQL = "WITH RECURSIVE recs(id, pid, name, path, path_names, item_type, children_types," +
            " optimization_type_id, construction_element_id, base_optimization_id)"
            + " AS "
            + "(SELECT tr.id, tr.pid, tr.name, array_to_string(tr.path, ',') AS path, tr.path_names, tr.item_type, tr.children_types, "
            + "tr.optimization_type_id, tr.construction_element_id, tr.base_optimization_id "
            + "FROM <TREE_PLACEHOLDER> tr WHERE tr.id = ? "
            + "UNION ALL "
            + "SELECT tr2.id, tr2.pid, tr2.name, array_to_string(tr2.path, ',') AS PATH, tr2.path_names, tr2.item_type, tr2.children_types, "
            + "tr2.optimization_type_id, tr2.construction_element_id, tr2.base_optimization_id "
            + "FROM <TREE_PLACEHOLDER> AS tr2, recs AS r "
            + "WHERE tr2.pid = r.id) "
            + "SELECT * FROM recs "
            + "WHERE NOT recs.id = ? "
            + "ORDER BY name, item_type";

    private static final String GET_OPTIMIZATION_FOLDERS_ASCENDING_BRANCH =
            "WITH RECURSIVE recs(ordering, id, pid, name, path, path_names, item_type, children_types, optimization_type_id, construction_element_id, base_optimization_id) "
                    + "AS "
                    + "(SELECT 0 AS ordering, tr.id, tr.pid, tr.name, array_to_string(tr.path, ',') AS path, tr.path_names, tr.item_type, tr.children_types, "
                    + "tr.optimization_type_id, tr.construction_element_id, tr. base_optimization_id "
                    + "FROM optimizations_tree_join tr WHERE tr.id = ? "
                    + "UNION ALL "
                    + "SELECT r.ordering + 1, tr2.id, tr2.pid, tr2.name, array_to_string(tr2.path, ',') AS PATH, tr2.path_names, tr2.item_type, "
                    + "tr2.children_types, tr2.optimization_type_id, tr2.construction_element_id, tr2.base_optimization_id "
                    + "FROM optimizations_tree_join AS tr2, recs AS r "
                    + "WHERE tr2.id = r.pid) "
                    + "SELECT * FROM recs "
                    + "WHERE recs.item_type = 'OptimizationFolder'"
                    + "ORDER BY ordering DESC, name, item_type";

    private static final String GET_MATERIAL_NODES_BY_PARAM_ID_SQL = "SELECT t.* FROM tree t "
            + "JOIN   parameters_materials "
            + "ON     t.id = material_id "
            + "WHERE  parameter_id = ?";

    @Autowired
    private MaterialsTreeDao materialsTreeDao;

    @Autowired
    private CompositeMaterialsTreeDao compositeMaterialsTreeDao;

    @Autowired
    private TechnologiesTreeDao technologiesTreeDao;

    @Autowired
    private ConstructionElementsTreeDao constructionElementsTreeDao;

    @Autowired
    private OptimizationsTreeDao optimizationsTreeDao;

    private Map<TreeType, BaseSpaceDao<? extends TreeNode>> treeDaos;

    private Map<TreeType, String> treePlaceholders;

    @PostConstruct
    public void init() {
        treeDaos = new HashMap<>();
        treeDaos.put(TreeType.materials, materialsTreeDao);
        treeDaos.put(TreeType.compositeMaterials, compositeMaterialsTreeDao);
        treeDaos.put(TreeType.technologies, technologiesTreeDao);
        treeDaos.put(TreeType.constructionElements, constructionElementsTreeDao);
        treeDaos.put(TreeType.optimizations, optimizationsTreeDao);

        treePlaceholders = new EnumMap<>(TreeType.class);
        treePlaceholders.put(TreeType.materials, "materials_tree_join");
        treePlaceholders.put(TreeType.compositeMaterials, "composite_materials_tree_join");
        treePlaceholders.put(TreeType.technologies, "technologies_tree");
        treePlaceholders.put(TreeType.constructionElements, "construction_elements_tree_join");
        treePlaceholders.put(TreeType.optimizations, "optimizations_tree_join");
    }


    public List<? extends TreeNode> getRootNodes(TreeType treeType) {
        return checkDaoExists(treeType)
                .find("WHERE pid IS NULL");
    }


    private <T extends TreeNode> BaseSpaceDao<T> checkDaoExists(TreeType treeType) {
        if (!treeDaos.containsKey(treeType)) {
            throw new NotFoundException("Tree with type " + treeType + " is not found");
        }
        return (BaseSpaceDao<T>) treeDaos.get(treeType);
    }


    public List<TreeNode> findMaterialsByParameterId(Integer parameterId, UserEntity currentUser) {
        return jdbcTemplate().query(GET_MATERIAL_NODES_BY_PARAM_ID_SQL, getRowMapper(), parameterId);
    }


    public boolean hasChildren(int parentId) {
        return !find("WHERE pid = ?", parentId).isEmpty();
    }


    public List<? extends TreeNode> getAllNodesByTreeType(TreeType treeType) {
        return checkDaoExists(treeType)
                .findAll();
    }


    public List<? extends TreeNode> getSubTree(Integer rootId, TreeType treeType) {
        BaseSpaceDao<TreeNode> dao = checkDaoExists(treeType);
        if (treeType == TreeType.optimizations) {
            return dao.jdbcTemplate().query(GET_OPTIMIZATION_SUBNODES_SQL.replace("<TREE_PLACEHOLDER>", treePlaceholders.get(treeType)), dao.getRowMapper(), rootId, rootId);
        } else {
            return dao.jdbcTemplate().query(GET_SUBNODES_SQL.replace("<TREE_PLACEHOLDER>", treePlaceholders.get(treeType)), dao.getRowMapper(), rootId, rootId);
        }
    }

    public OptimizationsTreeNode getOptimizationFolder(Integer id) {
        return optimizationsTreeDao.findOne(id);
    }

    public List<OptimizationsTreeNode> getOptimizationFoldersByOptimizationId(int baseOptimizationId, UserEntity unused) {
        return optimizationsTreeDao.find("WHERE base_optimization_id = ?", baseOptimizationId);
    }

    //! order is important
    public List<OptimizationsTreeNode> getOptimizationFoldersAscendingBranch(Integer rootId) {
        return optimizationsTreeDao.jdbcTemplate().query(GET_OPTIMIZATION_FOLDERS_ASCENDING_BRANCH,
                optimizationsTreeDao.getRowMapper(), rootId);
    }


    public TreeNode insertAndReadFolder(TreeType treeType, TreeNode newFolder, int ownerId, boolean isConstructionElementFolder) {
        Map<String, Object> params = createFolderParamsMap(treeType, newFolder, ownerId);
        String sql = INSERT_FOLDER_SQL;
        if (isConstructionElementFolder) {
            sql = INSERT_CONSTRUCTION_ELEMENT_FILES_FOLDER_SQL;
        }
        int id = insertAndRead(sql, params);
        newFolder.setId(id);
        return checkDaoExists(treeType).findOne(id);
    }

    private Map<String, Object> createFolderParamsMap(TreeType treeType, TreeNode folder, int ownerId) {
        Map<String, Object> params = new HashMap<>();
        params.put("ownerId", ownerId);
        params.put("pid", folder.getPid());
        params.put("name", folder.getName());
        params.put("tree_type", treeType.toString());
        return params;
    }


    public void deleteFolder(int folderId) {
        jdbcTemplate().update(DELETE_FOLDER_SQL, folderId);
    }


    public TreeNode insertAndReadOptimizationFolder(OptimizationsTreeNode newOptimizationFolder, int ownerId) {
        Map<String, Object> params = createOptimizationFolderParamsMap(newOptimizationFolder, ownerId);
        int id = insertAndRead(INSERT_OPTIMIZATION_FOLDER_SQL, params);
        newOptimizationFolder.setId(id);
        return checkDaoExists(TreeType.optimizations).findOne(id);
    }

    private Map<String, Object> createOptimizationFolderParamsMap(OptimizationsTreeNode optimizationFolder,
                                                                  int ownerId) {
        Map<String, Object> params = new HashMap<>();
        params.put("constructionElementId", optimizationFolder.getConstructionElementId());
        Integer parentFolderId = optimizationFolder.getConstructionElementId().equals(optimizationFolder.getPid()) ? null
                : optimizationFolder.getPid();
        params.put("parentFolderId", parentFolderId);
        params.put("name", optimizationFolder.getName());
        params.put("optimizationTypeId", optimizationFolder.getOptimizationTypeId());
        params.put("baseOptimizationId", optimizationFolder.getBaseOptimizationId());
        params.put("ownerId", ownerId);
        return params;
    }


    public TreeNode findOne(TreeType treeType, int nodeId) {
        return checkDaoExists(treeType).findOne(nodeId);
    }


    public <T extends TreeNode> TreeNode updateAndRead(TreeType treeType, T nodeToUpdate) {
        return checkDaoExists(treeType)
                .updateAndRead(nodeToUpdate);
    }


    public void deleteOptimizationFolder(int optimizationFolderId) {
        jdbcTemplate().update(DELETE_OPTIMIZATION_FOLDER_SQL, optimizationFolderId);
    }


    public TreeNode getConstructionElementFileFolder(int constructionElementId, FileType type) {
        String folderName = null;
        switch (type) {
            case CADFile:
                folderName = "CAD";
                break;
            case CAEFile:
                folderName = "CAE";
                break;
            case CAOFile:
                folderName = "CAO";
                break;
            case DesignSolutionFile:
                folderName = "Design solutions";
            default:
                break;
        }
        if (folderName == null) {
            return null;
        }
        return findOne("WHERE name = ? AND pid = ?", folderName, constructionElementId);
    }

    public List<TreeNode> findByPathNames(ItemType itemType, String pathNames) {
        return find("WHERE path_names = ? AND item_type = ? :: tree_node_type", pathNames, itemType.toString());
    }

    public List<TreeNode> findByName(ItemType itemType, String name) {
        return  find("WHERE name = ? AND item_type = ? :: tree_node_type", name, itemType.toString());

    }
}
