package ru.crp.lvms.optimizations.mf.controllers;

import com.fasterxml.jackson.databind.node.IntNode;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.multipart.MultipartFile;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.model.AccessRights;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.optimizations.mf.managers.MFOptimizationSettingsManager;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;

import javax.annotation.Nonnull;
import javax.servlet.http.HttpServletRequest;
import javax.validation.Valid;
import javax.validation.constraints.NotNull;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import static ru.crp.lvms.utils.URLs.MF_SCRIPT_URL;


@RestController
@RequestMapping(value = MF_SCRIPT_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class MFScriptController {
    public static class UpdateScriptRequest {
        @NotNull
        private Integer paramId;
        @NotNull
        private String string;

        public Integer getParamId() {
            return paramId;
        }

        public void setParamId(Integer paramId) {
            this.paramId = paramId;
        }

        public String getString() {
            return string;
        }

        public void setString(String string) {
            this.string = string;
        }
    }

    private static final Logger LOGGER = LoggerFactory.getLogger(MFScriptController.class);

    @Autowired
    private FileManager fileManager;

    @Autowired
    private OptimizationManager optimizationManager;

    @Autowired
    private MFOptimizationSettingsManager mfOptimizationSettingsManager;

    @Nonnull
    @RequestMapping(method = RequestMethod.GET)
    public ResponseEntity<Map<String, Object>> getScript(@PathVariable("optimizationId") Integer optimizationId,
                                                         @CurrentUser PortalUser currentUser) throws IOException {
        Map<String, Object> response = new HashMap<>(2);

        response.put("code", null);
        response.put("optimizationId", new IntNode(optimizationId));
        Optional<FileEntity> script = getScriptEntity(optimizationId, currentUser);
        if (!script.isPresent()) {
            return ResponseEntity.ok(response);
        }
        Path pathToScript = fileManager.resolvePathToFile(script.get(), currentUser.getUser());
        if (pathToScript == null) {
            return new ResponseEntity<>(HttpStatus.NOT_FOUND);
        }
        response.put("code", Files.readAllLines(pathToScript, StandardCharsets.UTF_8));
        return ResponseEntity.ok(response);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @Nonnull
    @RequestMapping(method = RequestMethod.POST)
    public FileEntity replaceScript(HttpServletRequest request, @PathVariable("optimizationId") Integer optimizationId,
                                    @RequestPart("file") MultipartFile newModel, @CurrentUser PortalUser currentUser) {
        OptimizationEntity optimization = optimizationManager.checkEntityExistsAndAccesible(optimizationId, AccessRights.view, currentUser.getUser());
        List<FileEntity> scripts = optimizationManager.getOptimizationFiles(optimization, FileType.ModFrontierScript, currentUser.getUser());
        scripts.forEach(script -> optimizationManager.deleteModel(optimization, script, currentUser.getUser()));
        mfOptimizationSettingsManager.unselectVaryingParams(optimizationId);
        return optimizationManager.addModel(optimization, newModel, FileType.ModFrontierScript, currentUser.getUser());
    }


    @PreAuthorize("hasRole('ROLE_USER')")
    @Nonnull
    @RequestMapping(value = "/{lineNumber}", method = RequestMethod.PUT)
    @ResponseStatus(HttpStatus.OK)
    public void updateScript(@PathVariable("lineNumber") int lineNumber,
                             @PathVariable("optimizationId") Integer optimizationId,
                             @RequestBody @Valid UpdateScriptRequest updateScriptRequest,
                             @CurrentUser PortalUser currentUser) throws IOException {
        Optional<FileEntity> script = getScriptEntity(optimizationId, currentUser);
        if (!script.isPresent()) {
            throw new NotFoundException("Скрипт не найден");
        }
        Path pathToScript = fileManager.resolvePathToFile(script.get(), currentUser.getUser());
        if (pathToScript == null) {
            throw new NotFoundException("Скрипт не найден");
        }
        mfOptimizationSettingsManager.selectVaryingParam(optimizationId, updateScriptRequest.getParamId());
        List<String> strings = Files.readAllLines(pathToScript, StandardCharsets.UTF_8);
        strings.set(lineNumber, updateScriptRequest.string);
        Files.write(pathToScript, strings, StandardCharsets.UTF_8);
    }

    private Optional<FileEntity> getScriptEntity(int optimizationId, PortalUser currentUser) {
        OptimizationEntity optimization = optimizationManager.checkEntityExistsAndAccesible(optimizationId, AccessRights.view, currentUser.getUser());
        List<FileEntity> scripts = optimizationManager.getOptimizationFiles(optimization, FileType.ModFrontierScript, currentUser.getUser());
        if (scripts.size() > 1) {
            LOGGER.warn("More than 1 mod frontier script found for optimization " + optimizationId);
        }
        return scripts.stream().findFirst();
    }
}
