package ru.crp.lvms.optimizations.optimus.config;

import org.modelmapper.TypeToken;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import ru.crp.lvms.model.entities.misc.ParametersDictionaryType;
import ru.crp.lvms.model.entities.modfrontier.AlgorithmType;
import ru.crp.lvms.model.entities.modfrontier.DoeType;
import ru.crp.lvms.model.entities.modfrontier.ParametersDictionary;
import ru.crp.lvms.optimizations.base.dao.OptimizationJsonSettingsDao;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationJsonSetting;
import ru.crp.lvms.optimizations.base.entities.OptimizationListSetting;
import ru.crp.lvms.optimizations.optimus.config.xml.*;
import ru.crp.lvms.optimizations.optimus.entities.*;
import ru.crp.lvms.optimizations.optimus.enums.OptimusAlgorithmType;
import ru.crp.lvms.optimizations.optimus.enums.OptimusBaseSettingType;
import ru.crp.lvms.optimizations.optimus.dao.OptimusOptimizationCADVaryingParamsDao;
import ru.crp.lvms.optimizations.optimus.dao.OptimusOptimizationConstraintsDao;
import ru.crp.lvms.optimizations.optimus.dao.OptimusOptimizationTargetFuncsDao;
import ru.crp.lvms.optimizations.optimus.dao.OptimusOptimizationVaryingParamsDao;
import ru.crp.lvms.optimizations.optimus.enums.OptimusDoeType;
import ru.crp.lvms.optimizations.optimus.managers.OptimusOptimizationSettingsManager;
import ru.crp.lvms.utils.Mapper;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import static java.util.Arrays.asList;
import static java.util.stream.Collectors.toList;
import static ru.crp.lvms.utils.Mapper.map;
import static ru.crp.lvms.utils.Utils.mapWithIndex;

@Component
public class OptimusConfigGenerator {

    @Autowired
    private OptimusOptimizationVaryingParamsDao optimusOptimizationVaryingParamsDao;

    @Autowired
    private OptimusOptimizationCADVaryingParamsDao optimusOptimizationCADVaryingParamsDao;

    @Autowired
    private OptimusOptimizationConstraintsDao optimusOptimizationConstraintsDao;

    @Autowired
    private OptimusOptimizationTargetFuncsDao optimusOptimizationTargetFuncsDao;

    @Autowired
    private OptimusOptimizationSettingsManager optimusOptimizationSettingsManager;

    @Autowired
    private OptimizationJsonSettingsDao jsonSettingsDao;

    public Path generateConfig(OptimizationEntity optimization) throws IOException, JAXBException {
        OptimusConfigRoot optimusConfigRoot = composeXmlStructure(optimization);

        Path tempConfigFile = Files.createTempFile(null, null);

        JAXBContext jaxbContext = JAXBContext.newInstance(OptimusConfigRoot.class);
        Marshaller jaxbMarshaller = jaxbContext.createMarshaller();
        jaxbMarshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true);
        jaxbMarshaller.marshal(optimusConfigRoot, tempConfigFile.toFile());

        jaxbMarshaller.setProperty(Marshaller.JAXB_ENCODING, "Unicode");

        return tempConfigFile;
    }

    private OptimusConfigRoot composeXmlStructure(OptimizationEntity optimization) {
        OptimusConfigRoot optimusConfigRoot = new OptimusConfigRoot();


        Map<OptimusBaseSettingType, OptimusOptimizationBaseSetting> baseSettings = optimusOptimizationSettingsManager.findBaseSettings(optimization.getId(), null);
        optimusConfigRoot.setFactors(formFactors(optimization, baseSettings));
        optimusConfigRoot.setResponses(formResponses(optimization, baseSettings));
        optimusConfigRoot.setDependencies(formDependency(optimization, baseSettings));
        optimusConfigRoot.setExperiment(formExperiment(optimization, baseSettings));
        optimusConfigRoot.setSequence(formSequence(optimization, baseSettings));
        optimusConfigRoot.setDecks(formDecks(optimization, baseSettings));
//

        return optimusConfigRoot;
    }

    private FactorsType formFactors(OptimizationEntity optimization, Map<OptimusBaseSettingType, OptimusOptimizationBaseSetting> baseSettings) {
        List<FactorList> factors = getVaryingParams(optimization, baseSettings);
        FactorsType factorsType = new FactorsType();
        factorsType.setFactor(factors);
        return factorsType;
    }

    // returns used varying and cad varying params in single list
    private List<FactorList> getVaryingParams(OptimizationEntity optimization, Map<OptimusBaseSettingType, OptimusOptimizationBaseSetting> baseSettings) {
        Integer varyingParamsCount = baseSettings.get(OptimusBaseSettingType.NUM_CAO_VARYING_PARAMS).getCurrentValue();
        List<OptimusOptimizationVaryingParam> varyingParams = optimusOptimizationVaryingParamsDao
                .findForOptimization(optimization.getId()).stream()
                .limit(varyingParamsCount)
                .collect(toList());

        List<FactorList> factor = mapWithIndex(varyingParams, 1, i -> {
            FactorList vt = map(varyingParams.get(i - 1), FactorList.class);
            vt.setName("par" + i);
            if (varyingParams.get(i-1).getSampling() != 0 ) {
                vt.setRange("STEPPED");
                double step = (varyingParams.get(i-1).getUpperBound()-varyingParams.get(i-1).getLowerBound())/(varyingParams.get(i-1).getSampling()-1);
                vt.setStep(step);
            }

            return vt;
        }).collect(toList());
        if (optimization.getSolverCadId() != null) {
            Integer cadVaryingParamsCount = baseSettings.get(OptimusBaseSettingType.NUM_CAD_VARYING_PARAMS).getCurrentValue();
            List<OptimusOptimizationCADVaryingParam> cadVaryingParams = optimusOptimizationCADVaryingParamsDao
                    .findForOptimization(optimization.getId()).stream()
                    .limit(cadVaryingParamsCount)
                    .collect(toList());

            factor.addAll(mapWithIndex(cadVaryingParams, 1, i -> {
                FactorList vt = map(cadVaryingParams.get(i - 1), FactorList.class);
                vt.setName("gpar" + i);
                return vt;
            }).collect(toList()));
        }
        return factor;
    }



    private ResponsesType formResponses(OptimizationEntity optimization, Map<OptimusBaseSettingType, OptimusOptimizationBaseSetting> baseSettings) {
        ResponsesType responsesType = new ResponsesType();
        Integer targetFuncsCount = baseSettings.get(OptimusBaseSettingType.NUM_TARGET_FUNCS).getCurrentValue();
        List<OptimusOptimizationTargetFunc> targetFuncs = optimusOptimizationTargetFuncsDao.findForOptimization(optimization.getId()).stream().limit(targetFuncsCount).collect(toList());
        List<ResponseList> response = mapWithIndex(targetFuncs, 1, i -> {
            ResponseList vt = map(targetFuncs.get(i - 1), ResponseList.class);
            vt.setName("out" + i);
            return vt;
        }).collect(toList());

        Integer constraintsCount = baseSettings.get(OptimusBaseSettingType.NUM_CONSTRAINTS).getCurrentValue();
        List<OptimusOptimizationConstraint> optimizationConstrains = optimusOptimizationConstraintsDao.findForOptimization(optimization.getId()).stream().limit(constraintsCount).collect(toList());

        List<ResponseList> constrains = mapWithIndex(optimizationConstrains, 1, i -> {
            ResponseList var = map(optimizationConstrains.get(i - 1), ResponseList.class);
            var.setName("out" + (i + response.size()));
            if (optimizationConstrains.get(i-1).getFuncId() == 6 ) {
                Double limitvar=optimizationConstrains.get(i-1).getValue();
                var.setLow(limitvar);
            }
            if (optimizationConstrains.get(i-1).getFuncId() == 5 ) {
                Double limitvar=optimizationConstrains.get(i-1).getValue();
                var.setHigh(limitvar);
            }

            return var;
        }).collect(toList());


        responsesType.setResponse(asList(response, constrains).stream()
                .flatMap(List::stream).collect(toList()));
        return  responsesType;
    }

    private DependenciesType formDependency(OptimizationEntity optimization, Map<OptimusBaseSettingType, OptimusOptimizationBaseSetting> baseSettings) {
        DependenciesType dependenciesType = new DependenciesType();
        DependencyType dependencyType = new DependencyType();
        Integer varyingParamsCount = baseSettings.get(OptimusBaseSettingType.NUM_CAO_VARYING_PARAMS).getCurrentValue();
        List<OptimusOptimizationVaryingParam> varyingParams = optimusOptimizationVaryingParamsDao
                .findForOptimization(optimization.getId()).stream()
                .limit(varyingParamsCount)
                .collect(toList());

        List<VarList> var  = mapWithIndex(varyingParams, 1, i -> {
            VarList vt = map(varyingParams.get(i - 1), VarList.class);
            vt.setVar("par" + i);
            return vt;
        }).collect(toList());

        if (optimization.getSolverCadId() != null) {
            Integer cadVaryingParamsCount = baseSettings.get(OptimusBaseSettingType.NUM_CAD_VARYING_PARAMS).getCurrentValue();
            List<OptimusOptimizationCADVaryingParam> cadVaryingParams = optimusOptimizationCADVaryingParamsDao
                    .findForOptimization(optimization.getId()).stream()
                    .limit(cadVaryingParamsCount)
                    .collect(toList());

            var.addAll(mapWithIndex(cadVaryingParams, 1, i -> {
                VarList vt = map(cadVaryingParams.get(i - 1), VarList.class);
                vt.setVar("gpar" + i);
                return vt;
            }).collect(toList()));
        }

        Integer outVaryingParamsCount = baseSettings.get(OptimusBaseSettingType.NUM_OUTPUT_PARAMS).getCurrentValue();


        List<DepVarList> depvar  = IntStream.range(1, outVaryingParamsCount+1)
                .mapToObj(i -> {
                    DepVarList vt = new DepVarList();
                    vt.setDepvar("out" + i);
                    return vt;
                }).collect(toList());


        dependencyType.setVar(var);
        dependencyType.setDepvar(depvar);
        dependenciesType.setDependency(dependencyType);
        return dependenciesType;
    }


    private ExperimentType formExperiment(OptimizationEntity optimization, Map<OptimusBaseSettingType, OptimusOptimizationBaseSetting> baseSettings) {
        ExperimentType experimentType = new ExperimentType();
        Integer selectedAlgorithmId = optimusOptimizationSettingsManager.findSelectedAlgorithmId(optimization.getId());
        ParametersDictionary selectedAlgorithm = optimusOptimizationSettingsManager.findByTypeNameAndId(ParametersDictionaryType.OptimusOptimizationMethod, selectedAlgorithmId);
        if (selectedAlgorithm.getName().equals("EGO")) {
            experimentType.setGoal("expression");
            experimentType.setObjective("$out1$");
            experimentType.setAlgorithm(OptimusAlgorithmType.EGO.toString());
            experimentType.setType("OPTIMIZATION");
            experimentType.setOpt("0");
        }
        else if (selectedAlgorithm.getName().equals("SIMAN")) {
            experimentType.setGoal("expression");
            experimentType.setObjective("$out1$");
            experimentType.setAlgorithm(OptimusAlgorithmType.SIMAN.toString());
            experimentType.setType("OPTIMIZATION");
        }
        else if (selectedAlgorithm.getName().equals("DOE")) {
            Integer selectedDoeId = optimusOptimizationSettingsManager.findSelectedDoeId(optimization.getId());
            ParametersDictionary selectedDoe = optimusOptimizationSettingsManager.findByTypeNameAndId(ParametersDictionaryType.OptimusDoEType, selectedDoeId);
            experimentType.setMethod(selectedDoe.getName().toString());
            experimentType.setType("DOE");
            jsonSettingsDao.findSettingsByParent(optimization.getId(), selectedDoeId).stream()
                    .forEach(ds-> {
                        switch (ds.getXmlName()) {
                            case "numpoints":
                                experimentType.setNumpoints(ds.getPlainValue());
                                return;
                            case "seed":
                                experimentType.setSeed(ds.getPlainValue());
                                return;
                        }
                    });
        }
        List<OptimizationJsonSetting> test = jsonSettingsDao.findSettingsByParent(optimization.getId(), selectedAlgorithmId).stream().collect(toList());
        jsonSettingsDao.findSettingsByParent(optimization.getId(), selectedAlgorithmId).stream()
                .forEach(s -> {
                    switch (s.getXmlName()) {
                        case "nrexpfirstlh":
                            experimentType.setNrexpfirstlh(s.getPlainValue());
                            return;
                        case "seed":
                            experimentType.setSeed(s.getPlainValue());
                            return;
                        case "maxnumeval":
                            experimentType.setMaxnumeval(s.getPlainValue());
                            return;
                        case "tolerance":
                            experimentType.setTolerance(s.getPlainValue());
                            return;
                        case "modeltype":
                            s.getListValue().forEach(ls->{
                                if (ls.isSelected()) {
                                    experimentType.setModeltype(ls.getValue());
                                    if (ls.getXmlName().equals("Krigging")) experimentType.setTheta("0");
                                }});
                            return;
                        case "nugget":
                            experimentType.setNugget(s.getPlainValue());
                            return;
                        case "intensivesearch":
                            if (s.getBoolValue()) experimentType.setIntensivesearch("1");
                            else experimentType.setIntensivesearch("0");
                            return;
                        case "updatefreq":
                            experimentType.setUpdatefreq(s.getPlainValue());
                            return;
                        case "nrsigma":
                            experimentType.setNrsigma(s.getPlainValue());
                            return;
                        case "savelastkrigingmodel":
                            if (s.getBoolValue()) experimentType.setSavelastkrigingmodel("1");
                            else experimentType.setSavelastkrigingmodel("0");
                            return;
                        case "inittemp":
                            experimentType.setInittemp(s.getPlainValue());
                            return;
                        case "annealfactor":
                            experimentType.setAnnealfactor(s.getPlainValue());
                            return;
                        case "stepwidthadjfactor":
                            experimentType.setStepwidthadjfactor(s.getPlainValue());
                            return;
                        case "numcycles":
                            experimentType.setNumcycles(s.getPlainValue());
                            return;
                        case "iterattconst":
                            experimentType.setIterattconst(s.getPlainValue());
                            return;
                        case "avgstopwidth":
                            experimentType.setAvgstopwidth(s.getPlainValue());
                            return;
                        case "maxeval":
                            experimentType.setMaxeval(s.getPlainValue());
                            return;
                        default:
                            throw new IllegalArgumentException("Unexepected enum value: " + s.getType());
                    }
                });
        return experimentType;
    }


        private SequenceType formSequence(OptimizationEntity optimization, Map<OptimusBaseSettingType, OptimusOptimizationBaseSetting> baseSettings) {
            SequenceType sequenceType = new SequenceType();
            CommandsType commandsType = new CommandsType();
            Needed_forType needed_forType =new Needed_forType();
            MakeType makeType = new MakeType();
            MakeCreoType makeCreoType = new MakeCreoType();
            GroupCreoType groupCreoType = new GroupCreoType();
            GroupType groupType = new GroupType();
            TimeoutType timeoutType = new TimeoutType();
            GetType getType = new GetType();

            Integer outVaryingParamsCount = baseSettings.get(OptimusBaseSettingType.NUM_OUTPUT_PARAMS).getCurrentValue();

            List<VarList> var  = IntStream.range(1, outVaryingParamsCount+1)
            .mapToObj(i -> {
                VarList vt = new VarList();
                vt.setVar("out" + i);
                return vt;
            }).collect(toList());



            List<MakeType> lmt = new ArrayList<MakeType>();
            List<MakeCreoType> clmt = new ArrayList<MakeCreoType>();

            Boolean isCreo = false;
            if (isCreo) {
                makeCreoType.setGid("creoPar.txt");
                makeCreoType.setDeck("creoPar.txt");
                makeCreoType.setFile("creoPar.txt");
                clmt.add(makeCreoType);

            }

            if (optimization.getCaeName().equals("ANSYS")) {
                makeType.setGid("uniform.mac");
                makeType.setDeck("uniform.mac");
                makeType.setFile("uniform.mac");
                lmt.add(makeType);
            }
            else {
                makeType.setGid("uniform.py");
                makeType.setDeck("uniform.py");
                makeType.setFile("uniform.py");
                lmt.add(makeType);
            }


            //Group
            List<GroupType> lgt = new ArrayList<GroupType>();
            List<GroupCreoType> clgt = new ArrayList<GroupCreoType>();
            List<TimeoutType> to = new ArrayList<TimeoutType>();
           // List<OptimizationJsonSetting> test = jsonSettingsDao.findRunSettings(optimization.getId()).stream().filter(s -> s.getXmlName().equals("timeout")).collect(toList());
            String timeoutValue = jsonSettingsDao.findRunSettings(optimization.getId()).stream()
                    .filter(s -> s.getXmlName().equals("timeout"))
                    .collect(toList())
                    .get(0)
                    .getPlainValue();



            if (isCreo) {
                groupCreoType.setGid("CREO");
                timeoutType.setSeconds("360");
                if (timeoutValue.equals("")) {
                    timeoutType.setSeconds("10e9");
                }
                else {
                    timeoutType.setSeconds(timeoutValue);

                }
                timeoutType.setSystem("\"%CREO_EXEC_PATH%\" -np 4 -b -i uniform.mac -o .\\programm_output.txt");
                to.add(timeoutType);
                clgt.add(groupCreoType);
            }

            if (optimization.getCaeName().equals("ANSYS")) {
                groupType.setGid("ANSYS");
                if (timeoutValue.equals("")) {
                    timeoutType.setSeconds("10e9");
                    }
                else {
                    timeoutType.setSeconds(timeoutValue);

                }
                timeoutType.setSystem("\"%ANSYS_EXEC_PATH%\" -np 4 -b -i uniform.mac -o .\\programm_output.txt" );
                to.add(timeoutType);
                lgt.add(groupType);
            }
            else {
                groupType.setGid("ABAQUS");
                if (timeoutValue.equals("")) {
                    timeoutType.setSeconds("10e9");
                }
                else {
                    timeoutType.setSeconds(timeoutValue);

                }
                timeoutType.setSystem("\"%ABAQ_EXEC_SCRIPT%\" cae noGUI=uniform.py" );
                to.add(timeoutType);
                lgt.add(groupType);
            }


            //GET

            List<GetType> get  = IntStream.range(1, outVaryingParamsCount+1)
                    .mapToObj(i -> {
                        GetType gout = new GetType();
                        gout.setVar("out" + i);
                        gout.setFile("output.txt");
                        Integer offs=i-1;
                        gout.setOffset(offs);
                        gout.setKey("#1");
                        return gout;
                    }).collect(toList());



            needed_forType.setVar(var);
            commandsType.setNeeded_for(needed_forType);
            commandsType.setMake(lmt);
            commandsType.setCreoMake(clmt);
            commandsType.setCreoGroup(clgt);
            commandsType.setGroup(lgt);
            groupType.setTimeout(to);
            groupCreoType.setTimeout(to);

            commandsType.setGet(get);
            sequenceType.setCommands(commandsType);


        return sequenceType;
    }

    private DecksType formDecks(OptimizationEntity optimization, Map<OptimusBaseSettingType, OptimusOptimizationBaseSetting> baseSettings) {
        DecksType decksType = new DecksType();
        DeckType deckType = new DeckType();
        DeckCreoType deckCreoType = new DeckCreoType();

        List<DeckType> ldt = new ArrayList<DeckType>();
        List<DeckCreoType> cldt = new ArrayList<DeckCreoType>();

        Boolean isCreo = false;
        if (isCreo) {
            deckCreoType.setName("creoPar.txt");
            deckCreoType.setRead("creoPar.txt.tpl");
            deckCreoType.setSubst("$");
            cldt.add(deckCreoType);

        }

        if (optimization.getCaeName().equals("ANSYS")) {
            deckType.setName("uniform.mac");
            deckType.setRead("uniform.mac.tpl");
            deckType.setSubst("$");
            ldt.add(deckType);
        }
        else {
            deckType.setName("uniform.py");
            deckType.setRead("uniform.py.tpl");
            deckType.setSubst("$");
            ldt.add(deckType);
        }

        decksType.setDeck(ldt);
        decksType.setCreoDeck(cldt);
        //decksType.setDeck(decksType);

        return decksType;
    }

}
