package ru.crp.lvms.optimizations.optistruct.controllers;

import org.apache.commons.lang3.tuple.Pair;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.validation.BeanPropertyBindingResult;
import org.springframework.validation.Errors;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.RestController;
import ru.crp.cmls.api.TaskParentType;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.solving.TaskStartingException;
import ru.crp.lvms.model.AccessRights;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.enums.OptimizationCao;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.optimizations.optistruct.tcl.OptistructFemScriptGenerator;
import ru.crp.lvms.optimizations.optistruct.tcl.OptistructOssScriptGenerator;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.validators.task.optimization.OptistructTaskParamsValidator;

import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.nio.file.Path;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import static java.util.Arrays.asList;
import static java.util.stream.Collectors.toMap;
import static ru.crp.cmlb.utils.ZipUtil.createZip;
import static ru.crp.lvms.utils.URLs.OPTISTRUCT_CONFIG_URL;
import static ru.crp.lvms.web.controllers.misc.FileHelper.setHeadersForDownloading;

@RestController
@RequestMapping(value = OPTISTRUCT_CONFIG_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class OptistructConfigController {

    @Autowired
    private OptimizationManager optimizationManager;

    @Autowired
    private OptistructTaskParamsValidator optistructTaskParamsValidator;

    @Autowired
    private OptistructFemScriptGenerator optistructFemScriptGenerator;

    @Autowired
    private OptistructOssScriptGenerator optistructOssScriptGenerator;

    @RequestMapping(method = RequestMethod.GET)
    public void exportScripts(@PathVariable("optimizationId") int optimizationId,
                              HttpServletResponse response,
                              @CurrentUser PortalUser currentUser) throws IOException {
        OptimizationEntity optimization = optimizationManager.checkEntityExistsAndAccesible(optimizationId, AccessRights.view, currentUser.getUser());
        if (optimization.getSolverCaoId() != OptimizationCao.Optistruct.getId()) {
            throw new BadRequestException(String.format("Настройки '%s' недоступны", OptimizationCao.Optistruct));
        }
        TaskEntity task = new TaskEntity();
        task.setParentId(optimizationId);
        task.setSolverName(OptimizationCao.Optistruct.getSolverConfigName());
        task.setParentType(TaskParentType.optimization);
        Errors errors = new BeanPropertyBindingResult(task, task.getClass().getName());
        optistructTaskParamsValidator.validate(task, errors);
        if (errors.hasErrors() && !(errors.getAllErrors().size() == 1 && errors.getAllErrors().get(0).getCode().equals("solvingAlreadyStarted"))) {
            throw new TaskStartingException("Не удалось создать файл", errors);
        }
        List<Pair<Path, String>> settings = asList(
                Pair.of(optistructFemScriptGenerator.generateConfig(optimization), "generate_fem.tcl"),
                Pair.of(optistructOssScriptGenerator.generateConfig(optimization), "oss.tcl"));
        List<Path> files = settings
                .stream()
                .map(pair -> pair.getKey())
                .collect(Collectors.toList());
        Map<String, String> nameReplacements = settings
                .stream()
                .collect(toMap(pair -> pair.getKey().getFileName().toString(), pair -> pair.getValue()));
        setHeadersForDownloading(response, "configs.zip");
        createZip(response.getOutputStream(), files, nameReplacements);
    }
}
