package ru.crp.lvms.optimizations.optistruct.tcl;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import ru.crp.lvms.exceptions.ValidationException;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.entities.tree.TechnologicalConstraintTreeNodeValue;
import ru.crp.lvms.optimizations.optistruct.dao.OptistructLoadStepDao;
import ru.crp.lvms.optimizations.optistruct.dao.OptistructOptimizationAdditionalParametersDao;
import ru.crp.lvms.optimizations.optistruct.entities.*;
import ru.crp.lvms.optimizations.optistruct.enums.OptistructBaseSettingType;
import ru.crp.lvms.optimizations.optistruct.enums.OptistructCollectorType;
import ru.crp.lvms.optimizations.optistruct.enums.OptistructLoadStepCalculationType;
import ru.crp.lvms.optimizations.optistruct.enums.OptistructSettingsType;
import ru.crp.lvms.optimizations.optistruct.managers.OptistructOptimizationSettingsManager;
import ru.crp.lvms.optimizations.optistruct.managers.OptistructTechnologicalConstraintTreeValueManager;
import ru.crp.lvms.optimizations.optistruct.tcl.tc.*;
import ru.crp.lvms.utils.Utils;

import javax.annotation.PostConstruct;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.*;
import java.util.function.BiFunction;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static java.util.Collections.nCopies;
import static java.util.stream.Collectors.joining;
import static java.util.stream.Collectors.toList;

@Component
public class OptistructFemScriptGenerator {
    public static final String DEFAULT_SET_NAME = "DefaultSet";
    private static final int DESMAX = 1;
    private static final int DELTOP = 2;
    private static final int OBJTOL = 3;

    private interface ConfigGenerationFunction extends
            BiFunction<OptimizationEntity, Map<OptistructBaseSettingType, OptistructOptimizationBaseSetting>, String> {
    }

    @Autowired
    private OptistructOptimizationSettingsManager optistructOptimizationSettingsManager;

    @Autowired
    private OptistructLoadStepDao optistructLoadStepDao;

    @Autowired
    private OptistructOptimizationAdditionalParametersDao optistructOptimizationAdditionalParametersDao;

    @Autowired
    private OptistructTechnologicalConstraintTreeValueManager optistructTechnologicalConstraintTreeValueManager;


    @Autowired
    private OptistructTFAndConstraintsGenerator optistructTFAndConstraintsGenerator;

    private List<OptistructTCConverter> tcConverters;

    @PostConstruct
    public void init() {
        tcConverters = new ArrayList<>();
        tcConverters.add(new DefaultTCConverter(optistructTechnologicalConstraintTreeValueManager));
        tcConverters.add(new SymmetryConditionTCConverter());
        tcConverters.add(new CastTCConverter());
        tcConverters.add(new ExtrusionTCConverter());
    }


    public Path generateConfig(OptimizationEntity optimization) throws IOException {
        Path tempConfigFile = Files.createTempFile(null, null);
        Map<OptistructBaseSettingType, OptistructOptimizationBaseSetting> baseSettings =
                optistructOptimizationSettingsManager.findBaseSettings(optimization.getId(), null);
        Stream<ConfigGenerationFunction> funcs = Stream.of(
                this::formStart,
                this::formOptimizationSpace,
                this::formTC,
                this::formLoadSteps,
                optistructTFAndConstraintsGenerator::formTF,
                optistructTFAndConstraintsGenerator::formConstraints,
                this::formAdditionalParameters,
                this::formEnding
        );
        String data = funcs.map(fun -> fun.apply(optimization, baseSettings))
                .filter(s -> !s.isEmpty())
                .collect(Collectors.joining("\n"));
        Files.write(tempConfigFile, data.getBytes());
        return tempConfigFile;
    }

    private String formOptimizationSpace(OptimizationEntity optimization, Map<OptistructBaseSettingType, OptistructOptimizationBaseSetting> baseSettings) {
        List<OptistructSet> oss = optistructOptimizationSettingsManager.findSelectedSets(optimization.getId(), null).stream()
                .filter(OptistructSet::getIsOptimizationSpace)
                .collect(toList());
        if (oss.isEmpty()) {
            throw new ValidationException("Пространство оптимизации не выбрано");
        }
        String optSpaces = oss.stream().map(os -> "\"" + os.getName() + "\"").collect(Collectors.joining(" "));
        Integer componentType = oss.get(0).getComponentTypeId();
        if (oss.stream().anyMatch(os -> !Objects.equals(os.getComponentTypeId(), componentType))) {
            throw new ValidationException("Пространства оптимизации имеют разные типы компонентов");
        }
        String mark = "*createmark properties 1 " + optSpaces;
        String topology = String.format("*topologydesvarcreate 1 \"%s\" 0 0 %s", DEFAULT_SET_NAME, String.valueOf(componentType));
        return mark + "\n" + topology;
    }

    private String formTC(OptimizationEntity optimization, Map<OptistructBaseSettingType, OptistructOptimizationBaseSetting> baseSettings) {
        Integer tcCount = baseSettings.get(OptistructBaseSettingType.NUM_TECH_CONSTRAINTS).getCurrentValue();
        if (tcCount == 0) {
            return "";
        }
        List<OptistructOptimizationTechnologicalConstraint> tcs = optistructOptimizationSettingsManager
                .findOptimizationSettings(optimization.getId(), OptistructSettingsType.TechConstraint, null)
                .stream()
                .limit(tcCount)
                .map(abstr -> (OptistructOptimizationTechnologicalConstraint) abstr)
                .collect(toList());
        OptistructSet set = optistructOptimizationSettingsManager
                .findSelectedSets(optimization.getId(), null)
                .stream()
                .findAny()
                .orElseThrow(() -> new ValidationException("Пространство оптимизации не выбрано"));

        List<String> convertedTCs = new ArrayList<>();
        for (OptistructOptimizationTechnologicalConstraint currentTC : tcs) {
            List<TechnologicalConstraintTreeNodeValue> currentTCValues = optistructTechnologicalConstraintTreeValueManager.findValuesOfTC(currentTC.getId());
            OptistructTCConverter converter = tcConverters.stream()
                    .filter(c -> c.canConvert(currentTC)).findAny().get();
            Stream<String> convertedCurrentTC = converter.convert(currentTC, currentTCValues, tcs, DEFAULT_SET_NAME, convertedTCs);
            convertedTCs.addAll(convertedCurrentTC.collect(toList()));
        }
        return convertedTCs.stream().collect(joining("\n"));
    }

    private String formAdditionalParameters(OptimizationEntity optimization, Map<OptistructBaseSettingType, OptistructOptimizationBaseSetting> baseSettings) {
        Map<Integer, String> ap =
                optistructOptimizationAdditionalParametersDao.findForOptimization(optimization.getId())
                        .stream()
                        .filter(p -> p.getValue() != null)
                        .collect(Collectors.toMap(
                                OptistructOptimizationAdditionalParameter::getParameterId,
                                OptistructOptimizationAdditionalParameter::getValue));
        String desmax = ap.getOrDefault(DESMAX, "0");
        String deltop = ap.getOrDefault(DELTOP, "0");
        String objtol = ap.getOrDefault(OBJTOL, "0");
        return "*createarray 0\n" +
                "*createdoublearray 0\n" +
                "*opticontrolcreate 1 " + desmax + " 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 " + objtol + " 0 0 0 0 1 " + deltop + " 0 0 0 0 0 0 1 0 1 0";
    }

    public String formLoadSteps(OptimizationEntity optimization, Map<OptistructBaseSettingType,
            OptistructOptimizationBaseSetting> baseSettings) {
        Integer lsCount = baseSettings.get(OptistructBaseSettingType.NUM_LOAD_STEPS).getCurrentValue();

        List<OptistructCollector> collectors = optistructOptimizationSettingsManager
                .findCollectorsForOptimizationByType(optimization.getId(), null);

        Integer maxCollectorId = collectors.stream().mapToInt(OptistructCollector::getScriptId).max().getAsInt();

        Map<String, List<OptistructCollector>> collectorsByDataType = collectors.stream()
                .collect(Collectors.groupingBy(OptistructCollector::getDataType));

        Map<Integer, Integer> collectorsById = collectors.stream()
                .collect(Collectors.toMap(OptistructCollector::getId, OptistructCollector::getScriptId));

        Optional<OptistructCollector> inertiaCollector =
                Optional.ofNullable(collectorsByDataType.get(OptistructCollectorType.Inertia.name()))
                        .map(list -> list.get(0));

        Optional<OptistructCollector> modalCollector =
                Optional.ofNullable(collectorsByDataType.get(OptistructCollectorType.Modal.name()))
                        .map(list -> list.get(0));

        Optional<OptistructCollector> constCollector =
                Optional.ofNullable(collectorsByDataType.get(OptistructCollectorType.Fixation.name()))
                        .map(list -> list.get(0));

        return Utils.zipWithIndex(
                optistructLoadStepDao.findForOptimization(optimization.getId()), 1
        ).limit(lsCount).map(loadStepWithIdx -> {

            // Variables used in generation, extracted for shortage
            Integer idx = loadStepWithIdx.getKey();
            Integer collectorIdx = idx + maxCollectorId;
            OptistructOptimizationLoadStep loadStep = loadStepWithIdx.getValue();
            Integer loadsNumber = loadStep.getSelectedLoadIds().size();
            String loadIds = loadStep.getSelectedLoadIds().stream()
                    .map(collectorsById::get)
                    .map(String::valueOf)
                    .collect(Collectors.joining(" "));
            String onesArray = nCopies(loadsNumber, "1").stream().collect(Collectors.joining(" "));

            String loadCollectorCommand = "";
            String loadStepCommand = "";
            if (loadStep.getCalculationType() == OptistructLoadStepCalculationType.STATIC) {

                loadCollectorCommand = "*createentity loadcols name=LOADCOL" + collectorIdx + "\n" +
                        "*setvalue loadcols id=" + collectorIdx + " cardimage=\"LOAD\"\n" +
                        "*attributeupdateint loadcols " + collectorIdx + " 3236 1 0 0 " + loadsNumber + "\n" +
                        "*createdoublearray " + loadsNumber + " " + onesArray + "\n" +
                        "*attributeupdatedoublearray loadcols " + collectorIdx + " 380 1 2 0 1 " + loadsNumber + "\n" +
                        "*createarray " + loadsNumber + " " + loadIds + "\n" +
                        "*attributeupdateentityidarray loadcols " + collectorIdx + " 383 1 2 0 loadcols 1 " + loadsNumber + "\n";

                loadStepCommand = "*createmark loadcols 1 \"spc\" \"LOADCOL" + collectorIdx + "\"" +
                        inertiaCollector.map(collector -> " \"" + collector.getName() + "\"\n").orElse("\n") +
                        "*createmark outputblocks 1\n" +
                        "*createmark groups 1\n" +
                        "*loadstepscreate \"" + loadStep.getName() + "\" " + idx + "\n" +
                        "*attributeupdateint loadsteps " + idx + " 4143 1 1 0 1\n" +
                        "*attributeupdateint loadsteps " + idx + " 4709 1 1 0 1\n" +
                        "*attributeupdateentity loadsteps " + idx + " 4147 1 1 0 loadcols " + collectorIdx + "\n" +
                        "*attributeupdateint loadsteps " + idx + " 9224 1 1 0 0\n" +
                        "*attributeupdatestring loadsteps " + idx + " 4060 1 1 0 \"STATIC\"\n";

                if (loadStep.getIsFixed()) {
                    if (!constCollector.isPresent()) {
                        throw new ValidationException("Коллектор закреплений не найден в .fem файле");
                    }

                    loadStepCommand += "*attributeupdateentity loadsteps " + idx + " 4145 1 1 0 loadcols " + constCollector.get().getScriptId() + "\n";
                }

                loadStepCommand += inertiaCollector.map(collector ->
                        "*attributeupdateentity loadsteps " + loadStep.getScriptId() + " 4149 1 1 0 loadcols " + collector.getScriptId() + "\n"
                ).orElse("");

            } else {
                // Modal calculations has no loads and hence has no load collectors

                if (!modalCollector.isPresent()) {
                    throw new ValidationException("Коллектор с методом расчета для модального анализа не найден в .fem файле");
                }

                loadStepCommand = "*loadstepscreate \"" + loadStep.getName() + "\" " + idx + "\n" +
                        "*attributeupdateint loadsteps " + idx + " 4143 1 1 0 1\n" +
                        "*attributeupdateint loadsteps " + idx + " 4709 1 1 0 3\n" +
                        "*attributeupdateint loadsteps " + idx + " 9224 1 1 0 0\n" +
                        "*attributeupdateentity loadsteps " + idx + " 4966 1 1 0 loadcols " + modalCollector.get().getScriptId() + "\n" +
                        "*attributeupdatestring loadsteps " + idx + " 4060 1 1 0 \"MODES\"\n";
            }

            return loadCollectorCommand + loadStepCommand;
        }).collect(Collectors.joining());
    }

    private String formEnding(OptimizationEntity optimization, Map<OptistructBaseSettingType, OptistructOptimizationBaseSetting> baseSettings) {
        return "*cardcreate \"DENSITY\" \n" +
                "*createmark cards 1 \"DENSITY\" \n" +
                "*attributeupdatestring cards 1 4157 1 2 0 \"OPTI\" \n" +
                "*attributeupdatestring cards 1 4158 1 2 0 \"DES\" \n" +
                "*attributeupdatestring cards 1 4125 1 2 0 \"YES\" \n" +
                "*createstringarray 1 \"CONNECTORS_SKIP\"\n" +
                "*feoutputwithdata \"C:/Program Files/Altair/13.0/templates/feoutput/optistruct/optistruct\" " +
                "\"OUTPUT_FILES_PATH/file_name2.fem\" 1 0 1 1 1";
    }

    private String formStart(OptimizationEntity optimization, Map<OptistructBaseSettingType, OptistructOptimizationBaseSetting> baseSettings) {
        return "*templatefileset \"C:/Program Files/Altair/13.0/templates/feoutput/optistruct/optistruct\"\n" +
                "*createstringarray 8 \"OptiStruct\" \" \" \"ANSA\" \"PATRAN\" \"ASSIGNPROP_BYHMCOMMENTS\" \"LOADCOLS_DISPLAY_SKIP \" \"CONTACTSURF_DISPLAY_SKIP\" \"SYSTCOLS_DISPLAY_SKIP\"\n" +
                "*feinputwithdata2 \"#optistruct\\\\optistruct\" \"OUTPUT_FILES_PATH/file_name.fem\" 0 0 0 0 0 1 5 1 0\n";
    }

}
