package ru.crp.lvms.optimizations.tosca;

import org.springframework.stereotype.Component;
import ru.crp.lvms.optimizations.base.enums.OptimizationCae;
import ru.crp.lvms.optimizations.tosca.entities.ToscaSet;
import ru.crp.lvms.optimizations.tosca.enums.SetType;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Stream;

import static java.util.stream.Collectors.toList;

@Component
public class ToscaSetsParser {
    private static final String START_SET_KEY = "CMBLOCK";
    private static final int SET_NAME_INDEX = 1;
    private static final int SET_TYPE_INDEX = 2;

    public static final Pattern ABAQUS_ELEMENTS_SET_PATTERN = Pattern.compile("^.*ELSET,\\s+ELSET=(\\S+).*$");
    public static final Pattern ABAQUS_NODES_SET_PATTERN = Pattern.compile("^.*NSET,\\s+NSET=(\\S+).*$");

    public List<ToscaSet> extractSets(int optimizationId, OptimizationCae caeSolver, Path pathToFile) throws IOException {
        if (!Files.isRegularFile(pathToFile)) {
            throw new IllegalArgumentException(String.format("File %s is not regular file", pathToFile));
        }
        if (caeSolver == OptimizationCae.ANSYS) {
            return extractSetsFromAnsysModel(optimizationId, pathToFile);
        } else if (caeSolver == OptimizationCae.ABAQUS) {
            return extractSetsFromAbaqusModel(optimizationId, pathToFile);
        } else {
            throw new IllegalArgumentException(String.format("Неподдерживаемый тип CAE Решателя '%s'", caeSolver));
        }
    }


    private List<ToscaSet> extractSetsFromAnsysModel(int optimizationId, Path pathToFile) throws IOException {
        try (Stream<String> lines = Files.lines(pathToFile)) {
            return lines.filter(line -> line.startsWith(START_SET_KEY))
                    .map(line -> line.split(","))
                    .peek(splitted -> {
                        if (splitted.length < 3) {
                            throw new IllegalArgumentException(String.format("Unexpected file format. For string %s", Arrays.toString(splitted)));
                        }
                    })
                    .map(splitted -> new ToscaSet(optimizationId, SetType.fromCAOFileValue(splitted[SET_TYPE_INDEX]), splitted[SET_NAME_INDEX].trim()))
                    .collect(toList());
        }
    }


    private List<ToscaSet> extractSetsFromAbaqusModel(int optimizationId, Path pathToFile) throws IOException {
        List<ToscaSet> sets = new ArrayList<>();
        try (Stream<String> lines = Files.lines(pathToFile)) {
            lines.forEach(line -> {
                Matcher elementsSetMatcher = ABAQUS_ELEMENTS_SET_PATTERN.matcher(line);
                if (elementsSetMatcher.find()) {
                    String elementSetName = elementsSetMatcher.group(1);
                    if (elementSetName.endsWith(",")) {
                        elementSetName = elementSetName.substring(0, elementSetName.length() - 1);
                    }
                    sets.add(new ToscaSet(optimizationId, SetType.Elements, elementSetName));
                }
                Matcher nodesSetMatcher = ABAQUS_NODES_SET_PATTERN.matcher(line);
                if (nodesSetMatcher.find()) {
                    String nodeSetName = nodesSetMatcher.group(1);
                    if (nodeSetName.endsWith(",")) {
                        nodeSetName = nodeSetName.substring(0, nodeSetName.length() - 1);
                    }
                    sets.add(new ToscaSet(optimizationId, SetType.Nodes, nodeSetName));
                }
            });
        }
        return sets;
    }
}
