package ru.crp.lvms.optimizations.tosca.config;

import org.apache.commons.lang3.tuple.Pair;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import ru.crp.lvms.model.dao.FileDao;
import ru.crp.lvms.model.entities.FileEntityTable;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.optimizations.base.dao.OptimizationJsonSettingsDao;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationJsonSetting;
import ru.crp.lvms.optimizations.base.entities.OptimizationListSetting;
import ru.crp.lvms.optimizations.misc.AbstractOptimizationSettings;
import ru.crp.lvms.optimizations.tosca.dao.ToscaOptimizationSetsDao;
import ru.crp.lvms.optimizations.tosca.entities.*;
import ru.crp.lvms.optimizations.tosca.enums.*;
import ru.crp.lvms.optimizations.tosca.managers.ToscaOptimizationSettingsManager;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.*;
import java.util.function.BiFunction;
import java.util.function.Predicate;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import static java.util.Arrays.asList;
import static java.util.stream.Collectors.joining;
import static java.util.stream.Collectors.toList;
import static ru.crp.lvms.optimizations.tosca.managers.ToscaDefaultParamsCreator.FORMAT_SETTING_XML_NAME;
import static ru.crp.lvms.optimizations.tosca.managers.ToscaDefaultParamsCreator.MULTIPLE_VALUES_SEPARATOR;
import static ru.crp.lvms.utils.Utils.mapWithIndex;
import static ru.crp.lvms.utils.Utils.zipWithIndex;

@Component
public class ToscaConfigGenerator {

    private static final String FEM_INPUT_ID_PREFIX = "CAEFile";
    private static final String DV_ID_PREFIX = "CAEFile";
    private static final String DRESP_CONSTR_ID_PREFIX = "ConstrDresp";
    private static final String DRESP_TF_ID_PREFIX = "TFDresp";
    private static final String CONSTR_ID_PREFIX = "Constr";
    private static final String OPTIMIZE_ID_PREFIX = "Optimize";
    private static final String OBJ_FUNC_ID_PREFIX = "TFObjFunc";
    private static final String OPT_PARAM_ID_PREFIX = "OptParams";
    private static final String STOP_ID_PREFIX = "Stop";
    private static final String MAX_ITER_PARAM_NAME = "ITER_MAX";

    public static final Set<String> SKIPPED_OPT_PARAMS = new HashSet<>(Collections.singletonList(MAX_ITER_PARAM_NAME));

    @Autowired
    private FileDao fileDao;

    @Autowired
    private ToscaOptimizationSettingsManager toscaOptimizationSettingsManager;

    @Autowired
    private OptimizationJsonSettingsDao optimizationJsonSettingsDao;

    @Autowired
    private ToscaOptimizationSetsDao toscaOptimizationSetsDao;

    @Autowired
    private ToscaTechnologicalConstraintsGenerator toscaTechnologicalConstraintsGenerator;


    public Path generateConfig(OptimizationEntity optimization) throws IOException {
        if (toscaOptimizationSettingsManager.findSelectedOptimizationMethodId(optimization.getId()) == null) {
            throw new InternalError("Optimization method not specified");
        }
        Map<ToscaBaseSettingType, ToscaOptimizationBaseSetting> baseSettings =
                toscaOptimizationSettingsManager.findBaseSettings(optimization.getId(), null);
        Path tempConfigFile = Files.createTempFile(null, null);
        Stream<BiFunction<OptimizationEntity, Map<ToscaBaseSettingType, ToscaOptimizationBaseSetting>, String>> funcs = Stream.of(
                this::formFemInput,
                this::formOptSpace,
                this::formDvCon,
                this::formDrespForConstr,
                this::formDrespForTF,
                this::formObjFunc,
                this::formConstraint,
                this::formOptimize,
                this::formOptParam,
                this::formStop,
                this::formSmooth
        );
        String data = funcs.map(fun -> fun.apply(optimization, baseSettings)).collect(Collectors.joining("\n"));
        Files.write(tempConfigFile, data.getBytes());
        return tempConfigFile;
    }


    private String formFemInput(OptimizationEntity optimization, Map<ToscaBaseSettingType, ToscaOptimizationBaseSetting> baseSettings) {
        List<String> names = fileDao.findOptimizationFiles(optimization.getId(), FileType.CAEFile.getId())
                .stream()
                .map(FileEntityTable::getName)
                .collect(toList());
        return zipWithIndex(names)
                .map(pair -> new Command("FEM_INPUT")
                        .addParameter("ID_NAME", FEM_INPUT_ID_PREFIX + pair.getKey())
                        .addParameter("FILE", pair.getValue())
                        .toString()
                )
                .collect(joining("\n"));
    }


    private String formOptSpace(OptimizationEntity optimization, Map<ToscaBaseSettingType, ToscaOptimizationBaseSetting> baseSettings) {
        OptimizationTypeParameters parameters = OptimizationTypeParameters.valueOf(optimization);
        String elgroup = toscaOptimizationSettingsManager
                .findOptimizationSettings(optimization.getId(), ToscaSettingsType.Set, null)
                .stream()
                .filter(set -> ((ToscaSet) set).getIsSelected())
                .map(abstr -> (ToscaSet) abstr)
                .filter(ToscaSet::getIsOptimizationSpace)
                .filter(parameters.getPredicate())
                .map(ToscaSet::getName)
                .findAny()
                .orElse("ALL_ELEMENTS");
        String ndgroup = toscaOptimizationSettingsManager
                .findOptimizationSettings(optimization.getId(), ToscaSettingsType.Set, null)
                .stream()
                .filter(set -> ((ToscaSet) set).getIsSelected())
                .map(abstr -> (ToscaSet) abstr)
                .filter(ToscaSet::getIsOptimizationSpace)
                .filter(parameters.getPredicate())
                .map(ToscaSet::getName)
                .findAny()
                .orElse("ALL_NODES");
        Command optSpaceCommand = new Command("DV" + parameters.getDvName());
        optSpaceCommand.addParameter("ID_NAME", DV_ID_PREFIX);
        switch (parameters.getGroupName()) {
            case "ND_GROUP":
                optSpaceCommand.addParameter(parameters.getGroupName(), ndgroup);
                break;
            case "EL_GROUP":
                optSpaceCommand.addParameter(parameters.getGroupName(), elgroup);
                break;
            default:
                throw new IllegalArgumentException(String.format("Unexpected node or element type '%s'", parameters.getGroupName()));
        }
        return optSpaceCommand.toString();
    }


    public String formDvCon(OptimizationEntity optimization, Map<ToscaBaseSettingType, ToscaOptimizationBaseSetting> baseSettings) {
        OptimizationTypeParameters typeParameters = OptimizationTypeParameters.valueOf(optimization);
        List<ToscaOptimizationTechnologicalConstraint> constraints = getUsedTC(optimization, baseSettings);
        List<Command> commands = zipWithIndex(constraints)
                .flatMap(pair -> {
                    Integer idx = pair.getKey();
                    ToscaOptimizationTechnologicalConstraint tc = pair.getValue();
                    String elspace = Optional.ofNullable(tc.getSetId())
                            .map(toscaOptimizationSetsDao::findOne)
                            .filter(typeParameters.getPredicate())
                            .map(ToscaSet::getName)
                            .orElse("ALL_ELEMENTS");
                    String ndspace = Optional.ofNullable(tc.getSetId())
                            .map(toscaOptimizationSetsDao::findOne)
                            .filter(typeParameters.getPredicate())
                            .map(ToscaSet::getName)
                            .orElse("ALL_NODES");
                    switch (typeParameters.getGroupName()) {
                        case "ND_GROUP":
                            return toscaTechnologicalConstraintsGenerator.formCommandsForTC(typeParameters, tc, idx, ndspace);
                        case "EL_GROUP":
                            return toscaTechnologicalConstraintsGenerator.formCommandsForTC(typeParameters, tc, idx, elspace);
                        default:
                            throw new IllegalArgumentException(String.format("Unexpected node or element type '%s'", typeParameters.getGroupName()));
                    }
                })
                .collect(toList());
        return mapWithIndex(commands, 1, i -> commands.get(i - 1).toString())
                .collect(joining("\n"));
    }


    private List<ToscaOptimizationTechnologicalConstraint> getUsedTC(OptimizationEntity optimization, Map<ToscaBaseSettingType, ToscaOptimizationBaseSetting> baseSettings) {
        List<? extends AbstractOptimizationSettings> settingsById =
                toscaOptimizationSettingsManager.findOptimizationSettings(optimization.getId(),
                        ToscaSettingsType.TechnologicalConstraint, null);
        return settingsById
                .stream()
                .limit(baseSettings.get(ToscaBaseSettingType.NUM_TECH_CONSTRAINTS).getCurrentValue())
                .map(abstr -> (ToscaOptimizationTechnologicalConstraint) abstr)
                .collect(toList());
    }


    private String formDresp(OptimizationEntity optimization, String namePrefix,
                             Map<ToscaBaseSettingType, ToscaOptimizationBaseSetting> baseSettings,
                             ToscaSettingsType type) {
        long limit;
        switch (type) {
            case TargetFunc:
                limit = baseSettings.get(ToscaBaseSettingType.NUM_TARGET_FUNCS).getCurrentValue();
                break;
            case Constraint:
                limit = baseSettings.get(ToscaBaseSettingType.NUM_CONSTRAINTS).getCurrentValue();
                break;
            default:
                throw new IllegalArgumentException(String.format("Unexpected settings type '%s'", type));
        }
        List<? extends AbstractOptimizationSettings> entities =
                toscaOptimizationSettingsManager
                        .findOptimizationSettings(optimization.getId(), type, null);
        List<ToscaTargetFunction> tfs = entities.stream()
                .limit(limit)
                // constrains extend target funcstions so cast is ok
                .map(abstr -> (ToscaTargetFunction) abstr)
                .collect(toList());
        return zipWithIndex(tfs)
                .map(pair -> {
                    ToscaTargetFunction c = pair.getValue();
                    List<Pair<String, String>> params = new ArrayList<>(asList(Pair.of("DEF_TYPE", "SYSTEM"),
                            Pair.of("TYPE", c.getDrespType())));

                    String elGroup = Optional.ofNullable(c.getElementsGroupId())
                            .map(toscaOptimizationSetsDao::findOne)
                            .filter(set -> set.getSetTypeId() == SetType.Elements.getId())
                            .map(ToscaSet::getName)
                            .orElse("ALL_ELEMENTS");

                    String ndGroup = Optional.ofNullable(c.getNodesGroupId())
                            .map(toscaOptimizationSetsDao::findOne)
                            .filter(set -> set.getSetTypeId() == SetType.Nodes.getId())
                            .map(ToscaSet::getName)
                            .orElse("ALL_NODES");

                    switch (NodeOrElem.valueOf(c.getNodeOrElementId())) {
                        case ELEMENT:
                            params.add(Pair.of("EL_GROUP", elGroup));
                            break;
                        case NODE:
                            params.add(Pair.of("ND_GROUP", ndGroup));
                            break;
                        case NODE_AND_ELEMENT:
                            params.add(Pair.of("EL_GROUP", elGroup));
                            params.add(Pair.of("ND_GROUP", ndGroup));
                            break;
                        case NODE_OR_ELEM:
                            if (!ndGroup.equals("ALL_NODES")) {
                                params.add(Pair.of("ND_GROUP", ndGroup));
                            }
                            else {
                                params.add(Pair.of("EL_GROUP", elGroup));
                            }
                            break;
                        case SCALAR:
                            params.add(Pair.of("EL_GROUP", c.getScalar()));
                            break;
                        case NONE:
                            break;
                        default:
                            throw new IllegalArgumentException(String.format("Unexpected node or element type '%s'", NodeOrElem.valueOf(c.getNodeOrElementId())));
                    }

                    return new Command("DRESP", params).addParameter("ID_NAME", namePrefix + pair.getKey()).toString();
                }).collect(joining("\n"));
    }


    private String formDrespForConstr(OptimizationEntity optimization,
                                      Map<ToscaBaseSettingType, ToscaOptimizationBaseSetting> baseSettings) {
        return formDresp(optimization, DRESP_CONSTR_ID_PREFIX, baseSettings, ToscaSettingsType.Constraint);
    }


    private String formDrespForTF(OptimizationEntity optimization,
                                  Map<ToscaBaseSettingType, ToscaOptimizationBaseSetting> baseSettings) {
        return formDresp(optimization, DRESP_TF_ID_PREFIX, baseSettings, ToscaSettingsType.TargetFunc);
    }


    private String formObjFunc(OptimizationEntity optimization,
                               Map<ToscaBaseSettingType, ToscaOptimizationBaseSetting> baseSettings) {
        String target = toscaOptimizationSettingsManager
                .findOptimizationSettings(optimization.getId(), ToscaSettingsType.TargetFunc, null)
                .stream()
                .limit(baseSettings.get(ToscaBaseSettingType.NUM_TARGET_FUNCS).getCurrentValue())
                .findAny()
                .map(abstr -> (ToscaTargetFunction) abstr)
                .map(ToscaTargetFunction::getTypeId)
                .map(ToscaTargetFunctionType::valueOf)
                .map(ToscaTargetFunctionType::getName)
                .orElse("MIN");
        Command command = new Command("OBJ_FUNC").addParameter("ID_NAME", OBJ_FUNC_ID_PREFIX);
        command.addParameter("DRESP", DRESP_TF_ID_PREFIX + 0);
        command.addParameter("TARGET", target);
        return command.toString();
    }


    private String formConstraint(OptimizationEntity optimization,
                                  Map<ToscaBaseSettingType, ToscaOptimizationBaseSetting> baseSettings) {
        List<ToscaConstraint> constraints =
                (List<ToscaConstraint>) toscaOptimizationSettingsManager
                        .findOptimizationSettings(optimization.getId(), ToscaSettingsType.Constraint, null)
                        .stream().limit(baseSettings.get(ToscaBaseSettingType.NUM_CONSTRAINTS).getCurrentValue())
                        .collect(toList());
        return mapWithIndex(constraints, 0, i -> {
            Command command = new Command("CONSTRAINT")
                    .addParameter("ID_NAME", CONSTR_ID_PREFIX + i)
                    .addParameter("DRESP", DRESP_CONSTR_ID_PREFIX + i);
            ToscaConstraint c = constraints.get(i);
            String value = c.getValue();
            if (c.getIsAbsolute()) {
                command = command.addParameter("MAGNITUDE", "ABS");
            } else {
                command = command.addParameter("MAGNITUDE", "REL");
                if (value != null) {
                    value = (Double.parseDouble(value) * 0.01) + "";
                }
            }
            switch (ToscaConstraintType.valueOf(c.getTypeId())) {
                case EQUALS:
                    command = command.addParameter("EQ_VALUE", value);
                    break;
                case LESS_OR_EQUAL:
                    command = command.addParameter("LE_VALUE", value);
                    break;
                case GREATER_OR_EQUAL:
                    command = command.addParameter("GE_VALUE", value);
                    break;
                default:
                    throw new IllegalArgumentException(String.format("Unexpected tosca constraint type '%s'", ToscaConstraintType.valueOf(c.getTypeId())));
            }
            return command.toString();
        }).collect(joining("\n"));
    }


    private String formOptimize(OptimizationEntity optimization,
                                Map<ToscaBaseSettingType, ToscaOptimizationBaseSetting> baseSettings) {
        Command command = new Command("OPTIMIZE")
                .addParameter("ID_NAME", OPTIMIZE_ID_PREFIX)
                .addParameter("OBJ_FUNC", OBJ_FUNC_ID_PREFIX)
                .addParameter("DV", DV_ID_PREFIX);
        List<Pair<String, String>> tcParams = toscaTechnologicalConstraintsGenerator.formOptimize(optimization, baseSettings);
        List<? extends AbstractOptimizationSettings> cSettings = toscaOptimizationSettingsManager
                .findOptimizationSettings(optimization.getId(), ToscaSettingsType.Constraint, null)
                .stream().limit(baseSettings.get(ToscaBaseSettingType.NUM_CONSTRAINTS).getCurrentValue()).
                        collect(Collectors.toList());
        List<Pair<String, String>> cParams = IntStream.range(0, cSettings.size())
                .mapToObj(i -> Pair.of("CONSTRAINT", CONSTR_ID_PREFIX + i)).collect(toList());
        String strategy = ToscaJsonParametersParents.valueOf(
                toscaOptimizationSettingsManager.findSelectedOptimizationMethodId(optimization.getId())
        ).getName();
        return command.addParameters(tcParams).addParameters(cParams).addParameter("STRATEGY", strategy).toString();
    }


    private String formOptParam(OptimizationEntity optimization,
                                Map<ToscaBaseSettingType, ToscaOptimizationBaseSetting> baseSettings) {
        Command command = new Command("OPT_PARAM")
                .addParameter("ID_NAME", OPT_PARAM_ID_PREFIX)
                .addParameter("OPTIMIZE", OPTIMIZE_ID_PREFIX);
        ToscaJsonParametersParents selectedOptimizationMethod = ToscaJsonParametersParents.valueOf(
                toscaOptimizationSettingsManager.findSelectedOptimizationMethodId(optimization.getId()));

        Predicate<OptimizationJsonSetting> predicate = param -> !SKIPPED_OPT_PARAMS.contains(param.getXmlName());

        Integer parentId = selectedOptimizationMethod.getId();

        List<OptimizationJsonSetting> settings =
                optimizationJsonSettingsDao.findSettingsByParent(optimization.getId(), parentId);
        List<Pair<String, String>> args =
                settings.stream()
                        // we don't need to add parameters with default values to generated file (STOP_CRITERION_ITER is exception)
                        .filter(s -> {
                            if (selectedOptimizationMethod == ToscaJsonParametersParents.TopoSensitivity &&
                                    "STOP_CRITERION_ITER".equals(s.getXmlName())) {
                                return true;
                            }
                            return isValueDifferFromDefault(s);
                        })
                        .filter(predicate)
                        .flatMap(v -> extractJsonValue(v, false))
                        .collect(toList());
        return command.addParameters(args).toString();
    }

    private boolean isValueDifferFromDefault(OptimizationJsonSetting setting) {
        switch (setting.getType()) {
            case list:
                return !setting.getListValue().stream().filter(OptimizationListSetting::isSelected).findFirst()
                        .map(selected -> Objects.equals(selected, setting.getDefaultValue()))
                        .orElse(true);
            case composite:
                return setting.getCompositeValue().stream()
                        .map(this::isValueDifferFromDefault).reduce(false, (a, b) -> a || b);
            default:
                return !Objects.equals(setting.getValue(), setting.getDefaultValue());
        }
    }

    private Stream<Pair<String, String>> extractJsonValue(OptimizationJsonSetting setting, boolean isCompositeParamChild) {
        switch (setting.getType()) {
            case plain:
                if (setting.getPlainValue() == null || setting.getPlainValue().isEmpty()) {
                    return isCompositeParamChild ? Stream.of(Pair.of(setting.getXmlName(), "0")) : Stream.empty();
                } else {
                    return Stream.of(setting.getPlainValue().split(MULTIPLE_VALUES_SEPARATOR))
                            .map(value -> Pair.of(setting.getXmlName(), value));
                }
            case bool:
                Boolean boolValue = setting.getBoolValue();
                if (boolValue == null) {
                    boolValue = false;
                }
                return Stream.of(Pair.of(setting.getXmlName(), Objects.toString(boolValue)));
            case list:
                if (setting.getIsMultipleSelectableInList()) {
                    return setting.getListValue().stream()
                            .filter(OptimizationListSetting::isSelected)
                            .map(OptimizationListSetting::getXmlName)
                            .map(xmlName -> Pair.of(setting.getXmlName(), xmlName));
                } else {
                    Optional<OptimizationListSetting> selectedSetting = setting.getListValue().stream()
                            .filter(OptimizationListSetting::isSelected)
                            .findFirst();
                    if (!selectedSetting.isPresent()) {
                        return Stream.of(Pair.of(setting.getXmlName(), ""));
                    }
                    if (!selectedSetting.get().isCustom()) {
                        return Stream.of(Pair.of(selectedSetting.get().getXmlName(), selectedSetting.get().getName()));
                    }
                    String selectedValue = selectedSetting.get().getValue();
                    return Stream.of(Pair.of(
                            selectedSetting.get().getXmlName(),
                            selectedValue == null ? "" : selectedValue)
                    );
                }
            case composite:
                String value = setting.getCompositeValue().stream()
                        .flatMap(v -> extractJsonValue(v, true))
                        .map(Pair::getValue)
                        .map(val -> val == null ? "0" : val)
                        .collect(Collectors.joining(","));
                return Stream.of(Pair.of(setting.getXmlName(), value));
            default:
                throw new IllegalArgumentException("Unexepected enum value: " + setting.getType());
        }
    }


    private String formStop(OptimizationEntity optimization, Map<ToscaBaseSettingType, ToscaOptimizationBaseSetting> baseSettings) {
        Integer parentId = ToscaJsonParametersParents.valueOf(
                toscaOptimizationSettingsManager.findSelectedOptimizationMethodId(optimization.getId())
        ).getId();
        List<OptimizationJsonSetting> settings =
                optimizationJsonSettingsDao.findSettingsByParent(optimization.getId(), parentId);
        OptimizationJsonSetting iterMax =
                settings.stream().filter(setting -> Objects.equals(setting.getXmlName(), MAX_ITER_PARAM_NAME))
                        .findAny().get();
        return new Command("STOP")
                .addParameter("ID_NAME", STOP_ID_PREFIX)
                .addParameter("ITER_MAX", iterMax.getPlainValue()).toString();
    }

    private String formSmooth(OptimizationEntity optimization, Map<ToscaBaseSettingType, ToscaOptimizationBaseSetting> baseSettings) {
        Predicate<OptimizationJsonSetting> predicate = (anything) -> true;
        List<OptimizationJsonSetting> settings =
                optimizationJsonSettingsDao.findSettingsByParent(optimization.getId(),
                        ToscaJsonParametersParents.PostprocessingParameters.getId());
        List<Pair<String, String>> args = extractParametersFromJson(settings, predicate);
//        if (settings.stream().anyMatch(setting -> Objects.equals(setting.getXmlName(), "GROUP_INCLUDE"))) {
//            args.add(Pair.of("GROUP_EXCLUDE", "ALL_ELEMENTS"));
//        }
        return new Command("SMOOTH")
                .addParameter("ID_NAME", "mySmooth")
                .addParameters(args).toString();
    }

    private List<Pair<String, String>> extractParametersFromJson(List<OptimizationJsonSetting> settings,
                                                                 Predicate<OptimizationJsonSetting> predicate) {
        return settings.stream()
                .filter(predicate)
                .flatMap(setting -> {
                    switch (setting.getType()) {
                        case plain:
                            if (setting.getPlainValue() == null || setting.getPlainValue().trim().isEmpty()) {
                                return Stream.empty();
                            } else {
                                return Stream.of(setting.getPlainValue().split(MULTIPLE_VALUES_SEPARATOR))
                                        .map(value -> Pair.of(setting.getXmlName(), value));
                            }
                        case bool:
                            return Stream.of(Pair.of(setting.getXmlName(), setting.getBoolValue() + ""));
                        case list:
                            if (setting.getIsMultipleSelectableInList()) {
                                return setting.getListValue().stream()
                                        .filter(OptimizationListSetting::isSelected)
                                        .map(OptimizationListSetting::getXmlName)
                                        .map(xmlName -> Pair.of(setting.getXmlName(), xmlName));
                            } else {
                                Optional<String> settingName = setting.getListValue().stream()
                                        .filter(OptimizationListSetting::isSelected)
                                        .map(OptimizationListSetting::getXmlName)
                                        .findFirst();
                                return settingName.isPresent() ? Stream.of(Pair.of(setting.getXmlName(), settingName.get())) : Stream.empty();
                            }
                        case composite:
                            if (Objects.equals(setting.getParentId(), ToscaJsonParametersParents.PostprocessingParameters.getId()) && setting.getXmlName().equals(FORMAT_SETTING_XML_NAME)) {
                                return extractFormatParams(setting);
                            }
                            String value = setting.getCompositeValue().stream()
                                    .flatMap(v -> extractJsonValue(v, true))
                                    .map(Pair::getValue)
                                    .map(val -> val == null ? "0" : val)
                                    .collect(Collectors.joining(","));
                            return Stream.of(Pair.of(setting.getXmlName(), value));
                        default:
                            throw new IllegalArgumentException("Unexepected enum value: " + setting.getType());
                    }
                }).collect(toList());
    }

    private Stream<Pair<String, String>> extractFormatParams(OptimizationJsonSetting setting) {
        return setting.getCompositeValue().stream()
                .flatMap(v -> extractJsonValue(v, true))
                .filter(p -> p.getValue().equals("true"))
                .map(Pair::getKey)
                .flatMap(xmlName -> Stream.of(xmlName.split(MULTIPLE_VALUES_SEPARATOR)))
                .map(format -> Pair.of(setting.getXmlName(), format))
                .distinct();
    }
}
