package ru.crp.lvms.pdf.reports;

import org.apache.fop.apps.FOPException;
import org.apache.fop.apps.Fop;
import org.apache.fop.apps.FopFactory;
import org.apache.fop.apps.MimeConstants;
import org.jetbrains.annotations.NotNull;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartUtilities;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.annotations.XYShapeAnnotation;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import org.xml.sax.SAXException;
import ru.crp.lvms.chart.ChartsSet;
import ru.crp.lvms.chart.LineChart;
import ru.crp.lvms.exceptions.ReportGenerationException;
import ru.crp.lvms.pdf.reports.vo.ReportInnerECParameters;
import ru.crp.lvms.pdf.reports.vo.ReportInnerSimulationParameters;
import ru.crp.lvms.utils.Mapper;

import javax.annotation.PostConstruct;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.sax.SAXResult;
import javax.xml.transform.stream.StreamSource;
import java.awt.*;
import java.awt.geom.Rectangle2D;
import java.io.*;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;

@Component
public class ReportGenerator {

    private static final int CHART_WIDTH = 640;
    private static final int CHART_HEIGHT = 480;

    private FopFactory fopFactory;

    private TransformerFactory transformerFactory;

    @Value("${reportsConfigFile}")
    private String stringPathToConfig;

    @Value("${ecReportsTemplateFile}")
    private String stringPathToEcTemplate;

    private File ecTemplateFile;

    @Value("${simReportsTemplateFile}")
    private String stringPathToSimTemplate;

    private File simTemplateFile;

    @PostConstruct
    private void init() throws IOException, SAXException {
        transformerFactory = TransformerFactory.newInstance();
        Path pathToConfig = Paths.get(stringPathToConfig);
        Path pathToEcTemplate = Paths.get(stringPathToEcTemplate);
        Path pathToSimTemplate = Paths.get(stringPathToSimTemplate);
        if (Files.notExists(pathToConfig) || Files.notExists(pathToEcTemplate) || Files.notExists(pathToSimTemplate)) {
            throw new ReportGenerationException("Файлы шаблонов для отчетов не найдены");
        }

        fopFactory = FopFactory.newInstance(pathToConfig.toFile());
        ecTemplateFile = pathToEcTemplate.toFile();
        simTemplateFile = pathToSimTemplate.toFile();
    }

    public Path generateReport(ECReportInput ecReportInput) {
        ReportInnerECParameters reportInnerEcParameters =
                Mapper.map(ecReportInput, ReportInnerECParameters.class);
        String xmlData;
        try {
            xmlData = parametersToXml(reportInnerEcParameters, ReportInnerECParameters.class);
            return writeDataToFile(xmlData, ecTemplateFile);
        } catch (JAXBException | IOException | FOPException | TransformerException e) {
            throw new ReportGenerationException("Ошибка при генерации отчета", e);
        }
    }

    public Path generateReport(SimulationReportInput simulationReportInput) {
        ReportInnerSimulationParameters reportInnerSimulationParameters =
                Mapper.map(simulationReportInput, ReportInnerSimulationParameters.class);
        String xmlData;
        try {
            xmlData = parametersToXml(reportInnerSimulationParameters, ReportInnerSimulationParameters.class);
            return writeDataToFile(xmlData, simTemplateFile);
        } catch (JAXBException | IOException | FOPException | TransformerException e) {
            throw new ReportGenerationException("Ошибка при генерации отчета", e);
        }
    }

    public Path generatePlot(ChartsSet curve) throws IOException {
        XYSeries data = new XYSeries(curve.getTitle());
        LineChart resultChart = (LineChart) curve.getCharts().get(0);
        resultChart.getPoints().forEach(p -> data.add(p.getX(), p.getY()));
        JFreeChart chart = ChartFactory.createXYLineChart(curve.getTitle(), curve.getxAxisLabel() + ", " + curve.getxDimension(),
                curve.getyAxisLabel() + ", " + curve.getyDimension(), new XYSeriesCollection(data), PlotOrientation.VERTICAL, false, false, false);
        XYPlot plot = (XYPlot) chart.getPlot();
        double coefY = (data.getMaxY() - data.getMinY()) / (data.getMaxX() - data.getMinX());
        XYShapeAnnotation annotation = new XYShapeAnnotation(
                new Rectangle2D.Double(curve.getHighlightPoint().getX() - 2, curve.getHighlightPoint().getY() - 2 * coefY, 4, 4 * coefY),
                new BasicStroke(2.0f), Color.red);
        plot.getRendererForDataset(plot.getDataset(0)).setSeriesPaint(0, Color.blue);
        plot.setBackgroundPaint(Color.white);
        plot.setDomainGridlinePaint(Color.darkGray);
        plot.setRangeGridlinePaint(Color.darkGray);
        plot.addAnnotation(annotation);
        Path output = Files.createTempFile(null, ".png");
        ChartUtilities.saveChartAsPNG(output.toFile(), chart, CHART_WIDTH, CHART_HEIGHT);
        return output;
    }

    @NotNull
    private Path writeDataToFile(String xmlData, File template) throws FOPException, TransformerException, IOException {
        // Replacing null's with special whitespace character
        xmlData = xmlData.replace("null", "&#160;");
        Path tempReportFile = Files.createTempFile(null, null);
        OutputStream out = new BufferedOutputStream(new FileOutputStream(tempReportFile.toFile()));
        Fop fop = fopFactory.newFop(MimeConstants.MIME_PDF, out);
        Source templateSource = new StreamSource(template);
        Source xmlDataSource = new StreamSource(new StringReader(xmlData));
        Result output = new SAXResult(fop.getDefaultHandler());
        transformerFactory.newTransformer(templateSource).transform(xmlDataSource, output);
        out.flush();
        return tempReportFile;
    }

    private String parametersToXml(Object params, Class classOfObject) throws JAXBException {
        JAXBContext jaxbContext = JAXBContext.newInstance(classOfObject);
        Marshaller jaxbMarshaller = jaxbContext.createMarshaller();
        jaxbMarshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true);
        StringWriter stringWriter = new StringWriter();
        jaxbMarshaller.marshal(params, stringWriter);
        return stringWriter.toString();
    }
}
