package ru.crp.lvms.tasks.fileGenerators;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import ru.crp.cmlb.model.entities.FileData;
import ru.crp.cmlb.model.entities.ResourcePreparationException;
import ru.crp.cmls.api.TaskParentType;
import ru.crp.cmls.calculations.solvers.xml.config.DataType;
import ru.crp.lvms.exceptions.solving.TaskStartingException;
import ru.crp.lvms.managers.CompositeMaterialManager;
import ru.crp.lvms.model.entities.effective.characteristics.ECDefectEntity;
import ru.crp.lvms.model.entities.effective.characteristics.ECEntity;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.model.entities.technology.TechnologyType;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

@Component
public class ECInputFileGenerator extends BaseInputFileGenerator {
    private static final String EC_DEFAULT_PARAMS = "1.50 1.50 2.50 0 0.05 0.06 0.8 0.5 9";

    @Autowired
    private CompositeMaterialManager compositeMaterialManager;

    @Value("${ecMacrosesDir}")
    private String pathToECMacrosesStr;

    @Value("${ec3CMMacrosesDir}")
    private String pathToECMacrosesFor3CMMaterialsStr;


    public List<FileData> getInputFilesForEffectiveCharacteristics(TaskEntity taskEntity) throws IOException, ResourcePreparationException {
        if (taskEntity.getParentType() != TaskParentType.effectiveCharacteristic) {
            throw new TaskStartingException("Unable to find parameters for Effective Characteristics");
        }
        ArrayList<FileData> fileData = new ArrayList<>();

        ECEntity entity = ecManager.extend(ecManager.findOne(taskEntity.getParentId(), null), null);
        TechnologyType technologyType = compositeMaterialManager.findOne(entity.getCompositeMaterialId(), null).getTechnologyType();

        generateECInputFile(fileData, entity, technologyType);
        if (technologyType == TechnologyType.ADDITIVE_3CM) {
            addFilesFromDir(fileData, pathToECMacrosesFor3CMMaterialsStr);
        } else {
            addFilesFromDir(fileData, pathToECMacrosesStr);
        }

        return fileData;
    }

    protected void generateECInputFile(List<FileData> fileData, ECEntity entity, TechnologyType technologyType) throws IOException {
        ECParameters ecParameters = ecManager.getEffectiveCharacteristicsParameters(entity.getId());
        if (!ecParameters.isAllParametersPresent()) {
            throw new TaskStartingException("Unable to find parameters for Effective Characteristics");
        }

        if (technologyType == TechnologyType.ADDITIVE_3CM) {
            Path inputFile = Files.createTempFile(null, null);
            String params = String.join(" ",
                    commasToDots(ecParameters.getMatrixElacticityModulus(), "0"),
                    commasToDots(ecParameters.getMatrixPoissonRatio(), "0"),
                    commasToDots(ecParameters.getFiberElacticityModulus(), "0"),
                    commasToDots(ecParameters.getFiberPoissonRatio(), "0"),
                    commasToDots(ecParameters.getEpoxideElacticityModulus(), "0"),
                    commasToDots(ecParameters.getEpoxidePoissonRatio(), "0"),
                    commasToDots(ecParameters.getFiberConcentration(), "0"),
                    commasToDots(ecParameters.getFiberDiameter(), "0"),
                    commasToDots(ecParameters.getValueA(), "0"),
                    commasToDots(ecParameters.getValueB(), "0"),
                    commasToDots(ecParameters.getValueL(), "0"),
                    commasToDots(ecParameters.getAverageLayerThickness(), "0"),
                    prepareDefectString(entity.getDefects()),
                    commasToDots(ecParameters.getMatrixStrengthLimit(), "0"),
                    commasToDots(ecParameters.getFiberStrengthLimit(), "0"),
                    commasToDots(ecParameters.getEpoxideStrengthLimit(), "0")
            );
            Files.write(inputFile, params.getBytes(StandardCharsets.UTF_8));

            fileData.add(FileData.fromFile("ef_input1.txt", DataType.SUBMODEL, inputFile.toFile()));
        } else {
            Path inputFile = Files.createTempFile(null, null);
            String params = String.join(" ",
                    commasToDots(ecParameters.getMatrixElacticityModulus(), "0"),
                    commasToDots(ecParameters.getMatrixPoissonRatio(), "0"),
                    commasToDots(ecParameters.getFiberElacticityModulus(), "0"),
                    commasToDots(ecParameters.getFiberPoissonRatio(), "0"),
                    commasToDots(ecParameters.getFiberConcentration(), "0"),
                    commasToDots(ecParameters.getFiberDiameter(), "0"),
                    EC_DEFAULT_PARAMS,
                    commasToDots(ecParameters.getMatrixStrengthLimit(), "0"),
                    commasToDots(ecParameters.getFiberStrengthLimit(), "0"));

            Files.write(inputFile, params.getBytes(StandardCharsets.UTF_8));

            fileData.add(FileData.fromFile("ef_input1.txt", DataType.SUBMODEL, inputFile.toFile()));
        }
    }

    private String prepareDefectString(List<ECDefectEntity> ecDefectEntities) {
        return ecDefectEntities.stream()
                .flatMap(defect -> Stream.of(defect.getPercentage(), defect.getThickness(), defect.getWidth(), defect.getLength()))
                .map(value -> value == null ? "0" : commasToDots(String.valueOf(value), "0"))
                .collect(Collectors.joining(" "));
    }
}
