package ru.crp.lvms.tasks.fileGenerators;

import org.apache.commons.collections.map.MultiKeyMap;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import org.springframework.validation.BeanPropertyBindingResult;
import org.springframework.validation.Errors;
import ru.crp.cmlb.model.entities.FileData;
import ru.crp.cmls.calculations.solvers.xml.config.DataType;
import ru.crp.lvms.exceptions.InternalException;
import ru.crp.lvms.exceptions.solving.TaskStartingException;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.enums.OptimizationCad;
import ru.crp.lvms.optimizations.base.enums.OptimizationCae;
import ru.crp.lvms.optimizations.base.enums.OptimizationCao;
import ru.crp.lvms.optimizations.mf.config.MFConfigGenerator;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationBaseSetting;
import ru.crp.lvms.optimizations.mf.enums.MFBaseSettingType;
import ru.crp.lvms.validators.task.optimization.MFTaskParamsValidator;

import javax.annotation.PostConstruct;
import javax.xml.bind.JAXBException;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.stream.Stream;

import static java.util.stream.Collectors.joining;
import static java.util.stream.Collectors.toList;
import static ru.crp.lvms.optimizations.mf.enums.MFBaseSettingType.*;

@Component
public class MFInputFileGenerator extends BaseInputFileGenerator {
    private static final String DEFAULT_CAD_MODEL_NAME = "CAD.SLDPRT";
    private static final String GENERATED_XML_FILE_NAME = "settings.xml";

    @Value("${modFrontierAbaqusProjectsDir}")
    private String pathToModFrontierAbaqusProjects;

    @Value("${modFrontierAbaqusSolidWorksProjectsDir}")
    private String pathToModFrontierAbaqusSolidWorksProjects;

    @Value("${modFrontierAnsysProjectsDir}")
    private String pathToModFrontierAnsysProjects;

    @Value("${modFrontierAnsysSolidWorksProjectsDir}")
    private String pathToModFrontierAnsysSolidWorksProjects;

    @Value("${modFrontierMacrosesDir}")
    private String pathToModFrontierMacroses;

    @Autowired
    private MFConfigGenerator mFConfigGenerator;

    @Autowired
    private FileManager fileManager;

    @Autowired
    private MFTaskParamsValidator mfTaskParamsValidator;

    private String[][] projectFilesMap = new String[NUM_TARGET_FUNCS.getMaxValue()][NUM_CONSTRAINTS.getMaxValue()];

    private MultiKeyMap projectDirs = new MultiKeyMap();

    @PostConstruct
    public void init() throws IOException {
        projectDirs.put(OptimizationCae.ABAQUS.getId(), null, Paths.get(pathToModFrontierAbaqusProjects));
        projectDirs.put(OptimizationCae.ABAQUS.getId(), OptimizationCad.SolidWorks.getId(), Paths.get(pathToModFrontierAbaqusSolidWorksProjects));

        projectDirs.put(OptimizationCae.ANSYS.getId(), null, Paths.get(pathToModFrontierAnsysProjects));
        projectDirs.put(OptimizationCae.ANSYS.getId(), OptimizationCad.SolidWorks.getId(), Paths.get(pathToModFrontierAnsysSolidWorksProjects));

        projectDirs.values()
                .forEach(pathToProjectFilesDir -> {
                    if (!Files.exists((Path) pathToProjectFilesDir)) {
                        throw new InternalException("Не найдены файлы проектов для ModFrontier в папке " + pathToProjectFilesDir);
                    }
                });
        for (int i = 0; i < projectFilesMap.length; ++i) {
            for (int j = 0; j < projectFilesMap[0].length; ++j) {
                projectFilesMap[i][j] = "obj" + (i + 1) + "c" + (j + 1) + ".prj";
            }
        }
    }


    public List<FileData> getInputFilesForOptimization(OptimizationEntity optimization) throws IOException {
        ArrayList<FileData> fileData = new ArrayList<>();

        fileData.addAll(collectCAEFiles(optimization));
        if (optimization.getSolverCadId() != null) {
            fileData.add(getCADFile(optimization));
        }
        fileData.add(getModFrontierScript(optimization));
        fileData.add(getGeneratedXml(optimization));

        addFilesFromDir(fileData, pathToModFrontierMacroses);
        addProjAndGeneratedFiles(fileData, optimization);
        return fileData;
    }


    private List<FileData> collectCAEFiles(OptimizationEntity optimization) {
        return optimizationManager.getOptimizationFiles(optimization, FileType.CAEFile, null).stream()
                .map(fileEntity -> {
                    Path pathToFile = fileManager.resolvePathToFile(fileEntity, null);
                    if (pathToFile == null || !Files.exists(pathToFile)) {
                        throw new InternalException(String.format("File %d is present in db but not in file system", fileEntity.getId()));
                    }
                    return FileData.fromFile(fileEntity.getName(), DataType.SUBMODEL, pathToFile.toFile());
                })
                .collect(toList());
    }


    private FileData getCADFile(OptimizationEntity optimization) {
        Pair<FileEntity, Path> cadModel = getSingleFile(optimization, FileType.CADFile, "CAD модель не найдена");
        return FileData.fromFile(DEFAULT_CAD_MODEL_NAME, DataType.SUBMODEL, cadModel.getRight().toFile());
    }


    private FileData getModFrontierScript(OptimizationEntity optimization) throws IOException {
        Pair<FileEntity, Path> script = getSingleFile(optimization, FileType.ModFrontierScript, "script-файл для ModeFrontier не найден");
        String scriptName = "script." + FilenameUtils.getExtension(script.getLeft().getName());
        return FileData.fromFile(scriptName, DataType.SESSION_FILE, script.getRight().toFile());
    }


    private FileData getGeneratedXml(OptimizationEntity optimization) throws IOException {
        try {
            Path pathToSettings = mFConfigGenerator.generateConfig(optimization);
            return FileData.fromFile(GENERATED_XML_FILE_NAME, DataType.SESSION_FILE, pathToSettings.toFile());
        } catch (JAXBException ex) {
            throw new TaskStartingException(String.format("Ошибка создания файла '%s'", GENERATED_XML_FILE_NAME), ex);
        }
    }


    public void addProjAndGeneratedFiles(List<FileData> files, OptimizationEntity optimization) {
        Map<MFBaseSettingType, MFOptimizationBaseSetting> baseSettings = MFOptimizationSettingsManager.findBaseSettings(optimization.getId(), null);

        Errors errors = new BeanPropertyBindingResult(new TaskEntity(), TaskEntity.class.getName());
        mfTaskParamsValidator.validateBaseSettings(optimization, baseSettings, errors);
        if (errors.hasErrors()) {
            throw new TaskStartingException("Обнаружены ошибки валидации:", errors);
        }

        String fileContent = Stream.of(NUM_TARGET_FUNCS, NUM_CAO_VARYING_PARAMS, NUM_OUTPUT_PARAMS, NUM_CONSTRAINTS)
                .map(type -> "" + baseSettings.get(type).getCurrentValue())
                .collect(joining(System.lineSeparator()));
        fileContent = fileContent + System.lineSeparator() + "OUTPUT_FILES_PATH";

        try {
            Path tempFile = Files.createTempFile(null, null);
            Files.write(tempFile, fileContent.getBytes(StandardCharsets.UTF_8));
            files.add(FileData.fromFile("inputISKPI.txt", DataType.SESSION_FILE, tempFile.toFile()));

            int numTargetFuncs = baseSettings.get(NUM_TARGET_FUNCS).getCurrentValue();
            int numConstraints = baseSettings.get(NUM_CONSTRAINTS).getCurrentValue();

            Path pathToProjFile;
            String projFileName = projectFilesMap[numTargetFuncs - 1][numConstraints - 1];
            Integer selectedCAE = optimization.getSolverCaeId();
            Integer selectedCAD = optimization.getSolverCadId();
            Object projDir = projectDirs.get(selectedCAE, selectedCAD);
            if (projDir != null) {
                pathToProjFile = ((Path) projDir).resolve(projFileName);
            } else {
                throw new TaskStartingException(String.format("Не найдены файлы проекта для выбранной комбинации решателей: CAO = %s, CAE = %s, CAD = %s",
                        OptimizationCao.ModeFrontier,
                        selectedCAE == null ? "-" : OptimizationCae.valueOf(selectedCAE),
                        selectedCAD == null ? "-" : OptimizationCad.valueOf(selectedCAD)));
            }
            if (!Files.exists(pathToProjFile)) {
                throw new TaskStartingException(String.format("Proj-файл не найден: '%s'", pathToProjFile));
            }
            files.add(FileData.fromFile(pathToProjFile.getFileName().toString(), DataType.SUBMODEL, pathToProjFile.toFile()));
        } catch (IOException ex) {
            throw new TaskStartingException("Ошибка при создании файла с базовыми настройками", ex);
        }
    }
}
