package ru.crp.lvms.tasks.fileGenerators;

import org.apache.commons.collections.map.MultiKeyMap;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import org.springframework.validation.BeanPropertyBindingResult;
import org.springframework.validation.Errors;
import ru.crp.cmlb.model.entities.FileData;
import ru.crp.cmls.calculations.solvers.xml.config.DataType;
import ru.crp.cmls.calculations.solvers.xml.config.Type;
import ru.crp.lvms.exceptions.InternalException;
import ru.crp.lvms.exceptions.solving.TaskStartingException;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.enums.OptimizationCad;
import ru.crp.lvms.optimizations.base.enums.OptimizationCae;
import ru.crp.lvms.optimizations.base.enums.OptimizationCao;
import ru.crp.lvms.optimizations.mf.config.MFConfigGenerator;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationBaseSetting;
import ru.crp.lvms.optimizations.mf.enums.MFBaseSettingType;
import ru.crp.lvms.optimizations.optimus.config.OptimusConfigGenerator;
import ru.crp.lvms.optimizations.optimus.entities.OptimusOptimizationBaseSetting;
import ru.crp.lvms.optimizations.optimus.enums.OptimusBaseSettingType;
import ru.crp.lvms.optimizations.optimus.tpl.OptimusTemplateGenerator;
import ru.crp.lvms.optimizations.optistruct.enums.OptistructConvertDirection;
import ru.crp.lvms.validators.task.optimization.MFTaskParamsValidator;
import ru.crp.lvms.validators.task.optimization.OptimusTaskParamsValidator;

import javax.annotation.PostConstruct;
import javax.xml.bind.JAXBException;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static java.util.stream.Collectors.joining;
import static java.util.stream.Collectors.toList;
import static ru.crp.lvms.optimizations.mf.enums.MFBaseSettingType.*;

@Component
public class OptimusInputFileGenerator extends BaseInputFileGenerator {
    private static final String GENERATED_TPL_ANSYS_FILE_NAME = "uniform.mac.tpl";
    private static final String GENERATED_TPL_ABAQUS_FILE_NAME = "uniform.py.tpl";
    private static final String ANSYS_SCRIPT_NAME = "'script.mac'";
    private static final String ABAQUS_SCRIPT_NAME = "execfile('script.py')";
    private static final String DEFAULT_CAD_MODEL_NAME = "unum";
    private static final String GENERATED_XML_FILE_NAME = "settings.seq";


    @Value("${optimusMacrosesDir}")
    private String pathToOptimusMacroses;

    @Autowired
    private OptimusTemplateGenerator optimusTemplateGenerator;

    @Autowired
    private OptimusConfigGenerator optimusConfigGenerator;

    @Autowired
    private FileManager fileManager;



    public List<FileData> getInputFilesForOptimization(OptimizationEntity optimization) throws IOException {
        ArrayList<FileData> fileData = new ArrayList<>();

        fileData.addAll(collectCAEFiles(optimization));
        if (optimization.getSolverCadId() != null) {
            fileData.add(getCADFile(optimization));
        }
        fileData.add(getOptimusScript(optimization));
        fileData.add(getGeneratedXml(optimization));
        fileData.add(getGeneratedTpl(optimization));

       // addFilesFromDir(fileData, pathToOptimusMacroses);

        return fileData;
    }


    private List<FileData> collectCAEFiles(OptimizationEntity optimization) {
        return optimizationManager.getOptimizationFiles(optimization, FileType.CAEFile, null).stream()
                .map(fileEntity -> {
                    Path pathToFile = fileManager.resolvePathToFile(fileEntity, null);
                    if (pathToFile == null || !Files.exists(pathToFile)) {
                        throw new InternalException(String.format("File %d is present in db but not in file system", fileEntity.getId()));
                    }
                    return FileData.fromFile(fileEntity.getName(), DataType.SUBMODEL, pathToFile.toFile());
                })
                .collect(toList());
    }


    private FileData getCADFile(OptimizationEntity optimization) {
        Pair<FileEntity, Path> cadModel = getSingleFile(optimization, FileType.CADFile, "CAD модель не найдена");
        return FileData.fromFile(DEFAULT_CAD_MODEL_NAME, DataType.SUBMODEL, cadModel.getRight().toFile());
    }


    private FileData getOptimusScript(OptimizationEntity optimization) throws IOException {
        Pair<FileEntity, Path> script = getSingleFile(optimization, FileType.OptimusScript, "script-файл для Optimus не найден");
        String scriptName = "script." + FilenameUtils.getExtension(script.getLeft().getName());
        return FileData.fromFile(scriptName, DataType.SESSION_FILE, script.getRight().toFile());
    }


    private FileData getGeneratedXml(OptimizationEntity optimization) throws IOException {
        try {
            Path pathToSettings = optimusConfigGenerator.generateConfig(optimization);
            return FileData.fromFile(GENERATED_XML_FILE_NAME, DataType.SESSION_FILE, pathToSettings.toFile());
        } catch (JAXBException ex) {
            throw new TaskStartingException(String.format("Ошибка создания файла '%s'", GENERATED_XML_FILE_NAME), ex);
        }
    }

    private FileData getGeneratedTpl(OptimizationEntity optimization) throws IOException {
        try {
            Path pathToTpl = optimusTemplateGenerator.generateConfig(optimization);
            switch (optimization.getCaeName()) {
                case "ANSYS":
                    Files.write(pathToTpl, ("\n" + ANSYS_SCRIPT_NAME).getBytes(), StandardOpenOption.APPEND);
                    return FileData.fromFile(GENERATED_TPL_ANSYS_FILE_NAME, DataType.SESSION_FILE, pathToTpl.toFile());
                case "ABAQUS":
                    Files.write(pathToTpl, ("\n"+ABAQUS_SCRIPT_NAME).getBytes(), StandardOpenOption.APPEND);
                    return FileData.fromFile(GENERATED_TPL_ABAQUS_FILE_NAME, DataType.SESSION_FILE, pathToTpl.toFile());
                default:
                    throw new IllegalArgumentException("Unexepected CAE system: " + optimization.getCaeName());
            }
        } catch (JAXBException ex) {
            throw new TaskStartingException(String.format("Ошибка создания файла tpl-файла"), ex);
        }
    }
}
