package ru.crp.lvms.tasks.fileGenerators;

import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import ru.crp.cmlb.model.entities.FileData;
import ru.crp.cmls.calculations.solvers.xml.config.DataType;
import ru.crp.cmls.calculations.solvers.xml.config.Type;
import ru.crp.lvms.exceptions.InternalException;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.optistruct.enums.OptistructConvertDirection;
import ru.crp.lvms.optimizations.optistruct.tcl.OptistructBatchGenerator;
import ru.crp.lvms.optimizations.optistruct.tcl.OptistructConvertScriptGenerator;
import ru.crp.lvms.optimizations.optistruct.tcl.OptistructFemScriptGenerator;
import ru.crp.lvms.optimizations.optistruct.tcl.OptistructOssScriptGenerator;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static java.util.Collections.singletonList;
import static ru.crp.lvms.model.mappers.PgArrayPropertyEditor.LOGGER;


@Component
public class OptistructInputFileGenerator extends BaseInputFileGenerator {

    // MUST be synchronized with argument value in optistruct-config.xml
    public static final String CONVERT_TO_FILE_NAME = "convert_to.tcl";
    public static final String CONVERT_FROM_FILE_NAME = "convert_from.tcl";
    public static final String GENERATE_FEM_FILE_NAME = "generate_fem.tcl";
    public static final String OSS_FILE_NAME = "oss.tcl";
    public static final String BATCH_FILE_NAME = "start.bat";

    @Autowired
    private FileManager fileManager;

    @Autowired
    private OptistructConvertScriptGenerator optistructConvertScriptGenerator;

    @Autowired
    private OptistructFemScriptGenerator optistructFemScriptGenerator;

    @Autowired
    private OptistructOssScriptGenerator optistructOssScriptGenerator;

    @Autowired
    private OptistructBatchGenerator optistructBatchGenerator;

    public List<FileData> getInputFilesForOptimization(Type type, OptimizationEntity optimization) throws IOException {
        ArrayList<FileData> fileData = new ArrayList<>();

        fileData.addAll(optimizationManager.getInputFilesForOptimization(optimization, singletonList((ru.crp.lvms.model.entities.misc.FileType.LoadCondition))));

        FileData optimizationFile = getOptimizationFile(optimization);
        if (type == Type.optistructGenerateFEM) {
            fileData.add(optimizationFile);
        } else {
            fileData.add(getFemFile(optimization));
        }

        fileData.addAll(getGeneratedFiles(type, optimization, OptistructConvertDirection.fromExtension(optimizationFile.getName())));
        return fileData;
    }


    private FileData getFemFile(OptimizationEntity optimization) {
        List<FileEntity> generatedFEMFiles = optimizationManager.getOptimizationFemFiles(optimization);
        if (generatedFEMFiles.size() != 1) {
            throw new InternalException(String.format("Неверное количество .fem файлов - %d", generatedFEMFiles.size()));
        }
        Path pathToFEMFile = fileManager.resolvePathToFile(generatedFEMFiles.get(0), null);
        return FileData.fromFile(generatedFEMFiles.get(0).getName(), DataType.CUSTOM_INPUT_FILE, pathToFEMFile.toFile());
    }


    // todo: add all files
    // Currently adding only one file due to converting complications:
    // every file needs to be converted by specific convert script back and forth
    private FileData getOptimizationFile(OptimizationEntity optimization) {
        return optimizationManager.getOptimizationFiles(optimization, null)
                .stream()
                .filter(fileEntity -> !fileEntity.getName().endsWith("fem"))
                .map(fileEntity -> {
                    Path pathToFile = fileManager.resolvePathToFile(fileEntity, null);
                    if (pathToFile == null || !Files.exists(pathToFile)) {
                        LOGGER.error("File {} is present in db but not in file system", fileEntity.getId());
                        return null;
                    }
                    return FileData.fromFile("file_name." + FilenameUtils.getExtension(fileEntity.getName()), DataType.SUBMODEL, pathToFile.toFile());
                })
                .filter(Objects::nonNull)
                .findAny().orElseThrow(() -> new InternalException("Файлы оптимизации не найдены"));
    }


    private List<FileData> getGeneratedFiles(Type taskType, OptimizationEntity optimization,
                                             OptistructConvertDirection direction) throws IOException {
        Stream<Pair<String, Path>> generatedFiles;
        if (taskType == Type.optistructGenerateFEM) {
            generatedFiles = Stream.of(Pair.of(CONVERT_TO_FILE_NAME, optistructConvertScriptGenerator.generateConfig(optimization, direction)),
                    Pair.of(BATCH_FILE_NAME, optistructBatchGenerator.generateConfig(CONVERT_TO_FILE_NAME)));
        } else {
            generatedFiles = Stream.of(
                    Pair.of(CONVERT_FROM_FILE_NAME, optistructConvertScriptGenerator.generateConfig(optimization, direction.backward())),
                    Pair.of(GENERATE_FEM_FILE_NAME, optistructFemScriptGenerator.generateConfig(optimization)),
                    Pair.of(OSS_FILE_NAME, optistructOssScriptGenerator.generateConfig(optimization)),
                    Pair.of(BATCH_FILE_NAME, optistructBatchGenerator.generateConfig(GENERATE_FEM_FILE_NAME, OSS_FILE_NAME, CONVERT_FROM_FILE_NAME)));
        }
        return generatedFiles
                .map(pair -> FileData.fromFile(pair.getKey(), DataType.GENERATED, pair.getValue().toFile()))
                .collect(Collectors.toList());
    }
}
