package ru.crp.lvms.tasks.slurm;

import net.schmizz.sshj.SSHClient;
import net.schmizz.sshj.common.IOUtils;
import net.schmizz.sshj.connection.channel.direct.Session;
import net.schmizz.sshj.sftp.RemoteResourceInfo;
import net.schmizz.sshj.sftp.SFTPClient;
import net.schmizz.sshj.transport.verification.PromiscuousVerifier;
import net.schmizz.sshj.xfer.FileSystemFile;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import ru.crp.lvms.cluster.model.SlurmClusterConfigEntity;
import ru.crp.lvms.model.dao.NodesDao;
import ru.crp.lvms.tasks.slurm.exceptions.SlurmConnectionException;
import ru.crp.lvms.tasks.slurm.exceptions.SlurmNonZeroExitCodeException;
import ru.crp.lvms.tasks.slurm.requests.AbstractRequest;

import java.io.IOException;
import java.nio.file.Path;
import java.util.Comparator;
import java.util.PriorityQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;

@Service
public class SlurmConnectionManager {
    public static final Logger LOGGER = LoggerFactory.getLogger(SlurmConnectionManager.class);

    @Autowired
    private NodesDao nodeDao;

    private AtomicBoolean slurmConnectionAlive = new AtomicBoolean(false);


    public <T> T execCommand(SlurmClusterConfigEntity config, AbstractRequest<T> command) {
        SSHClient ssh = new SSHClient();
        try {
            ssh.loadKnownHosts();
            ssh.addHostKeyVerifier(new PromiscuousVerifier());
            ssh.setTimeout((int) TimeUnit.SECONDS.toMillis(config.getConnectionTimeout()));
            ssh.setConnectTimeout((int) TimeUnit.SECONDS.toMillis(config.getConnectionTimeout()));
            ssh.connect(config.getAddress());
            ssh.authPublickey(config.getUser(), config.getPathToPrivateKey());
            Session session = ssh.startSession();
            try {
                final Session.Command cmd = session.exec(command.getCommandLine());
                cmd.join(config.getConnectionTimeout(), TimeUnit.SECONDS);
                setSlurmConnectionAlive(config.getClusterId(), true);
                if (cmd.getExitStatus() != 0) {
                    String response = IOUtils.readFully(cmd.getErrorStream()).toString();
                    throw new SlurmNonZeroExitCodeException(cmd.getExitStatus(), response);
                }
                String response = IOUtils.readFully(cmd.getInputStream()).toString();
                return command.parseResponse(response);
            } finally {
                session.close();
            }
        } catch (IOException ex) {
            setSlurmConnectionAlive(config.getClusterId(), false);
            throw new SlurmConnectionException(ex);
        } finally {
            try {
                ssh.disconnect();
            } catch (IOException ex) {
                throw new SlurmConnectionException(ex);
            }
        }
    }


    @Transactional
    public void setSlurmConnectionAlive(int clusterId, boolean newState) {
        slurmConnectionAlive.set(newState);
        if (!newState) {
            nodeDao.updateAllNodesAliveStatus(clusterId, newState);
        }
    }

    public boolean isSlurmConnectionAlive() {
        return slurmConnectionAlive.get();
    }

    public void uploadFile(SlurmClusterConfigEntity config, Path pathToFile, String pathOnServer) {
        SSHClient ssh = new SSHClient();
        try {
            ssh.loadKnownHosts();
            ssh.addHostKeyVerifier(new PromiscuousVerifier());
            ssh.useCompression();
            ssh.connect(config.getAddress());
            ssh.authPublickey(config.getUser(), config.getPathToPrivateKey());
            ssh.newSCPFileTransfer()
                    .upload(new FileSystemFile(pathToFile.toAbsolutePath().toFile()), pathOnServer);
            setSlurmConnectionAlive(config.getClusterId(), true);
        } catch (IOException ex) {
            setSlurmConnectionAlive(config.getClusterId(), false);
            throw new SlurmConnectionException(ex);
        } finally {
            try {
                ssh.disconnect();
            } catch (IOException ex) {
                throw new SlurmConnectionException(ex);
            }
        }
    }

    public void downloadResults(SlurmClusterConfigEntity config, String outputDirOnServer, Path resultsDir) {
        if (!outputDirOnServer.endsWith("/")) {
            outputDirOnServer = outputDirOnServer + "/";
        }
        outputDirOnServer = outputDirOnServer.substring(0, outputDirOnServer.length() - 1);
        SSHClient ssh = new SSHClient();
        try {
            ssh.loadKnownHosts();
            ssh.addHostKeyVerifier(new PromiscuousVerifier());
            ssh.useCompression();
            ssh.connect(config.getAddress());
            ssh.authPublickey(config.getUser(), config.getPathToPrivateKey());
            LOGGER.info(String.format("Downloading from: '%s' to '%s'", outputDirOnServer, resultsDir.toAbsolutePath().toString()));

            SFTPClient sftpClient = ssh.newSFTPClient();
            PriorityQueue<String> d3dumps = new PriorityQueue<>(Comparator.reverseOrder());
            for (RemoteResourceInfo rs : sftpClient.ls(outputDirOnServer)) {
                String name = rs.getName();
                // all d3dumps are included to d3full so we don't need to download any of them
                if (name.startsWith("d3dump")) {
                    continue;
                }
                if (name.startsWith("d3full")) {
                    d3dumps.add(rs.getPath());
                    continue;
                }
                sftpClient.get(rs.getPath(), new FileSystemFile(resultsDir.toAbsolutePath().toFile()));
            }
            if (!d3dumps.isEmpty()) {
                sftpClient.get(d3dumps.peek(), new FileSystemFile(resultsDir.toAbsolutePath().toFile()));
            }
            setSlurmConnectionAlive(config.getClusterId(), true);
        } catch (IOException ex) {
            setSlurmConnectionAlive(config.getClusterId(), false);
            throw new SlurmConnectionException(ex);
        } finally {
            try {
                ssh.disconnect();
            } catch (IOException ex) {
                throw new SlurmConnectionException(ex);
            }
        }
    }
}
