package ru.crp.lvms.tasks.slurm;


import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import ru.crp.cmls.api.NodeStatus;
import ru.crp.cmls.calculations.solvers.dto.SolverPresenceRequest;
import ru.crp.lvms.cluster.model.SlurmClusterConfigEntity;
import ru.crp.lvms.model.dao.NodesDao;
import ru.crp.lvms.model.entities.nodes.NodeEntity;
import ru.crp.lvms.tasks.slurm.enums.SlurmNodeStatus;

import java.util.Date;
import java.util.List;
import java.util.Set;

import static java.util.Arrays.asList;
import static java.util.Collections.emptySet;
import static java.util.stream.Collectors.toSet;

@Service
public class SlurmNodeInfoMapper {
    private static final Logger LOGGER = LoggerFactory.getLogger(SlurmNodeInfoMapper.class);

    private static final int EXPECTED_NODE_PROPERTIES_ARRAY_LENGTH = 6;
    public static final String NODES_INFO_FORMAT = "cpus,freemem,features,nodeaddr,nodelist,statelong";

    @Autowired
    private NodesDao nodeDao;

    public static final SlurmNodeInfo parseInfo(String nodeInfo) {
        SlurmNodeInfo node = new SlurmNodeInfo();
        String[] nodeProperties = nodeInfo.split("\\s+");
        if (nodeProperties.length != EXPECTED_NODE_PROPERTIES_ARRAY_LENGTH) {
            throw new IllegalStateException(String.format("Unexpected node info format: '%s'", nodeInfo));
        }
        try {
            node.setCpus(Integer.parseInt(nodeProperties[0]));
        } catch (NumberFormatException ex) {
            node.setCpus(-1);
        }
        try {
            node.setFreeMem(Long.parseLong(nodeProperties[1]));
        } catch (NumberFormatException ex) {
            node.setFreeMem(-1L);
        }
        node.setFeatures(nodeProperties[2]);
        node.setNodeAddr(nodeProperties[3]);
        node.setNodeName(nodeProperties[4]);
        node.setNodeStatus(SlurmNodeStatus.toNodeStatus(nodeProperties[5]));

        return node;
    }

    public NodeStatus mapToNodeStatus(SlurmNodeInfo slurmNode, List<SolverPresenceRequest> solversToCheck) {
        NodeStatus nodeStatus = new NodeStatus();
        nodeStatus.setAvailableProcessors(slurmNode.getCpus());
        nodeStatus.setFreePhysicalMemorySize(slurmNode.getFreeMem());

        nodeStatus.setTotalPhysicalMemorySize(nodeStatus.getFreePhysicalMemorySize());
        nodeStatus.setServiceUptimeSeconds(0);
        nodeStatus.setSystemCpuLoad(0);

        Set<String> solverNames = solversToCheck.stream().map(SolverPresenceRequest::getSolverName).collect(toSet());
        Set<String> nodeSolvers = extractNodeSolvers(slurmNode, solverNames);
        nodeStatus.setSolvers(nodeSolvers);
        return nodeStatus;
    }


    public Set<String> extractNodeSolvers(SlurmNodeInfo slurmNode, Set<String> solverNames) {
        String features = slurmNode.getFeatures();
        if (features == null || features.isEmpty()) {
            return emptySet();
        }
        List<String> featuresList = asList(features.split(","));
        return solverNames.stream()
                .filter(featuresList::contains)
                .collect(toSet());
    }


    public NodeEntity mapToNode(SlurmClusterConfigEntity config, SlurmNodeInfo slurmNode, List<String> busySlurmNodeNames) {
        NodeEntity node = nodeDao.findByNameAndCluster(slurmNode.getNodeName(), config.getClusterId());
        if (node == null) {
            node = new NodeEntity();
        }
        node.setName(slurmNode.getNodeName());
        node.setOwnerId(config.getNodesOwnerId());

        SlurmNodeStatus nodeStatus = slurmNode.getNodeStatus();
        node.setIsAlive(nodeStatus.isAlive());
        if (nodeStatus.isBusy()) {
            node.setIsEnabled(busySlurmNodeNames.contains(node.getName()));
        } else {
            node.setIsEnabled(true);
        }
        node.setMemory(slurmNode.getFreeMem());
        node.setModificationDate(new Date());

        node.setNumOfCores(slurmNode.getCpus());

        node.setUrl(slurmNode.getNodeAddr());
        if (!node.getUrl().endsWith("/")) {
            node.setUrl(node.getUrl() + "/");
        }
        node.setDescription("Slurm node");
        node.setClusterId(config.getClusterId());
        return node;
    }
}
