package ru.crp.lvms.utils;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Component;
import ru.crp.lvms.exceptions.GifCreateException;
import ru.crp.lvms.exceptions.InternalException;

import javax.imageio.*;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.metadata.IIOMetadataNode;
import javax.imageio.stream.FileImageOutputStream;
import javax.imageio.stream.ImageOutputStream;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.nio.file.Path;
import java.util.Iterator;
import java.util.List;

@Component
public class GifSequenceWriter {
    private static final Logger LOGGER = LoggerFactory.getLogger(GifSequenceWriter.class);

    private static final int TIME_BETWEEN_FRAMES_MS = 500;
    private static final int TIME_TO_SHOW_FIRST_PICTURE_MS = 3000;
    private static final int TIME_TO_SHOW_LAST_PICTURE_MS = 3000;

    public void makeGifFromImages(List<Path> images, Path outputFile, boolean loopContinuously) {
        if (images.isEmpty()) {
            LOGGER.warn("There must be at least two images to create gif");
            throw new GifCreateException("There must be at least two images to create gif");
        }

        try(ImageOutputStream outputStream =
                    new FileImageOutputStream(outputFile.toFile())) {
            BufferedImage firstImage = ImageIO.read(images.get(0).toFile());

            ImageWriter gifWriter = getWriter();
            ImageWriteParam imageWriteParam = gifWriter.getDefaultWriteParam();
            ImageTypeSpecifier imageTypeSpecifier =
                    ImageTypeSpecifier.createFromBufferedImageType(firstImage.getType());

            IIOMetadata imageMetaData = gifWriter.getDefaultImageMetadata(imageTypeSpecifier, imageWriteParam);

            String metaFormatName = imageMetaData.getNativeMetadataFormatName();

            IIOMetadataNode root = (IIOMetadataNode) imageMetaData.getAsTree(metaFormatName);

            IIOMetadataNode graphicsControlExtensionNode = getMetadataNode(root, "GraphicControlExtension");

            graphicsControlExtensionNode.setAttribute("disposalMethod", "none");
            graphicsControlExtensionNode.setAttribute("userInputFlag", "FALSE");
            graphicsControlExtensionNode.setAttribute("transparentColorFlag", "FALSE");
            graphicsControlExtensionNode.setAttribute("delayTime", Integer.toString(TIME_BETWEEN_FRAMES_MS / 10));
            graphicsControlExtensionNode.setAttribute("transparentColorIndex", "0");

            IIOMetadataNode commentsNode = getMetadataNode(root, "CommentExtensions");
            commentsNode.setAttribute("CommentExtension", "Created by LVM-Space");

            IIOMetadataNode appEntensionsNode = getMetadataNode(root, "ApplicationExtensions");

            IIOMetadataNode child = new IIOMetadataNode("ApplicationExtension");

            child.setAttribute("applicationID", "NETSCAPE");
            child.setAttribute("authenticationCode", "2.0");

            int loop = loopContinuously ? 0 : 1;

            child.setUserObject(new byte[]{ 0x1, (byte) (loop & 0xFF), (byte)
                    ((loop >> 8) & 0xFF)});
            appEntensionsNode.appendChild(child);

            imageMetaData.setFromTree(metaFormatName, root);

            gifWriter.setOutput(outputStream);

            gifWriter.prepareWriteSequence(null);

            gifWriter.writeToSequence(new IIOImage(firstImage, null, imageMetaData), imageWriteParam);

            if (images.size() > 1) {
                for (int i = 1; i < TIME_TO_SHOW_FIRST_PICTURE_MS / TIME_BETWEEN_FRAMES_MS; i++) {
                    gifWriter.writeToSequence(new IIOImage(firstImage, null, imageMetaData), imageWriteParam);
                }
                for (int i = 1; i <= images.size() - 1; i++) {
                    BufferedImage nextImage = ImageIO.read(images.get(i).toFile());
                    gifWriter.writeToSequence(new IIOImage(nextImage, null, imageMetaData), imageWriteParam);
                }
                BufferedImage lastImage = ImageIO.read(images.get(images.size() - 1).toFile());
                for (int i = 1; i < TIME_TO_SHOW_LAST_PICTURE_MS / TIME_BETWEEN_FRAMES_MS; i++) {
                    gifWriter.writeToSequence(new IIOImage(lastImage, null, imageMetaData), imageWriteParam);
                }
            }

            gifWriter.endWriteSequence();
        } catch (IOException e) {
            LOGGER.warn("Gif conversion failed");
            throw new InternalException("Failed to create gif", e);
        }

    }

    private static ImageWriter getWriter() throws IIOException {
        Iterator<ImageWriter> iter = ImageIO.getImageWritersBySuffix("gif");
        if(!iter.hasNext()) {
            throw new IIOException("No GIF Image Writers Exist");
        } else {
            return iter.next();
        }
    }

    private static IIOMetadataNode getMetadataNode(IIOMetadataNode rootNode, String nodeName) {
        int nNodes = rootNode.getLength();
        for (int i = 0; i < nNodes; i++) {
            if (rootNode.item(i).getNodeName().compareToIgnoreCase(nodeName)
                    == 0) {
                return((IIOMetadataNode) rootNode.item(i));
            }
        }
        IIOMetadataNode node = new IIOMetadataNode(nodeName);
        rootNode.appendChild(node);
        return(node);
    }
}