package ru.crp.lvms.utils;

import org.apache.commons.lang3.tuple.Pair;
import org.joda.time.LocalDate;
import org.joda.time.format.DateTimeFormat;
import org.joda.time.format.DateTimeFormatter;
import org.modelmapper.AbstractConverter;
import org.modelmapper.Condition;
import org.modelmapper.TypeToken;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import ru.crp.cmlb.model.entities.base.Named;
import ru.crp.lvms.model.dao.UserDao;
import ru.crp.lvms.model.entities.effective.characteristics.ECDefectEntity;
import ru.crp.lvms.model.entities.effective.characteristics.ECParamEntity;
import ru.crp.lvms.model.entities.technology.TechnologyEntity;
import ru.crp.lvms.optimizations.mf.enums.MFConstraintsType;
import ru.crp.lvms.optimizations.mf.enums.MFTargetFuncsType;
import ru.crp.lvms.pdf.reports.vo.Defect;
import ru.crp.lvms.pdf.reports.vo.Parameter;
import ru.crp.lvms.pdf.reports.vo.SimpleParameter;
import ru.crp.lvms.web.dto.user.UserRole;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.servlet.ServletContext;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.nio.file.Path;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.stream.Collectors;

import static java.util.Arrays.asList;

@Component
public class MappingHelpers {
    private static final List<String> IGNORED_PROPERTIES = asList("id", "creationDate", "modificationDate", "ownerId");

    private DateTimeFormatter reportDateFormat = DateTimeFormat.mediumDate().withLocale(Locale.forLanguageTag("ru"));

    @Autowired
    private UserDao userDao;

    @Autowired
    private ServletContext servletContext;

    public static <T> TypeToken<List<T>> createListOfType(Class<T> clazz) {
        return TypeToken.of(new ParameterizedType() {
            @Nonnull
            @Override
            public Type[] getActualTypeArguments() {
                return new Type[]{clazz};
            }

            @Nonnull
            @Override
            public Type getRawType() {
                return List.class;
            }

            @Nullable
            @Override
            public Type getOwnerType() {
                return null;
            }
        });
    }


    public AbstractConverter<Integer, String> idToUser = new AbstractConverter<Integer, String>() {
        @Override
        protected String convert(@Nullable Integer id) {
            if (id == null) {
                return "";
            }
            return checkEntity(userDao.findOne(id));
        }
    };

    public Condition<Integer, Integer> skipNotAccessibleReference =
            mappingContext -> mappingContext.getSource() != null && !mappingContext.getSource().equals(-1);

    public final AbstractConverter<Integer, String> idToUserRoleName = new AbstractConverter<Integer, String>() {
        @Override
        protected String convert(Integer id) {
            return UserRole.valueOf(id).name();
        }
    };

    public final AbstractConverter<Integer, String> idToUserRoleDisplayName = new AbstractConverter<Integer, String>() {
        @Override
        protected String convert(Integer id) {
            return UserRole.valueOf(id).getDisplayName();
        }
    };

    // make mapper skip "secure" properties
    public Condition skipSecureFields = context -> {
        String destPropertyName = context.getMapping().getLastDestinationProperty().getName();
        return !IGNORED_PROPERTIES.contains(destPropertyName);
    };


    private String checkEntity(@Nullable Named entity) {
        return entity == null ? "" : entity.getName();
    }

    public AbstractConverter<List<ECParamEntity>, List<Parameter>> toCompositeParameters = new AbstractConverter<List<ECParamEntity>, List<Parameter>>() {
        @Override
        protected List<Parameter> convert(List<ECParamEntity> params) {
            return Parameter.handleCompositeParameters(params);
        }
    };

    public AbstractConverter<List<Pair<String, String>>, List<SimpleParameter>> fromPairsList = new AbstractConverter<List<Pair<String, String>>, List<SimpleParameter>>() {
        @Override
        protected List<SimpleParameter> convert(List<Pair<String, String>> source) {
            return Parameter.fromPairsList(source);
        }
    };

    public AbstractConverter<Date, String> toReportDate = new AbstractConverter<Date, String>() {
        @Override
        protected String convert(Date source) {
            return LocalDate.fromDateFields(source).toString(reportDateFormat);
        }
    };

    public AbstractConverter<List<ECDefectEntity>, List<Defect>> fromECDefectEntity = new AbstractConverter<List<ECDefectEntity>, List<Defect>>() {
        @Override
        protected List<Defect> convert(List<ECDefectEntity> source) {
            return source.stream().map(Defect::fromECDefectEntity).collect(Collectors.toList());
        }
    };

    public AbstractConverter<Path, String> pathToStringAbsolute = new AbstractConverter<Path, String>() {
        @Override
        protected String convert(@Nullable Path source) {
            if (source != null) {
                source = source.toAbsolutePath();
            }
            return String.valueOf(source);
        }
    };

    public AbstractConverter<Object, String> toString = new AbstractConverter<Object, String>() {
        @Override
        protected String convert(@Nullable Object source) {
            return String.valueOf(source);
        }
    };

    public AbstractConverter<Integer, String> targetFuncToXml = new AbstractConverter<Integer, String>() {
        @Override
        protected String convert(@Nullable Integer source) {
            if (source == null) {
                return MFTargetFuncsType.NONE.getXmlValue();
            }
            return MFTargetFuncsType.valueOf(source).getXmlValue();
        }
    };

    public AbstractConverter<Integer, String> constraintToXml = new AbstractConverter<Integer, String>() {
        @Override
        protected String convert(@Nullable Integer source) {
            if (source == null) {
                return MFConstraintsType.NONE.getXmlValue();
            }
            return MFConstraintsType.valueOf(source).getXmlValue();
        }
    };


    @Nullable
    public String getPathToTechnologyImage(TechnologyEntity technology) {
        return String.format("%s/rest/technologies/%d/imageFile", servletContext.getContextPath(), technology.getId());
    }

}
