package ru.crp.lvms.validators.task;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.validation.Errors;
import ru.crp.cmls.api.TaskParentType;
import ru.crp.cmls.calculations.solvers.xml.config.ResultType;
import ru.crp.lvms.managers.CompositeMaterialManager;
import ru.crp.lvms.managers.ResultsManager;
import ru.crp.lvms.managers.SimulationManager;
import ru.crp.lvms.managers.effective.characteristics.ECManager;
import ru.crp.lvms.managers.tasks.ResultsParent;
import ru.crp.lvms.model.entities.ResultEntity;
import ru.crp.lvms.model.entities.effective.characteristics.ECEntity;
import ru.crp.lvms.model.entities.simulations.SimulationEntity;
import ru.crp.lvms.model.entities.simulations.SolvingStatus;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.tasks.fileGenerators.ECParameters;
import ru.crp.lvms.tasks.fileGenerators.ThermalSimulationParameters;

import java.util.List;

import static ru.crp.lvms.model.entities.simulations.SimulationType.*;

@Component
public class SimulationTaskParamsValidator extends BaseTaskParamsValidator<SimulationEntity> {
    @Autowired
    private SimulationManager simulationManager;

    @Autowired
    private ECManager ecManager;

    @Autowired
    private ResultsManager resultsManager;

    @Autowired
    private CompositeMaterialManager compositeMaterialManager;

    @Override
    protected void doValidate(TaskEntity task, SimulationEntity taskParent, Errors errors) {
        if (task.getParentType() != TaskParentType.simulation) {
            throw new IllegalArgumentException("Wrong validator is used for task with parent type " + task.getParentType());
        }
        validateParentCompositeMaterial(taskParent.getCompositeMaterialId(), errors);
        if (errors.hasErrors()) {
            return;
        }
        ECEntity ec = ecManager.findOne(taskParent.getEcId(), null);
        if (ec == null) {
            errors.reject("initialECNotFound");
            return;
        }
        validateInitialEcResults(ec, errors);

        validateMaterialParams(taskParent, errors);
        validateSimulationParams(taskParent, errors);
    }


    private void validateSimulationParams(SimulationEntity taskParent, Errors errors) {
        int simTypeId = taskParent.getTypeId();
        if (simTypeId == Extension.getId() || simTypeId == Bend.getId() || simTypeId == Compression.getId()) {
            if (taskParent.getManualSizeB() == null || taskParent.getManualSizeH() == null || taskParent.getManualSizeL() == null) {
                errors.reject("notEnoughSimParams");
            }
        }
    }

    private void validateMaterialParams(SimulationEntity taskParent, Errors errors) {
        boolean isEpoxidePresent = compositeMaterialManager.findOne(taskParent.getCompositeMaterialId(), null).isEpoxidePresent();

        int simTypeId = taskParent.getTypeId();
        if (simTypeId == ThermalExpansion.getId() || simTypeId == ThermalConductivityCoeff.getId()) {
            ThermalSimulationParameters parameters = ecManager.getThermalSimulationParameters(taskParent.getEcId(), isEpoxidePresent);
            if (!parameters.isMatrixParametersPresent()) {
                errors.reject("notEnoughMatrixParams");
            }
            if (!parameters.isFiberParametersPresent()) {
                errors.reject("notEnoughFiberParams");
            }
            if (isEpoxidePresent && !parameters.isEpoxideParametersPresent()) {
                errors.reject("notEnoughEpoxideParams");
            }
            if (!parameters.isAllParametersPresent()) {
                return;
            }
        }
        if (simTypeId == Compression.getId()) {
            ECParameters ecParameters = ecManager.getEffectiveCharacteristicsParameters(taskParent.getEcId());
            if (!ecParameters.isAllMatrixParametersPresent() || !ecParameters.isAllFiberParametersPresent()) {
                errors.reject("notEnoughMaterialParams");
                return;
            }
            if (!ecParameters.isAllECParametersPresent()) {
                errors.reject("notEnoughECParams");
                return;
            }
        }
    }

    private void validateInitialEcResults(ECEntity ec, Errors errors) {
        if (ec.getStatusId() != SolvingStatus.Completed.getId()) {
            errors.reject("initialECNotSolved", new Object[]{String.format("'%s'", ec.getName())}, null);
            return;
        }
        ResultsParent resultParent = new ResultsParent(ec.getId(), TaskParentType.effectiveCharacteristic);

        List<ResultEntity> databases = resultsManager.getResultByParent(resultParent, ResultType.database);
        if (databases.size() != 2) {
            errors.reject("ecDatabasesNotFound");
        }

        List<ResultEntity> microstrParams = resultsManager.getResultByParent(resultParent, ResultType.microstrParams);
        if (microstrParams.isEmpty()) {
            errors.reject("ecMicrostructureNotFound");
        }

        List<ResultEntity> mezostrParams = resultsManager.getResultByParent(resultParent, ResultType.mezostrParams);
        if (mezostrParams.isEmpty()) {
            errors.reject("ecMezostructureNotFound");
        }
    }

    @Override
    protected SimulationEntity getTaskParent(TaskEntity task) {
        return simulationManager.findOne(task.getParentId(), null);
    }
}
