package ru.crp.lvms.validators.task.optimization;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.validation.Errors;
import ru.crp.cmls.calculations.solvers.xml.config.Type;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.enums.OptimizationCao;
import ru.crp.lvms.optimizations.optistruct.entities.*;
import ru.crp.lvms.optimizations.optistruct.enums.OptistructBaseSettingType;
import ru.crp.lvms.optimizations.optistruct.enums.OptistructSettingsType;
import ru.crp.lvms.optimizations.optistruct.managers.OptistructOptimizationSettingsManager;
import ru.crp.lvms.utils.NumericalInputConverter;

import java.util.List;
import java.util.Map;
import java.util.Objects;

import static java.util.stream.Collectors.toList;
import static org.springframework.util.StringUtils.isEmpty;
import static ru.crp.lvms.optimizations.optistruct.enums.OptistructBaseSettingType.*;
import static ru.crp.lvms.optimizations.optistruct.managers.OptistructAdditionalParametersManager.PLAIN_TYPE_ID;

@Component
public class OptistructTaskParamsValidator extends BaseOptimizationTaskParamsValidator {

    public static final int MAX_ALLOWED_ATTI_ID_COUNT = 10;

    @Autowired
    private OptistructOptimizationSettingsManager optistructOptimizationSettingsManager;

    @Autowired
    private NumericalInputConverter numericalInputConverter;

    @Override
    protected void doValidate(TaskEntity task, OptimizationEntity taskParent, Errors errors) {
        if (!Objects.equals(taskParent.getSolverCaoId(), OptimizationCao.Optistruct.getId())) {
            throw new IllegalArgumentException(String.format("Wrong validator is used for task with solver %s", OptimizationCao.valueOf(taskParent.getSolverCaoId())));
        }
        super.doValidate(task, taskParent, errors);
        if (errors.hasErrors()) {
            return;
        }

        // no additional validation is needed because we just use default tcl script to convert model to .fem format
        if (task.getType() == Type.optistructGenerateFEM) {
            return;
        }
        List<OptistructSet> oss = optistructOptimizationSettingsManager.findSelectedSets(taskParent.getId(), null).stream()
                .filter(OptistructSet::getIsOptimizationSpace)
                .collect(toList());
        if (oss.isEmpty()) {
            errors.reject("notEnoughOptimizationSpaces");
        }
        Map<OptistructBaseSettingType, OptistructOptimizationBaseSetting> baseSettings = optistructOptimizationSettingsManager.findBaseSettings(taskParent.getId(), null);
        validateBaseSettings(baseSettings, errors);
        if (errors.hasErrors()) {
            return;
        }

        List<OptistructOptimizationTF> targetFunctions = (List<OptistructOptimizationTF>)
                optistructOptimizationSettingsManager.findOptimizationSettings(taskParent.getId(), OptistructSettingsType.TargetFunction, null);
        validateTargetFunctions(targetFunctions, baseSettings, errors);

        List<OptistructOptimizationConstraint> constraints = (List<OptistructOptimizationConstraint>)
                optistructOptimizationSettingsManager.findOptimizationSettings(taskParent.getId(), OptistructSettingsType.Constraint, null);
        validateConstraints(constraints, baseSettings, errors);

        List<OptistructOptimizationTechnologicalConstraint> techConstraints = (List<OptistructOptimizationTechnologicalConstraint>)
                optistructOptimizationSettingsManager.findOptimizationSettings(taskParent.getId(), OptistructSettingsType.TechConstraint, null);
        validateTechConstraints(techConstraints, baseSettings, errors);
    }

    private void validateTechConstraints(List<OptistructOptimizationTechnologicalConstraint> techConstraints, Map<OptistructBaseSettingType, OptistructOptimizationBaseSetting> baseSettings, Errors errors) {
        Integer expectedTechConstraintsCount = baseSettings.get(NUM_TECH_CONSTRAINTS).getCurrentValue();
        if (techConstraints.size() < expectedTechConstraintsCount) {
            errors.reject("notEnoughTechConstraints");
            return;
        }
    }

    private void validateConstraints(List<OptistructOptimizationConstraint> constraints, Map<OptistructBaseSettingType, OptistructOptimizationBaseSetting> baseSettings, Errors errors) {
        Integer expectedConstraintsCount = baseSettings.get(NUM_CONSTRAINTS).getCurrentValue();
        if (constraints.size() < expectedConstraintsCount) {
            errors.reject("notEnoughConstraints");
            return;
        }
        constraints
                .stream()
                .limit(expectedConstraintsCount)
                .forEach(c -> {
                            if (isEmpty(c.getLowerBound()) && isEmpty(c.getUpperBound())) {
                                errors.reject("cParamNotSet"); }
                                if (c.getAttiId()!= null && c.getAttiValue()==null) {
                                    errors.reject("cLNSetNotSet");
                            }
                        }
                );
    }

    private void validateTargetFunctions(List<OptistructOptimizationTF> targetFunctions, Map<OptistructBaseSettingType, OptistructOptimizationBaseSetting> baseSettings, Errors errors) {
        Integer expectedTargetFuncsCount = baseSettings.get(NUM_TARGET_FUNCS).getCurrentValue();
        if (targetFunctions.size() < expectedTargetFuncsCount) {
            errors.reject("notEnoughTargetFuncs");
            return;
        }
        targetFunctions
                .stream()
                .limit(expectedTargetFuncsCount)
                .forEach(tf -> {
                    if (numericalInputConverter.count(tf.getAttiValue()) > MAX_ALLOWED_ATTI_ID_COUNT) {
                        errors.reject("tooManyAttiIds", new Object[]{MAX_ALLOWED_ATTI_ID_COUNT}, null);
                    }
                    if (Objects.equals(tf.getAttaId(), PLAIN_TYPE_ID) && isEmpty(tf.getAttaValue())) {
                        errors.reject("plainValueNotSet", new Object[]{"Atta"}, null);
                    }
                    if (Objects.equals(tf.getAttiId(), PLAIN_TYPE_ID) && isEmpty(tf.getAttiValue())) {
                        errors.reject("plainValueNotSet", new Object[]{"Atti"}, null);
                    }
                    if (tf.getAttiId()!= null && tf.getAttiValue()==null) {
                        errors.reject("tfLNSetNotSet");
                    }
                });
    }

    private void validateBaseSettings(Map<OptistructBaseSettingType, OptistructOptimizationBaseSetting> baseSettings, Errors errors) {
        baseSettings.forEach((type, settings) -> {
            Integer settingsValue = settings.getCurrentValue();
            if (settingsValue == null) {
                errors.reject("baseParamValueIsNotSet", new Object[]{settings.getName()}, null);
            } else if (settingsValue > type.getMaxValue() || settingsValue < type.getMinValue()) {
                errors.reject("baseParamValueInvalid", new Object[]{settings.getName(), type.getMaxValue() + 1}, null);
            }
        });
    }
}
