package ru.crp.lvms.validators.task.optimization;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.util.StringUtils;
import org.springframework.validation.Errors;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.enums.OptimizationCao;
import ru.crp.lvms.optimizations.tosca.entities.ToscaConstraint;
import ru.crp.lvms.optimizations.tosca.entities.ToscaOptimizationBaseSetting;
import ru.crp.lvms.optimizations.tosca.entities.ToscaOptimizationTechnologicalConstraint;
import ru.crp.lvms.optimizations.tosca.entities.ToscaTargetFunction;
import ru.crp.lvms.optimizations.tosca.enums.NodeOrElem;
import ru.crp.lvms.optimizations.tosca.enums.ToscaBaseSettingType;
import ru.crp.lvms.optimizations.tosca.enums.ToscaSettingsType;
import ru.crp.lvms.optimizations.tosca.managers.ToscaOptimizationSettingsManager;

import java.util.List;
import java.util.Map;
import java.util.Objects;

@Component
public class ToscaTaskParamsValidator extends BaseOptimizationTaskParamsValidator {
    @Autowired
    private ToscaOptimizationSettingsManager toscaOptimizationSettingsManager;

    @Override
    protected void doValidate(TaskEntity task, OptimizationEntity taskParent, Errors errors) {
        if (!Objects.equals(taskParent.getSolverCaoId(), OptimizationCao.Tosca.getId())) {
            throw new IllegalArgumentException(String.format("Wrong validator is used for task with solver %s", OptimizationCao.valueOf(taskParent.getSolverCaoId())));
        }
        super.doValidate(task, taskParent, errors);
        if (errors.hasErrors()) {
            return;
        }

        Map<ToscaBaseSettingType, ToscaOptimizationBaseSetting> baseSettings = toscaOptimizationSettingsManager.findBaseSettings(taskParent.getId(), null);
        validateBaseSettings(baseSettings, errors);
        if (errors.hasErrors()) {
            return;
        }

        List<ToscaTargetFunction> targetFunctions = (List<ToscaTargetFunction>) toscaOptimizationSettingsManager.findOptimizationSettings(taskParent.getId(), ToscaSettingsType.TargetFunc, null);
        validateTargetFunctions(targetFunctions, baseSettings, errors);

        List<ToscaConstraint> constraints = (List<ToscaConstraint>) toscaOptimizationSettingsManager.findOptimizationSettings(taskParent.getId(), ToscaSettingsType.Constraint, null);
        validateConstraints(constraints, baseSettings, errors);

        List<ToscaOptimizationTechnologicalConstraint> techConstraints = (List<ToscaOptimizationTechnologicalConstraint>) toscaOptimizationSettingsManager.findOptimizationSettings(taskParent.getId(), ToscaSettingsType.TechnologicalConstraint, null);
        validateTechConstraints(techConstraints, baseSettings, errors);
    }

    private void validateTargetFunctions(List<ToscaTargetFunction> targetFuncs, Map<ToscaBaseSettingType, ToscaOptimizationBaseSetting> baseSettings, Errors errors) {
        Integer expectedTargetFuncsCount = baseSettings.get(ToscaBaseSettingType.NUM_TARGET_FUNCS).getCurrentValue();
        if (targetFuncs.size() < expectedTargetFuncsCount) {
            errors.reject("notEnoughTargetFuncs");
            return;
        }
        for (int i = 0; i < expectedTargetFuncsCount; ++i) {
            ToscaTargetFunction funcToCheck = targetFuncs.get(i);
            if (funcToCheck.getOptimizationId() == 0 ||
                    StringUtils.isEmpty(funcToCheck.getName()) ||
                    funcToCheck.getTypeId() == null ||
                    funcToCheck.getParameterId() == null ||
                    (funcToCheck.getNodeOrElementId() == NodeOrElem.SCALAR.getId() && StringUtils.isEmpty(funcToCheck.getScalar()))) {
                errors.reject("notEnoughTargetFuncs");
                break;
            }
        }
    }

    private void validateConstraints(List<ToscaConstraint> constraints, Map<ToscaBaseSettingType, ToscaOptimizationBaseSetting> baseSettings, Errors errors) {
        Integer expectedConstraintsCount = baseSettings.get(ToscaBaseSettingType.NUM_CONSTRAINTS).getCurrentValue();
        if (constraints.size() < expectedConstraintsCount) {
            errors.reject("notEnoughConstraints");
            return;
        }
        for (int i = 0; i < expectedConstraintsCount; ++i) {
            ToscaConstraint constraintToCheck = constraints.get(i);
            if (constraintToCheck.getOptimizationId() == 0 ||
                    StringUtils.isEmpty(constraintToCheck.getName()) ||
                    constraintToCheck.getTypeId() == null ||
                    constraintToCheck.getParameterId() == null ||
                    constraintToCheck.getValue() == null ||
                    constraintToCheck.getIsAbsolute() == null) {
                errors.reject("notEnoughConstraints");
                break;
            }
        }
    }

    private void validateTechConstraints(List<ToscaOptimizationTechnologicalConstraint> techConstraints, Map<ToscaBaseSettingType, ToscaOptimizationBaseSetting> baseSettings, Errors errors) {
        Integer expectedTechConstraintsCount = baseSettings.get(ToscaBaseSettingType.NUM_TECH_CONSTRAINTS).getCurrentValue();
        if (techConstraints.size() < expectedTechConstraintsCount) {
            errors.reject("notEnoughTechConstraints");
            return;
        }
        for (int i = 0; i < expectedTechConstraintsCount; ++i) {
            ToscaOptimizationTechnologicalConstraint constraintToCheck = techConstraints.get(i);
            if (constraintToCheck.getOptimizationId() == 0 ||
                    StringUtils.isEmpty(constraintToCheck.getName()) ||
                    constraintToCheck.getTechnologicalConstraintId() == null) {
                errors.reject("notEnoughTechConstraints");
                break;
            }
        }
    }


    private void validateBaseSettings(Map<ToscaBaseSettingType, ToscaOptimizationBaseSetting> baseSettings, Errors errors) {
        if (baseSettings.size() != ToscaBaseSettingType.values().length) {
            errors.reject("notEnoughBaseParams");
            return;
        }
        baseSettings.forEach((type, settings) -> {
            if (type == ToscaBaseSettingType.NUM_SETS) {
                return;
            }
            Integer settingsValue = settings.getCurrentValue();
            if (settingsValue == null) {
                errors.reject("baseParamValueIsNotSet", new Object[]{settings.getName()}, null);
            } else if (settingsValue > type.getMaxValue() || settingsValue < type.getMinValue()) {
                errors.reject("baseParamValueInvalid", new Object[]{settings.getName(), type.getMaxValue() + 1}, null);
            }
        });
    }
}
