package ru.crp.lvms.web.controllers.construction.element;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.http.MediaType;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.multipart.MultipartFile;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.managers.construction.elements.ConstructionElementManager;
import ru.crp.lvms.managers.image.ImageHandler;
import ru.crp.lvms.model.AccessRights;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.construction.elements.ConstructionElementEntity;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.web.controllers.misc.ControllerImageNew;
import ru.crp.lvms.web.controllers.misc.FileHelper;
import ru.crp.lvms.web.dto.pagination.DataRequest;
import ru.crp.lvms.web.dto.pagination.PageingResponse;

import javax.annotation.Nonnull;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.UncheckedIOException;
import java.util.List;

import static ru.crp.lvms.utils.URLs.CONSTRUCTION_ELEMENTS_URL;
import static ru.crp.lvms.web.controllers.misc.CrudController.deleteSeveralEntities;

@RestController
@RequestMapping(value = CONSTRUCTION_ELEMENTS_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class ConstructionElementFilesController implements FileHelper, ControllerImageNew<ConstructionElementEntity> {
    @Autowired
    private FileManager fileManager;

    @Autowired
    private ConstructionElementsController constructionElementsController;

    @Autowired
    private ConstructionElementManager constructionElementManager;


    // CAD files api
    @RequestMapping(value = "/{id}/initialCadModel", method = RequestMethod.GET)
    public void getInitialCADModel(HttpServletResponse response,
                                   @PathVariable("id") Integer constructionElementId,
                                   @CurrentUser PortalUser currentUser) throws IOException {
        ConstructionElementEntity constructionElement = constructionElementManager
                .checkEntityExistsAndAccesible(constructionElementId, AccessRights.view, currentUser.getUser());
        if (constructionElement.getInitialCadModelId() == null) {
            throw new NotFoundException("Файл с исходной CAD моделью не найден");
        }
        FileEntity file = fileManager.findOne(constructionElement.getInitialCadModelId(), currentUser.getUser());
        writeFileToResponse(file, response, currentUser.getUser());
    }


    @PreAuthorize("hasRole('ROLE_USER')")
    @Nonnull
    @RequestMapping(value = "/{id}/initialCadModel", method = RequestMethod.POST)
    public ConstructionElementEntity updateInitialCADModelContent(HttpServletRequest request,
                                                                  @PathVariable("id") Integer entityId,
                                                                  @RequestPart MultipartFile file,
                                                                  @CurrentUser PortalUser currentUser) {
        constructionElementManager.updateFile(entityId, FileType.CADFile, file, currentUser.getUser());
        return constructionElementsController.get(request, entityId, currentUser);
    }


    @PreAuthorize("hasRole('ROLE_USER')")
    @Nonnull
    @RequestMapping(value = "/{id}/initialCadModel", method = RequestMethod.DELETE)
    public ConstructionElementEntity deleteInitialCADModel(HttpServletRequest request,
                                                           @PathVariable("id") Integer entityId,
                                                           @CurrentUser PortalUser currentUser) {
        ConstructionElementEntity constructionElementEntity = constructionElementManager.
                checkEntityExistsAndAccesible(entityId, AccessRights.update, currentUser.getUser());
        if (constructionElementEntity.getInitialCadModelId() != null) {
            constructionElementManager.deleteInitialCadModel(entityId,
                    constructionElementEntity.getInitialCadModelId(), currentUser.getUser());
        }
        return constructionElementsController.get(request, entityId, currentUser);
    }


    // CAD files
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/CADFiles", method = RequestMethod.POST)
    public FileEntity createCADFile(@PathVariable("id") Integer constructionElementId,
                                    @RequestPart MultipartFile file,
                                    @CurrentUser PortalUser currentUser) throws IOException {
        return constructionElementManager
                .createNewFile(constructionElementId, file, FileType.CADFile, currentUser.getUser());
    }


    @RequestMapping(value = "/{id}/CADFiles", method = RequestMethod.GET)
    public List<FileEntity> getCADFiles(@PathVariable("id") Integer constructionElementId,
                                        @CurrentUser PortalUser currentUser) throws IOException {
        return constructionElementManager.findConstructionElementFiles(constructionElementId, FileType.CADFile,
                currentUser.getUser());
    }


    @RequestMapping(value = "/{id}/CADFiles/list", method = RequestMethod.POST)
    public PageingResponse<FileEntity> getCADFilesPage(@PathVariable("id") Integer constructionElementId,
                                                       @RequestBody DataRequest dataRequest,
                                                       @CurrentUser PortalUser currentUser) throws IOException {
        Pageable pageable = dataRequest.getSpringPageable();
        PageImpl<FileEntity> page = constructionElementManager.findConstructionElementFilesPage(constructionElementId,
                FileType.CADFile, dataRequest.getFilters(), pageable, currentUser.getUser());
        return new PageingResponse<>(page.getContent(), dataRequest.getFilters(), pageable, page.getTotalElements());
    }


    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/CADFiles/{fileId}", method = RequestMethod.DELETE)
    public void deleteCADFile(@PathVariable("id") Integer constructionElementId,
                              @PathVariable("fileId") Integer fileId,
                              @CurrentUser PortalUser currentUser) throws IOException {
        constructionElementManager.deleteFile(constructionElementId, fileId, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/CADFiles", method = RequestMethod.DELETE)
    public void deleteSeveralCADFiles(@PathVariable("id") Integer constructionElementId,
                                      @RequestParam("ids") List<Integer> filesIds,
                                      @CurrentUser PortalUser currentUser) throws BadRequestException {
        deleteSeveralEntities(FileEntity.class, filesIds, (id) -> {
            try {
                deleteCADFile(constructionElementId, id, currentUser);
            } catch (IOException e) {
                throw new UncheckedIOException(e);
            }
        });
    }

    // CAE files
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/CAEFiles", method = RequestMethod.POST)
    public FileEntity createCAEFile(@PathVariable("id") Integer constructionElementId,
                                    @RequestPart MultipartFile file,
                                    @CurrentUser PortalUser currentUser) throws IOException {
        return constructionElementManager
                .createNewFile(constructionElementId, file, FileType.CAEFile, currentUser.getUser());
    }


    @RequestMapping(value = "/{id}/CAEFiles", method = RequestMethod.GET)
    public List<FileEntity> getCAEFiles(@PathVariable("id") Integer constructionElementId,
                                        @CurrentUser PortalUser currentUser) throws IOException {
        return constructionElementManager.findConstructionElementFiles(constructionElementId, FileType.CAEFile,
                currentUser.getUser());
    }


    @RequestMapping(value = "/{id}/CAEFiles/list", method = RequestMethod.POST)
    public PageingResponse<FileEntity> getCAEFilesPage(@PathVariable("id") Integer constructionElementId,
                                                       @RequestBody DataRequest dataRequest,
                                                       @CurrentUser PortalUser currentUser) throws IOException {
        Pageable pageable = dataRequest.getSpringPageable();
        PageImpl<FileEntity> page = constructionElementManager.findConstructionElementFilesPage(constructionElementId,
                FileType.CAEFile, dataRequest.getFilters(), pageable, currentUser.getUser());
        return new PageingResponse<>(page.getContent(), dataRequest.getFilters(), pageable, page.getTotalElements());
    }


    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/CAEFiles/{fileId}", method = RequestMethod.DELETE)
    public void deleteCAEFile(@PathVariable("id") Integer constructionElementId,
                              @PathVariable("fileId") Integer fileId,
                              @CurrentUser PortalUser currentUser) throws IOException {
        constructionElementManager.deleteFile(constructionElementId, fileId, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/CAEFiles", method = RequestMethod.DELETE)
    public void deleteSeveralCAEFiles(@PathVariable("id") Integer constructionElementId,
                                      @RequestParam("ids") List<Integer> filesIds,
                                      @CurrentUser PortalUser currentUser) throws BadRequestException {
        deleteSeveralEntities(FileEntity.class, filesIds, (id) -> {
            try {
                deleteCAEFile(constructionElementId, id, currentUser);
            } catch (IOException e) {
                throw new UncheckedIOException(e);
            }
        });
    }


    // CAO files
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/CAOFiles", method = RequestMethod.POST)
    public FileEntity createCAOFile(@PathVariable("id") Integer constructionElementId,
                                    @RequestPart MultipartFile file,
                                    @CurrentUser PortalUser currentUser) throws IOException {
        return constructionElementManager
                .createNewFile(constructionElementId, file, FileType.CAOFile, currentUser.getUser());
    }


    @RequestMapping(value = "/{id}/CAOFiles", method = RequestMethod.GET)
    public List<FileEntity> getCAOFiles(@PathVariable("id") Integer constructionElementId,
                                        @CurrentUser PortalUser currentUser) throws IOException {
        return constructionElementManager.findConstructionElementFiles(constructionElementId, FileType.CAOFile,
                currentUser.getUser());
    }


    @RequestMapping(value = "/{id}/CAOFiles/list", method = RequestMethod.POST)
    public PageingResponse<FileEntity> getCAOFilesPage(@PathVariable("id") Integer constructionElementId,
                                                       @RequestBody DataRequest dataRequest,
                                                       @CurrentUser PortalUser currentUser) throws IOException {
        Pageable pageable = dataRequest.getSpringPageable();
        PageImpl<FileEntity> page = constructionElementManager.findConstructionElementFilesPage(constructionElementId,
                FileType.CAOFile, dataRequest.getFilters(), pageable, currentUser.getUser());
        return new PageingResponse<>(page.getContent(), dataRequest.getFilters(), pageable, page.getTotalElements());
    }


    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/CAOFiles/{fileId}", method = RequestMethod.DELETE)
    public void deleteCAOFile(@PathVariable("id") Integer constructionElementId,
                              @PathVariable("fileId") Integer fileId,
                              @CurrentUser PortalUser currentUser) throws IOException {
        constructionElementManager.deleteFile(constructionElementId, fileId, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/CAOFiles", method = RequestMethod.DELETE)
    public void deleteSeveralCAOFiles(@PathVariable("id") Integer constructionElementId,
                                      @RequestParam("ids") List<Integer> filesIds,
                                      @CurrentUser PortalUser currentUser) throws BadRequestException {
        deleteSeveralEntities(FileEntity.class, filesIds, (id) -> {
            try {
                deleteCAOFile(constructionElementId, id, currentUser);
            } catch (IOException e) {
                throw new UncheckedIOException(e);
            }
        });
    }

    // design solutions
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/designSolutions", method = RequestMethod.POST)
    public FileEntity createDesignSolution(@PathVariable("id") Integer constructionElementId,
                                           @RequestPart MultipartFile file,
                                           @CurrentUser PortalUser currentUser) throws IOException {
        return constructionElementManager
                .createNewFile(constructionElementId, file, FileType.DesignSolutionFile, currentUser.getUser());
    }


    @RequestMapping(value = "/{id}/designSolutions", method = RequestMethod.GET)
    public List<FileEntity> getDesignSolutions(@PathVariable("id") Integer constructionElementId,
                                               @CurrentUser PortalUser currentUser) throws IOException {
        return constructionElementManager.findConstructionElementFiles(constructionElementId,
                FileType.DesignSolutionFile, currentUser.getUser());
    }


    @RequestMapping(value = "/{id}/designSolutions/list", method = RequestMethod.POST)
    public PageingResponse<FileEntity> getDesignSolutionsPage(@PathVariable("id") Integer constructionElementId,
                                                              @RequestBody DataRequest dataRequest,
                                                              @CurrentUser PortalUser currentUser) throws IOException {
        Pageable pageable = dataRequest.getSpringPageable();
        PageImpl<FileEntity> page = constructionElementManager.findConstructionElementFilesPage(constructionElementId,
                FileType.DesignSolutionFile, dataRequest.getFilters(), pageable, currentUser.getUser());
        return new PageingResponse<>(page.getContent(), dataRequest.getFilters(), pageable, page.getTotalElements());
    }


    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/designSolutions/{fileId}", method = RequestMethod.DELETE)
    public void deleteDesignSolution(@PathVariable("id") Integer constructionElementId,
                                     @PathVariable("fileId") Integer fileId,
                                     @CurrentUser PortalUser currentUser) throws IOException {
        constructionElementManager.deleteFile(constructionElementId, fileId, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/designSolutions", method = RequestMethod.DELETE)
    public void deleteSeveralDesignSolutions(@PathVariable("id") Integer constructionElementId,
                                             @RequestParam("ids") List<Integer> filesIds,
                                             @CurrentUser PortalUser currentUser) throws BadRequestException {
        deleteSeveralEntities(FileEntity.class, filesIds, (id) -> {
            try {
                deleteDesignSolution(constructionElementId, id, currentUser);
            } catch (IOException e) {
                throw new UncheckedIOException(e);
            }
        });
    }


    // load conditions

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/loadConditions", method = RequestMethod.POST)
    public FileEntity createLoadCondition(@PathVariable("id") Integer constructionElementId,
                                          @RequestPart MultipartFile file,
                                          @CurrentUser PortalUser currentUser) throws IOException {
        return constructionElementManager
                .createNewFile(constructionElementId, file, FileType.LoadCondition, currentUser.getUser());
    }


    @RequestMapping(value = "/{id}/loadConditions", method = RequestMethod.GET)
    public List<FileEntity> getLoadConditions(@PathVariable("id") Integer constructionElementId,
                                              @CurrentUser PortalUser currentUser) throws IOException {
        return constructionElementManager
                .findConstructionElementFiles(constructionElementId, FileType.LoadCondition, currentUser.getUser());
    }


    @RequestMapping(value = "/{id}/loadConditions/list", method = RequestMethod.POST)
    public PageingResponse<FileEntity> getLoadConditionsPage(@PathVariable("id") Integer constructionElementId,
                                                             @RequestBody DataRequest dataRequest,
                                                             @CurrentUser PortalUser currentUser) throws IOException {
        Pageable pageable = dataRequest.getSpringPageable();
        PageImpl<FileEntity> page = constructionElementManager.findConstructionElementFilesPage(constructionElementId,
                FileType.LoadCondition, dataRequest.getFilters(), pageable, currentUser.getUser());
        return new PageingResponse<>(page.getContent(), dataRequest.getFilters(), pageable, page.getTotalElements());
    }


    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/loadConditions/{fileId}", method = RequestMethod.DELETE)
    public void deleteLoadCondition(@PathVariable("id") Integer constructionElementId,
                                    @PathVariable("fileId") Integer fileId,
                                    @CurrentUser PortalUser currentUser) throws IOException {
        constructionElementManager.deleteFile(constructionElementId, fileId, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/loadConditions", method = RequestMethod.DELETE)
    public void deleteSeveralLoadConditions(@PathVariable("id") Integer constructionElementId,
                                            @RequestParam("ids") List<Integer> filesIds,
                                            @CurrentUser PortalUser currentUser) throws BadRequestException {
        deleteSeveralEntities(FileEntity.class, filesIds, (id) -> {
            try {
                deleteLoadCondition(constructionElementId, id, currentUser);
            } catch (IOException e) {
                throw new UncheckedIOException(e);
            }
        });
    }

    @Override
    public FileManager getFileManager() {
        return fileManager;
    }

    @Override
    public ImageHandler imageHandler() {
        return constructionElementManager;
    }

    @Override
    public ConstructionElementEntity getEntity(HttpServletRequest request, Integer entityId, PortalUser currentUser) {
        return constructionElementsController.get(request, entityId, currentUser);
    }
}
