package ru.crp.lvms.web.controllers.construction.element;

import org.jetbrains.annotations.NotNull;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.multipart.MultipartFile;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.managers.construction.elements.AdjacentElementManager;
import ru.crp.lvms.managers.construction.elements.ConstructionElementDocsManager;
import ru.crp.lvms.managers.construction.elements.ConstructionElementManager;
import ru.crp.lvms.managers.misc.BaseManagerSuggest;
import ru.crp.lvms.managers.misc.EventManager;
import ru.crp.lvms.model.entities.construction.elements.AdjacentElementsEntityTable;
import ru.crp.lvms.model.entities.construction.elements.ConstructionElementDocsEntity;
import ru.crp.lvms.model.entities.construction.elements.ConstructionElementEntity;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.web.controllers.misc.CrudControllerSuggest;

import javax.annotation.Nonnull;
import javax.servlet.http.HttpServletRequest;
import javax.validation.Valid;
import java.util.List;

import static ru.crp.lvms.utils.URLs.CONSTRUCTION_ELEMENTS_URL;

@RestController
@RequestMapping(value = CONSTRUCTION_ELEMENTS_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class ConstructionElementsController
        extends CrudControllerSuggest<ConstructionElementEntity, ConstructionElementEntity, ConstructionElementEntity> {

    @Autowired
    private ConstructionElementManager constructionElementManager;

    @Autowired
    private ConstructionElementDocsManager constructionElementDocsManager;

    @Autowired
    private EventManager eventManager;

    @Autowired
    private AdjacentElementManager adjacentElementManager;

    @Override
    protected BaseManagerSuggest<ConstructionElementEntity> getManager() {
        return constructionElementManager;
    }

    @Override
    protected EventManager getEventsManager() {
        return eventManager;
    }

    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(method = RequestMethod.POST)
    public ConstructionElementEntity createConstructionElement(@Valid @RequestBody ConstructionElementEntity in,
                                                               @CurrentUser PortalUser currentUser) {
        in.setOwnerId(currentUser.getUser().getId());
        return constructionElementManager.createAndRead(in, currentUser.getUser());
    }


    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(method = RequestMethod.POST, consumes = MediaType.MULTIPART_FORM_DATA_VALUE)
    public ConstructionElementEntity createConstructionElement(@RequestParam("name") String name,
                                                               @RequestParam("pid") Integer pid,
                                                               @RequestParam(required = false) MultipartFile image,
                                                               @RequestParam("file") MultipartFile cad,
                                                               @CurrentUser PortalUser currentUser) {
        ConstructionElementEntity newCE = new ConstructionElementEntity();
        newCE.setName(name);
        newCE.setPid(pid);
        constructionElementManager.validate(newCE, currentUser.getUser());
        return constructionElementManager.createAndRead(newCE, image, cad, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @NotNull
    @Override
    @RequestMapping(value = "/{id}", method = RequestMethod.DELETE)
    public ConstructionElementEntity delete(HttpServletRequest request, @PathVariable("id") int id, @CurrentUser PortalUser currentUser) {
        return super.delete(request,id,currentUser);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @Override
    @RequestMapping(method = RequestMethod.DELETE)
    public void deleteSeveral(HttpServletRequest request,
                              @RequestParam("ids") List<Integer> ids,
                              @CurrentUser PortalUser currentUser) throws BadRequestException {
        super.deleteSeveral(request,ids,currentUser);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @NotNull
    @RequestMapping(value = "/{id}", method = RequestMethod.PUT)
    public ConstructionElementEntity update(HttpServletRequest request, @PathVariable("id") int id,
                                          @Valid @RequestBody ConstructionElementEntity updateVO,
                                          @CurrentUser PortalUser user) {
        return super.update(request,id,updateVO,user);
    }

    @Nonnull
    @RequestMapping(value = "/{id}/descriptionFiles", method = RequestMethod.GET)
    public List<ConstructionElementDocsEntity> getDescriptionFiles(@PathVariable("id") Integer entityId,
                                                                   @CurrentUser PortalUser currentUser) {
        return constructionElementDocsManager.getFiles(entityId, currentUser.getUser());
    }


    @PreAuthorize("hasRole('ROLE_USER')")
    @Nonnull
    @RequestMapping(value = "/{id}/descriptionFiles/{docId}", method = RequestMethod.POST)
    public ConstructionElementDocsEntity updateDescriptionFile(@PathVariable("id") Integer entityId,
                                                               @PathVariable("docId") Integer docId,
                                                               @RequestPart MultipartFile file,
                                                               @CurrentUser PortalUser currentUser) {
        return constructionElementDocsManager.updateFile(entityId, docId, file, currentUser.getUser());
    }


    @PreAuthorize("hasRole('ROLE_USER')")
    @Nonnull
    @RequestMapping(value = "/{id}/descriptionFiles", method = RequestMethod.POST)
    public ConstructionElementDocsEntity createDescriptionFile(@PathVariable("id") Integer entityId,
                                                               @RequestPart MultipartFile file,
                                                               @CurrentUser PortalUser currentUser) {
        return constructionElementDocsManager.createFile(entityId, file, currentUser.getUser());
    }


    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/descriptionFiles/{docId}", method = RequestMethod.DELETE)
    public void deleteDescriptionFile(@PathVariable("id") Integer entityId,
                                      @PathVariable("docId") Integer docId,
                                      @CurrentUser PortalUser currentUser) {
        constructionElementDocsManager.deleteFile(entityId, docId, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/descriptionFiles", method = RequestMethod.DELETE)
    public void deleteSeveralDescriptionFiles(@PathVariable("id") Integer entityId,
                                             @RequestParam("ids") List<Integer> filesIds,
                                             @CurrentUser PortalUser currentUser) throws BadRequestException {
        deleteSeveralEntities(ConstructionElementDocsEntity.class, filesIds, (id) -> {
            deleteDescriptionFile(entityId, id, currentUser);
        });
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @Nonnull
    @RequestMapping(value = "/{id}/adjacents/{adjId}", method = RequestMethod.POST)
    public AdjacentElementsEntityTable createAdjacentConstruction(@PathVariable("id") Integer entityId,
                                                                  @PathVariable("adjId") Integer adjId,
                                                                  @CurrentUser PortalUser currentUser) {
        return adjacentElementManager.adjacentConstructions(entityId, adjId, currentUser.getUser());
    }


    @Nonnull
    @RequestMapping(value = "/{id}/adjacents", method = RequestMethod.GET)
    public List<AdjacentElementsEntityTable> getAdjacentConstructions(@PathVariable("id") Integer entityId,
                                                                      @CurrentUser PortalUser currentUser) {
        return adjacentElementManager.getAdjacents(entityId, currentUser.getUser());
    }


    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/adjacents/{adjId}", method = RequestMethod.DELETE)
    public void deleteAdjacentConstruction(@PathVariable("id") Integer entityId,
                                           @PathVariable("adjId") Integer adjId,
                                           @CurrentUser PortalUser currentUser) {
        adjacentElementManager.deleteAdjacent(entityId, adjId, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{id}/adjacents", method = RequestMethod.DELETE)
    public void deleteSeveralAdjacentConstructions(@PathVariable("id") Integer entityId,
                                                   @RequestParam("ids") List<Integer> adjIds,
                                                   @CurrentUser PortalUser currentUser) throws BadRequestException {
        deleteSeveralEntities(AdjacentElementsEntityTable.class, adjIds, (id) -> {
            deleteAdjacentConstruction(entityId, id, currentUser);
        });
    }
}
