package ru.crp.lvms.web.controllers.materials;

import org.jetbrains.annotations.NotNull;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.ServletRequestUtils;
import org.springframework.web.bind.annotation.*;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.managers.MaterialManager;
import ru.crp.lvms.managers.misc.BaseManagerSuggest;
import ru.crp.lvms.managers.misc.EventManager;
import ru.crp.lvms.managers.misc.PreferencesManager;
import ru.crp.lvms.model.entities.MaterialEntity;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.utils.Mapper;
import ru.crp.lvms.utils.URLs;
import ru.crp.lvms.web.controllers.misc.CrudControllerSuggest;
import ru.crp.lvms.web.dto.MaterialIn;
import ru.crp.lvms.web.dto.pagination.DataRequest;
import ru.crp.lvms.web.dto.pagination.PageingResponse;

import javax.annotation.Nonnull;
import javax.servlet.http.HttpServletRequest;
import javax.validation.Valid;
import java.util.List;

import static ru.crp.lvms.utils.URLs.MATERIALS_URL;

@RestController
@RequestMapping(value = MATERIALS_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class MaterialController extends CrudControllerSuggest<MaterialEntity, MaterialIn, MaterialEntity> {
    @Autowired
    private MaterialManager materialManager;

    @Autowired
    private EventManager eventManager;

    @Autowired
    private PreferencesManager preferencesManager;

    @Override
    public PreferencesManager getPreferencesManager() {
        return preferencesManager;
    }

    @Override
    protected BaseManagerSuggest<MaterialEntity> getManager() {
        return materialManager;
    }

    @Override
    protected EventManager getEventsManager() {
        return eventManager;
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @Nonnull
    @RequestMapping(method = RequestMethod.POST)
    public MaterialEntity createMaterial(@Valid @RequestBody MaterialIn materialIn, @CurrentUser PortalUser currentUser) {
        MaterialEntity newMaterial = Mapper.map(materialIn, entityClass);
        newMaterial.setOwnerId(currentUser.getUser().getId());
        return materialManager.createAndRead(newMaterial, currentUser.getUser());
    }


    @Nonnull
    @Override
    public List<MaterialEntity> getAll(HttpServletRequest request, @CurrentUser PortalUser currentUser) {
        List<MaterialEntity> all = super.getAll(request, currentUser);
        all.forEach(material -> extendMaterialIfNeed(request, material, currentUser));
        return all;
    }


    private void extendMaterialIfNeed(HttpServletRequest request, MaterialEntity material, PortalUser currentUser) {
        boolean useExtended = ServletRequestUtils.getBooleanParameter(request, URLs.EXTENDED_PARAM_NAME, false);
        if (useExtended) {
            materialManager.extendMaterial(material, currentUser.getUser());
        }
    }


    @Nonnull
    @Override
    public PageingResponse<MaterialEntity> page(HttpServletRequest request, @RequestBody DataRequest dataRequest,
                                                @CurrentUser PortalUser currentUser) {
        PageingResponse<MaterialEntity> page = super.page(request, dataRequest, currentUser);
        page.getContent().forEach(material -> extendMaterialIfNeed(request, material, currentUser));
        return page;
    }


    @Nonnull
    @Override
    public MaterialEntity get(HttpServletRequest request, @PathVariable("id") int id,
                              @CurrentUser PortalUser currentUser) {
        MaterialEntity materialOut = super.get(request, id, currentUser);
        extendMaterialIfNeed(request, materialOut, currentUser);
        return materialOut;
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @NotNull
    @Override
    @RequestMapping(value = "/{id}", method = RequestMethod.DELETE)
    public MaterialEntity delete(HttpServletRequest request, @PathVariable("id") int id, @CurrentUser PortalUser currentUser) {
        return super.delete(request,id,currentUser);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @Override
    @RequestMapping(method = RequestMethod.DELETE)
    public void deleteSeveral(HttpServletRequest request,
                              @RequestParam("ids") List<Integer> ids,
                              @CurrentUser PortalUser currentUser) throws BadRequestException {
        super.deleteSeveral(request,ids,currentUser);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @NotNull
    @RequestMapping(value = "/{id}", method = RequestMethod.PUT)
    public MaterialEntity update(HttpServletRequest request, @PathVariable("id") int id,
                        @Valid @RequestBody MaterialIn updateVO,
                        @CurrentUser PortalUser user) {
        return super.update(request,id,updateVO,user);
    }
}
