package ru.crp.lvms.web.controllers.technology;

import org.jetbrains.annotations.NotNull;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.http.MediaType;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.servlet.HandlerMapping;
import ru.crp.cmlb.model.SuggestRequest;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.managers.EquipmentManager;
import ru.crp.lvms.managers.TechnologyManager;
import ru.crp.lvms.managers.misc.BaseManager;
import ru.crp.lvms.managers.misc.EventManager;
import ru.crp.lvms.managers.misc.PreferencesManager;
import ru.crp.lvms.model.entities.EquipmentEntity;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.web.controllers.misc.ColumnsController;
import ru.crp.lvms.web.controllers.misc.CrudControllerColumns;
import ru.crp.lvms.web.dto.pagination.DataRequest;
import ru.crp.lvms.web.dto.pagination.PageingResponse;

import javax.annotation.Nonnull;
import javax.servlet.http.HttpServletRequest;
import javax.validation.Valid;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import static ru.crp.lvms.utils.URLs.EQUIPMENT_URL;
import static ru.crp.lvms.utils.URLs.TECHNOLOGY_URL;

@RestController
@RequestMapping(value = EQUIPMENT_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class EquipmentController extends CrudControllerColumns<EquipmentEntity, EquipmentEntity, EquipmentEntity> {
    @RestController
    @RequestMapping(value = TECHNOLOGY_URL + "/equipments", produces = MediaType.APPLICATION_JSON_VALUE)
    public static class EquipmentsColumnsController implements ColumnsController {
        @Autowired
        private PreferencesManager preferencesManager;

        @Override
        public Class getOutVoClass() {
            return EquipmentEntity.class;
        }

        @Override
        public PreferencesManager getPreferencesManager() {
            return preferencesManager;
        }
    }

    @Autowired
    private EquipmentManager equipmentManager;

    @Autowired
    private TechnologyManager technologyManager;

    @Autowired
    private EventManager eventManager;


    @Nonnull
    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(method = RequestMethod.POST)
    public EquipmentEntity createEquipment(HttpServletRequest request,
                                           @Valid @RequestBody EquipmentEntity in,
                                           @CurrentUser PortalUser currentUser) {
        EquipmentEntity newEquipment = equipmentManager.createAndRead(in, currentUser.getUser());
        Map pathVariables = (Map) request.getAttribute(HandlerMapping.URI_TEMPLATE_VARIABLES_ATTRIBUTE);
        int technologyId = Integer.parseInt((String) pathVariables.getOrDefault("technologyId", "-1"));
        technologyManager.addLinkToEquipment(technologyId, newEquipment.getId());
        return newEquipment;
    }


    @RequestMapping(value = "/suggest", method = RequestMethod.POST)
    public List<EquipmentEntity> suggest(@PathVariable("technologyId") Integer technologyId,
                                         @RequestBody @Valid SuggestRequest suggestRequest,
                                         @CurrentUser PortalUser currentUser) {
        suggestRequest.upsertQuery("technologyId", technologyId.toString());
        List<EquipmentEntity> suggest = equipmentManager.suggest(suggestRequest, currentUser.getUser());
        List<Integer> technologyEq = equipmentManager.getTechnologyEquipment(technologyId, currentUser.getUser()).stream()
                .map(EquipmentEntity::getId)
                .collect(Collectors.toList());
        return suggest.stream()
                .filter(equipment -> technologyEq.contains(equipment.getId()))
                .collect(Collectors.toList());

    }


    @NotNull
    @Override
    public List<EquipmentEntity> getAll(HttpServletRequest request, @CurrentUser PortalUser currentUser) {
        Map pathVariables = (Map) request.getAttribute(HandlerMapping.URI_TEMPLATE_VARIABLES_ATTRIBUTE);
        int technologyId = Integer.parseInt((String) pathVariables.getOrDefault("technologyId", "-1"));
        return equipmentManager.getTechnologyEquipment(technologyId, currentUser.getUser());
    }


    @NotNull
    @Override
    public PageingResponse<EquipmentEntity> page(HttpServletRequest request,
                                                 @RequestBody DataRequest dataRequest,
                                                 @CurrentUser PortalUser currentUser) {
        Map pathVariables = (Map) request.getAttribute(HandlerMapping.URI_TEMPLATE_VARIABLES_ATTRIBUTE);
        int technologyId = Integer.parseInt((String) pathVariables.getOrDefault("technologyId", "-1"));
        Pageable pageable = dataRequest.getSpringPageable();
        PageImpl<EquipmentEntity> page = equipmentManager
                .getTechnologyEquipmentPage(technologyId, dataRequest.getFilters(), pageable, currentUser.getUser());
        return new PageingResponse<>(page.getContent(), dataRequest.getFilters(), pageable, page.getTotalElements());
    }


    @PreAuthorize("hasRole('ROLE_USER')")
    @NotNull
    @Override
    public EquipmentEntity delete(HttpServletRequest request, @PathVariable("id") int id, @CurrentUser PortalUser user) {
        Map pathVariables = (Map) request.getAttribute(HandlerMapping.URI_TEMPLATE_VARIABLES_ATTRIBUTE);
        int technologyId = Integer.parseInt((String) pathVariables.getOrDefault("technologyId", "-1"));
        equipmentManager.deleteFromTechnology(id, technologyId, user.getUser());
        return get(request, id, user);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @Override
    public void deleteSeveral(HttpServletRequest request,
                              @RequestParam("ids") List<Integer> ids,
                              @CurrentUser PortalUser currentUser) throws BadRequestException {
        deleteSeveralEntities(EquipmentEntity.class, ids, (id) -> {
            delete(request, id, currentUser);
        });
    }

    @Override
    protected BaseManager<EquipmentEntity> getManager() {
        return equipmentManager;
    }


    @Override
    protected EventManager getEventsManager() {
        return eventManager;
    }
}
