CREATE TABLE technology_simulations (
  id            SERIAL,
  technology_id INTEGER NOT NULL,
  name          VARCHAR(255) NOT NULL,
  description   TEXT,
  owner_id      INTEGER                                           NOT NULL,
  date_created  TIMESTAMP WITHOUT TIME ZONE DEFAULT now()         NOT NULL,
  date_modified TIMESTAMP WITHOUT TIME ZONE DEFAULT now()         NOT NULL,


  PRIMARY KEY (id),
  FOREIGN KEY (owner_id) REFERENCES users (id),
  FOREIGN KEY (technology_id) REFERENCES technologies(id),
  UNIQUE (name)
);

CREATE TRIGGER technology_simulations_update
  BEFORE UPDATE ON technology_simulations
    FOR EACH ROW
      WHEN (OLD.* IS DISTINCT FROM NEW.*)
        EXECUTE PROCEDURE update_last_modified();


CREATE VIEW technology_simulations_join AS (
  SELECT ts.*,
  u.name AS owner,
  tsm.name AS technology_name

  FROM technology_simulations ts
  JOIN users u ON u.id = ts.owner_id
  JOIN technologies tsm ON tsm.id = ts.technology_id
);


-- technology simulations parameters
CREATE TABLE technology_simulations_parameters (
  id SERIAL,
  name VARCHAR NOT NULL,
  parent_param_id INTEGER,

  PRIMARY KEY (id),
  FOREIGN KEY (parent_param_id) REFERENCES technology_simulations_parameters (id) ON DELETE CASCADE,
  UNIQUE (name)
);

INSERT INTO technology_simulations_parameters(name) VALUES
 ('Радиус'),
 ('Количество слоев'),
 ('Шаг по вертикали'),
 ('Скорость по прямой'),
 ('Экструзия по прямой'),
 ('Скорость по дуге'),
 ('Экструзия по дуге'),
 ('Координаты области печати xmin'),
 ('Координаты области печати ymin'),
 ('Координаты области печати xmax'),
 ('Координаты области печати ymax'),
 ('Координаты рабочей области печати xmin'),
 ('Координаты рабочей области печати ymin'),
 ('Координаты рабочей области печати xmax'),
 ('Координаты рабочей области печати ymax');



CREATE TABLE technology_simulations_parameters_join (
  id SERIAL,
  technology_simulations_id INTEGER NOT NULL,
  param_id INTEGER NOT NULL,
  param_value VARCHAR,

  PRIMARY KEY (id),
  FOREIGN KEY (technology_simulations_id) REFERENCES technology_simulations (id) ON DELETE CASCADE,
  UNIQUE (param_id, technology_simulations_id)
);

CREATE VIEW technology_simulations_parameters_view AS (
SELECT
    tsp.*,
    params.name AS param_name,
    params.parent_param_id AS parent_param_id,
    parents.name AS parent_param_name

    FROM technology_simulations_parameters_join tsp
    INNER JOIN technology_simulations_parameters params ON tsp.param_id = params.id
    LEFT JOIN technology_simulations_parameters parents ON params.parent_param_id = parents.id
);

CREATE OR REPLACE RULE ts_parameters_CREATE AS ON INSERT TO technology_simulations_parameters_view DO INSTEAD (
  INSERT INTO technology_simulations_parameters_join (technology_simulations_id, param_id, param_value) VALUES
    (NEW.technology_simulations_id, NEW.param_id, NEW.param_value)
  RETURNING
    technology_simulations_parameters_join.*,
    NULL :: VARCHAR AS param_name,
    NULL :: INTEGER AS parent_param_id,
    NULL :: VARCHAR AS parent_param_name);

CREATE OR REPLACE RULE ts_parameters_UPDATE AS ON UPDATE TO technology_simulations_parameters_view DO INSTEAD (
  UPDATE technology_simulations_parameters_join
  SET param_value = NEW.param_value, param_id = NEW.param_id
  WHERE id = NEW.id );

CREATE OR REPLACE RULE ts_parameters_DELETE AS ON DELETE TO technology_simulations_parameters_view DO INSTEAD (
  DELETE FROM technology_simulations_parameters_join
  WHERE id = OLD.id
);



-- ribs parameters

CREATE TABLE technology_simulations_ribs_parameters (
  id                    SERIAL,
  position              INTEGER        NOT NULL,
  origin_x              DOUBLE PRECISION,
  origin_y DOUBLE PRECISION,
  angle DOUBLE PRECISION,
  steps DOUBLE PRECISION,
  technology_simulations_id INTEGER        NOT NULL,

  PRIMARY KEY (id),
  FOREIGN KEY (technology_simulations_id) REFERENCES technology_simulations (id),
  CONSTRAINT unique_order_technology UNIQUE (position, technology_simulations_id)
);

ALTER TABLE technology_simulations ADD COLUMN status_id INTEGER NOT NULL REFERENCES virtual_simulation_statuses(id) DEFAULT 1;

DROP VIEW technology_simulations_join;

CREATE VIEW technology_simulations_join AS (
  SELECT ts.*,
  u.name AS owner,
  tsm.name AS technology_name

  FROM technology_simulations ts
  JOIN users u ON u.id = ts.owner_id
  JOIN technologies tsm ON tsm.id = ts.technology_id
);

CREATE RULE ts_CREATE AS ON INSERT TO technology_simulations_join DO INSTEAD (
  INSERT INTO technology_simulations ( technology_id,name,owner_id) VALUES
    ( NEW.technology_id , NEW.name, NEW.owner_id)
  RETURNING
    technology_simulations.*,
    NULL :: VARCHAR(255) AS name,
    NULL :: VARCHAR AS owner);


CREATE RULE ts_UPDATE AS ON UPDATE TO technology_simulations_join DO INSTEAD (
  UPDATE technology_simulations
  SET technology_id = NEW.technology_id
  WHERE id = NEW.id);

CREATE RULE ts_DELETE AS ON DELETE TO technology_simulations_join DO INSTEAD (
  DELETE FROM technology_simulations
  WHERE id = OLD.id
);





