(function () {

  "use strict";

  window.App.Views.ReferencesView = Backbone.Marionette.LayoutView.extend({
    template: "./views/templates/project/base/references.html.ejs",
    className: "references-view",

    ui: {
      nodeItemText: ".js-toggle-full-info",
      graphRegion:  ".graph-region"
    },

    events: {
      "click @ui.nodeItemText": "onNodeItemTextClick"
    },

    initialize: function(options) {
      this.options = options;
    },

    onShow: function() {
      this.drawGraph();
    },

    drawGraph: function() {
      var that = this;
      this.$(".graph-region").css({ height: window.outerHeight - 400 });
      d3.json("rest/referenceGraph?id=" + this.options.id + "&type=" + this.options.type, function(error, treeData) {

        // Calculate total nodes, max label length
        var totalNodes = 0;
        var maxLabelLength = 0;
        var i = 0;
        var duration = 750;
        var root;

        // size of the diagram
        var viewerWidth = that.$(".graph-region").width();
        var viewerHeight = that.$(".graph-region").height();

        var tree = d3.layout.tree()
          .size([viewerHeight, viewerWidth]);

        // define a d3 diagonal projection for use by the node paths later on.
        var diagonal = d3.svg.diagonal()
          .projection(function(d) {
            return [d.y, d.x * 3];
          });

        // A recursive helper function for performing some setup by walking through all nodes
        function visit(parent, visitFn, childrenFn) {
          if (!parent) return;

          visitFn(parent);

          var children = childrenFn(parent);
          if (children) {
            var count = children.length;
            for (var i = 0; i < count; i++) {
              visit(children[i], visitFn, childrenFn);
            }
          }
        }

        // Call visit function to establish maxLabelLength
        visit(treeData, function(d) {
          totalNodes++;
          maxLabelLength = Math.max(d.name.length, maxLabelLength);
        }, function(d) {
          return d.children && d.children.length > 0 ? d.children : null;
        });


        // Define the zoom function for the zoomable tree
        function zoom() {
          svgGroup.attr("transform", "translate(" + d3.event.translate + ")scale(" + d3.event.scale + ")");
        }

        // define the zoomListener which calls the zoom function on the "zoom" event constrained within the scaleExtents
        var zoomListener = d3.behavior.zoom().scaleExtent([0.5, 3]).on("zoom", zoom);

        // define the baseSvg, attaching a class for styling and the zoomListener
        var baseSvg = d3.select(".graph-region").append("svg")
          .attr("width", viewerWidth)
          .attr("height", viewerHeight)
          .attr("class", "overlay")
          .call(zoomListener);

        // Helper functions for collapsing and expanding nodes.
        function collapse(d) {
          if (d.children) {
            d._children = d.children;
            d._children.forEach(collapse);
            d.children = null;
          }
        }

        function expand(d) {
          if (d._children) {
            d.children = d._children;
            d.children.forEach(expand);
            d._children = null;
          }
        }

        // Function to center node when clicked so node doesn't get lost when collapsing/moving with large amount of children.
        function centerNode(source) {
          var scale = zoomListener.scale();
          var x = -source.y0;
          var y = -source.x0;
          var x = x * scale + viewerWidth / 2;
          var y = y * scale + viewerHeight / 2;
          d3.select('g').transition()
            .duration(duration)
            .attr("transform", "translate(" + x + "," + y + ")scale(" + scale + ")");
          zoomListener.scale(scale);
          zoomListener.translate([x, y]);
        }

        // Toggle children function
        function toggleChildren(d) {
          if (d.children) {
            d._children = d.children;
            d.children = null;
          } else if (d._children) {
            d.children = d._children;
            d._children = null;
          }
          return d;
        }

        // Toggle children on click.
        function click(d) {
          if (d3.event.defaultPrevented) return; // click suppressed
          d = toggleChildren(d);
          update(d);
          centerNode(d);
        }

        // Open object in new tab
        function textClick(d) {
          window.open("#project/" + that.options.type.toLowerCase() + "s/" + d.id + "/parameters/view");
        }

        function searchTree(element){
          if(element.entrance == true){
            return element;
          } else if (element.children != null){
            var result = null;
            for(i=0; result == null && i < element.children.length; i++){
              result = searchTree(element.children[i]);
            }
            return result;
          }
          return null;
        }

        function update(source) {
          var levelWidth = [1];
          var childCount = function(level, n) {

            if (n.children && n.children.length > 0) {
              if (levelWidth.length <= level + 1) levelWidth.push(0);
              levelWidth[level + 1] += n.children.length;
              n.children.forEach(function(d) {
                childCount(level + 1, d);
              });
            }
          };

          childCount(0, root);

          var newHeight = d3.max(levelWidth) * 25; // 25 pixels per line
          tree = tree.size([newHeight, viewerWidth]);

          // Compute the new tree layout.
          var nodes = tree.nodes(root).reverse(),
            links = tree.links(nodes);


          nodes.forEach(function(d) {
            d.y = d.depth * maxLabelLength * 8;
          });

          var node = svgGroup.selectAll("g.node")
            .data(nodes, function(d) {
              return d.id || (d.id = ++i);
            });

          var nodeEnter = node.enter().append("g")
            .attr("class", function(d) {
              var nodeClass = d.entrance ? "node js-node-item ui-node-entrance" : "node js-node-item";
              return nodeClass;
            })
            .attr("transform", function(d) {
              return "translate(" + source.y0 + "," + source.x0 + ")";
            });

          nodeEnter.append("circle")
            .attr('class', function(d) {
              return d.entrance ? "nodeCircle entrance-circle" : "nodeCircle";
            })
            .attr("r", 0)
            .on('click', click);

          nodeEnter.append("text")
            .attr("class", "js-toggle-full-info")
            .attr("x", function(d) { return d.children || d._children ? -2 : 2; })
            .attr("dy", function(d) {
              return d.entrance ? "-1em" : "-.75em";
            })
            .attr('class', "nodeText")
            .attr("text-anchor", function(d) {
              return d.children ? "end" : "start";
            })
            .text(function(d) {
              return d.name;
            })
            .style("fill-opacity", 0)
            .on("click", textClick);

          // Update the text to reflect whether node has children or not.
          node.select('text')
            .attr("x", function(d) { return d.children || d._children ? -2 : 2; })
            .attr("text-anchor", function(d) {
              return d.children ? "end" : "start";
            })
            .text(function(d) {
              return d.name;
            });

          // Change the circle fill depending on whether it has children and is collapsed
          node.select("circle.nodeCircle")
            .attr("r", function(d) { return d.entrance ? 9 : 6.5 })
            .attr("fill", function(d) {
              var emptyColor = d.entrance ? "#FFDE75" : "#FFF";
              return d._children ? "lightsteelblue" : emptyColor;
            });

          // Transition nodes to their new position.
          var nodeUpdate = node.transition()
            .duration(duration)
            .attr("transform", function(d) {
              return "translate(" + d.y + "," + d.x * 3 + ")";
            });

          // Fade the text in
          nodeUpdate.select("text")
            .style("fill-opacity", 1);

          // Transition exiting nodes to the parent's new position.
          var nodeExit = node.exit().transition()
            .duration(duration)
            .attr("transform", function(d) {
              return "translate(" + source.y + "," + source.x + ")";
            })
            .remove();

          nodeExit.select("circle")
            .attr("r", 0);

          nodeExit.select("text")
            .style("fill-opacity", 0);

          // Update the links…
          var link = svgGroup.selectAll("path.link")
            .data(links, function(d) {
              return d.target.id;
            });

          // Enter any new links at the parent's previous position.
          link.enter().insert("path", "g")
            .attr("class", function(d) {
              return d.target.mainPath ? "bold-link link" : "link"
            })
            .attr("d", function(d) {
              var o = {
                x: source.x0 * 3,
                y: source.y0
              };
              return diagonal({
                source: o,
                target: o
              });
            });

          // Transition links to their new position.
          link.transition()
            .duration(duration)
            .attr("d", diagonal);

          // Transition exiting nodes to the parent's new position.
          link.exit().transition()
            .duration(duration)
            .attr("d", function(d) {
              var o = {
                x: source.x * 3,
                y: source.y
              };
              return diagonal({
                source: o,
                target: o
              });
            })
            .remove();

          // Stash the old positions for transition.
          nodes.forEach(function(d) {
            d.x0 = d.x * 3;
            d.y0 = d.y;
          });
        }

        // Append a group which holds all nodes and which the zoom Listener can act upon.
        var svgGroup = baseSvg.append("g");

        // Define the root
        root = treeData;
        root.x0 = viewerHeight * 3/ 2;
        root.y0 = 0;

        var entrance = searchTree(treeData)

        // Layout the tree initially and center on the root node.
        update(root);
        centerNode(entrance);
      });

    },

    onNodeItemClick: function(e) {
      var $target = $(e.target).closest(".node");
      var id = $target.data("id");
    }

  });
}());