package ru.crp.lvms;

import org.junit.After;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Value;
import ru.crp.lvms.security.PortalUser;

import javax.imageio.ImageIO;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Random;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.fail;
import static ru.crp.lvms.utils.TestUtils.*;
import static ru.crp.lvms.utils.URLs.USER_AVATAR_URL;


public class UsersAvatarsIT extends BaseIT {

    private Integer currentUserId;

    @Value("${avatars.path}")
    private String avatarsFolder;
    private Path pathToUserAvatar;
    private Path pathToDefaultAvatar;

    private BufferedImage userAvatar;
    private BufferedImage defaultAvatar;

    private String avatarUrl;

    @Override
    public void setup() throws Exception {
        super.setup();
        currentUserId = PortalUser.getCurrentUser().getUser().getId();
        avatarUrl = getAvatarUrl();

        Path pathToAvatars = Paths.get(avatarsFolder);
        if (!Files.exists(pathToAvatars)) {
            Files.createDirectories(pathToAvatars);
        }
        pathToDefaultAvatar = pathToAvatars.resolve("default.png");
        defaultAvatar = createTestImage();
        ImageIO.write(defaultAvatar, "png", pathToDefaultAvatar.toFile());

        pathToUserAvatar = pathToAvatars.resolve("" + currentUserId + ".png");
        userAvatar = createTestImage();
        ImageIO.write(userAvatar, "png", pathToUserAvatar.toFile());
    }


    @After
    public void cleanUp() throws IOException {
        Files.deleteIfExists(pathToDefaultAvatar);
        Files.deleteIfExists(pathToUserAvatar);
    }


    @Test
    public void itShouldBeAbleToGetUserAvatar() throws Exception {
        BufferedImage returnedUserAvatar = toImage(GETFile(mockMvc, avatarUrl));
        assertImageEquals(userAvatar, returnedUserAvatar);
    }


    @Test
    public void itShouldBeAbleToChangeUserAvatar() throws Exception {
        BufferedImage newAvatar = createTestImage();
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        ImageIO.write(newAvatar, "png", baos);

        UPLOADFile(mockMvc, avatarUrl, baos.toByteArray(), ".png", null);

        BufferedImage returnedUserAvatar = toImage(GETFile(mockMvc, avatarUrl));
        assertImageEquals(newAvatar, returnedUserAvatar);
    }


    @Test
    public void itShouldBeAbleToDeleteUserAvatar() {
        DELETEobject(mockMvc, avatarUrl);
        assertFalse(Files.exists(pathToUserAvatar));
    }


    @Test
    public void defaultAvatarShouldBeUsedIfNoCustomPresent() throws Exception {
        Files.deleteIfExists(pathToUserAvatar);
        BufferedImage returnedUserAvatar = toImage(GETFile(mockMvc, avatarUrl));
        assertImageEquals(defaultAvatar, returnedUserAvatar);
    }


    private BufferedImage createTestImage() {
        Random random = new Random();
        BufferedImage newImage = new BufferedImage(100, 100, BufferedImage.TYPE_INT_RGB);
        Graphics2D g = newImage.createGraphics();
        g.setColor(new Color(random.nextInt(255), random.nextInt(255), random.nextInt(255)));
        g.fillRect(0, 0, 100, 100);
        g.dispose();
        return newImage;
    }


    private String getAvatarUrl() {
        return USER_AVATAR_URL.replace("{id}", String.valueOf(currentUserId));
    }


    private void assertImageEquals(BufferedImage expected, BufferedImage actual) {
        if (expected.getWidth() != actual.getWidth()
                || expected.getHeight() != actual.getHeight()) {
            fail("Expected and actual avatar is not equals");
        }
        int width = expected.getWidth();
        int height = expected.getHeight();

        for (int y = 0; y < height; y++) {
            for (int x = 0; x < width; x++) {
                if (expected.getRGB(x, y) != actual.getRGB(x, y)) {
                    fail("Expected and actual avatar are not equals");
                }
            }
        }
    }


    private BufferedImage toImage(byte[] bytes) throws IOException {
        return ImageIO.read(new ByteArrayInputStream(bytes));
    }
}
