package ru.crp.lvms.managers.tasks.resultsHandlers;

import com.fasterxml.jackson.databind.ObjectMapper;
import org.apache.commons.csv.CSVFormat;
import org.apache.commons.csv.CSVPrinter;
import org.apache.commons.io.FileUtils;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import ru.crp.lvms.chart.ChartsSet;
import ru.crp.lvms.chart.LineChart;
import ru.crp.lvms.managers.tasks.ResultsParent;
import ru.crp.lvms.model.entities.ResultEntity;

import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.Random;

import static org.junit.Assert.*;
import static ru.crp.cmls.api.TaskParentType.effectiveCharacteristic;
import static ru.crp.cmls.calculations.solvers.xml.config.ResultType.curves;
import static ru.crp.lvms.managers.tasks.resultsHandlers.CurvesResultHandler.*;
import static ru.crp.lvms.model.entities.misc.CreatedBy.Auto;

public class CurvesResultHandlerTest {

    private static final int NUM_RECORDS = 100;
    private CurvesResultHandler resultHandler;
    private Path tempDir;
    private Path rawCurveData;
    private ObjectMapper objectMapper;
    private ResultsParent testResultsParent;

    @Before
    public void setup() throws IOException {
        tempDir = Files.createTempDirectory(null);
        objectMapper = new ObjectMapper();
        resultHandler = new CurvesResultHandler(tempDir, objectMapper);

        testResultsParent = new ResultsParent();
        testResultsParent.setType(effectiveCharacteristic);
        testResultsParent.setId(1);
        testResultsParent.setOwnerId(1);
        testResultsParent.setName("test-name");
    }

    @After
    public void cleanup() throws IOException {
        if (Files.isDirectory(tempDir)) {
            FileUtils.deleteDirectory(tempDir.toFile());
        }
    }

    // simple format means 4 columns in record
    @Test
    public void itShouldBeAbleToParseCSVInSimpleFormat() throws Exception {
        initCurvesData(true);
        testParsing(NUM_RECORDS);

    }

    // extended format means 8 columns in record
    @Test
    public void itShouldBeAbleToParseCSVInExtendedFormat() throws Exception {
        initCurvesData(false);
        testParsing(NUM_RECORDS * 2);
    }

    private void initCurvesData(boolean isSimpleFormat) throws IOException {
        rawCurveData = Files.createTempFile(tempDir, null, null);
        try (Writer out = Files.newBufferedWriter(rawCurveData)) {
            CSVFormat csvFormat = CSVFormat.DEFAULT.withDelimiter(',');
            CSVPrinter csvPrinter = new CSVPrinter(out, csvFormat);
            Random r = new Random();
            for (int i = 0; i < NUM_RECORDS; ++i) {
                if (isSimpleFormat) {
                    csvPrinter.printRecord(r.nextDouble(), r.nextDouble(), r.nextDouble(), r.nextDouble());
                } else {
                    csvPrinter.printRecord(r.nextDouble(), r.nextDouble(), r.nextDouble(), r.nextDouble(),
                            r.nextDouble(), r.nextDouble(), r.nextDouble(), r.nextDouble());
                }
            }
        }
    }

    private void testParsing(int expectedRecords) throws IOException {
        List<ResultEntity> results = resultHandler.doHandleResultsFile(rawCurveData, testResultsParent, Auto);
        // one for fiber and one for matrix. Expects fiber to be first
        assertEquals(2, results.size());
        testSingleResult(results.get(0), true, expectedRecords);
        testSingleResult(results.get(1), false, expectedRecords);
    }

    private void testSingleResult(ResultEntity result, boolean isFiber, int expectedRecords) throws IOException {
        assertEquals(result.getCreatedBy(), Auto);
        assertEquals(result.getType(), curves);
        Path parsedCurve = assertJsonFileExists(isFiber);
        try (Reader r = Files.newBufferedReader(parsedCurve)) {
            ChartsSet curveData = objectMapper.readValue(r, ChartsSet.class);

            assertNotNull(curveData.getCharts());
            LineChart chart = (LineChart) curveData.getCharts().get(0);
            assertEquals(expectedRecords, chart.getPoints().size());

            assertEquals(X_LABEL, curveData.getxAxisLabel());
            assertEquals(X_DIMENSION, curveData.getxDimension());
            if (isFiber) {
                assertEquals(FIBER_CURVE_NAME, curveData.getTitle());
                assertEquals(Y_FIBER_LABEL, curveData.getyAxisLabel());
                assertEquals(Y_FIBER_DIMENSION, curveData.getyDimension());
            } else {
                assertEquals(MATRIX_CURVE_NAME, curveData.getTitle());
                assertEquals(Y_MATRIX_LABEL, curveData.getyAxisLabel());
                assertEquals(Y_MATRIX_DIMENSION, curveData.getyDimension());
            }
        }
    }

    private Path assertJsonFileExists(boolean isFiber) {
        String parsedCurveName = CurvesResultHandler.getJsonFileNameByCsv(rawCurveData.getFileName().toString(), isFiber);
        Path parsedCurve = rawCurveData.resolveSibling(parsedCurveName);
        assertTrue(Files.isRegularFile(parsedCurve));
        return parsedCurve;
    }
}
