package ru.crp.lvms.optimizations;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.JsonNode;
import org.junit.Before;
import org.junit.Test;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.request.MockHttpServletRequestBuilder;
import ru.crp.lvms.BaseIT;
import ru.crp.lvms.model.entities.MaterialEntity;
import ru.crp.lvms.model.entities.technology.TechnologyEntityTable;
import ru.crp.lvms.optimizations.base.entities.OptimizationDependency;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.enums.OptimizationCao;
import ru.crp.lvms.utils.TestUtils;
import ru.crp.lvms.utils.URLs;

import java.io.IOException;
import java.util.List;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.delete;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.post;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.content;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

public class OptimizationIT extends BaseIT {
    private OptimizationEntity optimization;
    private String technologiesUrl;
    private String materialsUr;

    @Override
    @Before
    public void setup() throws Exception {
        super.setup();
        optimization = projectCreator.createOptimization();
        technologiesUrl = URLs.OPTIMIZATIONS_URL + "/" + optimization.getId() + "/technologies";
        materialsUr = URLs.OPTIMIZATIONS_URL + "/" + optimization.getId() + "/materials";
    }

    @Test
    public void itMustDeleteMFOptimization() {
        OptimizationEntity optimization = projectCreator.createOptimization(OptimizationCao.ModeFrontier);
        TestUtils.DELETEobject(mockMvc, URLs.OPTIMIZATIONS_URL + "/" + optimization.getId());
    }

    @Test
    public void itMustDeleteToscaOptimization() {
        OptimizationEntity optimization = projectCreator.createOptimization(OptimizationCao.Tosca);
        TestUtils.DELETEobject(mockMvc, URLs.OPTIMIZATIONS_URL + "/" + optimization.getId());
    }

    @Test
    public void itMustDeleteOptistructOptimization() {
        OptimizationEntity optimization = projectCreator.createOptimization(OptimizationCao.Optistruct);
        TestUtils.DELETEobject(mockMvc, URLs.OPTIMIZATIONS_URL + "/" + optimization.getId());
    }

    @Test
    public void itShouldBeAbleToAddTechnologyToOptimization() throws Exception {
        TechnologyEntityTable technology = projectCreator.getTechnology();
        addDependencyToOptimization(technologiesUrl, "entityId", technology.getId());
        List<OptimizationDependency> optimizationTechnologies = getOptimizationDependencies(technologiesUrl);
        assertEquals(1, optimizationTechnologies.size());
        assertObjectPropertiesEquals(optimization.getId(), technology.getId(), technology.getName(), optimizationTechnologies.get(0));
    }

    @Test
    public void itShouldBeAbleToRemoveTechnologyFromOptimization() throws Exception {
        TechnologyEntityTable technology = projectCreator.getTechnology();
        addDependencyToOptimization(technologiesUrl, "entityId", technology.getId());
        deleteObjectFromOptimization(technologiesUrl, technology.getId());
        List<OptimizationDependency> optimizationTechnologies = getOptimizationDependencies(technologiesUrl);
        assertTrue(optimizationTechnologies.isEmpty());
    }

    @Test
    public void addingNotExistentTechnologyToOptimizationShouldCause400() throws Exception {
        mockMvc.perform(delete(technologiesUrl + "/" + -1))
                .andExpect(status().isBadRequest())
                .andExpect(content().contentType(MediaType.APPLICATION_JSON))
                .andReturn();
    }


    @Test
    public void itShouldBeAbleToAddMaterialToOptimization() throws Exception {
        MaterialEntity material = projectCreator.getMaterial();
        addDependencyToOptimization(materialsUr, "entityId", material.getId());
        List<OptimizationDependency> optimizationMaterials = getOptimizationDependencies(materialsUr);
        assertEquals(1, optimizationMaterials.size());
        assertObjectPropertiesEquals(optimization.getId(), material.getId(), material.getName(), optimizationMaterials.get(0));
    }

    @Test
    public void itShouldBeAbleToRemoveMaterialFromOptimization() throws Exception {
        MaterialEntity material = projectCreator.getMaterial();
        addDependencyToOptimization(materialsUr, "entityId", material.getId());
        deleteObjectFromOptimization(materialsUr, material.getId());
        List<OptimizationDependency> optimizationMaterials = getOptimizationDependencies(materialsUr);
        assertTrue(optimizationMaterials.isEmpty());
    }

    @Test
    public void addingNotExistentMaterialToOptimizationShouldCause400() throws Exception {
        mockMvc.perform(delete(materialsUr + "/" + -1))
                .andExpect(status().isBadRequest())
                .andExpect(content().contentType(MediaType.APPLICATION_JSON))
                .andReturn();
    }

    private void addDependencyToOptimization(String url, String paramName, int objectId) throws Exception {
        MockHttpServletRequestBuilder builder = post(url)
                .param(paramName, "" + objectId)
                .contentType(MediaType.APPLICATION_JSON);
        mockMvc.perform(builder)
                .andExpect(status().isOk())
                .andReturn();
    }

    private void deleteObjectFromOptimization(String url, int objectId) throws Exception {
        mockMvc.perform(delete(url + "/" + objectId))
                .andExpect(status().isOk())
                .andReturn();
    }

    private List<OptimizationDependency> getOptimizationDependencies(String url) throws IOException {
        JsonNode response = TestUtils.GETobjects(mockMvc, url);
        return jacksonMapper.readValue(response.toString(), new TypeReference<List<OptimizationDependency>>() {
        });
    }

    private void assertObjectPropertiesEquals(int optimizationId, int dependencyId, String dependencyName, OptimizationDependency actualDependency) {
        assertEquals(optimizationId, actualDependency.optimizationId);
        assertEquals(dependencyId, actualDependency.id);
        assertEquals(dependencyName, actualDependency.name);
    }
}
