package ru.crp.lvms.optimizations.mf.results;

import com.fasterxml.jackson.databind.ObjectMapper;
import org.apache.commons.io.FileUtils;
import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import ru.crp.lvms.managers.tasks.ResultsParent;
import ru.crp.lvms.model.entities.misc.CreatedBy;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationConstraint;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationTargetFunc;
import ru.crp.lvms.optimizations.mf.enums.MFTargetFuncsType;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import static java.util.Arrays.asList;
import static java.util.Collections.singletonList;
import static org.junit.Assert.assertEquals;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
import static ru.crp.cmls.api.TaskParentType.optimization;

public class ModfrontierParamsHandlerTest {
    private ModeFrontierParamsHandler resultHandler;
    private Path tempDir;
    private ResultsParent testResultsParent;


    @Before
    public void init() throws IOException {
        tempDir = Files.createTempDirectory(null);
        ObjectMapper objectMapper = new ObjectMapper();
        resultHandler = new ModeFrontierParamsHandler(tempDir, objectMapper, (id) -> {
            MFOptimizationTargetFunc mock = mock(MFOptimizationTargetFunc.class);
            when(mock.getFuncId()).thenReturn(MFTargetFuncsType.MAX.getId());
            return singletonList(mock);
        }, (id) -> {
            MFOptimizationConstraint mock = mock(MFOptimizationConstraint.class);
            return singletonList(mock);
        });

        testResultsParent = new ResultsParent();
        testResultsParent.setType(optimization);
        testResultsParent.setId(1);
        testResultsParent.setOwnerId(1);
        testResultsParent.setName("test-name");
    }

    @Test
    public void testParsing() throws URISyntaxException {
        URI pathToTestScript = getClass().getClassLoader().getResource("test.des").toURI();
        resultHandler.doHandleResultsFile(Paths.get(pathToTestScript), testResultsParent, CreatedBy.Auto);
    }

    @After
    public void cleanup() throws IOException {
        if (Files.isDirectory(tempDir)) {
            FileUtils.deleteDirectory(tempDir.toFile());
        }
    }

    @Test
    public void testParetoSet() {
        String[] headers = new String[]{"<ID>", "<RID>", "par1", "out1", "obj1", "obj2", "c1", "<ERROR>", "<VIRTUAL>", "<FEASIBLE>", "<MARKED>", "<CATEGORY>"};
        ModFrontierParams params = new ModFrontierParams(headers);
        params.addResultRow(new String[]{"0", "0", "0", "0", "1", "1", "0", "false", "false", "true", "false", "1"});
        params.addResultRow(new String[]{"0", "0", "0", "0", "2", "1", "0", "false", "false", "true", "false", "2"});
        params.addResultRow(new String[]{"0", "0", "0", "0", "1", "2", "0", "false", "false", "true", "false", "3"});
        List<ModFrontierParamsRow> paretoSet = params.getParetoSet(asList(Double::compare, Double::compare));
        assertEquals(2, paretoSet.size());


        headers = new String[]{"<ID>", "<RID>", "par1", "out1", "obj1", "obj2", "c1", "<ERROR>", "<VIRTUAL>", "<FEASIBLE>", "<MARKED>", "<CATEGORY>"};
        params = new ModFrontierParams(headers);
        params.addResultRow(new String[]{"0", "0", "0", "0", "1", "1", "0", "false", "false", "true", "false", "1"});
        params.addResultRow(new String[]{"0", "0", "0", "0", "1", "1", "0", "false", "false", "true", "false", "2"});
        params.addResultRow(new String[]{"0", "0", "0", "0", "1", "1", "0", "false", "false", "true", "false", "3"});
        paretoSet = params.getParetoSet(asList(Double::compare, Double::compare));
        assertEquals(3, paretoSet.size());

        headers = new String[]{"<ID>", "<RID>", "par1", "out1", "obj1", "obj2", "c1", "<ERROR>", "<VIRTUAL>", "<FEASIBLE>", "<MARKED>", "<CATEGORY>"};
        params = new ModFrontierParams(headers);
        params.addResultRow(new String[]{"0", "0", "0", "0", "1", "1", "0", "false", "false", "true", "false", "1"});
        params.addResultRow(new String[]{"0", "0", "0", "0", "1", "1", "0", "false", "false", "true", "false", "2"});
        params.addResultRow(new String[]{"0", "0", "0", "0", "2", "2", "0", "false", "false", "true", "false", "3"});
        paretoSet = params.getParetoSet(asList(Double::compare, Double::compare));
        assertEquals(1, paretoSet.size());


        headers = new String[]{"<ID>", "<RID>", "par1", "out1", "obj1", "obj2", "c1", "<ERROR>", "<VIRTUAL>", "<FEASIBLE>", "<MARKED>", "<CATEGORY>"};
        params = new ModFrontierParams(headers);
        params.addResultRow(new String[]{"0", "0", "0", "0", "1", "1", "0", "false", "false", "true", "false", "1"});
        params.addResultRow(new String[]{"0", "0", "0", "0", "1", "1", "0", "false", "false", "true", "false", "2"});
        params.addResultRow(new String[]{"0", "0", "0", "0", "2", "2", "0", "false", "false", "true", "false", "3"});
        paretoSet = params.getParetoSet(asList(((Comparator<Double>) Double::compare).reversed(), Double::compare));
        assertEquals(3, paretoSet.size());
    }

    private List<String> getNumberedHeaders(String prefix, int count) {
        return getIndicesOrderedStream(count).
                map(index -> prefix + index)
                .collect(Collectors.toList());
    }

    private Stream<String> getIndicesOrderedStream(int count) {
        return IntStream.rangeClosed(1, count)
                .mapToObj(Integer::toString)
                .sorted();
    }

    private String[] generateHeaders(int numberedHeadersCount) {
        List<String> headers = new ArrayList<>(numberedHeadersCount * 4 + 7);
        headers.add("<ID>");
        headers.add("<RID>");
        headers.addAll(getNumberedHeaders("par", numberedHeadersCount));
        headers.addAll(getNumberedHeaders("out", numberedHeadersCount));
        headers.addAll(getNumberedHeaders("obj", numberedHeadersCount));
        headers.addAll(getNumberedHeaders("c", numberedHeadersCount));
        headers.add("<ERROR>");
        headers.add("<VIRTUAL>");
        headers.add("<FEASIBLE>");
        headers.add("<MARKED>");
        headers.add("<CATEGORY>");
        String[] array = new String[headers.size()];
        return headers.toArray(array);
    }

    private Double[] generateValuesList(double start, double step, int count) {
        Double[] values = new Double[count];
        for (int i = 0; i < count; i++) {
            values[i] = start + i * step;
        }
        return values;
    }

    private String[] generateResultRow(Double[] parValues, Double[] outValues, Double[] objValues, Double[] cValues, int numberedHeadersCount) {
        List<Integer> indices = getIndicesOrderedStream(numberedHeadersCount)
                .map(Integer::parseInt)
                .collect(Collectors.toList());
        List<String> results = new ArrayList<>();
        results.add("0");
        results.add("0");
        results.addAll(indices.stream()
                .map(i -> parValues[i - 1])
                .map(Object::toString)
                .collect(Collectors.toList()));
        results.addAll(indices.stream()
                .map(i -> outValues[i - 1])
                .map(Object::toString)
                .collect(Collectors.toList()));
        results.addAll(indices.stream()
                .map(i -> objValues[i - 1])
                .map(Object::toString)
                .collect(Collectors.toList()));
        results.addAll(indices.stream()
                .map(i -> cValues[i - 1])
                .map(Object::toString)
                .collect(Collectors.toList()));
        results.add("false");
        results.add("false");
        results.add("false");
        results.add("false");
        results.add("1");
        String[] array = new String[results.size()];
        return results.toArray(array);
    }

    @Test
    public void testParsingColumnsOrdering() {
        int numberedHeadersCount = 22;
        ModFrontierParams params = new ModFrontierParams(generateHeaders(numberedHeadersCount));
        Double[] parValues = generateValuesList(1, 4, numberedHeadersCount);
        Double[] outValues = generateValuesList(2, 4, numberedHeadersCount);
        Double[] objValues = generateValuesList(3, 4, numberedHeadersCount);
        Double[] cValues = generateValuesList(4, 4, numberedHeadersCount);
        params.addResultRow(generateResultRow(parValues, outValues, objValues, cValues, numberedHeadersCount));
        ModFrontierParamsRow row = params.getRows().get(0);
        Assert.assertArrayEquals(parValues, row.getVaryingParams().toArray());
    }
}
