package ru.crp.lvms.utils;

import com.fasterxml.jackson.databind.JsonNode;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.mock.web.MockMultipartFile;
import org.springframework.stereotype.Component;
import org.springframework.test.web.servlet.MockMvc;
import ru.crp.cmlb.model.SuggestRequest;
import ru.crp.cmlb.model.entities.CalculationStatus;
import ru.crp.cmls.api.TaskParentType;
import ru.crp.cmls.calculations.solvers.xml.config.Type;
import ru.crp.lvms.managers.ParametersManager;
import ru.crp.lvms.managers.TechnologyManager;
import ru.crp.lvms.managers.solvers.SolverConfigManager;
import ru.crp.lvms.model.dao.*;
import ru.crp.lvms.model.dao.effective.characteristics.ECParamDao;
import ru.crp.lvms.model.entities.*;
import ru.crp.lvms.model.entities.compositeMaterials.CompositeMaterialEntity;
import ru.crp.lvms.model.entities.construction.elements.ConstructionElementEntity;
import ru.crp.lvms.model.entities.effective.characteristics.ECEntity;
import ru.crp.lvms.model.entities.effective.characteristics.ECParamEntity;
import ru.crp.lvms.model.entities.misc.*;
import ru.crp.lvms.model.entities.simulations.SimulationEntity;
import ru.crp.lvms.model.entities.simulations.SimulationType;
import ru.crp.lvms.model.entities.simulations.SolvingStatus;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.model.entities.technology.TechnologyEntity;
import ru.crp.lvms.model.entities.technology.TechnologyEntityTable;
import ru.crp.lvms.model.entities.technology.TechnologyParametersEntity;
import ru.crp.lvms.model.entities.technology.TechnologyType;
import ru.crp.lvms.model.entities.tree.DefectNameEntity;
import ru.crp.lvms.model.entities.tree.OptimizationsTreeNode;
import ru.crp.lvms.model.entities.tree.TreeNode;
import ru.crp.lvms.optimizations.base.dao.OptimizationTypeDao;
import ru.crp.lvms.optimizations.base.dao.OptimizationTypeSolverCombinationDao;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationType;
import ru.crp.lvms.optimizations.base.entities.OptimizationTypeSolverCombination;
import ru.crp.lvms.optimizations.base.enums.OptimizationCao;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.web.dto.CompositeMaterialIn;
import ru.crp.lvms.web.dto.DefaultParameterValue;
import ru.crp.lvms.web.dto.MaterialIn;
import ru.crp.lvms.web.dto.pagination.DataRequest;
import ru.crp.lvms.web.dto.pagination.Pagination;

import javax.annotation.Nullable;
import javax.annotation.PostConstruct;
import java.math.BigDecimal;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.*;

import static java.util.Collections.*;
import static org.junit.Assert.assertFalse;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;
import static ru.crp.lvms.utils.TestUtils.*;
import static ru.crp.lvms.utils.URLs.*;

@SuppressWarnings("LocalVariableOfConcreteClass")
@Component
public class ProjectCreator {
    @Autowired
    private ParameterValuesDao parameterValuesDao;

    @Autowired
    private TechnologyDao technologyDao;

    @Autowired
    private FileDao fileDao;

    @Autowired
    private OptimizationTypeDao optimizationTypeDao;

    @Autowired
    private TechnologyManager technologyManager;

    @Autowired
    private DefectNameDao defectNameDao;

    @Autowired
    private ParametersDao parametersDao;

    @Autowired
    private ECParamDao ecParamDao;

    @Autowired
    private OptimizationTypeSolverCombinationDao optimizationTypeSolverCombinationDao;

    private MockMvc mockMvc;

    private TreeNode rootMaterialFolder;
    private TreeNode rootCompositeMaterialFolder;
    private TreeNode rootConstructionElementFolder;

    private MaterialEntity material;
    private CompositeMaterialEntity compositeMaterial;
    private TechnologyEntityTable technology;
    private ConstructionElementEntity constructionElementEntity;
    private OptimizationsTreeNode optimizationFolder;
    private DefectNameEntity defectName;
    private ECEntity ecEntity;
    private FileEntity sampleFile;
    private SimulationEntity simulationEntity;
    private OptimizationEntity optimizationEntity;

    @Value("${storage.root}")
    private String materialFilesFolder;

    private Path pathToMaterialFiles;

    private int testFileId;

    private List<OptimizationType> optimizationTypes;

    @PostConstruct
    public void init() {
        pathToMaterialFiles = Paths.get(materialFilesFolder);
    }

    public void setMockMvc(MockMvc mockMvc) {
        this.mockMvc = mockMvc;
    }

    public void createProjectData() throws Exception {
        rootMaterialFolder = createFolder(TreeType.materials);
        rootCompositeMaterialFolder = createFolder(TreeType.compositeMaterials);
        rootConstructionElementFolder = createFolder(TreeType.constructionElements);
        defectName = createDefectName();
        material = createMaterial();
        technology = createTechnologyEntity();
        compositeMaterial = createCompositeMaterial();
        optimizationTypes = optimizationTypeDao.findAll();
        assertFalse(optimizationTypes.isEmpty());
        constructionElementEntity = createConstructionElementEntity();
        ecEntity = createEffectiveCharacteristicEntity();
        createMaterialFile(false);

        sampleFile = createMaterialFile(true, FileType.SampleFile, newMaterialFile(true));

        testFileId = fileDao.findAll().get(0).getId();
        optimizationFolder = createOptimizationFolder();
        simulationEntity = createSimulation();
        optimizationEntity = createOptimization();
    }

    private DefectNameEntity createDefectName() {
        DefectNameEntity newDefect = new DefectNameEntity();
        newDefect.setName(unique("name"));
        newDefect = defectNameDao.insertAndRead(newDefect);
        return newDefect;
    }

    public List<OptimizationType> getOptimizationTypes() {
        return optimizationTypes;
    }

    public ParameterSample newParameterPoint(@Nullable BigDecimal temp, @Nullable BigDecimal value) {
        ParameterSample newPoint;
        try {
            newPoint = SetterChecker.createProxy(ParameterSample.class);
        } catch (IllegalAccessException | InstantiationException | NoSuchMethodException e) {
            throw new RuntimeException("Unable to CREATE ParameterSample", e);
        }
        if (temp == null) {
            temp = BigDecimal.valueOf(1d);
        }
        if (value == null) {
            value = BigDecimal.valueOf(1d);
        }
        newPoint.setTemperature(temp);
        newPoint.setParamValue(value);
        newPoint.setFuncParamValue(null);
        newPoint.setSimParamValue(null);
        return newPoint;
    }

    public ParameterSample addParameterToMaterial(Integer parameterId, Integer materialId) {
        ParameterSample newPoint = newParameterPoint(null, null);
        parameterValuesDao.insertParameterValuesForMaterial(materialId, parameterId, singletonList(newPoint));
        return newPoint;
    }

    public ParameterSample addParameterToMaterial(int parameterId, int materialId, ParameterSample newPoint) {
        parameterValuesDao.insertParameterValuesForMaterial(materialId, parameterId, singletonList(newPoint));
        return newPoint;
    }

    public List<ParameterSample> getParameterPoints(boolean isComposite, int materialId, int paramId) {
        String url = createMaterialParametersUrl(isComposite, materialId, paramId, null);
        JsonNode responseJson = GETobjects(mockMvc, url);
        return parameterSamplesFromJson(responseJson);
    }

    public List<DefaultParameterValue> getDefaultParameters(boolean isComposite, int materialId) {
        String url = createMaterialParametersUrl(isComposite, materialId, null, null);
        JsonNode responseJson = GETobjects(mockMvc, url);
        return defaultParameterValuesFromJson(responseJson);
    }

    public MaterialEntity getMaterial() {
        return material;
    }

    public TechnologyEntityTable getTechnology() {
        return technology;
    }

    public CompositeMaterialEntity getCompositeMaterial() {
        return compositeMaterial;
    }

    public MaterialIn newMaterial() {
        MaterialIn materialIn;
        try {
            materialIn = SetterChecker.createProxy(MaterialIn.class);
        } catch (Exception e) {
            throw new RuntimeException("Unable to INIT MaterialIn", e);
        }
        materialIn.setPid(rootMaterialFolder.getId());
        materialIn.setName(unique("material"));
        materialIn.setSource(unique("material-data"));
        materialIn.setDescription(unique("material-description"));
        materialIn.setLock(false);
        materialIn.setManufacture(unique("material-manufacturer"));
        materialIn.setTypeId(1);
        return materialIn;
    }

    public MaterialEntity createMaterial() {
        try {
            return createMaterial(null);
        } catch (Exception e) {
            throw new RuntimeException("Unable to CREATE ParametersEntityTable", e);
        }
    }

    public MaterialEntity createMaterial(@Nullable MaterialIn customMaterialIn) throws Exception {
        if (customMaterialIn == null) {
            customMaterialIn = newMaterial();
        }
        MaterialEntity materialEntity = CREATEobject(mockMvc, MATERIALS_URL, customMaterialIn, MaterialEntity.class);
        ParameterSample parameterSample = new ParameterSample();
        parameterSample.setTemperature(ParametersManager.DEFAULT_TEMPERATURE);
        parameterSample.setParamValue(BigDecimal.valueOf(0.111));
        parametersDao.findAll().stream().map(AuditableImpl::getId).forEach(i ->
                parameterValuesDao.upsertParameterValue(materialEntity.getId(), i, parameterSample)
        );
        return materialEntity;
    }

    public CompositeMaterialIn newCompositeMaterial() {
        CompositeMaterialIn materialIn;
        try {
            materialIn = SetterChecker.createProxy(CompositeMaterialIn.class);
        } catch (Exception e) {
            throw new RuntimeException("Unable to INIT MaterialIn", e);
        }
        materialIn.setName(unique("composite-material"));
        materialIn.setPid(rootCompositeMaterialFolder.getId());
        materialIn.setLock(false);
        materialIn.setTechnologyId(technology.getId());
        materialIn.setEquipmentId(null);
        materialIn.setFiberConcentration(20D);
        materialIn.setFiberId(material.getId());
        materialIn.setMatrixId(material.getId());
        materialIn.setEpoxideId(material.getId());
        materialIn.setSymmetricalLayers(true);
        materialIn.setDescription(unique("description"));
        materialIn.setManufacture(unique("manufacture"));
        materialIn.setTypeId(0);
        materialIn.setSource(unique("source"));
        return materialIn;
    }

    public CompositeMaterialEntity createCompositeMaterial() {
        try {
            return createCompositeMaterial(null);
        } catch (Exception e) {
            throw new RuntimeException("Unable to CREATE CompositeMaterialEntity", e);
        }
    }

    public CompositeMaterialEntity createCompositeMaterial(@Nullable CompositeMaterialIn customRequest) throws Exception {
        if (customRequest == null) {
            customRequest = newCompositeMaterial();
        }
        return CREATEobject(mockMvc, COMPOSITE_MATERIALS_URL, customRequest, CompositeMaterialEntity.class);
    }


    public DataRequest newDataRequest() throws Exception {
        DataRequest dataRequest = SetterChecker.createProxy(DataRequest.class);
        dataRequest.setFilters(emptyMap());
        dataRequest.setSort(emptyList());
        dataRequest.setPageable(new Pagination(1, 10000));
        return dataRequest;
    }

    public SuggestRequest newSuggestRequest(String column, String value) throws Exception {
        SuggestRequest suggestRequest = SetterChecker.createProxy(SuggestRequest.class);
        suggestRequest.setPageLimit(100);
        suggestRequest.setExcludedIds(new ArrayList<>());
        List<SuggestRequest.Entry> query = new ArrayList<>();
        query.add(new SuggestRequest.Entry(column, value));
        suggestRequest.setQueries(query);
        return suggestRequest;
    }

    public MaterialEntity newMaterialEntity() throws Exception {
        MaterialEntity materialEntity = SetterChecker.createProxy(MaterialEntity.class);
        materialEntity.setPid(100);
        materialEntity.setTypeId(1);
        materialEntity.setType("");
        materialEntity.setParameterValues(emptyList());
        materialEntity.setMaterialFiles(emptyList());
        materialEntity.setOwner("");
        materialEntity.setLock(false);
        materialEntity.setOwnerId(PortalUser.getCurrentUser().getUser().getId());
        materialEntity.setDescription("");
        materialEntity.setManufacture("");
        materialEntity.setSource("");

        materialEntity.setName(unique("material-name"));

        return materialEntity;
    }

    public FileEntity newMaterialFile(boolean isComposite) throws Exception {
        FileEntity materialFile = SetterChecker.createProxy(FileEntity.class);
        materialFile.setFileType(FileType.MaterialFile.name());
        materialFile.setFileTypeId(FileType.MaterialFile.getId());
        materialFile.setPid(isComposite ? compositeMaterial.getId() : material.getId());

        materialFile.setReferenceId(null);
        materialFile.setReference(null);

        materialFile.setName(unique("name"));
        materialFile.setDescription(unique("material-file-descr"));

        materialFile.setOwner(null);
        materialFile.setOwnerId(PortalUser.getCurrentUser().getUser().getId());

        materialFile.setFlag(StatusFlag.none);

        String fileName = unique("material-name");
        Path pathToFile = pathToMaterialFiles.resolve(fileName);
        Files.createFile(pathToFile);
        pathToFile.toFile().deleteOnExit();
        materialFile.setPathToFile(fileName);

        return materialFile;
    }


    public FileEntity createMaterialFile(boolean isComposite) throws Exception {
        return createMaterialFile(isComposite, null);
    }

    public FileEntity createMaterialFile(boolean isComposite, int pid) throws Exception {
        FileEntity fileEntity = newMaterialFile(isComposite);
        fileEntity.setPid(pid);
        return createMaterialFile(isComposite, fileEntity);
    }

    public FileEntity createMaterialFile(boolean isComposite, @Nullable FileEntity customMaterialFile) throws Exception {
        return createMaterialFile(isComposite, null, customMaterialFile);
    }

    public FileEntity createMaterialFile(boolean isComposite, @Nullable FileType fileType, @Nullable FileEntity customMaterialFile) throws Exception {
        if (customMaterialFile == null) {
            customMaterialFile = newMaterialFile(isComposite);
        }
        if (fileType == null) {
            fileType = isComposite ? FileType.CompositeMaterialFile : FileType.MaterialFile;
        }
        customMaterialFile.setFileTypeId(fileType.getId());
        return fileDao.insertAndRead(customMaterialFile);
    }


    public ParametersEntityTable newParameter() {
        ParametersEntityTable pet;
        try {
            pet = SetterChecker.createProxy(ParametersEntityTable.class);
        } catch (Exception e) {
            throw new RuntimeException("Unable to INIT ParametersEntityTable", e);
        }
        pet.setType(1);
        pet.setMaterialRelation(2);
        pet.setOwnerId(1);
        pet.setName(unique("name"));
        return pet;
    }


    public ParametersEntityTable createParameter() {
        try {
            return createParameter(null);
        } catch (Exception e) {
            throw new RuntimeException("Unable to CREATE ParametersEntityTable", e);
        }
    }


    public ParametersEntityTable createParameter(@Nullable ParametersEntityTable custom) throws Exception {
        if (custom == null) {
            custom = newParameter();
        }
        return CREATEobject(mockMvc, PARAMETERS_URL, custom, ParametersEntityTable.class);
    }


    public TreeNode newFolder() throws Exception {
        TreeNode newFolder = SetterChecker.createProxy(TreeNode.class);
        newFolder.setPid(null);
        newFolder.setItemType(ItemType.Folder);
        newFolder.setName(unique("folder-name"));
        return newFolder;
    }

    public TreeNode createFolder(TreeType treeType) throws Exception {
        return createFolder(treeType, null);
    }

    public TreeNode createFolder(TreeType treeType, @Nullable TreeNode customFolder) throws Exception {
        if (customFolder == null) {
            customFolder = newFolder();
        }
        return CREATEobject(mockMvc, TREE_URL + "/" + treeType, customFolder, TreeNode.class);
    }

    public TechnologyEntity newTechnology() {
        TechnologyEntity ti;
        try {
            ti = SetterChecker.createProxy(TechnologyEntity.class);
        } catch (Exception e) {
            throw new RuntimeException("Unable to INIT TechnologyEntity", e);
        }
        ti.setName(unique("name"));
        ti.setVideoUrl(unique("videoUrl"));
        ti.setDescriptionFileId(null);
        ti.setDescription(null);
        ti.setImageFileId(null);
        ti.setPathToImage(null);
        ti.setDescriptionFileName(null);
        ti.setLock(false);
        ti.setOwnerId(PortalUser.getCurrentUser().getUser().getId());
        ti.setDescriptionFileSource(null);
        ti.setEquipments(null);
        ti.setImageFileSource(null);
        ti.setType(TechnologyType.ADDITIVE_3CM);
        return ti;
    }

    public TechnologyEntity createTechnologyEntity() {
        String imageContent = "imageContent";
        String descrContent = "descrContent";
        MockMultipartFile image = new MockMultipartFile("imageFile", unique("file-name") + ".jpeg", null, imageContent.getBytes());
        MockMultipartFile description = new MockMultipartFile("descriptionFile", unique("file-name"), null, descrContent.getBytes());
        try {
            Map<String, String> params = new HashMap<>();
            params.put("name", unique("name"));
            params.put("technologyType", TechnologyType.ADDITIVE_3CM.name());
            return TestUtils.performRequestWithFiles(TECHNOLOGY_URL, mockMvc, params, null,
                    Arrays.asList(image, description), status().isOk(), TechnologyEntity.class);
        } catch (Exception e) {
            throw new RuntimeException("Unable to CREATE TechnologyEntity", e);
        }
    }

    public StandardEntity newStandard() {
        try {
            StandardEntity newStandard = SetterChecker.createProxy(StandardEntity.class);
            newStandard.setName(unique("standard-name"));
            newStandard.setOwnerId(PortalUser.getCurrentUser().getUser().getId());
            newStandard.setLoadFileId(null);
            newStandard.setStandardFileId(null);
            newStandard.setDescription(null);
            newStandard.setCreator("");
            return newStandard;
        } catch (Exception e) {
            throw new RuntimeException("Unable to INIT MaterialIn", e);
        }
    }

    public StandardEntity createStandard() {
        try {
            return createStandard(null);
        } catch (Exception e) {
            throw new RuntimeException("Unable to CREATE StandardEntity", e);
        }
    }

    public StandardEntity createStandard(@Nullable StandardEntity customStandard) throws Exception {
        if (customStandard == null) {
            customStandard = newStandard();
        }
        return CREATEobject(mockMvc, STANDARDS_URL, customStandard, StandardEntity.class);

    }

    public TechnologyParametersEntity newTechnologyParametersEntity() {
        TechnologyParametersEntity pet;
        try {
            pet = SetterChecker.createProxy(TechnologyParametersEntity.class);
        } catch (Exception e) {
            throw new RuntimeException("Unable to INIT TechnologyParametersEntity", e);
        }
        pet.setName(unique("name"));
        pet.setTechnologyId(technology.getId());
        pet.setValue(unique("value"));
        return pet;
    }


    public TechnologyParametersEntity createTechnologyParametersEntity() {
        try {
            return createTechnologyParametersEntity(null);
        } catch (Exception e) {
            throw new RuntimeException("Unable to CREATE ParametersEntityTable", e);
        }
    }

    public TechnologyParametersEntity createTechnologyParametersEntity(@Nullable TechnologyParametersEntity custom) throws Exception {
        if (custom == null) {
            custom = newTechnologyParametersEntity();
        }
        return CREATEobject(mockMvc, getUrlForTechnologyParameters(),
                custom, TechnologyParametersEntity.class);
    }

    public String getUrlForTechnologyParameters() {
        return TECHNOLOGY_URL + "/" + technology.getId() + "/parameters";
    }

    public PossibleDefectsEntity newPossibleDefectsEntity() {
        PossibleDefectsEntity pet;
        try {
            pet = SetterChecker.createProxy(PossibleDefectsEntity.class);
        } catch (Exception e) {
            throw new RuntimeException("Unable to INIT TechnologyParametersEntity", e);
        }
        pet.setDefectNameId(defectName.getId());
        pet.setName(unique("name"));
        pet.setTechnologyId(technology.getId());
        pet.setValue(unique("value"));
        pet.setActName(unique("actName"));
        pet.setDefectFileId(null);
        return pet;
    }


    public PossibleDefectsEntity createPossibleDefectsEntity() {
        try {
            return createPossibleDefectsEntity(null);
        } catch (Exception e) {
            throw new RuntimeException("Unable to CREATE ParametersEntityTable", e);
        }
    }

    public PossibleDefectsEntity createPossibleDefectsEntity(@Nullable PossibleDefectsEntity custom) throws Exception {
        if (custom == null) {
            custom = newPossibleDefectsEntity();
        }
        return CREATEobject(mockMvc, getUrlForPossibleDefects(),
                custom, PossibleDefectsEntity.class);
    }

    public String getUrlForPossibleDefects() {
        return TECHNOLOGY_URL + "/" + technology.getId() + "/defects";
    }


    public EquipmentEntity newEquipment() {
        EquipmentEntity pet;
        try {
            pet = SetterChecker.createProxy(EquipmentEntity.class);
        } catch (Exception e) {
            throw new RuntimeException("Unable to INIT TechnologyParametersEntity", e);
        }
        pet.setName(unique("name"));
        pet.setManufacturer(unique("manufacturer"));
        return pet;
    }


    public EquipmentEntity createEquipment() {
        try {
            return createEquipment(null);
        } catch (Exception e) {
            throw new RuntimeException("Unable to CREATE ParametersEntityTable", e);
        }
    }

    public EquipmentEntity createEquipment(@Nullable EquipmentEntity custom) throws Exception {
        if (custom == null) {
            custom = newEquipment();
        }
        EquipmentEntity newEquipment = CREATEobject(mockMvc, getUrlForEquipments(),
                custom, EquipmentEntity.class);
        technologyManager.addLinkToEquipment(technology.getId(), newEquipment.getId());
        return newEquipment;
    }

    public String getUrlForEquipments() {
        return TECHNOLOGY_URL + "/" + technology.getId() + "/equipments";
    }


    public ConstructionElementEntity newConstructionElementEntity() {
        ConstructionElementEntity pet = null;
        try {
            pet = SetterChecker.createProxy(ConstructionElementEntity.class);
        } catch (Exception e) {
            throw new RuntimeException("Unable to INIT ConstructionElementEntity", e);
        }
        pet.setName(unique("name"));
        pet.setOwnerId(PortalUser.getCurrentUser().getUser().getId());
        pet.setPid(rootConstructionElementFolder.getId());
        pet.setInitialCadModelId(testFileId);
        pet.setImageFileId(testFileId);
        pet.setDescription(null);
        pet.setInitialCadModelName(null);
        return pet;
    }


    public ConstructionElementEntity createConstructionElementEntity() {
        try {
            return createConstructionElementEntity(null);
        } catch (Exception e) {
            throw new RuntimeException("Unable to CREATE ConstructionElementEntity", e);
        }
    }

    public ConstructionElementEntity createConstructionElementEntity(@Nullable ConstructionElementEntity custom) throws Exception {
        if (custom == null) {
            custom = newConstructionElementEntity();
        }
        String imageContent = unique("imageContent");
        String cadContent = unique("cadContent");
        MockMultipartFile image = new MockMultipartFile("image", unique("image") + ".jpeg", null, imageContent.getBytes());
        MockMultipartFile cad = new MockMultipartFile("file", unique("cad"), null, cadContent.getBytes());
        try {
            Map<String, String> params = new HashMap<>();
            params.put("name", custom.getName());
            params.put("pid", custom.getPid().toString());

            return TestUtils.performRequestWithFiles(CONSTRUCTION_ELEMENTS_URL, mockMvc, params, null,
                    Arrays.asList(image, cad), status().isOk(), ConstructionElementEntity.class);
        } catch (Exception e) {
            throw new RuntimeException("Unable to CREATE ConstructionElementEntity", e);
        }
    }

    public OptimizationEntity newOptimization() {
        try {
            OptimizationTypeSolverCombination cadCaeCao =
                    optimizationTypeSolverCombinationDao.findAvailableCombinations().get(0);

            OptimizationEntity newOptimisationEntity = SetterChecker.createProxy(OptimizationEntity.class);
            newOptimisationEntity.setName(unique("OptimizationEntity"));
            newOptimisationEntity.setSolverCaoId(cadCaeCao.getCaoId());
            newOptimisationEntity.setSolverCaeId(cadCaeCao.getCaeId());
            newOptimisationEntity.setSolverCadId(cadCaeCao.getCadId());
            newOptimisationEntity.setCaoName(null);
            newOptimisationEntity.setCaeName(null);
            newOptimisationEntity.setCadName(null);
            newOptimisationEntity.setPid(optimizationFolder.getId());
            newOptimisationEntity.setOptimizationTypeId(cadCaeCao.getOptimizationTypeId());
            newOptimisationEntity.setOwnerId(PortalUser.getCurrentUser().getUser().getId());
            newOptimisationEntity.setPostprocessingStatusId(PostprocessingStatus.NotExecuted.getId());
            newOptimisationEntity.setSolvingStatusId(SolvingStatus.Created.getId());
            return newOptimisationEntity;
        } catch (IllegalAccessException | InstantiationException | NoSuchMethodException e) {
            throw new RuntimeException(e);
        }
    }

    public OptimizationEntity createOptimization() {
        return createOptimization(OptimizationCao.ModeFrontier);
    }

    public OptimizationEntity createOptimization(OptimizationCao selectedCAOSolver) {
        OptimizationEntity optimizationEntity = newOptimization();
        optimizationEntity.setSolverCaoId(selectedCAOSolver.getId());
        OptimizationTypeSolverCombination combination =
                optimizationTypeSolverCombinationDao.findAvailableForCao(selectedCAOSolver.getId()).get(0);
        optimizationEntity.setSolverCaeId(combination.getCaeId());
        optimizationEntity.setSolverCadId(combination.getCadId());
        OptimizationsTreeNode parentWithCorrectType = createOptimizationFolderWithType(combination.getOptimizationTypeId());
        optimizationEntity.setPid(parentWithCorrectType.getId());
        optimizationEntity.setOptimizationTypeId(combination.getOptimizationTypeId());
        return createOptimization(optimizationEntity);
    }

    public OptimizationEntity createOptimization(@Nullable OptimizationEntity customEntity) {
        if (customEntity == null) {
            customEntity = newOptimization();
        }

        try {
            return CREATEobject(mockMvc, OPTIMIZATIONS_URL, customEntity, OptimizationEntity.class);
        } catch (Exception ex) {
            throw new RuntimeException(ex);
        }
    }

    public OptimizationsTreeNode newOptimizationFolder() {
        try {
            OptimizationsTreeNode newFolder = SetterChecker.createProxy(OptimizationsTreeNode.class);
            newFolder.setName(unique("optimization-folder-name"));
            newFolder.setConstructionElementId(constructionElementEntity.getId());
            newFolder.setItemType(ItemType.OptimizationFolder);
            newFolder.setPid(constructionElementEntity.getId());
            newFolder.setOptimizationTypeId(optimizationTypes.get(0).getId());
            newFolder.setBaseOptimizationId(null);
            return newFolder;
        } catch (NoSuchMethodException | InstantiationException | IllegalAccessException e) {
            throw new RuntimeException(e);
        }
    }

    public OptimizationsTreeNode createOptimizationFolderWithType(Integer typeId) {
        OptimizationsTreeNode optimizationsTreeNode = newOptimizationFolder();
        optimizationsTreeNode.setOptimizationTypeId(typeId);
        return createOptimizationFolder(optimizationsTreeNode);
    }

    public OptimizationsTreeNode createOptimizationFolder() {
        return createOptimizationFolder(null);
    }


    public OptimizationsTreeNode createOptimizationFolder(@Nullable OptimizationsTreeNode customFolder) {
        if (customFolder == null) {
            customFolder = newOptimizationFolder();
        }
        try {
            return CREATEobject(mockMvc, TREE_URL + "/optimizations", customFolder, OptimizationsTreeNode.class);
        } catch (Exception e) {
            throw new RuntimeException();
        }
    }

    public ECEntity createEffectiveCharacteristicEntity() {
        try {
            return createEffectiveCharacteristicEntity(null);
        } catch (Exception e) {
            throw new RuntimeException("Unable to CREATE ECEntity", e);
        }
    }

    public ECEntity createEffectiveCharacteristicEntity(@Nullable ECEntity custom) throws Exception {
        if (custom == null) {
            custom = newEffectiveCharacteristicEntity();
        }
        ECEntity ecEntity = CREATEobject(mockMvc, URLs.EFFECTIVE_CHARACTERISTICS_URL,
                custom, ECEntity.class);
        ecParamDao.getAllParamIds().forEach(i -> {
            ECParamEntity ecParam = ecParamDao.getByCharacteristicIdAndParamId(ecEntity.getId(), i);
            if (ecParam == null) {
                ecParam = ecParamDao.insertAndRead(new ECParamEntity(ecEntity.getId(), i));
            }
            ecParam.setParameterValue("0");
            ecParamDao.update(ecParam);
        });
        return ecEntity;
    }

    public ECEntity newEffectiveCharacteristicEntity() {
        ECEntity newCharacteristic;
        try {
            newCharacteristic = SetterChecker.createProxy(ECEntity.class);
        } catch (Exception e) {
            throw new RuntimeException("Unable to INIT TechnologyParametersEntity", e);
        }
        newCharacteristic.setStatusId(SolvingStatus.Completed.getId());
        newCharacteristic.setParams(null);
        newCharacteristic.setDefects(null);
        newCharacteristic.setCompositeMaterialId(compositeMaterial.getId());
        newCharacteristic.setCompositeMaterialName(null);
        newCharacteristic.setOwnerId(PortalUser.getCurrentUser().getUser().getId());
        newCharacteristic.setMicrostructureFileName(null);
        newCharacteristic.setMicrostructureFileId(null);
        newCharacteristic.setName(unique("name"));
        newCharacteristic.setDescription(unique("description"));
        newCharacteristic.setResultsCount(null);
        newCharacteristic.setAverageThicknessIsSetable(false);
        return newCharacteristic;
    }

    public SimulationEntity newSimulation() {
        try {
            SimulationEntity newSimulation = SetterChecker.createProxy(SimulationEntity.class);
            newSimulation.setName(unique("standard-name"));
            newSimulation.setDescription("");
            newSimulation.setStatusId(SolvingStatus.Failed.getId());
            newSimulation.setTypeId(SimulationType.Bend.getId());
            newSimulation.setCompositeMaterialId(compositeMaterial.getId());
            newSimulation.setSampleFileId(sampleFile.getId());
            newSimulation.setEcId(ecEntity.getId());
            newSimulation.setStandardId(null);
            newSimulation.setOwnerId(PortalUser.getCurrentUser().getUser().getId());

            newSimulation.setExperimentFileId(null);
            newSimulation.setManualSizeB(0.0);
            newSimulation.setManualSizeH(0.0);
            newSimulation.setManualSizeL(0.0);
            newSimulation.setStatus(null);
            newSimulation.setCompositeMaterialName(null);
            newSimulation.setEcName(null);
            newSimulation.setExperimentFileName(null);
            newSimulation.setSampleFileName(null);
            newSimulation.setStandardName(null);
            newSimulation.setType(null);
            newSimulation.setResultsCount(null);
            return newSimulation;
        } catch (Exception e) {
            throw new RuntimeException("Unable to INIT SimulationEntity", e);
        }
    }

    public SimulationEntity createSimulation() {
        try {
            return createSimulation(null);
        } catch (Exception e) {
            throw new RuntimeException("Unable to CREATE SimulationEntity", e);
        }
    }

    public SimulationEntity createSimulation(@Nullable SimulationEntity customSimulation) throws Exception {
        if (customSimulation == null) {
            customSimulation = newSimulation();
        }
        return CREATEobject(mockMvc, URLs.SIMULATIONS_URL, customSimulation, SimulationEntity.class);

    }

    public TaskEntity createTask() {
        try {
            return createTask(null);
        } catch (Exception e) {
            throw new RuntimeException("Unable to CREATE TaskEntity", e);
        }
    }

    public TaskEntity createTask(@Nullable TaskEntity custom) throws Exception {
        if (custom == null) {
            custom = newTask();
        }
        return CREATEobject(mockMvc, TASKS_URL, custom, TaskEntity.class);
    }

    public TaskEntity newSolvingTask() {
        TaskEntity taskEntity = newTask();
        taskEntity.setParentId(simulationEntity.getId());
        taskEntity.setParentType(TaskParentType.simulation);
        return taskEntity;
    }

    public TaskEntity newTask() {
        TaskEntity newTask;
        try {
            newTask = SetterChecker.createProxy(TaskEntity.class);
        } catch (Exception e) {
            throw new RuntimeException("Unable to INIT TaskEntity", e);
        }
        ECEntity ec = createEffectiveCharacteristicEntity();
        newTask.setParentId(ec.getId());
        newTask.setParentType(TaskParentType.effectiveCharacteristic);
        newTask.setSolverName(SolverConfigManager.MOCK_SOLVER_NAME);
        newTask.setType(Type.solving);
        newTask.setOwnerId(PortalUser.getCurrentUser().getUser().getId());

        newTask.setStatusId(CalculationStatus.Waiting.getId());
        newTask.setQueuePriority(0);
        newTask.setProgress(null);
        newTask.setStartExecDate(null);
        newTask.setTaskPosition(null);
        newTask.setNumOfCores(null);
        newTask.setNode(null);
        newTask.setNodeId(null);
        newTask.setMemory(null);
        newTask.setStatus(null);
        newTask.setParentName(null);
        newTask.setAdditionalStartupArguments(null);
        newTask.setClusterId(1);
        newTask.setAutoSelectNode(false);

        return newTask;
    }

    public SupportLetterEntity newSupportLetter() {
        SupportLetterEntity supportLetterEntity;
        try {
            supportLetterEntity = SetterChecker.createProxy(SupportLetterEntity.class);
        } catch (Exception e) {
            throw new RuntimeException("Unable to INIT SupportLetterEntity", e);
        }
        supportLetterEntity.setUserName(null);
        supportLetterEntity.setIsProcessed(null);
        supportLetterEntity.setUserId(null);
        supportLetterEntity.setContent(unique("content"));
        supportLetterEntity.setPageUrl(unique("url"));
        return supportLetterEntity;
    }

    public SupportLetterEntity createSupportLetter() {
        try {
            return createSupportLetter(null);
        } catch (Exception e) {
            throw new RuntimeException("Unable to CREATE SupportLetterEntity", e);
        }
    }

    public SupportLetterEntity createSupportLetter(@Nullable SupportLetterEntity custom) throws Exception {
        if (custom == null) {
            custom = newSupportLetter();
        }
        return CREATEobject(mockMvc, SUPPORT_URL, custom, SupportLetterEntity.class);
    }
}
