package ru.crp.lvms.utils;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import org.springframework.http.MediaType;
import org.springframework.mock.web.MockMultipartFile;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.context.SecurityContext;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.test.util.ReflectionTestUtils;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.MvcResult;
import org.springframework.test.web.servlet.ResultMatcher;
import org.springframework.test.web.servlet.request.MockHttpServletRequestBuilder;
import org.springframework.test.web.servlet.request.MockMultipartHttpServletRequestBuilder;
import ru.crp.cmlb.model.entities.base.Entity;
import ru.crp.lvms.model.entities.*;
import ru.crp.lvms.model.entities.compositeMaterials.CompositeMaterialEntity;
import ru.crp.lvms.model.entities.construction.elements.AdjacentElementsEntityTable;
import ru.crp.lvms.model.entities.construction.elements.ConstructionElementDocsEntity;
import ru.crp.lvms.model.entities.construction.elements.ConstructionElementEntity;
import ru.crp.lvms.model.entities.effective.characteristics.ECEntity;
import ru.crp.lvms.model.entities.misc.BaseEntityTable;
import ru.crp.lvms.model.entities.misc.ItemType;
import ru.crp.lvms.model.entities.misc.TreeType;
import ru.crp.lvms.model.entities.simulations.SimulationEntity;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.model.entities.tasks.TaskIn;
import ru.crp.lvms.model.entities.technology.TechnologyEntity;
import ru.crp.lvms.model.entities.technology.TechnologyParametersEntity;
import ru.crp.lvms.model.entities.tree.TreeNode;
import ru.crp.lvms.optimizations.base.entities.OptimizationJsonSetting;
import ru.crp.lvms.optimizations.base.entities.OptimizationListSetting;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationBaseSetting;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationConstraint;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationTargetFunc;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationVaryingParam;
import ru.crp.lvms.optimizations.tosca.entities.ToscaConstraint;
import ru.crp.lvms.optimizations.tosca.entities.ToscaOptimizationTechnologicalConstraint;
import ru.crp.lvms.optimizations.tosca.entities.ToscaTargetFunction;
import ru.crp.lvms.optimizations.tosca.entities.ToscaTechnologicalConstraint;
import ru.crp.lvms.web.dto.DefaultParameterValue;
import ru.crp.lvms.web.dto.MaterialIn;
import ru.crp.lvms.web.dto.pagination.DataRequest;

import javax.annotation.Nullable;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.function.Function;

import static org.apache.commons.lang3.RandomStringUtils.randomAlphanumeric;
import static org.junit.Assert.assertTrue;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.content;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;
import static ru.crp.lvms.utils.URLs.COMPOSITE_MATERIALS_URL;
import static ru.crp.lvms.utils.URLs.MATERIALS_URL;

public class TestUtils {
    private static ObjectMapper jacksonMapper = new ObjectMapper();


    public static String unique(String prefix) {
        return prefix + randomAlphanumeric(5);
    }


    public static String toJson(@Nullable Object object) {
        try {
            if (object != null) {
                // dirty hack to handle @JsonGetter on TaskEntity.getType. TODO change it
                if (TaskEntity.class.isAssignableFrom(object.getClass())) {
                    TaskEntity task = (TaskEntity) object;
                    TaskIn taskIn = new TaskIn();

                    taskIn.setParentId(task.getParentId());
                    taskIn.setParentType(task.getParentType());
                    taskIn.setSolverName(task.getSolverName());
                    taskIn.setType(task.getType());
                    taskIn.setNumOfCores(task.getNumOfCores());
                    taskIn.setMemory(task.getMemory());
                    taskIn.setAdditionalStartupArguments(task.getAdditionalStartupArguments());
                    taskIn.setClusterId(task.getClusterId());
                    taskIn.setNodeId(task.getNodeId());

                    return jacksonMapper.writeValueAsString(taskIn);
                }
                assertTrue("You must use proxy to be sure that you filled all fields:\n" +
                                "Not used for type " + object.getClass() + ", do " +
                                "SetterChecker.createProxy(" + object.getClass().getSimpleName() + ".class)",
                        SetterChecker.checkIsProxy(object));
                assertTrue("Call all setters, you lazy:\n" +
                                "Not called for type " + object.getClass() + ": "
                                + SetterChecker.notCalledSetters(object),
                        SetterChecker.checkAllSettersCalled(object));

                return jacksonMapper.writeValueAsString(object);
            } else {
                return null;
            }
        } catch (JsonProcessingException e) {
            throw new RuntimeException("toJson failed", e);
        }

    }


    public static String toJsonNoProxy(@Nullable Object object) {
        try {
            if (object != null) {
                return jacksonMapper.writeValueAsString(object);
            } else {
                return null;
            }
        } catch (JsonProcessingException ex) {
            throw new RuntimeException(ex);
        }
    }


    public static <T> T GETobject(MockMvc mockMvc, String url, Class<T> responseClass) throws Exception {
        MvcResult result = mockMvc.perform(get(url))
                .andExpect(status().isOk())
                .andExpect(content().contentType(MediaType.APPLICATION_JSON))
                .andReturn();
        return jacksonMapper.readValue(result.getResponse().getContentAsString(), responseClass);
    }


    public static JsonNode GETPage(MockMvc mockMvc, String url, DataRequest dataRequest) throws Exception {
        MvcResult result = mockMvc.perform(post(url)
                .contentType(MediaType.APPLICATION_JSON)
                .content(toJson(dataRequest)))
                .andExpect(status().isOk())
                .andExpect(content().contentType(MediaType.APPLICATION_JSON))
                .andReturn();
        return jacksonMapper.readTree(result.getResponse().getContentAsString());
    }


    public static void GETAndExpectNotFound(MockMvc mockMvc, String url) throws Exception {
        mockMvc.perform(get(url))
                .andExpect(status().isNotFound())
                .andExpect(content().contentType(MediaType.APPLICATION_JSON))
                .andReturn();
    }


    public static MvcResult DELETEobject(MockMvc mockMvc, String url) {
        try {
            return mockMvc.perform(delete(url))
                    .andExpect(status().isOk())
                    .andReturn();
        } catch (Exception ex) {
            throw new RuntimeException(ex);
        }
    }


    public static void DELETEAndExpectConflict(MockMvc mockMvc, String url) throws Exception {
        mockMvc.perform(delete(url))
                .andExpect(status().isConflict())
                .andReturn();
    }


    public static JsonNode GETobjects(MockMvc mockMvc, String url) {
        try {
            MvcResult result = mockMvc.perform(get(url))
                    .andExpect(status().isOk())
                    .andExpect(content().contentType(MediaType.APPLICATION_JSON))
                    .andReturn();
            return jacksonMapper.readTree(result.getResponse().getContentAsString());
        } catch (Exception ex) {
            throw new RuntimeException(ex);
        }
    }


    public static Path downloadFile(MockMvc mockMvc, String url) throws Exception {
        MockHttpServletRequestBuilder builder = get(url).contentType(MediaType.APPLICATION_JSON);
        MvcResult result = mockMvc.perform(builder)
                .andExpect(status().isOk())
                .andExpect(content().contentType(MediaType.APPLICATION_OCTET_STREAM))
                .andReturn();
        Path tmpFile = Files.createTempFile(null, null);
        Files.write(tmpFile, result.getResponse().getContentAsByteArray());
        return tmpFile;
    }


    public static JsonNode CREATEobject(MockMvc mockMvc, String url, @Nullable Object object) throws Exception {
        MockHttpServletRequestBuilder builder = post(url).contentType(MediaType.APPLICATION_JSON);
        if (object != null) {
            builder = builder.content(toJsonNoProxy(object));
        }
        MvcResult result = mockMvc.perform(builder)
                .andExpect(status().isOk())
                .andExpect(content().contentType(MediaType.APPLICATION_JSON))
                .andReturn();
        return jacksonMapper.readTree(result.getResponse().getContentAsString());
    }


    public static <T> T CREATEobject(MockMvc mockMvc, String url, @Nullable Object object, Class<T> responseClass) throws Exception {
        MockHttpServletRequestBuilder builder = post(url).contentType(MediaType.APPLICATION_JSON);
        if (object != null) {
            builder = builder.content(toJson(object));
        }
        MvcResult result = mockMvc.perform(builder)
                .andExpect(status().isOk())
                .andExpect(content().contentType(MediaType.APPLICATION_JSON))
                .andReturn();
        return jacksonMapper.readValue(result.getResponse().getContentAsString(), responseClass);
    }


    public static void CREATEAndExpectBadRequest(MockMvc mockMvc, String url, Object object) throws Exception {
        mockMvc.perform(post(url)
                .contentType(MediaType.APPLICATION_JSON)
                .content(toJson(object)))
                .andExpect(status().isBadRequest())
                .andExpect(content().contentType(MediaType.APPLICATION_JSON))
                .andReturn();
    }


    public static <T> T performRequestWithFiles(String url, MockMvc mockMvc, @Nullable Map<String, String> params,
                                                @Nullable Object json,
                                                List<MockMultipartFile> files,
                                                ResultMatcher isOk,
                                                Class<T> responseClass) throws Exception {
        MockMultipartHttpServletRequestBuilder builder = buildMultipartRequest(url, params, json, files);
        MvcResult result = mockMvc.perform(builder)
                .andExpect(isOk)
                .andExpect(content().contentType(MediaType.APPLICATION_JSON))
                .andReturn();
        return jacksonMapper.readValue(result.getResponse().getContentAsString(), responseClass);
    }


    private static MockMultipartHttpServletRequestBuilder buildMultipartRequest(String url,
                                                                                @Nullable Map<String, String> params,
                                                                                @Nullable Object json,
                                                                                List<MockMultipartFile> files)
            throws JsonProcessingException, IllegalAccessException, InstantiationException, NoSuchMethodException {
        MockMultipartHttpServletRequestBuilder builder = fileUpload(url);
        files.forEach(builder::file);

        if (json != null) {
            MockMultipartFile objectIn = new MockMultipartFile("in", unique("fileName"),
                    MediaType.APPLICATION_JSON_VALUE, toJsonNoProxy(json).getBytes());
            builder.file(objectIn);
        }
        if (params != null) {
            params.forEach(builder::param);
        }
        return builder;
    }


    public static <T> T UPDATEobject(MockMvc mockMvc, String url, Object object, Class<T> responseClass, Function<Object, String> toJsoner) throws Exception {
        MvcResult result = mockMvc.perform(put(url)
                .contentType(MediaType.APPLICATION_JSON)
                .content(toJsoner.apply(object)))
                .andExpect(status().isOk())
                .andExpect(content().contentType(MediaType.APPLICATION_JSON))
                .andReturn();
        return jacksonMapper.readValue(result.getResponse().getContentAsString(), responseClass);
    }


    public static <T> T UPDATEobject(MockMvc mockMvc, String url, Object object, Class<T> responseClass) throws Exception {
        return UPDATEobject(mockMvc, url, object, responseClass, TestUtils::toJson);
    }


    public static void UPDATEAndExpectBadRequest(MockMvc mockMvc, String url, Object content) throws Exception {
        UPDATEAndExpectBadRequest(mockMvc, url, toJson(content));
    }


    public static void UPDATEAndExpectBadRequest(MockMvc mockMvc, String url, String content) throws Exception {
        mockMvc.perform(put(url)
                .contentType(MediaType.APPLICATION_JSON)
                .content(content))
                .andExpect(status().isBadRequest())
                .andExpect(content().contentType(MediaType.APPLICATION_JSON))
                .andReturn();
    }


    public static Map<String, String> toMap(MvcResult result) throws IOException {
        return jacksonMapper.readValue(result.getResponse().getContentAsString(),
                new TypeReference<Map<String, String>>() {
                });
    }


    public static void changeCurrentUserRole(String newRole) {
        SecurityContext context = SecurityContextHolder.getContext();
        Collection<? extends GrantedAuthority> authorities = context.getAuthentication().getAuthorities();
        GrantedAuthority next = authorities.iterator().next();
        ReflectionTestUtils.setField(next, "role", newRole);
    }

    public static String createObjectUrl(String baseUrl, Entity entity) {
        return createObjectUrl(baseUrl, entity.getId());
    }


    public static String createObjectUrl(String baseUrl, int id) {
        return String.join("/", baseUrl, String.valueOf(id));
    }


    public static String createPageUrl(String baseUrl) {
        return String.join("/", baseUrl, "list");
    }


    public static List<MaterialEntity> materialsFromJson(JsonNode materialsJson) {
        try {
            return jacksonMapper.readValue(materialsJson.toString(), new TypeReference<List<MaterialEntity>>() {
            });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse json to MaterialEntity", e);
        }
    }


    public static List<CompositeMaterialEntity> compositeMaterialsPageFromJson(JsonNode materialsJson) {
        try {
            return jacksonMapper.readValue(materialsJson.get("content").toString(),
                    new TypeReference<List<CompositeMaterialEntity>>() {
                    });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse json to CompositeMaterialEntity", e);
        }
    }


    public static List<CompositeMaterialEntity> compositeMaterialsFromJson(JsonNode materialsJson) {
        try {
            return jacksonMapper.readValue(materialsJson.toString(),
                    new TypeReference<List<CompositeMaterialEntity>>() {
                    });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse json to CompositeMaterialEntity", e);
        }
    }


    public static List<MaterialEntity> materialsPageFromJson(JsonNode materialsJson) {
        try {
            return jacksonMapper.readValue(materialsJson.get("content").toString(), new TypeReference<List<MaterialEntity>>() {
            });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse json to MaterialEntity", e);
        }
    }


    public static List<FileEntity> filesFromJson(JsonNode materialFilesJson) {
        try {
            return jacksonMapper.readValue(materialFilesJson.toString(), new TypeReference<List<FileEntity>>() {
            });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse json to FileEntity", e);
        }
    }


    public static List<FileEntity> filesPageFromJson(JsonNode materialsJson) {
        try {
            return jacksonMapper.readValue(materialsJson.get("content").toString(),
                    new TypeReference<List<FileEntity>>() {
                    });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse page json to FileEntity", e);
        }
    }


    public static List<? extends BaseEntityTable> parametersFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.toString(), new TypeReference<List<ParametersEntityTable>>() {
            });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse json to ParametersEntityTable", e);
        }
    }


    public static List<? extends BaseEntityTable> parametersPageFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.get("content").toString(),
                    new TypeReference<List<ParametersEntityTable>>() {
                    });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse page json to ParametersEntityTable", e);
        }
    }


    public static List<? extends BaseEntityTable> technologyFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.toString(), new TypeReference<List<TechnologyEntity>>() {
            });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse json to TechnologyEntity", e);
        }
    }


    public static List<? extends BaseEntityTable> technologyPageFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.get("content").toString(),
                    new TypeReference<List<TechnologyEntity>>() {
                    });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse page json to TechnologyEntity", e);
        }
    }


    public static List<? extends BaseEntityTable> technologyParametersFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.toString(), new TypeReference<List<TechnologyParametersEntity>>() {
            });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse json to TechnologyEntity", e);
        }
    }


    public static List<? extends BaseEntityTable> technologyParametersPageFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.get("content").toString(),
                    new TypeReference<List<TechnologyParametersEntity>>() {
                    });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse page json to TechnologyEntity", e);
        }
    }


    public static List<? extends BaseEntityTable> equipmentFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.toString(), new TypeReference<List<EquipmentEntity>>() {
            });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse json to EquipmentEntity", e);
        }
    }


    public static List<? extends BaseEntityTable> equipmentPageFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.get("content").toString(),
                    new TypeReference<List<EquipmentEntity>>() {
                    });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse page json to EquipmentEntity", e);
        }
    }


    public static List<? extends BaseEntityTable> possibleDefectsFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.toString(), new TypeReference<List<PossibleDefectsEntity>>() {
            });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse json to PossibleDefectsEntity", e);
        }
    }


    public static List<? extends BaseEntityTable> possibleDefectsPageFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.get("content").toString(),
                    new TypeReference<List<PossibleDefectsEntity>>() {
                    });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse page json to PossibleDefectsEntity", e);
        }
    }

    public static List<? extends BaseEntityTable> effectiveCharacteristicsFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.toString(), new TypeReference<List<ECEntity>>() {
            });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse json to PossibleDefectsEntity", e);
        }
    }

    public static List<? extends BaseEntityTable> effectiveCharacteristicPageFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.get("content").toString(),
                    new TypeReference<List<ECEntity>>() {
                    });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse page json to PossibleDefectsEntity", e);
        }
    }


    public static List<? extends BaseEntityTable> constructionElementFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.toString(), new TypeReference<List<ConstructionElementEntity>>() {
            });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse json to ConstructionElementEntity", e);
        }
    }

    public static List<? extends BaseEntityTable> constructionElementPageFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.get("content").toString(),
                    new TypeReference<List<ConstructionElementEntity>>() {
                    });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse page json to ConstructionElementEntity", e);
        }
    }


    public static List<DefaultParameterValue> defaultParameterValuesFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.toString(), new TypeReference<List<DefaultParameterValue>>() {
            });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse json to DefaultParameterValue", e);
        }
    }


    public static List<ParameterSample> parameterSamplesFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.toString(), new TypeReference<List<ParameterSample>>() {
            });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse json to ParameterSample", e);
        }
    }


    public static List<TreeNode> treeNodesFromJson(JsonNode materialsJson) throws IOException {
        return jacksonMapper.readValue(materialsJson.toString(), new TypeReference<List<TreeNode>>() {
        });
    }


    public static List<ItemType> itemTypesFromJson(JsonNode materialsJson) throws IOException {
        return jacksonMapper.readValue(materialsJson.toString(), new TypeReference<List<ItemType>>() {
        });
    }


    public static List<TreeType> treeTypesFromJson(JsonNode materialsJson) throws IOException {
        return jacksonMapper.readValue(materialsJson.toString(), new TypeReference<List<TreeType>>() {
        });
    }


    public static List<ConstructionElementDocsEntity> constructionElementDocsFromJson(JsonNode json) throws IOException {
        return jacksonMapper.readValue(json.toString(), new TypeReference<List<ConstructionElementDocsEntity>>() {
        });
    }


    public static List<AdjacentElementsEntityTable> adjacentElementsFromJson(JsonNode json) throws IOException {
        return jacksonMapper.readValue(json.toString(), new TypeReference<List<AdjacentElementsEntityTable>>() {
        });
    }


    public static List<StandardEntity> standardsFromJson(JsonNode materialsJson) {
        try {
            return jacksonMapper.readValue(materialsJson.toString(), new TypeReference<List<StandardEntity>>() {
            });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse json to StandardEntity", e);
        }
    }


    public static List<StandardEntity> standardsPageFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.get("content").toString(),
                    new TypeReference<List<StandardEntity>>() {
                    });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse page json to StandardEntity", e);
        }
    }


    public static TreeNode toTreeNode(MaterialIn materialIn) throws Exception {
        TreeNode treeNode = SetterChecker.createProxy(TreeNode.class);
        treeNode.setId(materialIn.getId());
        treeNode.setName(materialIn.getName());
        treeNode.setPid(materialIn.getPid());

        treeNode.setItemType(ItemType.Material);
        return treeNode;
    }


    public static TreeNode toTreeNode(FileEntity materialIn) throws Exception {
        TreeNode treeNode = SetterChecker.createProxy(TreeNode.class);
        treeNode.setId(materialIn.getId());
        treeNode.setName(materialIn.getName());
        treeNode.setPid(materialIn.getPid());

        treeNode.setItemType(ItemType.File);
        return treeNode;
    }


    public static String createMaterialParametersUrl(boolean isComposite, int materialId,
                                                     @Nullable Integer paramId, @Nullable Integer pointId) {
        String materialUrl = isComposite ? COMPOSITE_MATERIALS_URL : MATERIALS_URL;
        String paramsUrl = createObjectUrl(materialUrl, materialId) + "/parameters";
        if (paramId == null) {
            return paramsUrl;
        }
        paramsUrl = createObjectUrl(paramsUrl, paramId);
        if (pointId == null) {
            return paramsUrl;
        }
        return paramsUrl + "/point/" + pointId;
    }


    public static byte[] GETFile(MockMvc mockMvc, String url) throws Exception {
        MvcResult result = mockMvc.perform(get(url))
                .andExpect(status().isOk())
                .andReturn();
        return result.getResponse().getContentAsByteArray();
    }


    public static MvcResult UPLOADFile(MockMvc mockMvc, String url, byte[] content, @Nullable String fileExtension, @Nullable String name) throws Exception {
        if (fileExtension == null) {
            fileExtension = "";
        }
        if (name == null) {
            name = "file";
        }
        MockMultipartFile file = new MockMultipartFile(name, unique("file-name") + fileExtension, null, content);
        return mockMvc.perform(fileUpload(url).file(file))
                .andExpect(status().isOk())
                .andReturn();
    }

    public static List<SimulationEntity> simulationsFromJson(JsonNode simulationsJson) {
        try {
            return jacksonMapper.readValue(simulationsJson.toString(), new TypeReference<List<SimulationEntity>>() {
            });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse json to StandardEntity", e);
        }
    }


    public static List<SimulationEntity> simulationsPageFromJson(JsonNode simulationsJson) {
        try {
            return jacksonMapper.readValue(simulationsJson.get("content").toString(),
                    new TypeReference<List<SimulationEntity>>() {
                    });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse page json to StandardEntity", e);
        }
    }

    public static List<TaskEntity> tasksFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.toString(), new TypeReference<List<TaskEntity>>() {
            });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse json to TaskEntity", e);
        }
    }

    public static List<TaskEntity> tasksPageFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.get("content").toString(),
                    new TypeReference<List<TaskEntity>>() {
                    });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse page json to TaskEntity", e);
        }
    }

    public static List<MFOptimizationBaseSetting> optimizationBaseSettingsFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.toString(), new TypeReference<List<MFOptimizationBaseSetting>>() {
            });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse json to PossibleDefectsEntity", e);
        }
    }

    public static List<MFOptimizationVaryingParam> optimizationVaryingParamsFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.toString(), new TypeReference<List<MFOptimizationVaryingParam>>() {
            });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse json to PossibleDefectsEntity", e);
        }
    }

    public static List<MFOptimizationTargetFunc> optimizationTargetFuncsFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.toString(), new TypeReference<List<MFOptimizationTargetFunc>>() {
            });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse json to PossibleDefectsEntity", e);
        }
    }

    public static List<MFOptimizationConstraint> optimizationConstrainsFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.toString(), new TypeReference<List<MFOptimizationConstraint>>() {
            });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse json to PossibleDefectsEntity", e);
        }
    }

    public static List<ToscaConstraint> optimizationToskaConstraintFromJson(JsonNode ocJson) throws IOException {
        return jacksonMapper.readValue(ocJson.toString(), new TypeReference<List<ToscaConstraint>>() {
        });
    }

    public static List<ToscaTargetFunction> optimizationToskaTargetFunctionFromJson(JsonNode ocJson) throws IOException {
        return jacksonMapper.readValue(ocJson.toString(), new TypeReference<List<ToscaTargetFunction>>() {
        });
    }

    public static List<ToscaOptimizationTechnologicalConstraint> optimizationToskaTechnologicalConstraintFromJson(JsonNode ocJson) throws IOException {
        return jacksonMapper.readValue(ocJson.toString(), new TypeReference<List<ToscaOptimizationTechnologicalConstraint>>() {
        });
    }

    public static Map<String, List<ToscaTechnologicalConstraint>> optimizationToskaTechnologicalConstraintParameterFromJson(JsonNode ocJson) throws IOException {
        return jacksonMapper.readValue(ocJson.toString(), new TypeReference<Map<String, List<ToscaTechnologicalConstraint>>>() {
        });
    }

    public static List<OptimizationListSetting> optimizationListSettingFromJson(JsonNode ocJson) throws IOException {
        return jacksonMapper.readValue(ocJson.toString(), new TypeReference<List<OptimizationListSetting>>() {
        });
    }

    public static List<OptimizationJsonSetting> optimizationJsonSettingFromJson(JsonNode ocJson) throws IOException {
        return jacksonMapper.readValue(ocJson.toString(), new TypeReference<List<OptimizationJsonSetting>>() {
        });
    }

    public static List<SupportLetterEntity> supportLetterEntitiesFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.toString(), new TypeReference<List<SupportLetterEntity>>() {
            });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse json to SupportLetterEntity", e);
        }
    }

    public static List<SupportLetterEntity> supportLetterEntitiesPageFromJson(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.get("content").toString(),
                    new TypeReference<List<SupportLetterEntity>>() {
                    });
        } catch (IOException e) {
            throw new RuntimeException("Unable to parse page json to SupportLetterEntity", e);
        }
    }
}

