package ru.crp.lvms.managers;

import org.apache.commons.lang3.tuple.Pair;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.stereotype.Service;
import ru.crp.cmls.api.TaskParentType;
import ru.crp.cmls.calculations.solvers.xml.config.ResultType;
import ru.crp.lvms.chart.ChartsSet;
import ru.crp.lvms.chart.Point;
import ru.crp.lvms.exceptions.ReportGenerationException;
import ru.crp.lvms.managers.tasks.ResultsFilesManager;
import ru.crp.lvms.managers.tasks.ResultsParent;
import ru.crp.lvms.model.dao.ParameterValuesDao;
import ru.crp.lvms.model.entities.compositeMaterials.CompositeMaterialEntity;
import ru.crp.lvms.model.entities.ResultEntity;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.misc.CreatedBy;
import ru.crp.lvms.optimizations.base.dao.ResultsDao;

import java.io.IOException;
import java.math.BigDecimal;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import static java.util.Arrays.asList;
import static java.util.Collections.emptyList;
import static java.util.stream.Collectors.groupingBy;
import static java.util.stream.Collectors.toMap;

@Service
public class ResultsManager {
    private final static List<String> EC_PARAMS =
            Arrays.asList("Е x, МПа", "ν xy", "ν yx", "G xy",
                    "Е y, МПа", "ν xz", "ν zx", "G xz",
                    "Е z, МПа", "ν yz", "ν zy", "G yz");
    private final static List<String> SIM_PARAMS =
            Arrays.asList("Модуль упругости, МПа", "Предел прочности, МПа", "Относительное удлинение");
    public final static String MICRO_PICTURE_NAME = "file000.jpg";
    public final static String MEZO_PICTURE_NAME = "file001.jpg";

    public final static String FIBER_CURVE = "fiber";
    public final static String MATRIX_CURVE = "matrix";

    public static final int MATRIX_CURVE_VALUE_IDX = 4;
    public static final int FIBER_CURVE_VALUE_IDX = 3;

    public static final int STRENGTH_LIMIT_ID = 3;

    @Autowired
    private ResultsDao resultsDao;

    @Autowired
    private ResultsFilesManager resultsFilesManager;

    @Autowired
    private CompositeMaterialManager compositeMaterialManager;

    @Autowired
    private ParameterValuesDao parameterValuesDao;


    public Map<ResultType, List<ResultEntity>> getResultsMapByParent(int parentId, TaskParentType parentType, UserEntity currentUser) {
        Map<ResultType, List<ResultEntity>> results = findAllByParent(parentId, parentType).stream()
                .collect(groupingBy(ResultEntity::getType));
        Map<ResultType, List<ResultEntity>> emptyResults = Stream.of(ResultType.values())
                .filter(type -> !results.containsKey(type))
                .collect(toMap(Function.identity(), (unused) -> emptyList()));
        results.putAll(emptyResults);
        return results;
    }

    public Map<ResultType, List<ResultEntity>> getCustomResultsMapByParent(int parentId, TaskParentType parentType, ResultType resType, UserEntity currentUser) {
        Map<ResultType, List<ResultEntity>> results = findCustomResultsByParent(parentId, parentType,resType).stream()
                .collect(groupingBy(ResultEntity::getType));
        Map<ResultType, List<ResultEntity>> emptyResults = Stream.of(ResultType.values())
                .filter(type -> !results.containsKey(type))
                .collect(toMap(Function.identity(), (unused) -> emptyList()));
        results.putAll(emptyResults);
        return results;
    }


    public List<ResultEntity> findAllByParent(int parentId, TaskParentType parentType) {
        return resultsDao.findAllByParent(parentId, parentType);
    }

    public List<ResultEntity> findCustomResultsByParent(int parentId, TaskParentType parentType,ResultType type) {
        return resultsDao.findCustomResultsByParent(parentId, parentType,type);
    }


    public PageImpl<ResultEntity> findPageByParent(int parentId, TaskParentType parentType, Map<String, Object> filters,
                                                   Pageable page, UserEntity currentUser) {
        return resultsDao.findPageByParent(parentId, parentType, filters, page);
    }

    public PageImpl<ResultEntity> findCustomResultsPageByParent(int parentId, TaskParentType parentType, Map<String, Object> filters,
                                                   Pageable page,ResultType type, UserEntity currentUser) {
        return resultsDao.findCustomResultsPageByParent(parentId, parentType, filters, page, type);
    }

    public ResultEntity getResult(Integer resultId) {
        return resultsDao.findOne(resultId);
    }

    public void updateResult(ResultEntity updatedEntity) {
        resultsDao.update(updatedEntity);
    }

    // todo add deleting file from fs
    public void deleteResult(Integer resultId) {
        resultsDao.delete(resultId);
    }

    // Deletes previous results with same path
    public void saveAllWithoutDuplicates(List<ResultEntity> extractedResults) {
        resultsDao.deleteWithPaths(extractedResults.stream().map(ResultEntity::getPathToFile).collect(Collectors.toList()));
        resultsDao.batchInsert(extractedResults);
    }

    public List<ResultEntity> getResultByParent(ResultsParent resultParent, ResultType resultType) {
        return resultsDao.getResultByParentAndType(resultParent.getId(), resultParent.getType(), resultType);
    }

    public void deleteResults(ResultsParent parent, boolean deleteFiles, CreatedBy... createdBy) {
        resultsDao.deleteResults(parent.getId(), parent.getType(), asList(createdBy));
        if (deleteFiles) {
            resultsFilesManager.deleteResults(parent);
        }
    }


    public List<Pair<String, String>> getECResultValues(Integer id, ResultType resultType) throws IOException {
        ResultsParent resultParent = new ResultsParent(id, TaskParentType.effectiveCharacteristic);
        List<ResultEntity> params = getResultByParent(resultParent, resultType);
        if (params.isEmpty()) {
            throw new ReportGenerationException("Файл с параметрами не найден");
        }
        Path resultFile = resultsFilesManager.getPathToResult(params.get(0));
        String[] wordsInResultsFile = Files.lines(resultFile)
                .collect(Collectors.joining(""))
                .trim()
                .split("\\s+");
        return IntStream.range(0, Math.min(EC_PARAMS.size(), wordsInResultsFile.length))
                .mapToObj(i -> Pair.of(EC_PARAMS.get(i), wordsInResultsFile[i]))
                .collect(Collectors.toList());
    }


    public List<String> extractValuesLines(Integer id) throws IOException {
        ResultsParent resultParent = new ResultsParent(id, TaskParentType.simulation);
        List<ResultEntity> params = getResultByParent(resultParent, ResultType.values);
        if (params.isEmpty()) {
            throw new ReportGenerationException("Файл с параметрами не найден");
        }
        Path resultFile = resultsFilesManager.getPathToResult(params.get(0));
        List<String> workingLines = Files.readAllLines(resultFile);
        // First three lines are heading
        workingLines = workingLines.subList(3, workingLines.size());
        return workingLines.stream().map(p -> {
            String[] splitted = p.split(" ");
            return splitted[splitted.length - 1];
        }).collect(Collectors.toList());
    }

    public List<Pair<String, String>> getSimulationResultValues(List<String> values) throws IOException {
        return IntStream.range(0, Math.min(SIM_PARAMS.size(), values.size()))
                .mapToObj(i -> Pair.of(SIM_PARAMS.get(i), values.get(i)))
                .collect(Collectors.toList());
    }


    public ResultEntity getSimulationResultCurve(Integer simId, String curveType) {
        return resultsDao
                .getResultByParentAndType(simId, TaskParentType.simulation, ResultType.curves)
                .stream()
                .filter(re -> re.getName().contains(curveType))
                .findAny().orElseThrow(() -> new ReportGenerationException("Точки для графика не найдены"));
    }


    private BigDecimal getParameterValue(Integer materialId, Integer parameterId) {
        return parameterValuesDao.findByMaterialAndParameterIdsWithDefaultTemperature(materialId, parameterId);
    }


    public ChartsSet addHighlightPointToCurve(Integer simId, List<String> valuesLines, String curveType, ChartsSet curve) {
        CompositeMaterialEntity compositeMaterial = compositeMaterialManager.getCompositeMaterialUsingSimulation(simId);
        Double x = null;
        BigDecimal y = null;
        if (MATRIX_CURVE.equals(curveType)) {
            x = Double.valueOf(valuesLines.get(MATRIX_CURVE_VALUE_IDX));
            y = getParameterValue(compositeMaterial.getMatrixId(), STRENGTH_LIMIT_ID);
        } else if (FIBER_CURVE.equals(curveType)) {
            x = Double.valueOf(valuesLines.get(FIBER_CURVE_VALUE_IDX));
            y = getParameterValue(compositeMaterial.getFiberId(), STRENGTH_LIMIT_ID);
        }
        if (x != null && y != null) {
            curve.setHighlightPoint(Point.commonPoint(x, y.doubleValue()));
        }
        return curve;
    }


    public Path getResultPictureByName(Integer id, TaskParentType parentType, String name) {
        ResultsParent resultParent = new ResultsParent(id, parentType);
        List<ResultEntity> params = getResultByParent(resultParent, ResultType.pictures);
        if (params.isEmpty()) {
            throw new ReportGenerationException(String.format("Не найден результатс типом 'изображение' для объекта с номером %d", id));
        }
        return params.stream()
                .filter(p -> p.getPathToFile() != null)
                .map(resultsFilesManager::getPathToResult)
                .filter(p -> name.equals(p.getFileName().toString()))
                .findAny().orElseThrow(() -> new ReportGenerationException("Изображение не найдено"));
    }
}
