package ru.crp.lvms.managers;

import org.jetbrains.annotations.NotNull;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.MessageSource;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import ru.crp.cmlb.model.SuggestRequest;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.managers.image.ImageHandler;
import ru.crp.lvms.managers.misc.BaseManagerSuggest;
import ru.crp.lvms.managers.tasks.ResultsFilesManager;
import ru.crp.lvms.managers.tasks.ResultsParent;
import ru.crp.lvms.model.dao.BaseSpaceDao;
import ru.crp.lvms.model.dao.simulations.SimulationDao;
import ru.crp.lvms.model.dao.simulations.SimulationTypeDao;
import ru.crp.lvms.model.entities.compositeMaterials.CompositeMaterialEntity;
import ru.crp.lvms.model.entities.UserEntity;
import ru.crp.lvms.model.entities.misc.CreatedBy;
import ru.crp.lvms.model.entities.simulations.SimulationEntity;
import ru.crp.lvms.model.entities.simulations.SimulationTypeSuggest;
import ru.crp.lvms.pdf.reports.ReportGenerator;
import ru.crp.lvms.pdf.reports.SimulationReportInput;
import ru.crp.lvms.utils.Mapper;

import javax.annotation.Nullable;
import java.io.IOException;
import java.nio.file.Path;
import java.util.Date;
import java.util.List;
import java.util.Map;

import static ru.crp.lvms.model.AccessRights.view;
import static ru.crp.lvms.utils.MessageUtils.SIMULATION_WRONG_SAMPLE_FILE;
import static ru.crp.lvms.utils.MessageUtils.SIMULATION_WRONG_STANDARD;

@Service
public class SimulationManager extends BaseManagerSuggest<SimulationEntity> implements ImageHandler<SimulationEntity> {

    @Autowired
    private SimulationDao simulationDao;

    @Autowired
    private FileManager fileManager;

    @Autowired
    private CompositeMaterialManager compositeMaterialManager;

    @Autowired
    private MessageSource msgSrc;

    @Autowired
    private StandardManager standardManager;

    @Autowired
    private SimulationTypeDao simulationTypeDao;

    @Autowired
    private ResultsManager resultsManager;

    @Autowired
    private ResultsFilesManager resultsFilesManager;

    @Autowired
    private ReportGenerator reportGenerator;


    @Override
    public BaseSpaceDao<SimulationEntity> aliveDao() {
        return simulationDao;
    }


    @Override
    public void checkCreate(SimulationEntity newSimulation, UserEntity currentUser) {
        CompositeMaterialEntity compositeMaterialEntity = compositeMaterialManager
                .checkEntityExistsAndAccesible(newSimulation.getCompositeMaterialId(), view, currentUser);
        compositeMaterialManager.extendMaterial(compositeMaterialEntity, currentUser);
        if (newSimulation.getSampleFileId() != null && !compositeMaterialEntity.containsSampleFile(newSimulation.getSampleFileId())) {
            throw new BadRequestException(msgSrc.getMessage(SIMULATION_WRONG_SAMPLE_FILE, null, null));
        }
        if (newSimulation.getStandardId() != null && !standardManager.exists(newSimulation.getStandardId())) {
            throw new BadRequestException(msgSrc.getMessage(SIMULATION_WRONG_STANDARD, null, null));
        }
        super.checkCreate(newSimulation, currentUser);
    }


    @Transactional(readOnly = true)
    public List<SimulationEntity> getSimulationsByStandardId(Integer standardId, UserEntity currentUser) {
        return simulationDao.findByStandardId(standardId);
    }


    @Transactional(readOnly = true)
    public List<SimulationEntity> getSimulationsByMaterialId(Integer materialId, UserEntity currentUser) {
        return simulationDao.findByMaterialId(materialId);
    }

    public List<SimulationEntity> getSimulationsByECId(Integer ecId, UserEntity currentUser) {
        return simulationDao.findByECId(ecId);
    }

    public List<SimulationEntity> getSimulationsForTechnology(Integer technologyId, UserEntity currentUser) {
        return simulationDao.findSimulationsForTechnology(technologyId);
    }

    public PageImpl<SimulationEntity> getSimulationsPageForTechnology(Integer technologyId, Pageable page,
                                                                      Map<String, Object> filters, UserEntity currentUser) {
        return simulationDao.findSimulationsPageForTechnology(technologyId, page, filters);
    }

    public List<SimulationTypeSuggest> suggestSimulationType(SuggestRequest suggestRequest, UserEntity user) {
        return simulationTypeDao.suggest(suggestRequest, user.getId());
    }

    @Override
    public void delete(SimulationEntity simulationToDelete, @Nullable UserEntity currentUser) {
        checkDelete(simulationToDelete, currentUser);
        //todo delete experiment file id and image file id
        resultsManager.deleteResults(Mapper.map(simulationToDelete, ResultsParent.class), true, CreatedBy.Auto, CreatedBy.Manual);
        aliveDao().delete(simulationToDelete.getId());
    }

    @Override
    public FileManager fileManager() {
        return fileManager;
    }

    @NotNull
    public SimulationReportInput formSimulationReportInput(SimulationEntity simulationEntity) throws IOException {
        SimulationReportInput reportInfo = new SimulationReportInput();
        reportInfo.setTypeName(simulationEntity.getType());
        reportInfo.setStandartName(simulationEntity.getStandardName());
        reportInfo.setMaterialName(simulationEntity.getCompositeMaterialName());
        reportInfo.setEcName(simulationEntity.getEcName());
        reportInfo.setCreationDate(new Date());
        reportInfo.setUserName(simulationEntity.getOwner());
        reportInfo.setManualSizeH(simulationEntity.getManualSizeH());
        reportInfo.setManualSizeB(simulationEntity.getManualSizeB());
        reportInfo.setManualSizeL(simulationEntity.getManualSizeL());
        List<String> valuesLines = resultsManager.extractValuesLines(simulationEntity.getId());
        reportInfo.setResults(resultsManager.getSimulationResultValues(valuesLines));
        Path fiberChart = reportGenerator.generatePlot(
                resultsManager.addHighlightPointToCurve(simulationEntity.getId(), valuesLines, ResultsManager.FIBER_CURVE,
                        resultsFilesManager.getCurve(
                                resultsManager.getSimulationResultCurve(simulationEntity.getId(), ResultsManager.FIBER_CURVE)))
        );
        reportInfo.setFiberPic(fiberChart);
        Path matrixChart = reportGenerator.generatePlot(
                resultsManager.addHighlightPointToCurve(simulationEntity.getId(), valuesLines, ResultsManager.MATRIX_CURVE,
                        resultsFilesManager.getCurve(
                                resultsManager.getSimulationResultCurve(simulationEntity.getId(), ResultsManager.MATRIX_CURVE)))
        );
        reportInfo.setMatrixPic(matrixChart);
        return reportInfo;
    }
}
