package ru.crp.lvms.optimizations.mf.controllers;

import org.jetbrains.annotations.NotNull;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.web.bind.annotation.*;
import ru.crp.cmls.calculations.solvers.xml.config.ResultType;
import ru.crp.lvms.chart.ChartsSet;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.managers.ResultsManager;
import ru.crp.lvms.managers.tasks.ResultsFilesManager;
import ru.crp.lvms.managers.tasks.ResultsParent;
import ru.crp.lvms.model.AccessRights;
import ru.crp.lvms.model.entities.ResultEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationVaryingParam;
import ru.crp.lvms.optimizations.mf.entities.MFSliceChartParameters;
import ru.crp.lvms.optimizations.mf.entities.MFVaryingParamInterval;
import ru.crp.lvms.optimizations.mf.enums.MFSettingsType;
import ru.crp.lvms.optimizations.mf.managers.MFOptimizationSettingsManager;
import ru.crp.lvms.optimizations.mf.results.ModFrontierParams;
import ru.crp.lvms.optimizations.misc.AbstractOptimizationSettings;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.utils.Mapper;

import javax.annotation.Nonnull;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.Optional;

import static java.util.stream.Collectors.toList;
import static ru.crp.lvms.optimizations.mf.results.ModeFrontierParamsHandler.parseModFrontierParams;
import static ru.crp.lvms.utils.URLs.MF_SLICE_CHART_URL;

@RestController
@RequestMapping(value = MF_SLICE_CHART_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class MFSliceChartController {

    private ResultsManager resultsManager;

    private ResultsFilesManager resultsFilesManager;

    @Autowired
    private OptimizationManager optimizationManager;

    @Autowired
    private MFOptimizationSettingsManager mfOptimizationSettingsManager;

    private class NameAndIndex {
        private String name;
        private Integer index;

        public NameAndIndex(String name, Integer index) {
            this.name = name;
            this.index = index;
        }
    }

    @Nonnull
    @RequestMapping(method = RequestMethod.POST)
    public ChartsSet getSliceChart(@RequestBody MFSliceChartParameters sliceChartParameters,
                                   @PathVariable("optimizationId") Integer optimizationId,
                                   @CurrentUser PortalUser currentUser) throws IOException {

        OptimizationEntity optimization = optimizationManager
                .checkEntityExistsAndAccesible(optimizationId, AccessRights.view, currentUser.getUser());
        ResultsParent resultsParent = Mapper.map(optimization, ResultsParent.class);
        ResultEntity values = resultsManager.getResultByParent(resultsParent, ResultType.modfrontierParams).get(0);
        Path pathToResult = resultsFilesManager.getPathToResult(values);

        if (Files.notExists(pathToResult)) {
            throw new BadRequestException("Не найден файл результата для генерации графика");
        }

        ModFrontierParams modFrontierParams = parseModFrontierParams(pathToResult);

        NameAndIndex vpNameAndIndex = getNameAndIndex(optimizationId, MFSettingsType.VaryingParam,
                sliceChartParameters.getVaryingParamId(), currentUser);

        NameAndIndex tfNameAndIndex = getNameAndIndex(optimizationId, MFSettingsType.TargetFunc,
                sliceChartParameters.getTargetFuncId(), currentUser);

        List<MFOptimizationVaryingParam> existingVaryingParams =
                (List<MFOptimizationVaryingParam>) mfOptimizationSettingsManager
                        .findOptimizationSettings(optimizationId, MFSettingsType.VaryingParam, currentUser.getUser());

        List<Integer> constrainingVaryingParamsIds = existingVaryingParams
                .stream().map(AbstractOptimizationSettings::getId).collect(toList());

        List<MFVaryingParamInterval> ranges = sliceChartParameters.getIntervals().stream()
                .filter(range -> range.getId() != sliceChartParameters.getVaryingParamId())
                .map(range ->
                new MFVaryingParamInterval(range.getId(),
                        constrainingVaryingParamsIds.indexOf(range.getId()),
                        range.getLowerBound(),
                        range.getUpperBound())).collect(toList());

        checkVaryingParamsRangeNotExpanded(ranges, existingVaryingParams);

        return modFrontierParams.createSliceChart(vpNameAndIndex.name, tfNameAndIndex.name,
                vpNameAndIndex.index, tfNameAndIndex.index, ranges).toChartsSet();
    }

    private void checkVaryingParamsRangeNotExpanded(List<MFVaryingParamInterval> chartConstraints, List<MFOptimizationVaryingParam> constrainingVaryingParams) {
        for (MFOptimizationVaryingParam existing: constrainingVaryingParams) {
            for (MFVaryingParamInterval updated: chartConstraints) {
                if (existing.getId() == updated.getVaryingParamId()) {
                    if (updated.getLowerBound() < existing.getLowerBound() || existing.getUpperBound() < updated.getUpperBound()) {
                        throw new BadRequestException("Исходные границы варьируемого параметра " + existing.getName() +
                                " не должны быть расширены");
                    }
                    break;
                }
            }
        }
    }

    @NotNull
    private NameAndIndex getNameAndIndex(Integer optimizationId,
                                         MFSettingsType settingsType,
                                         Integer targetId,
                                         PortalUser currentUser) {
        List<? extends AbstractOptimizationSettings> settings = mfOptimizationSettingsManager
                .findOptimizationSettings(optimizationId, settingsType, currentUser.getUser());

        Optional<? extends AbstractOptimizationSettings> targetSetting = settings.stream()
                .filter(setting -> setting.getId() == targetId).findAny();

        if (!targetSetting.isPresent()) {
            throw new BadRequestException("Параметр с таким идентификатором не найден");
        }

        int targetIndex = settings.stream().map(AbstractOptimizationSettings::getId).collect(toList())
                .indexOf(targetId);

        return new NameAndIndex(targetSetting.get().getName(), targetIndex);
    }

    @Autowired
    public void setResultsManager(ResultsManager resultsManager) {
        this.resultsManager = resultsManager;
    }

    @Autowired
    public void setResultsFilesManager(ResultsFilesManager resultsFilesManager) {
        this.resultsFilesManager = resultsFilesManager;
    }
}
