package ru.crp.lvms.optimizations.optistruct.controllers;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.MediaType;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;
import ru.crp.cmlb.model.SuggestRequest;
import ru.crp.cmls.calculations.solvers.xml.config.ResultType;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.InternalException;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.managers.ResultsManager;
import ru.crp.lvms.managers.tasks.ResultsFilesManager;
import ru.crp.lvms.managers.tasks.ResultsParent;
import ru.crp.lvms.model.entities.ResultEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.misc.AbstractOptimizationSettings;
import ru.crp.lvms.optimizations.optistruct.entities.OptistructCollector;
import ru.crp.lvms.optimizations.optistruct.entities.OptistructLnSet;
import ru.crp.lvms.optimizations.optistruct.entities.OptistructSet;
import ru.crp.lvms.optimizations.optistruct.enums.ComponentType;
import ru.crp.lvms.optimizations.optistruct.enums.OptistructBaseSettingType;
import ru.crp.lvms.optimizations.optistruct.enums.OptistructSettingsType;
import ru.crp.lvms.optimizations.optistruct.managers.OptistructFemParser;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.utils.Mapper;
import ru.crp.lvms.utils.URLs;

import javax.annotation.Nullable;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import static ru.crp.lvms.model.mappers.PgArrayPropertyEditor.LOGGER;

@RestController
@RequestMapping(value = URLs.OPTISTRUCT_SETS_URL, produces = MediaType.APPLICATION_JSON_VALUE)
public class OptistructSetsController extends OptistructAbstractOptimizationSettingsController<OptistructSet> {

    @Autowired
    private FileManager fileManager;

    @Autowired
    private OptistructFemParser optistructFemParser;

    @Autowired
    private ResultsManager resultsManager;

    @Autowired
    private ResultsFilesManager resultsFilesManager;

    private Lock lock = new ReentrantLock();

    @Override
    protected OptistructSettingsType getSettingsType() {
        return OptistructSettingsType.Set;
    }

    @Override
    protected OptistructBaseSettingType getOptistructOptimizationBaseSettingType() {
        return OptistructBaseSettingType.NUM_SETS;
    }

    @Override
    public List<? extends AbstractOptimizationSettings> getAllSettings(@PathVariable("optimizationId") int optimizationId, @CurrentUser PortalUser currentUser) {
        checkOptimizationSolverType(optimizationId, currentUser);
        return getSettingsManager()
                .findOptimizationSettings(optimizationId, settingsType, currentUser.getUser());
    }

    @RequestMapping(value = "/selected", method = RequestMethod.GET)
    public List<OptistructSet> findSelectedSets(@PathVariable("optimizationId") Integer optimizationId,
                                                @Nullable @RequestParam(value = "type", required = false) ComponentType componentType,
                                                @CurrentUser PortalUser currentUser) {
        checkOptimizationSolverType(optimizationId, currentUser);
        return getSettingsManager().findSelectedSets(optimizationId, componentType);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/{setId}", method = RequestMethod.POST)
    public AbstractOptimizationSettings addToSelected(@PathVariable("optimizationId") Integer optimizationId,
                                                      @PathVariable("setId") Integer setId,
                                                      @RequestParam("isSelected") Boolean isSelected,
                                                      @CurrentUser PortalUser currentUser) {
        checkOptimizationSolverType(optimizationId, currentUser);
        OptistructSet set = (OptistructSet) getSettingsManager()
                .findOptimizationSettingsById(setId, OptistructSettingsType.Set, currentUser.getUser());
        set.setIsSelected(isSelected);
        return getSettingsManager().updateSettings(set, OptistructSettingsType.Set);
    }

    @RequestMapping(value = "/suggest", method = RequestMethod.POST)
    public List<OptistructSet> suggestSets(@PathVariable("optimizationId") Integer optimizationId,
                                           @RequestBody SuggestRequest suggestRequest,
                                           @CurrentUser PortalUser currentUser) {
        checkOptimizationSolverType(optimizationId, currentUser);
        return getSettingsManager().suggestSets(optimizationId, suggestRequest, currentUser.getUser());
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    // TODO: replace with parsing
    @RequestMapping(method = RequestMethod.POST)
    public OptistructSet createSet(@PathVariable("optimizationId") int optimizationId,
                                   @RequestBody OptistructSet newSet,
                                   @CurrentUser PortalUser currentUser) {
        checkOptimizationSolverType(optimizationId, currentUser);
        newSet.setOptimizationId(optimizationId);
        return getSettingsManager().createSet(newSet);
    }

    @PreAuthorize("hasRole('ROLE_USER')")
    @RequestMapping(value = "/extract", method = RequestMethod.POST)
    public void extractInfoFromFile(@PathVariable("optimizationId") Integer optimizationId,
                                    @CurrentUser PortalUser currentUser) {
        OptimizationEntity optimization = checkOptimizationSolverType(optimizationId, currentUser);
        List<ResultEntity> generatedFEMFiles = resultsManager.getResultByParent(Mapper.map(optimization, ResultsParent.class), ResultType.optistructFEMFile);
        if (generatedFEMFiles.isEmpty()) {
            throw new BadRequestException("Файл для извлечения наборов узлов и элементов не найден");
        }
        if (generatedFEMFiles.size() > 1) {
            throw new BadRequestException("Должен быть только один файл .fem");
        }
        Path pathToFEMFile = resultsFilesManager.getPathToResult(generatedFEMFiles.get(0));

        if (pathToFEMFile == null || !Files.isRegularFile(pathToFEMFile)) {
            LOGGER.error("Result is present in db but not in file system");
            throw new InternalException("FEM файл не найден на диске");
        }

        try {
            List<OptistructSet> newSets = optistructFemParser.extractSets(optimizationId, pathToFEMFile);
            getSettingsManager().resetSets(optimizationId, newSets);
            List<OptistructCollector> newCollectors = optistructFemParser.extractCollectors(optimizationId, pathToFEMFile);
            getSettingsManager().resetCollectors(optimizationId, newCollectors);
            List<OptistructLnSet> newLnSets = optistructFemParser.extractLnSets(optimizationId, pathToFEMFile);
            getSettingsManager().resetLnSets(optimizationId, newLnSets);
        } catch (IOException | IllegalArgumentException ex) {
            throw new InternalException(String.format("Не удалось извлечь наборы узлов и элементов из файла %s", pathToFEMFile.getFileName()), ex);
        }
    }

    @Override
    public AbstractOptimizationSettings updateSettings(@PathVariable("optimizationId") int optimizationId,
                                                       @PathVariable("id") int settingsId,
                                                       @RequestBody OptistructSet updatedSettings,
                                                       @CurrentUser PortalUser currentUser) {
        try {
            // to prevent transaction conflict. Todo fixme
            lock.lock();
            return super.updateSettings(optimizationId, settingsId, updatedSettings, currentUser);
        } finally {
            lock.unlock();

        }
    }
}
