package ru.crp.lvms.task;

import org.apache.commons.io.FilenameUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import ru.crp.lvms.exceptions.solving.TaskStartingException;
import ru.crp.lvms.model.entities.FileEntityTable;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.optimizations.base.dao.OptimizationJsonSettingsDao;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationJsonSetting;
import ru.crp.lvms.optimizations.tosca.enums.ToscaJsonParametersParents;
import ru.crp.lvms.tasks.fileGenerators.BaseInputFileGenerator;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static java.util.Arrays.asList;
import static ru.crp.lvms.optimizations.tosca.managers.ToscaDefaultParamsCreator.ANIMATION_NAME;
import static ru.crp.lvms.optimizations.tosca.managers.ToscaDefaultParamsCreator.SMOOTHING_NAME;

@Component
public class ToscaAnimationInputFileGenerator extends BaseInputFileGenerator {

    @Autowired
    private OptimizationJsonSettingsDao optimizationJsonSettingsDao;

    public static final String SOLVING_DESIGN_CYCLE_CFG = "solving/design_cycle.cfg";
    public static final String ITER_NUM_STRING_PREFIX = "DESIGN_CYCLE =";

    public static final Set<String> SKIPPED_PARAMS =
            new HashSet<>(asList(SMOOTHING_NAME, ANIMATION_NAME));

    public Path generateConfig(TaskEntity task, OptimizationEntity optimization) throws IOException {
        Path cfg = resultsFilesManager.getResultsDir(task).resolve(SOLVING_DESIGN_CYCLE_CFG);
        if (!Files.exists(cfg)) {
            throw new TaskStartingException("File " + cfg.toAbsolutePath() + " not found");
        }
        String iterationNumber;
        try (Stream<String> lines = Files.lines(cfg)) {
            iterationNumber = lines.filter(s -> s.startsWith(ITER_NUM_STRING_PREFIX))
                    .findAny()
                    .map(s -> s.substring(ITER_NUM_STRING_PREFIX.length()))
                    .orElseThrow(() -> new TaskStartingException("String with iterations number not found in file " +
                            SOLVING_DESIGN_CYCLE_CFG));
        }
        String caeFileName = optimizationManager.getOptimizationFiles(optimization, FileType.CAEFile, null)
                .stream()
                .map(FileEntityTable::getName)
                .findAny()
                .map(FilenameUtils::removeExtension)
                .orElseThrow(() -> new TaskStartingException("Unable to find CAE file"));
        List<OptimizationJsonSetting> settings =
                optimizationJsonSettingsDao.findSettingsByParent(optimization.getId(),
                        ToscaJsonParametersParents.AnimationParameters.getId());
        String params = settings.stream()
                .filter(param -> !SKIPPED_PARAMS.contains(param.getXmlName()))
                .map(param -> param.getXmlName() + "=" + param.getPlainValue())
                .collect(Collectors.joining("\n", "\n", "\n"));

        Path tempConfigFile = Files.createTempFile(null, null);
        String data = "iternum=" + iterationNumber + "\n" +
                "nameopttosca='solving'\n" +
                "namecae='" + caeFileName + "'" +
                params +
                "*use,anstoscapp.mac";
        Files.write(tempConfigFile, data.getBytes());
        return tempConfigFile;
    }
}
