package ru.crp.lvms.tasks.fileGenerators;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import ru.crp.cmlb.model.entities.FileData;
import ru.crp.cmls.calculations.solvers.xml.config.DataType;
import ru.crp.cmls.calculations.solvers.xml.config.Type;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.managers.FileManager;
import ru.crp.lvms.model.entities.FileEntity;
import ru.crp.lvms.model.entities.misc.FileType;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.tosca.config.ToscaConfigGenerator;
import ru.crp.lvms.task.ToscaAnimationInputFileGenerator;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.*;

import static ru.crp.lvms.model.mappers.PgArrayPropertyEditor.LOGGER;


@Component
public class ToscaInputFileGenerator extends BaseInputFileGenerator {
    public static final String ANSYS_CAE_FILE_EXT = ".cdb";
    public static final String ABAQUS_CAE_FILE_EXT = ".inp";
    // MUST be synchronized with argument value in tosca-config.xml
    public static final String SETTINGS_FILE_NAME = "solving.par";
    public static final String ANIMATION_SETTINGS_FILE_NAME = "postdata.txt";

    @Value("${toscaMacrosesDir}")
    private String pathToToscaMacroses;

    @Value("${toscaAnimationFilesDir}")
    private String toscaAnimationFilesDir;

    @Autowired
    private FileManager fileManager;

    @Autowired
    private ToscaConfigGenerator toscaConfigGenerator;

    @Autowired
    private ToscaAnimationInputFileGenerator toscaAnimationInputFileGenerator;

    public List<FileData> getInputFilesForOptimization(TaskEntity task, OptimizationEntity optimization) throws IOException {
        Set<FileData> fileData = new HashSet<>(); // Files from results dir must not have the same name
        addOptimizationFiles(fileData, optimization);
        addFilesFromDir(fileData, pathToToscaMacroses);

        switch (task.getType()) {
            case toscaSmoothing: // Smoothing uses same files, but looks only to smoothing part
            case toscaSolving:
                Path pathToSettings = toscaConfigGenerator.generateConfig(optimization);
                fileData.add(FileData.fromFile(SETTINGS_FILE_NAME, DataType.GENERATED, pathToSettings.toFile()));
                break;
            case toscaAnimation:
                addFilesFromDir(fileData, toscaAnimationFilesDir, DataType.CUSTOM_INPUT_FILE);
                Path config = toscaAnimationInputFileGenerator.generateConfig(task, optimization);
                fileData.add(FileData.fromFile(ANIMATION_SETTINGS_FILE_NAME, DataType.GENERATED, config.toFile()));
                break;
            default:
                throw new BadRequestException("Illegal type for Tosca calculation: " + task.getType());
        }

        if (task.getType() == Type.toscaSmoothing || task.getType() == Type.toscaAnimation) {
            Path resultsDir = resultsFilesManager.getResultsDir(task);
            addFilesFromDir(fileData, resultsDir, DataType.CUSTOM_INPUT_FILE);
        }

        return new ArrayList<>(fileData);
    }


    private void addOptimizationFiles(Set<FileData> fileData, OptimizationEntity optimization) {
        List<FileEntity> allOptimizationFiles = optimizationManager.getOptimizationFiles(optimization, null);
        allOptimizationFiles.stream()
                .filter(file -> file.getFileTypeId() == FileType.CAEFile.getId() || file.getFileTypeId() == FileType.LoadCondition.getId())
                .map(fileEntity -> {
                    Path pathToFile = fileManager.resolvePathToFile(fileEntity, null);
                    if (pathToFile == null || !Files.exists(pathToFile)) {
                        LOGGER.error("File " + fileEntity.getId() + " is present in db but not in file system");
                        return null;
                    }
                    if (!isToscaFile(fileEntity)) {
                        return null;
                    }
                    return FileData.fromFile(fileEntity.getName(), DataType.SUBMODEL, pathToFile.toFile());
                })
                .filter(Objects::nonNull)
                .forEach(fileData::add);
    }

    private boolean isToscaFile(FileEntity fileEntity) {
        if (fileEntity.getFileTypeId() == FileType.LoadCondition.getId()) {
            return true;
        }
        return fileEntity.getName().endsWith(ANSYS_CAE_FILE_EXT) || fileEntity.getName().endsWith(ABAQUS_CAE_FILE_EXT);
    }
}
