package ru.crp.lvms.tasks.space.commands;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.scheduling.annotation.Async;
import org.springframework.scheduling.annotation.AsyncResult;
import org.springframework.stereotype.Component;
import org.springframework.util.concurrent.ListenableFuture;
import ru.crp.cmlb.model.entities.ResourcePreparationException;
import ru.crp.cmls.calculations.solvers.xml.config.Type;
import ru.crp.lvms.cluster.model.ClusterEntity;
import ru.crp.lvms.managers.ResultsManager;
import ru.crp.lvms.managers.solvers.SolverConfigManager;
import ru.crp.lvms.managers.tasks.ResultsFilesManager;
import ru.crp.lvms.managers.tasks.ResultsParent;
import ru.crp.lvms.managers.tasks.TaskResultsExtractor;
import ru.crp.lvms.model.entities.misc.CreatedBy;
import ru.crp.lvms.model.entities.nodes.NodeEntity;
import ru.crp.lvms.model.entities.tasks.TaskEntity;
import ru.crp.lvms.optimizations.base.managers.OptimizationManager;
import ru.crp.lvms.tasks.misc.BaseTaskCommand;
import ru.crp.lvms.tasks.space.exceptions.SolverException;
import ru.crp.lvms.utils.Mapper;
import ru.crplab.downloader.FilesDownloadingManager;
import ru.crplab.downloader.collector.FSSubtreeAggregationInfo;

import java.nio.file.Path;
import java.util.concurrent.atomic.AtomicReference;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

@Component
public class SpaceDownloadResultsCommand extends BaseTaskCommand<Void, Void> {
    private static final Logger LOGGER = LoggerFactory.getLogger(SpaceDownloadResultsCommand.class);

    @Autowired
    private SolverConfigManager solverConfigManager;

    @Autowired
    private ResultsManager resultsManager;

    @Autowired
    private ResultsFilesManager resultsFilesManager;

    @Autowired
    private TaskResultsExtractor taskResultsExtractor;

    @Autowired
    private OptimizationManager optimizationManager;


    @Override
    public boolean isTaskReadyForProcessing(TaskEntity task, ClusterEntity taskCluster) {
        return super.isTaskReadyForProcessing(task, taskCluster) && checkNodeAlive(task);
    }


    @Async("downloadTasksResultExecutor")
    @Override
    public ListenableFuture<Void> processTask(TaskEntity task, Object unused) throws Exception {
        if (task.getPrevTaskId() == null && task.getType() != Type.toscaAnimation && task.getType() != Type.toscaSmoothing) {
            resultsManager.deleteResults(Mapper.map(task, ResultsParent.class), true, CreatedBy.Auto);
        }

        Path resultsDir = resultsFilesManager.getResultsDir(task);

        Lock lock = new ReentrantLock();
        Condition filesDownloaded = lock.newCondition();
        AtomicReference<Exception> downloadingException = new AtomicReference<>(null);

        NodeEntity workerNode = nodesManager.findOne(task.getNodeId(), null);
        startDownloader(lock, filesDownloaded, downloadingException, workerNode.getUrl(), task.getId(), resultsDir);
        lock.lock();
        filesDownloaded.await();
        lock.unlock();
        Exception ex = downloadingException.get();
        if (ex != null) {
            throw ex;
        }

        taskResultsExtractor.extractResults(task);

        if (task.getType() == Type.optistructGenerateFEM) {
            optimizationManager.copyResultsFemToOptimizationFiles(task);
        }
        return new AsyncResult<>(null);
    }


    private void startDownloader(Lock lock, Condition con, AtomicReference<Exception> ex, String nodeUrl, int taskId, Path resultsDir) throws SolverException, ResourcePreparationException {
        FSSubtreeAggregationInfo filesToDownload = solverConnectionManager.downloadResultsInfo(nodeUrl, taskId);
        FilesDownloadingManager downloader = new FilesDownloadingManager.Builder()
                .withTargetDir(resultsDir)
                .withFileDownloadedCallback((manager, _unused1, _unused2) -> {
                    if (manager.isFinished()) {
                        unlockMainMethod(lock, con);
                    }
                })
                .withDownloadingErrorCallback((manager, exception) -> {
                    ex.set(exception);
                    unlockMainMethod(lock, con);
                })
//                // move results from 'output' folder directly to target folder
//                .withTargetPathTransformer(pathToResult ->
//                        pathToResult.replace(OUTPUT_DIRECTORY_NAME + "/", "")
//                                .replace(OUTPUT_DIRECTORY_NAME.toUpperCase() + "/", ""))
                .withDownloader(solverConnectionManager.createFileDownloader(nodeUrl, taskId))
                .build();

        downloader.startDownloading(filesToDownload);
    }


    private void unlockMainMethod(Lock lock, Condition con) {
        lock.lock();
        con.signalAll();
        lock.unlock();
    }
}
