package ru.crp.lvms.web.controllers.misc;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.imgscalr.Scalr;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.MessageSource;
import org.springframework.http.HttpStatus;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.multipart.MultipartFile;
import ru.crp.cmlb.model.entities.base.Entity;
import ru.crp.lvms.exceptions.BadRequestException;
import ru.crp.lvms.exceptions.NotFoundException;
import ru.crp.lvms.managers.image.OldImageFilesManager;
import ru.crp.lvms.model.dao.BaseSpaceDao;
import ru.crp.lvms.security.CurrentUser;
import ru.crp.lvms.security.PortalUser;
import ru.crp.lvms.web.dto.misc.Auditable;

import javax.imageio.ImageIO;
import javax.servlet.http.HttpServletResponse;
import java.awt.image.BufferedImage;
import java.io.*;

import static ru.crp.lvms.utils.AuthUtils.ownerOrAdmin;
import static ru.crp.lvms.utils.MessageUtils.ENTITY_NOT_FOUND;
import static ru.crp.lvms.utils.MessageUtils.IMAGE_WRONG_FORMAT;

public class ControllerImage<E extends Entity & Auditable> {
    private static final int MAX_IMG_HEIGHT = 1000;
    private static final int MAX_IMG_WIDTH = 1000;

    private BaseSpaceDao<E> dao;

    private OldImageFilesManager fileManager;

    @Autowired
    private MessageSource msgSrc;

    public ControllerImage(BaseSpaceDao<E> dao, String imagesPath) {
        this.dao = dao;
        this.fileManager = new OldImageFilesManager(imagesPath);
    }


    @RequestMapping(method = RequestMethod.GET)
    public void get(@PathVariable("id") final int id, HttpServletResponse response) throws IOException {
        File avatar = fileManager.findFile(id);
        if (avatar == null) {
            avatar = fileManager.findDefaultFile();
        }

        if (avatar != null) {
            response.setHeader("Content-Type", "image/" + FilenameUtils.getExtension(avatar.getAbsolutePath()));
            FileUtils.copyFile(avatar, response.getOutputStream());
        } else {
            response.setStatus(HttpStatus.NOT_FOUND.value());
        }
    }


    @RequestMapping(method = RequestMethod.DELETE)
    public void delete(@PathVariable("id") final int id, HttpServletResponse response,
                       @CurrentUser PortalUser user) throws IOException {
        E entity = dao.findOne(id);
        if (entity != null) {
            if (fileManager.delete(id)) {
                response.setStatus(HttpStatus.OK.value());
            } else {
                response.setStatus(HttpStatus.NOT_FOUND.value());
            }
        } else {
            throw new NotFoundException(msgSrc.getMessage(ENTITY_NOT_FOUND, new Object[]{id}, null));
        }
    }


    @RequestMapping(method = RequestMethod.POST)
    public void save(@PathVariable("id") int id, @RequestParam("file") MultipartFile file,
                     @CurrentUser PortalUser user) {
        E entity = dao.findOne(id);
        if (entity == null) {
            throw new BadRequestException(msgSrc.getMessage(ENTITY_NOT_FOUND, new Object[]{id}, null));
        }
        ownerOrAdmin(entity.getOwnerId(), user.getUser());
        try (InputStream is = file.getInputStream();
             InputStream imageInputStream = processImage(is)) {
            fileManager.delete(id);
            fileManager.save(id, imageInputStream, "png");
        } catch (IOException ex) {
            throw new BadRequestException(msgSrc.getMessage(IMAGE_WRONG_FORMAT, null, null), ex);
        }
    }


    private InputStream processImage(InputStream inputStream) throws IOException {
        BufferedImage bufferedImage = ImageIO.read(inputStream);
        // check if image
        if (bufferedImage == null) {
            throw new BadRequestException(msgSrc.getMessage(IMAGE_WRONG_FORMAT, null, null));
        }
        // scale if too big
        if (bufferedImage.getHeight() > MAX_IMG_HEIGHT || bufferedImage.getWidth() > MAX_IMG_WIDTH) {
            bufferedImage = Scalr.resize(bufferedImage, MAX_IMG_WIDTH, MAX_IMG_HEIGHT);
        }
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        ImageIO.write(bufferedImage, "PNG", baos);
        return new ByteArrayInputStream(baos.toByteArray());
    }
}
