package ru.crp.lvms.optimizations;


import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.JsonNode;
import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.springframework.beans.factory.annotation.Autowired;
import ru.crp.lvms.BaseIT;
import ru.crp.lvms.chart.ChartsSet;
import ru.crp.lvms.chart.LineChart;
import ru.crp.lvms.chart.Point;
import ru.crp.lvms.managers.ResultsManager;
import ru.crp.lvms.managers.tasks.ResultsFilesManager;
import ru.crp.lvms.model.entities.ResultEntity;
import ru.crp.lvms.optimizations.base.entities.OptimizationEntity;
import ru.crp.lvms.optimizations.base.enums.OptimizationCao;
import ru.crp.lvms.optimizations.mf.controllers.MFSliceChartController;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationBaseSetting;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationTargetFunc;
import ru.crp.lvms.optimizations.mf.entities.MFOptimizationVaryingParam;
import ru.crp.lvms.optimizations.mf.entities.MFSliceChartParameters;
import ru.crp.lvms.optimizations.mf.enums.MFBaseSettingType;
import ru.crp.lvms.utils.TestUtils;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import static java.util.Arrays.asList;
import static java.util.Collections.nCopies;
import static org.mockito.Matchers.any;
import static org.mockito.Mockito.when;
import static ru.crp.lvms.optimizations.mf.enums.MFBaseSettingType.NUM_CAO_VARYING_PARAMS;
import static ru.crp.lvms.optimizations.mf.enums.MFBaseSettingType.NUM_TARGET_FUNCS;
import static ru.crp.lvms.utils.SetterChecker.createProxy;
import static ru.crp.lvms.utils.TestUtils.*;
import static ru.crp.lvms.utils.URLs.*;

public class MFSliceChartTest extends BaseIT {

    @Mock
    @Autowired
    private ResultsManager resultsManager;

    @Mock
    @Autowired
    private ResultsFilesManager resultsFilesManager;

    @InjectMocks
    @Autowired
    private MFSliceChartController mfSliceChartController;

    private Path desFile;

    @Before
    public void init() throws IOException {
        MockitoAnnotations.initMocks(this);
        desFile = Files.createTempFile(null, null);

        Files.write(desFile, asList(
                "<ID> <RID> par1 par2 par3 out1 obj1 c1 <ERROR>  <VIRTUAL>  <FEASIBLE>  <MARKED>  <CATEGORY>",
                "00000 []  +0.0E000 +5.00E000 +10.0E000 +0.0E001 +0.0E002 +0.0E003 false false true  false 00004",
                "00001 []  +1.0E000 +5.00E000 +10.0E000 +1.0E001 +1.0E002 +1.0E003 false false true  false 00004",
                "00002 []  +2.0E000 +5.00E000 +10.0E000 +2.0E001 +2.0E002 +2.0E003 false false false false 00004",
                "00003 []  +3.0E000 +10.0E000 +10.0E000 +3.0E001 +3.0E002 +3.0E003 false false false false 00004",
                "00004 []  +4.0E000 +10.0E000 +10.0E000 +4.0E001 +4.0E002 +4.0E003 false false true  false 00007",
                "00005 []  +5.0E000 +10.0E000 +20.0E000 +5.0E001 +5.0E002 +5.0E003 false false true  false 00007",
                "00006 []  +6.0E000 +15.0E000 +20.0E000 +6.0E001 +6.0E002 +6.0E003 false false false false 00007",
                "00007 []  +7.0E000 +15.0E000 +20.0E000 +7.0E001 +7.0E002 +7.0E003 false false true  false 00008",
                "00008 []  +8.0E000 +15.0E000 +20.0E000 +8.0E001 +8.0E002 +8.0E003 false false false false 00007",
                "00009 []  +9.0E000 +20.0E000 +20.0E000 +9.0E001 +9.0E002 +9.0E003 false false true  false 00008"
        ));
    }

    @After
    public void cleanup() throws IOException {
        Files.delete(desFile);
    }

    @Test
    public void itMustCreateSliceChart() throws Exception {
        ResultEntity result = new ResultEntity();
        when(resultsManager.getResultByParent(any(), any())).thenReturn(nCopies(5, result));
        when(resultsFilesManager.getPathToResult(result)).thenReturn(desFile);
        mfSliceChartController.setResultsManager(resultsManager);
        mfSliceChartController.setResultsFilesManager(resultsFilesManager);

        OptimizationEntity optimization = projectCreator.createOptimization(OptimizationCao.ModeFrontier);

        String basicURL = MF_BASE_SETTINGS_URL.replace("{optimizationId}", String.valueOf(optimization.getId()));
        JsonNode basicSettingsJson = GETobjects(mockMvc, basicURL);
        List<MFOptimizationBaseSetting> settings = jsonToOptimizationBaseSettingList(basicSettingsJson);

        Map<Integer, List<MFOptimizationBaseSetting>> tcSettings = settings.stream()
                .collect(Collectors.groupingBy(MFOptimizationBaseSetting::getTypeId));

        setBaseSettingValue(basicURL, tcSettings, NUM_CAO_VARYING_PARAMS, 3);
        setBaseSettingValue(basicURL, tcSettings, NUM_TARGET_FUNCS, 1);

        MFSliceChartParameters sliceChartParameters = createProxy(MFSliceChartParameters.class);

        List<MFOptimizationTargetFunc> targetFuncs = jsonToMFOptimizationTargetFuncList(
                GETobjects(mockMvc, MF_TARGET_FUNCS_URL.replace("{optimizationId}", String.valueOf(optimization.getId()))));
        sliceChartParameters.setTargetFuncId(targetFuncs.get(0).getId());

        String vpUrl = MF_VARYING_PARAMS_URL.replace("{optimizationId}", String.valueOf(optimization.getId()));
        List<MFOptimizationVaryingParam> varyingParams = jsonToMFOptimizationVaryingParamList(GETobjects(mockMvc, vpUrl));
        varyingParams.forEach(vp -> {
            vp.setLowerBound(0.0);
            vp.setUpperBound(100.0);
            try {
                UPDATEobject(mockMvc, vpUrl + "/" + vp.getId(), vp, MFOptimizationVaryingParam.class, TestUtils::toJsonNoProxy);
            } catch (Exception e) {
                e.printStackTrace();
            }
        });
        sliceChartParameters.setVaryingParamId(varyingParams.get(0).getId());

        sliceChartParameters.setIntervals(asList(
                new MFOptimizationVaryingParam(varyingParams.get(0).getId(), -10.0, 110.0),
                new MFOptimizationVaryingParam(varyingParams.get(1).getId(), 8.0, 17.0),
                new MFOptimizationVaryingParam(varyingParams.get(2).getId(), 15.0, 30.0)
        ));

        ChartsSet chartSet = CREATEobject(mockMvc, MF_SLICE_CHART_URL.replace("{optimizationId}", String.valueOf(optimization.getId())),
                sliceChartParameters, ChartsSet.class);
        LineChart chart = (LineChart) chartSet.getCharts().get(0);
        Assert.assertArrayEquals(new Double[]{50.0, 60.0, 70.0, 80.0}, chart.getPoints().stream().map(Point::getY).toArray());
    }

    private void setBaseSettingValue(String url, Map<Integer, List<MFOptimizationBaseSetting>> groupedSettings,
                                     MFBaseSettingType baseSettingType, Integer value) throws Exception {
        MFOptimizationBaseSetting baseSetting = groupedSettings.get(baseSettingType.getId()).get(0);
        baseSetting.setCurrentValue(value);
        TestUtils.UPDATEobject(mockMvc, url + "/" + baseSetting.getId(), baseSetting,
                MFOptimizationBaseSetting.class, TestUtils::toJsonNoProxy);
    }

    private List<MFOptimizationTargetFunc> jsonToMFOptimizationTargetFuncList(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.toString(), new TypeReference<List<MFOptimizationTargetFunc>>() {
            });
        } catch (IOException e) {
            Assert.fail("Unable to read value from json");
            return null;
        }
    }

    private List<MFOptimizationVaryingParam> jsonToMFOptimizationVaryingParamList(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.toString(), new TypeReference<List<MFOptimizationVaryingParam>>() {
            });
        } catch (IOException e) {
            Assert.fail("Unable to read value from json");
            return null;
        }
    }

    private List<MFOptimizationBaseSetting> jsonToOptimizationBaseSettingList(JsonNode json) {
        try {
            return jacksonMapper.readValue(json.toString(), new TypeReference<List<MFOptimizationBaseSetting>>() {
            });
        } catch (IOException e) {
            Assert.fail("Unable to read value from json");
            return null;
        }
    }

}
